#========================================================================================
# package      DE_EPAGES::Catalog::API::CSV::CatalogImportHandler;
# base         DE_EPAGES::ShopCSVExportImport::API::CSV::ShopObjectImportHandler
# state        public
#----------------------------------------------------------------------------------------
# description  This module implements a handler to import category product map base objects
#               See <package DE_EPAGES::Catalog::API::CSV::CatalogExportImport> for
#               usage examples.
#========================================================================================
package DE_EPAGES::Catalog::API::CSV::CatalogImportHandler;
use base qw ( DE_EPAGES::ShopCSVExportImport::API::CSV::ShopObjectImportHandler );

use strict;
use DE_EPAGES::Core::API::Warning qw( Warning );
use DE_EPAGES::Core::API::Error qw( Error ExistsError GetError);
use DE_EPAGES::Catalog::API::Constants;
use DE_EPAGES::CSVExportImport::API::Tools qw ( ParseValue);
use DE_EPAGES::Object::API::Factory qw( LoadClassByAlias );

#ERRORDEFINITION
sub ErrorDefinition {
    my $hSuperErrorDefinition = __PACKAGE__->SUPER::ErrorDefinition();
    return { %$hSuperErrorDefinition,
            'CATALOG_EMPTY_PRODUCT'     => 'line #Line, column Product: product has empty value',
            'CATALOG_NO_PRODUCT'        => 'line #Line, column Product: Product does not exist (#Product)',
            'CATALOG_EMPTY_CATEGORY'    => 'line #Line, column Category: category has empty value',
            'CATALOG_NO_CATEGORYOBJECT' => 'line #Line, column #ID: object is not from class category (#Category)',
            'CATALOG_NO_CATEGORY'       => 'line #Line, column Category: Category does not exist (#Category)',
        }
}
#ERRORDEFINITION

#========================================================================================
# function     analyzeAttributeLine
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $hAttributeValues = $self->analyzeAttributeLine();
#----------------------------------------------------------------------------------------
# return       $hAttributeValues | properties of a object of CSV file with following keys
#               <li>DirectAttributes  | direct attributes of the object with following keys
#                   <ul>
#                       <li>Alias | alias of the object | string
#                       <li>Class | class object of the object | object
#                       <li>Parent | parent object of the object | object
#                       <li> more direct attributes which should set via $Object->set method,
#                            or can used by $Class->insert($hAttributeValues->{'DirectAttributes'});
#                   </ul> | ref.hash
#               <li>DoNotImport |  0 or 1, set to 1 if there are errors to set the parsed values | boolean
#               <li>Object | object if it is found or undef if it doesn't exist | object
#               <li>LocalizedAttributes |localized attributes hash with language id as key | ref.hash
#               | ref.hash
# example      hAttributeValues='HASH
#                   'DirectAttributes'  =>'HASH'
#                       'Category'  =>'DE_EPAGES::Content::API::Object::ContentObject=HASH'
#                       'Class' =>'DE_EPAGES::Object::API::Object::Class=HASH'
#                       'Position'  =>'10'
#                       'Product'   =>'DE_EPAGES::Product::API::Object::Product=HASH'
#                   'DoNotImport'   =>'0'
#                   'Object'    =>'DE_EPAGES::Catalog::API::Object::CategoryProductMap=HASH'
#========================================================================================
sub analyzeAttributeLine {
    my $self = shift;

    my $DoNotImport      = 0;

    my $Item              = $self->topID('Item');
    my $ImportAttributes  = $self->topID('ImportAttributes');
    my $Shop = $self->object;

    my ($Product, $Category);
    my $ProductPath       = $Item->[$ImportAttributes->{'ProductColumn'}];
    $ProductPath =~ s/\s+$//; $ProductPath =~ s/^\s+//;
    if ($ProductPath eq '') {
        Warning( 'CATALOG_EMPTY_PRODUCT', { Line => $self->lineNumber , ID => $ImportAttributes->{'ProductColumn'}+1} );
        $DoNotImport = 1;
        return;
    } else {
        my $ProductFolder = $Shop->folder('Products');
        eval { $Product = $self->loadShopObject($ProductFolder, $ProductPath, 'Product'); };
        if (ExistsError()) {
            Warning('CATALOG_NO_PRODUCT', { Line => $self->lineNumber,
                                    Product => $ProductPath } );
            $DoNotImport = 1;
            return;
        }
    }

    # test if we have unity and try to set an own category to a subproduct
    if($Shop->get('IsUnityRegistered') and $Product->get('SuperProduct')){
        $DoNotImport = 1;
        return;
    }

    my $CategoryPath       = $Item->[$ImportAttributes->{'CategoryColumn'}];
    $CategoryPath =~ s/\s+$//; $CategoryPath =~ s/^\s+//;
    if ($CategoryPath eq '') {
        Warning( 'CATALOG_EMPTY_CATEGORY', {
            Line => $self->lineNumber ,
            ID => $ImportAttributes->{'CategoryColumn'}+1
        } );
        $DoNotImport = 1;
        return;
    } else {
        eval { $Category = $self->loadShopObject($Shop, $CategoryPath, 'Category'); };
        if (ExistsError()) {
            my $Error = GetError();
            Warning('CATALOG_NO_CATEGORYOBJECT', {
                Line => $self->lineNumber,
                Category => $CategoryPath
            } ) if $Error->code eq 'FORBIDDEN_OBJECT';
            Warning('CATALOG_NO_CATEGORY', {
                Line => $self->lineNumber,
                Category => $CategoryPath
            } ) if $Error->code eq 'UNKNOWN_OBJECT';
            $DoNotImport = 1;
            return;
        }
    }
    my $Class = LoadClassByAlias('CategoryProductMap');

    # exists category product map?
    my $aMaps = $Class->getObjects('ByCategoryIDProductID', [$Category->id, $Product->id]);
    my $CategoryProductMap = scalar @$aMaps ? $aMaps->[0] : undef;

    my %AttributeValues;
    my $hFormat = $self->locale->get([qw(LocaleID DecimalSep ThousandSep)]);

    for (my $i=0; $i<scalar(@$Item); $i++) {
        last if !exists $ImportAttributes->{$i}; # ignore extra columns

        my $hAttribute =  $ImportAttributes->{$i};
        next if ($hAttribute->{'Alias'} eq 'Product' or $hAttribute->{'Alias'} eq 'Category');

        my $Value =  $Item->[$i];
        eval { ParseValue($hAttribute->{'Alias'},$hAttribute->{'Type'}, \$Value, $hFormat) };
        if ( ExistsError() ) {
            Warning( 'VOID_FIELD', { Line => $self->lineNumber, ID => $i+1, Error => GetError()->shortMessage } );
            $DoNotImport = 1;
            next;
        }
        if (not defined $Value and $hAttribute->{'Attribute'}->get('IsMandatory')) {
            Warning( 'NO_MAND_VALUE', {
                Line => $self->lineNumber,
                Attribute => $hAttribute->{'Alias'}
            });
            $DoNotImport = 1;
            next;
        }
        my $AttributeType = $hAttribute->{'Type'};
        if ($AttributeType =~ /^Localized/) {
            $AttributeValues{'LocalizedAttributes'}{$hAttribute->{'LanguageID'}}{$hAttribute->{'Alias'}} = $Value;
        } else {
            $AttributeValues{'DirectAttributes'}{$hAttribute->{'Alias'}} = $Value;
        }
    }
    $AttributeValues{'DirectAttributes'}{'Class'} = $Class;
    $AttributeValues{'DirectAttributes'}{'Product'} = $Product;
    $AttributeValues{'DirectAttributes'}{'Category'} = $Category;

    $AttributeValues{'DoNotImport'}  = $DoNotImport;
    $AttributeValues{'Object'}       = $CategoryProductMap;
    return \%AttributeValues;
}

#========================================================================================
# function     testMissingFields
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->testMissingFields($hUsedAttributes);
#----------------------------------------------------------------------------------------
# description  test for mandatory columns Class,Parent called by analyzeHeaderLine
#               throws error MISSING_FIELD if they don't exist
#----------------------------------------------------------------------------------------
# input        $hUsedAttributes | used attributes | ref.hash
#========================================================================================
sub testMissingFields {
    my $self = shift;
    my ( $hUsedAttributes ) = @_;
    Error( 'MISSING_FIELD' => { 'Value' => 'Category', 'FileName' => $self->{'FileName'}})
        unless exists $hUsedAttributes->{'CategoryColumn'};
    Error( 'MISSING_FIELD' => { 'Value' => 'Product', 'FileName' => $self->{'FileName'}})
        unless exists $hUsedAttributes->{'ProductColumn'};
    Error( 'MISSING_FIELD' => { 'Value' => 'Position', 'FileName' => $self->{'FileName'}})
        unless exists $hUsedAttributes->{'PositionColumn'};
    return;
}

#========================================================================================
# function     modifyUsedAttributes
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->modifyUsedAttributes($hUsedAttributes, $ColumnIndex);
#----------------------------------------------------------------------------------------
# description  add keys:
#                 $hUsedAttributes->{$ColumnIndex}{'IsSuperProductAliasColumn'} =
#                   $hUsedAttributes->{$ColumnIndex}{'Alias'}  eq 'SuperProduct'  ? 1 : 0;
#                 $hUsedAttributes->{'SuperProductAliasColumn'} with value=$ColumnIndex
#                   if $hUsedAttributes->{$ColumnIndex}{'IsSuperProductAliasColumn'};
#----------------------------------------------------------------------------------------
# input        $hUsedAttributes | used attributes | ref.hash
# input        $ColumnIndex | current column | integer
#========================================================================================
sub modifyUsedAttributes {
    my $self = shift;
    my ( $hUsedAttributes, $ColumnIndex) = @_;
    $hUsedAttributes->{$ColumnIndex}{'IsCategoryColumn'} = $hUsedAttributes->{$ColumnIndex}{'Alias'} eq 'Category' ? 1 : 0;
    $hUsedAttributes->{$ColumnIndex}{'IsProductColumn'}  = $hUsedAttributes->{$ColumnIndex}{'Alias'} eq 'Product'  ? 1 : 0;
    $hUsedAttributes->{$ColumnIndex}{'IsPositionColumn'} = $hUsedAttributes->{$ColumnIndex}{'Alias'} eq 'Position' ? 1 : 0;
    $hUsedAttributes->{'CategoryColumn'} = $ColumnIndex if $hUsedAttributes->{$ColumnIndex}{'IsCategoryColumn'};
    $hUsedAttributes->{'ProductColumn'}  = $ColumnIndex if $hUsedAttributes->{$ColumnIndex}{'IsProductColumn'};
    $hUsedAttributes->{'PositionColumn'} = $ColumnIndex if $hUsedAttributes->{$ColumnIndex}{'IsPositionColumn'};
    return;
}

#========================================================================================
# function     baseAttributeClassAlias
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->baseAttributeClassAlias();
#----------------------------------------------------------------------------------------
# description  return Base AttributeClass Alias, here : 'ContentObject'
#----------------------------------------------------------------------------------------
# return        $BaseAttributeClassAlias | attribute class alias of normal attributes | string
#========================================================================================
sub baseAttributeClassAlias {
    my $self = shift;
    return 'CategoryProductMap';
}

1;

