#========================================================================================
# package      DE_EPAGES::Content::API::CSV::ContentExportImport;
# state        public
#----------------------------------------------------------------------------------------
# description  This module implements the high level interface for import and export of
#               categories and pages to CVS files.
#========================================================================================
package DE_EPAGES::Content::API::CSV::ContentExportImport;
use base DE_EPAGES::ShopCSVExportImport::API::CSV::ShopObjectExportImport;

use strict;
use DE_EPAGES::Content::API::Constants;
use DE_EPAGES::Content::API::Content qw (ContentChildClasses);
use DE_EPAGES::Content::API::CSV::ContentExportDriver;
use DE_EPAGES::Content::API::CSV::ContentImportHandler;
use DE_EPAGES::Core::API::PerlTools qw( ElementOf );
use DE_EPAGES::Core::API::Warning qw(Warning);
use DE_EPAGES::Object::API::Factory qw( LoadClassByAlias LoadObject );
use DE_EPAGES::Object::API::Table::Locale qw(GetInfoLocale);
use DE_EPAGES::Permission::API::Permission qw ( ActionByName ExistsActionByName );
use DE_EPAGES::Presentation::API::PageType qw (LoadPageTypeByViewAction);

#========================================================================================
# function     warningCodes
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aWarningCodes = $Driver->warningCodes();
#----------------------------------------------------------------------------------------
# description  Returns the warning codes for csv export.
#----------------------------------------------------------------------------------------
# return       $aWarningCodes | warning codes | array.string
#========================================================================================
sub warningCodes {
    my $self = shift;
    return [ @{$self->SUPER::warningCodes}, @CSV_WARNING_CODES_CONTENT ];
}

#========================================================================================
# function     exportDriver
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Driver->exportDriver();
#----------------------------------------------------------------------------------------
# description  Return driver package.
#----------------------------------------------------------------------------------------
# return       $DriverPackage  | package to driver | string
#========================================================================================
sub exportDriver {
    my $self = shift;
    return 'DE_EPAGES::Content::API::CSV::ContentExportDriver';
}

#========================================================================================
# function     importDriver
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Driver->importDriver();
#----------------------------------------------------------------------------------------
# description  Return driver package.
#----------------------------------------------------------------------------------------
# return       $DriverPackage  | package to driver | string
#========================================================================================
sub importDriver {
    my $self = shift;
    return 'DE_EPAGES::Content::API::CSV::ContentImportHandler';
}

#========================================================================================
# function     getExportObjects
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Driver->getExportObjects($Shop, $aObjects);
#----------------------------------------------------------------------------------------
# description  Returns list of objects to export.
#----------------------------------------------------------------------------------------
# input        $Shop           | alias of shop          | object
# input        $aObjects       | list of exportable objects | ref.array.integer
# return       $aExportObjects | edited list of exportable objects | ref.array.integer
#========================================================================================
sub getExportObjects {
    my $self = shift;
    my ($Shop, $aObjects) = @_;

    my $CategoryClass = LoadClassByAlias('Category');

    if (not defined $aObjects){
        my @Objects;
        my $BaseCategory = $Shop->get('Categories');
        push @Objects, $BaseCategory->id;
        $self->categoryChildren($CategoryClass, $BaseCategory, \@Objects);
        my $ItemCount = scalar(@Objects);
        Warning('NO_CONTENTITEMS') unless $ItemCount;

        return \@Objects;
    } else{
        #get all sub categories for selected objects
        my @Categories = @$aObjects;
        foreach my $ObjectID ( @Categories ) {
            my $Category = LoadObject($ObjectID);
            $self->categoryChildren($CategoryClass, $Category, $aObjects) if $Category->instanceOf($CategoryClass);
        }
        my $ItemCount = scalar(@$aObjects);
        Warning('NO_CONTENTITEMS') unless $ItemCount;

        return $aObjects;
    }
}

sub categoryChildren {
    my $self = shift;
    my ($CategoryClass, $Category, $aItems) =  @_;
    my $aChildObjectIDs = $Category->findChildren;
    push @$aItems, @$aChildObjectIDs;
    foreach my $ChildID ( @$aChildObjectIDs ) {
        my $ChildCategory = LoadObject($ChildID);
        $self->categoryChildren($CategoryClass, $ChildCategory, $aItems) if $ChildCategory->instanceOf($CategoryClass);
    }
    return;
}

#========================================================================================
# function     csvAttributes
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $aCSVAttributes = $Driver->csvAttributes($Shop, $LocaleID, $aLanguageCode2s);
#----------------------------------------------------------------------------------------
# description  Return list of exportable csv attributes.
#----------------------------------------------------------------------------------------
# input        $Shop | shop object | object
# input        $LocaleID | locale identifier (e.g. en_GB) | string
# input        $aLanguageCode2s | array of languages identified by code2 | ref.array.string
# return       $aAttributes | list of attributes | ref.array.hash
#========================================================================================
sub csvAttributes {
    my $self = shift;
    my ($Shop, $LocaleID, $aLanguageCode2s) = @_;

    my $hLocale = GetInfoLocale($LocaleID);
    my $ahLanguages = $self->getLanguages($Shop, $aLanguageCode2s);

    my @ContentAttributes = @{LoadClassByAlias('ContentObject')->attributes};
    my @SpecialClassAliases = qw ( Article Category Link );
    my @SpecialClassIDs;
    foreach my $ClassAlias (@SpecialClassAliases) {
        my $Class = LoadClassByAlias($ClassAlias);
        my @Attributes = @{$Class->directAttributes};
        push @ContentAttributes, @Attributes;
        push @SpecialClassIDs, $Class->id;
    }

    my @CSVAttributes;
    foreach my $Attribute ( sort {$a->get('AttributeClass')->id <=> $b->get('AttributeClass')->id or $a->get('Position') <=> $b->get('Position')} @ContentAttributes ) {
        next unless IsAttributeForCSV($Attribute, $Shop);
        my $AttributeClass =  $Attribute->get('AttributeClass');
        if (ElementOf($AttributeClass->id,  @SpecialClassIDs )) {
            push @CSVAttributes, @{
                $self->getAttributeInfo(
                    $Attribute, $hLocale->{'LanguageID'}, $ahLanguages, {
                        'ClassAlias' => $AttributeClass->alias,
                        'IsBaseClassAttribute' => 0
                    }
                )
            };
        } else {
            push @CSVAttributes, @{
                $self->getAttributeInfo(
                    $Attribute, $hLocale->{'LanguageID'}, $ahLanguages, {'ClassAlias' => 'ContentObject'}
                )
            };
        }
    }

    # Layout
    push @CSVAttributes, @{TemplateTypeAttributes($Shop, $hLocale)};

    my @FilteredCSVAttributes = @CSVAttributes;

    # filter unneeded attributes for unity shops (see: @UNITY_CONTENT_TYPE_ATTRIBUTES_IGNORE)
    if($Shop->get('IsUnityRegistered')){
        @FilteredCSVAttributes = ();
        foreach my $hAttribute (@CSVAttributes) {
            push @FilteredCSVAttributes, $hAttribute unless ElementOf($hAttribute->{'Alias'}, @UNITY_CONTENT_TYPE_ATTRIBUTES_IGNORE);
        }
    }

    return \@FilteredCSVAttributes;
    
}

sub IsAttributeForCSV {
    my ($Attribute, $Shop) = @_;
    my $Alias = $Attribute->alias;
    return 0 if ElementOf($Alias, @{$CSV_SPECIAL_CONTENTITEMS->{0}});
    return 0 if ($Alias eq 'FilePathPart' and not $Shop->get('IsFilePathPartUsed'));
    return 0 if $Attribute->get('IsArray') and not ElementOf($Alias, @{$CSV_SPECIAL_CONTENTITEMS->{1}}) ;
    # not this attribute if it is not exportable and not needed
    return 0 unless (  $Attribute->get('IsExportable') or ElementOf($Alias, @{$CSV_SPECIAL_CONTENTITEMS->{1}}) );
    return 1;
}

sub TemplateTypeAttributes {
    my ($Shop, $hLocale) = @_;

    my @CSVAttributes;
    my %Info;

    my $ClassStartPage = LoadClassByAlias('StartPage');
    my @ContentChildClasses = (LoadClassByAlias('StartPage'));
    push @ContentChildClasses, sort { $a->get('Position') <=> $b->get('Position') }
        @{ContentChildClasses($Shop, $ClassStartPage)};

    foreach my $ContentClass (@ContentChildClasses) {
        my $PageType = LoadPageTypeByViewAction( ActionByName( $ContentClass, 'View' ) );
        if ($PageType->existsChild('TemplateTypes') && ExistsActionByName($ContentClass, 'MBO-ViewDesign') ) {
            my $LayoutAction = ActionByName( $ContentClass, 'MBO-ViewDesign' );
            foreach my $TemplateType (@{$PageType->child('TemplateTypes')->children}) {
                my $AliasString = $ContentClass->alias . '/' . $TemplateType->alias;
                $Info{'Alias'} = $AliasString;
                my $NativeString = join '/', $LayoutAction->get('NameOrAlias', $hLocale->{'LanguageID'}),
                    $ContentClass->get('NameOrAlias', $hLocale->{'LanguageID'}),
                    $TemplateType->get('NameOrAlias', $hLocale->{'LanguageID'});
                $Info{'Name'} = $NativeString;
                $Info{'IsBaseClassAttribute'} = 0;
                $Info{'HeaderString'} = $NativeString. " [$AliasString]";
                $Info{'AliasString'}  = $AliasString;
                $Info{'Type'}  = 'Template';
                push @CSVAttributes, {%Info};
            }
        }
    }

    return \@CSVAttributes;
}

1;
