#========================================================================================
# package      DE_EPAGES::Core::API::File
# state        public
#----------------------------------------------------------------------------------------
# description  This module provides file and directory access.
#----------------------------------------------------------------------------------------
# example      use DE_EPAGES::Core::API::File qw( GetFileContent WriteFile ...);
#========================================================================================
package DE_EPAGES::Core::API::File;

use base qw( Exporter );

use strict;
our @EXPORT_OK = qw(
    RealPath
    AbsolutePath
    WindowsPath
    RelativePath

    MakePathWriteable
    MakePathReadable
    ResolveSymbolicLink
    fsplit
    SetAccessRights
    SetAccessRightsRecursive
    GetAccessRights
    GetExecBit
    IsInPath

    CreateDirectory
    DeleteDirectory
    DeleteStaticAtDirectory
    ExistsDirectory
    CopyDirectory
    BackupDirectory
    RebuildDirectory
    RenameDirectory
    GetInfoDirectory
    GetSubDirectories
    GetTrashRoot
    GetFileNames
    CompareDirectories
    GetDirectorySize
    CleanPath

    PathGlob2TypeGlob
    FilesOnlyInDir1
    GetDOSReadOnlyFiles

    MergeIniFiles
    WriteFile
    LockFile
    OpenFile
    MoveFile
    DeleteFile
    DeleteListedFiles
    CopyFile
    BackupFile
    MoveToDateDir
    GetFileContent
    GetFileContentLines
    GetMountPoint
    SameDevice
    GetFileInfo
    GetFileSize
    GetFileMD5Digest
    GetFileSHA1Digest
    GetSimilarFileName
    CompareMD5File
    CompareMD5TextFile
    CompareFileWithMD5File
    CompareDirectoryWithFilelist
    ExistsFile
    ExistsFileCaseSensitive
    GetCachedFile
    IsTextFile
    IsImageFile
    IsCustomizedFile
    Text2Unix
    Text2Dos
    DetectFileEncoding
    GetBOM
    IsFileNewer

    ConvertPathChars
    CorrectPathSlashes

    RunOnDirectory
    RunOnGlob
    CaptureStream

    CreateSymLink
    DeleteSymLink
    IsSymLink
    GetLocalFSTypes
    GetFSType
);

use Config;
use Cwd;
use Encode qw( encode );
use Fcntl qw(:DEFAULT :flock);
use File::Basename  qw ( dirname basename );
use File::Path      qw ( mkpath rmtree );
use File::Compare   qw ( compare );
use File::Copy      qw ( copy move );
use File::Find      qw ( find );
use File::Spec;
use Config::IniFiles;
use DateTime;
use Digest::MD5;
use Digest::SHA1;
use Time::HiRes qw (time);
use DE_EPAGES::Core::API::Constants qw (ALL_RW ALL_RWS ALL_RWX ALL_X BASE_MASK
                                        BASE_RW BASE_RWX GROUP_R GROUP_RX OWNER_RW
                                        USER_R USER_RW USER_RWX USER_RX SET_EFFECTIVE
                                        BOM DEFAULT_ENCODING SECONDS_PER_DAY
                                        IMAGE_FILE_TYPES TEXT_FILE_NAMES TEXT_FILE_TYPES);
use DE_EPAGES::Core::API::Error         qw ( Error ExistsError GetError HandleError );
use DE_EPAGES::Core::API::Log           qw ( GetLog );
use DE_EPAGES::Core::API::MimeTypes     qw ( GetMimeTypeByExtension );
use DE_EPAGES::Core::API::PerlTools     qw ( GetLocalTimeZone ElementOf GetMD5Digest GenerateGUID UsePackage );
use DE_EPAGES::Core::API::Statistics    qw ( AddStatistic );
use DE_EPAGES::Core::API::TestParameter qw ( TestString TestInteger );
use DE_EPAGES::Core::API::Warning       qw ( Warning );

# unluckily we can't use SystemCommand here
# because SystemCommand and File require each other
# use DE_EPAGES::Core::API::SystemCommand qw(SystemCommand);

##no critic qw(RequireLocalizedPunctuationVars)
${^WIDE_SYSTEM_CALLS} = 1;
$ENV{EPAGES_WEBSECURE} //= "$ENV{EPAGES_SHARED}/WebSecure";
##use critic (RequireLocalizedPunctuationVars)

#ERRORDEFINITION
sub ErrorDefinition () { return {
    'FILE_CHMOD_FAILED'     => 'couldn\'t chmod entry (#FileName) because #Code.',
    'FILE_CHOWN_FAILED'     => 'couldn\'t chown entry (#FileName) because #Code.',
    'FILE_UTIME_FAILED'     => 'couldn\'t utime entry (#FileName) because #Code.',
    'FILE_COPY_FAILED'      => 'couldn\'t copy file, because #Code (#Source -> #Target).',
    'FILE_LINK_FAILED'      => 'couldn\'t link file, because #Code (#Source -> #Target).',
    'CROSS_DEVICE_FAILURE'  => 'couldn\'t link file, because #Code (#Source -> #Target).',
    'FILE_MOVE_FAILED'      => 'couldn\'t move file, because #Code (#Source -> #Target).',
    'FILE_RENAME_FAILED'    => 'couldn\'t rename file, because #Code (#Source -> #Target).',
    'FILE_OPEN_FAILED'      => 'couldn\'t open file (#FileName) because #Code.',
    'FILE_CLOSE_FAILED'     => 'couldn\'t close file (#FileName) because #Code.',
    'FILE_REMOVE_FAILED'    => 'couldn\'t remove file (#FileName) because #Code.',
    'FILE_ACCESS_FAILED'    => 'couldn\'t change access of file (#FileName).',
    'FILE_NOT_FOUND'        => 'file not found (#FileName).',
    'FILE_STATISTIC_FAILED' => 'couldn\'t get statistic of file (#FileName) because #Code.',
    'FILE_LOCK_FAILED'      => 'couldn\'t lock file using #Method, because #Code.',
    'FILE_UNLOCK_FAILED'    => 'couldn\'t unlock file using #Method, because #Code.',
    'DIR_OPEN_FAILED'       => 'couldn\'t open directory (#Directory) because #Code.',
    'DIR_CLOSE_FAILED'      => 'couldn\'t close directory (#Directory) because #Code.',
    'DIR_CREATE_FAILED'     => 'couldn\'t create directory (#Directory) because #Code.',
    'DIR_REMOVE_FAILED'     => 'couldn\'t remove directory (#Directory) because #Code.',
    'DIR_RENAME_FAILED'     => 'couldn\'t rename directory (#Source -> #Target) because #Code.',
    'SYSTEM_CMD_ERROR'      => "system command '#Command' failed (exit code: #Code):\n#Message",
    'SYSTEM_CMD_TERMINATED' => 'system command \'#Command\' terminated with signal #Signal.',
    'SYSTEM_CMD_FAILED'     => 'system command \'#Command\' failed (exit code: #Code):\n#Message',
    'SYSTEM_CMD_MISSING'    => 'system command \'#Command\' not executable on operating system \'#OS\'.',
    'NO_TRASH_DIR'          => 'there is no trash for directory (#Path)',
    'CHANGE_DIR_FAILED'     => 'couldn\'t change to directory \'#Directory\' because #Code.',
    'SYMLINK_REMOVE_FAILED' => 'couldn\'t remove symlink (#FileName) because #Code.',
    'SYMLINK_CREATE_FAILED' => 'couldn\'t create symlink (#FileName) because #Code.',
                         }
}
#ERRORDEFINITION END

#========================================================================================
# function     RealPath
#----------------------------------------------------------------------------------------
# syntax       $DetPath = RealPath($Path);
#----------------------------------------------------------------------------------------
# description  Converts paths like 'C:/sub1/sub2/..' to 'C:/sub1'.
#----------------------------------------------------------------------------------------
# input        $Path | directory path | string
# return       $DetPath | directory path | string
#========================================================================================
sub RealPath ($) {
    my ($Path) = @_;
    TestString('Path', $Path);

    # replace all backslashes with forward slashes
    $Path =~ s/\\/\//go;

    # process/remove all .. and . from path
    my (@dirs, @pre_dirs);
    map {
        if    ($_ eq '.')  {}
        elsif ($_ eq '..') {
            if (scalar @dirs) {
                pop @dirs;
            } else {
                push @pre_dirs, $_;
            }
        }
        elsif ($_ ne '')   { push @dirs, $_; }

    } split(/\//, $Path);

    # now rebuild path - rebuild up to 2 leading slashes
    $Path = (($Path =~ /^\/\//) ? '//' : (($Path =~ /^\//) ? '/' : '')) . join ('/', @pre_dirs, @dirs);
    return $Path;
}

#========================================================================================
# function     AbsolutePath
#----------------------------------------------------------------------------------------
# syntax       $AbsPath = AbsolutePath($RelPath);
# syntax       $AbsPath = AbsolutePath($RelPath, $BaseDir);
#----------------------------------------------------------------------------------------
# description  Converts relative paths (as sub1/sub2) to absolute paths
#               (as C:temp\sub1\sub2 or /tmp/sub1/sub2)
#               by prepending $BaseDir (if given) or getcwd().
#----------------------------------------------------------------------------------------
# input        $RelPath | relative path | string
# input        $BaseDir | (optional, default=getcwd()) base directory | string
# return       $AbsPath | absolute path | string
#========================================================================================
sub AbsolutePath ($;$) {
    my ($Path, $BaseDir) = @_;
    TestString('Path', $Path);

    return File::Spec->rel2abs( $Path, $BaseDir );
}

#========================================================================================
# function     WindowsPath
#----------------------------------------------------------------------------------------
# syntax       $WindowsPath = WindowsPath($AnyPath);
#----------------------------------------------------------------------------------------
# description  Replaces slashes of an unix path with back-slashes.
#----------------------------------------------------------------------------------------
# input        $AnyPath | directory/file path (unix/windows format) | string
# return       $WindowsPath | directory/file path (windows format) | string
#========================================================================================
sub WindowsPath ($) {
    my ($Path) = @_;
    TestString('Path', $Path);

    # replace all forward slashes with backslashes
    $Path =~ s/\//\\/go;
    return $Path;
}

#========================================================================================
# function     CorrectPathSlashes
#----------------------------------------------------------------------------------------
# syntax       $CorrectedPath = CorrectPathSlashes($AnyPath);
#----------------------------------------------------------------------------------------
# description  Execute WindowsPath($AnyPath) if running on windows, else do nothing.
#----------------------------------------------------------------------------------------
# input        $AnyPath | directory/file path (unix/windows format) | string
# return       $WindowsPath | directory/file path (right format) | string
#========================================================================================
sub CorrectPathSlashes ($) {
    my ($Path) = @_;
    TestString('Path', $Path);

    # unix: do not change $Path
    return $Path if $Config{'osname'} ne 'MSWin32';

    # windows: replace all slashes with backslashes in $Path
    return WindowsPath($Path);
}

#========================================================================================
# function     GetDOSReadOnlyFiles
#----------------------------------------------------------------------------------------
# syntax       $aReadOnlyFiles = GetDOSReadOnlyFiles($AnyPath);
#----------------------------------------------------------------------------------------
# description  Returns a ref.array of DOS files with read-only flag
#----------------------------------------------------------------------------------------
# input        $AnyPath | directory/file path (unix/windows format) | string
# return       $aReadOnlyFiles | list of files with read-only flag | ref.array
#========================================================================================
sub GetDOSReadOnlyFiles ($) {
    my ($Path) = @_;
    TestString('Path', $Path);

    return [] unless $Config{'osname'} eq 'MSWin32';
    return [] unless -d $Path;

    $Path = WindowsPath($Path);
    my $ReadOnlyFiles = `dir /S /A:R /B $Path\\\* 2>nul`;
    my @ReadOnlyList = split /\r?\n/, $ReadOnlyFiles;

    return \@ReadOnlyList;
}

#========================================================================================
# function     MakePathWriteable
#----------------------------------------------------------------------------------------
# syntax       ($ExistingMode, $ExistingPath) = MakePathWriteable($Path);
# syntax       ($ExistingMode, $ExistingPath) = MakePathWriteable($Path, $ReadOnly);
#----------------------------------------------------------------------------------------
# description  Returns existing part of $Path and makes it writeable.
#----------------------------------------------------------------------------------------
# input        $Path | directory/file path | string
# input        $ReadOnly | make it readable, not writeable | string
# return       $ExistingMode | file mode | string
# return       $ExistingPath | directory/file path | string
#========================================================================================
sub MakePathWriteable ($;$) {
    my ($Path, $ReadOnly) = @_;
    TestString('Path', $Path);

    if ($Config{'osname'} eq 'MSWin32') {
        $Path = WindowsPath($Path);
        return (ALL_RWX, $Path) if $ReadOnly;

        if ( -f $Path ) {
            system( "attrib -R \"$Path\"" );
        } elsif ( -d $Path) {
            system( "attrib -R /S \"$Path\\\*\"" );
        }
        return (ALL_RWX, $Path);
    }

    my( $ExistingPath, $ExistingMode, $MyPath, $MyMode ) = ('', '', '', '');

    $Path = AbsolutePath($Path);

    if ( -e $Path ) {
        $ExistingPath = $Path;
    } else {
      FOREACH_DIR:
        $ExistingPath = '';
        $Path =~ s,^/+,,;
        foreach my $Dir (split /\//, $Path) {
            my $ResolvedLink = ResolveSymbolicLink("$ExistingPath/$Dir");
            if ($ResolvedLink ne "$ExistingPath/$Dir") {
                $Path = "/$Path";
                $Path =~ s,^\Q$ExistingPath/$Dir\E,$ResolvedLink,;
                goto FOREACH_DIR;
            }
            last unless -d "$ExistingPath/$Dir";
            $ExistingPath .= "/$Dir";
        }
        $ExistingPath = '/' if $ExistingPath eq '';
    }

    $MyMode = USER_RX; # owner-readable
    foreach my $Dir (split /\//, dirname($ExistingPath)) {
        $MyPath .= "/$Dir";
        $ExistingMode = (stat $MyPath)[2] & ALL_RWS;
        chmod($ExistingMode | $MyMode, $MyPath);
    }

    if ($ReadOnly) {
        $MyMode = -f $ExistingPath ? USER_R : USER_RX; # owner-readable
    } else {
        $MyMode = -f $ExistingPath ? USER_RW : USER_RWX; # owner-writeable
    }
    $ExistingMode = (stat $ExistingPath)[2] & ALL_RWS;
    chmod($ExistingMode | $MyMode, $ExistingPath);

    return ($ExistingMode, $ExistingPath);
}

#========================================================================================
# function     MakePathReadable
#----------------------------------------------------------------------------------------
# syntax       ($ExistingMode, $ExistingPath) = MakePathReadable($Path);
#----------------------------------------------------------------------------------------
# description  Returns existing part of $Path and makes it readable.
#----------------------------------------------------------------------------------------
# input        $Path | directory/file path | string
# return       $ExistingMode | file mode | string
# return       $ExistingPath | directory/file path | string
#========================================================================================
sub MakePathReadable ($) {
    my ($Path) = @_;
    return MakePathWriteable($Path, 1);
}

#========================================================================================
# function     ResolveSymbolicLink
#----------------------------------------------------------------------------------------
# syntax       $ResolvedLink = ResolveSymbolicLink($Path);
#----------------------------------------------------------------------------------------
# description  Returns resolved link of path if it exists, otherwise it returns path.
#----------------------------------------------------------------------------------------
# input          $Path | directory/file path | string
# return $SymbolicPart | directory/file path | string
#========================================================================================
sub ResolveSymbolicLink ($) {
    my ($Path) = @_;

    return $Path if $Config{'osname'} eq 'MSWin32';
    return $Path unless -l $Path;

    my $Link = readlink($Path);
    return $Path if $Link eq '';

    $Link = AbsolutePath(dirname($Path)). "/$Link" unless $Link =~ /^\//;
    $Link = AbsolutePath($Link);
    return $Link ne '' ? $Link : $Path;
}

#========================================================================================
# function     fsplit
#----------------------------------------------------------------------------------------
# syntax       ($Path, $FileName, $Extension) = fsplit($FilePath);
#----------------------------------------------------------------------------------------
# description  This function splits a path into its parts, this works for unix and dos paths.
#               Parts, not present, are returned empty or undef.
# example      ('/', 'roedel', '.txt') = fsplit('/roedel.txt');
# example      ('/', 'roedel', undef ) = fsplit('/roedel');
# example      ('.', 'roedel', '.txt') = fsplit('roedel.txt');
# example      ('c:/epages', 'roedel', '.txt') = fsplit('c:/epages/roedel.txt');
# example      ('\\server\share', 'roedel', '.txt') = fsplit('\\server\share\roedel.txt');
#----------------------------------------------------------------------------------------
# input        $FilePath | A complete file path.
# return       $Path   | path to file | string
# return       $File | file name without extension | string
# return       $Extension |  file extension (includes '.'), '' if the file name does
#                             not contain a dot | string
#========================================================================================
sub fsplit ($) {
    my ($FilePath) = @_;
    TestString('FilePath', $FilePath);

    # define file name
    my $FileName = $FilePath;
    $FileName =~ s/\\/\//go; # required for splitting DOS names on Unix systems
    $FileName = basename($FileName);

    # define path
    $FilePath =~ s/\Q$FileName\E$//;
    $FilePath =~ s/(.)[\\\/]+$/$1/x;

    # define extension
    my $Extension = $FileName =~ s/(\.[^\.]*)$//x ? $1 : '';

    return ($FilePath, $FileName, $Extension);
}

#========================================================================================
# function     SetAccessRights
#----------------------------------------------------------------------------------------
# syntax       SetAccessRights($FileName, $hFlags);
#----------------------------------------------------------------------------------------
# description  Changes access rights (chmod and chown) for unix
#               fallbacks for undefined fields:
#               - UserID: get UID from User
#               - User: default owner (depends on path of $FileName)
#               - GroupID: get GID from Group
#               - Group: default group (depends on path of $FileName)
#               - ExecBit: -d $FileName ? 1 : 0
#               - Mode: default mode (depends on path of $FileName) plus ExecBit.
#               Does nothing if:
#               - function is executed under Windows
#               - $FileName does not exist
#               - none of $hFlags (UserID GroupID Mode aTime mTime) is defined AND
#                 (stat eq ignore OR $FileName is outside of $EPAGES)
#----------------------------------------------------------------------------------------
# input        $FileName | file to be modified | string
# input        $hFlags | flags
#                   <ul>
#                       <li>Mode      - constant as defd in DE_EPAGES::Core::API::Constants
#                       <li>ExecBit   - 0: remove EXECUTE_BIT
#                                     - 1: add EXECUTE_BIT
#                                     - 2: add EXECUTE_BIT (even if digit == 0)
#                       <li>UserID    - user ID  - integer
#                       <li>GroupID   - group ID - integer
#                       <li>User      - user name  - string
#                       <li>Group     - group name - string
#                       <li>aTime     - access time (seconds since 1.1.1970 UTC)
#                       <li>mTime     - modified time (seconds since 1.1.1970 UTC)
#                       <li>stat      - return if !$KnownAccessRights and stat eq ignore
#                   </ul>
#               | ref.hash
#========================================================================================
sub SetAccessRights ($;$) {
    my ($FileName, $hFlagsInput) = @_;

    return if $Config{'osname'} eq 'MSWin32';
    return unless -e $FileName;

    my $hFlags = defined $hFlagsInput ? { %$hFlagsInput } : {};

    $hFlags->{stat} = 'epages' if !defined $hFlags->{stat};

    my (%User, %Group);
    _define_users($hFlags, \%User, \%Group);

    my $KnownAccessRights = 0;
    do { $KnownAccessRights = 1 if defined $hFlags->{$_} } foreach (qw(UserID GroupID Mode aTime mTime));
    return if not $KnownAccessRights and $hFlags->{stat} eq 'ignore';

    foreach my $Var (qw(EPAGES_APACHE EPAGES_CONFIG EPAGES_LOG EPAGES_STATIC EPAGES_STORES
                        EPAGES_WEBROOT EPAGES_WEBSECURE
                        EPAGES_CARTRIDGES EPAGES_PERL EPAGES_SHARED EPAGES)) {
        next unless defined $ENV{$Var};
        $ENV{$Var} =~ s,([^/])//+,$1/,g;
        $ENV{$Var} =~ s,([^/])/+$,$1,g;
    }

    my $Path = AbsolutePath($FileName);
    my $epStat = {};

    ##no critic qw(ProhibitCascadingIfElse)
    if (_UnixPathStartsWith($Path, $ENV{EPAGES_APACHE})) {$epStat = _GET_EPAGES_APACHE_PERMS($Path, $hFlags, \%User, \%Group);}
    elsif (_UnixPathStartsWith($Path, $ENV{EPAGES_CONFIG})) {$epStat = _GET_EPAGES_CONFIG_PERMS($Path, $hFlags, \%User, \%Group);}
    elsif (_UnixPathStartsWith($Path, $ENV{EPAGES_LOG})) {$epStat = _GET_EPAGES_LOG_PERMS($Path, $hFlags, \%User, \%Group);}
    elsif (_UnixPathStartsWith($Path, $ENV{EPAGES_STATIC})) {$epStat = _GET_EPAGES_STATIC_PERMS($Path, $hFlags, \%User, \%Group);}
    elsif (_UnixPathStartsWith($Path, $ENV{EPAGES_STORES})) {$epStat = _GET_EPAGES_STORES_PERMS($Path, $hFlags, \%User, \%Group);}
    elsif (_UnixPathStartsWith($Path, $ENV{EPAGES_WEBROOT})) {$epStat = _GET_EPAGES_WEBROOT_PERMS($Path, $hFlags, \%User, \%Group);}
    elsif (_UnixPathStartsWith($Path, $ENV{EPAGES_WEBSECURE})) {$epStat = _GET_EPAGES_WEBSECURE_PERMS($Path, $hFlags, \%User, \%Group);}
    elsif (_UnixPathStartsWith($Path, $ENV{EPAGES_CARTRIDGES})) {$epStat = _GET_EPAGES_CARTRIDGES_PERMS($Path, $hFlags, \%User, \%Group);}
    elsif (_UnixPathStartsWith($Path, $ENV{EPAGES_PERL})) {$epStat = _GET_EPAGES_PERL_PERMS($Path, $hFlags, \%User, \%Group);}
    elsif (_UnixPathStartsWith($Path, $ENV{EPAGES_SHARED})) {$epStat = _GET_EPAGES_SHARED_PERMS($Path, $hFlags, \%User, \%Group);}
    elsif (_UnixPathStartsWith($Path, $ENV{EPAGES})) {$epStat = _GET_EPAGES_PERMS($Path, $hFlags, \%User, \%Group);}
    elsif (not $KnownAccessRights) { return; }
    ##use critic qw(ProhibitCascadingIfElse)

    my $Mode = _define_mode($Path, $hFlags, $epStat);

    do { $hFlags->{$_} = $epStat->{$_} if !defined $hFlags->{$_} } foreach (qw(UserID GroupID));
    my ($UserID, $GroupID, $aTime, $mTime) = @$hFlags{qw(UserID GroupID aTime mTime)};
    $UserID = '' unless defined $UserID;

    my @EntryStat = stat $Path;
    my $MyUserID = $<;

    my $DoWarn = ( grep { _UnixPathStartsWith($Path, $_) } ($ENV{EPAGES_LOG}, "$ENV{EPAGES_SHARED}/Monitor") )
        ? 0 : 1;

    GetLog->debug(sprintf('Me=%s, File=%s, Mode=0%o, UserID=%s, GroupID=%s',
                          $MyUserID, $Path, $Mode, $UserID, $GroupID));
    # chmod
    _chmod ($Path, $MyUserID, $Mode, $DoWarn, @EntryStat);

    # chown
    _chown ($Path, $MyUserID, $UserID, $DoWarn,@EntryStat);

    # chgrp
    _chgrp ($Path, $MyUserID, $DoWarn, $GroupID, @EntryStat);

    return unless defined $aTime && defined $mTime;

    # utime
    _utime ($Path, $MyUserID, $Mode, $mTime, $UserID, $aTime, $DoWarn, $GroupID, @EntryStat);

    return;
}

#========================================================================================
# function     SetAccessRightsRecursive
#----------------------------------------------------------------------------------------
# syntax       SetAccessRightsRecursive($FileName, $hFlags);
#----------------------------------------------------------------------------------------
# description  Changes access rights (chmod/chgrp/chown) for unix recursively.
#               Does nothing if:
#               - function is executed under Windows
#               - $FileName does not exist
#               - no $hFlags (Mode User Group) is defined
#----------------------------------------------------------------------------------------
# example      SetAccessRightsRecursive( "$ENV{EPAGES_WEBROOT}/StoreTypes",
#                                         { Mode => GROUP_R, User => $ENV{EPAGES_APPUSER},
#                                           Group => 'apache' } );
# example      SetAccessRightsRecursive( '~root', { User => 0 } );
#----------------------------------------------------------------------------------------
# input        $FileName | file to be modified | string
# input        $hFlags | flags
#                   <ul>
#                       <li>Mode      - octal number as defined in DE_EPAGES::Core::API::Constants
#                                       Mode is file mode; directory mode == Mode | ALL_X
#                       <li>User      - user ID or name  - string
#                       <li>Group     - group ID or name - string
#                   </ul>
#               | ref.hash
#========================================================================================
sub SetAccessRightsRecursive ($;$) {
    my ($FileName, $hFlags) = @_;

    return if $Config{osname} eq 'MSWin32';
    return unless -e $FileName;

    if (defined $hFlags->{User} and defined $hFlags->{Group}) {
        _UnixCommand("chown -RH $hFlags->{User}:$hFlags->{Group} \"$FileName\"", $hFlags);
    } elsif (defined $hFlags->{User} and not defined $hFlags->{Group}) {
        _UnixCommand("chown -RH $hFlags->{User} \"$FileName\"", $hFlags);
    } elsif (defined $hFlags->{Group} and not defined $hFlags->{User}) {
        _UnixCommand("chgrp -RH $hFlags->{Group} \"$FileName\"", $hFlags);
    }

    return unless defined $hFlags->{Mode};

    my $FileMode = sprintf('%o', $hFlags->{Mode});
    if (not -d $FileName) {
        _UnixCommand("chmod $FileMode \"$FileName\"", $hFlags);
        return;
    }

    RunOnDirectory($FileName, sub {
        my $DirMode = sprintf('%o', _define_mode($FileName, {}, $hFlags));
        _UnixCommand("chmod -R $DirMode .", $hFlags);
        _UnixCommand("find . -follow -type f -print0 | xargs -0 chmod $FileMode",
                     $hFlags);
    });

    return;
}

#========================================================================================
# function     GetAccessRights
#----------------------------------------------------------------------------------------
# syntax       $hAccess = GetAccessRights($FileName);
#----------------------------------------------------------------------------------------
# description  Returns ref.hash with mode, uid and gid for FileName (only unix).
#----------------------------------------------------------------------------------------
# input        $FileName | file to be modified | string
# return       $hAccess | flags
#                   <ul>
#                       <li>Mode      - constant as defd in DE_EPAGES::Core::API::Constants
#                       <li>UserID    - user ID  - integer
#                       <li>GroupID   - group ID - integer
#                       <li>aTime     - access time (seconds since 1.1.1970 UTC)
#                       <li>mTime     - modified time (seconds since 1.1.1970 UTC)
#                   </ul>
#               | ref.hash
#========================================================================================
sub GetAccessRights ($) {
    my ($FileName) = @_;
    my %Stat = ();

    return \%Stat if $Config{'osname'} eq 'MSWin32' or not -e $FileName;

    ($Stat{'Mode'}, $Stat{'UserID'}, $Stat{'GroupID'}, $Stat{'aTime'}, $Stat{'mTime'})
        = (stat $FileName)[2,4,5,8,9];
    $Stat{'Mode'} &= ALL_RWS;

    return \%Stat;
}

#========================================================================================
# function     GetExecBit
#----------------------------------------------------------------------------------------
# syntax       $hBits = GetExecBit($FileName);
#----------------------------------------------------------------------------------------
# description  Returns ref.hash with execbit (only unix).
#----------------------------------------------------------------------------------------
# input        $FileName | file to be modified | string
# return       $hBits | flags
#                   <ul>
#                       <li>ExecBit   - 0: not executable by owner
#                                     - 1: executable by owner
#                   </ul>
#               | ref.hash
#========================================================================================
sub GetExecBit ($) {
    my ($FileName) = @_;
    my %Stat = ();

    return \%Stat if $Config{'osname'} eq 'MSWin32' or not -e $FileName;

    my $Mode = (stat $FileName)[2] & ALL_RWX;

    $Stat{'ExecBit'} = 0;
    $Stat{'ExecBit'} = 1 if $Mode & ALL_X;

    return \%Stat;
}

#========================================================================================
# function     IsInPath
#----------------------------------------------------------------------------------------
# syntax       $IsInPath = IsInPath($DirectoryOrFile, $BaseDirectory);
#----------------------------------------------------------------------------------------
# description  Test if a file or directory is located at a base directory.
#----------------------------------------------------------------------------------------
# example      $IsInPath = IsInPath('C:\Programs\epages\..\..\Windows\some.conf',
#                                          'C:\Programs\epages'); // false
# example      $IsInPath = IsInPath('C:\Programs\epages\..\some.conf',
#                                          'C:\Programs'); // true
#----------------------------------------------------------------------------------------
# input        $PathToDirectoryOrFile | path to directory or file | string
# input        $AllowedPath | allowed path | string
# return       $IsInPath | PathToDirectoryOrFile is part of AllowedPath | boolean
#========================================================================================
sub IsInPath ($$) {
    my ($DirectoryOrFile, $BaseDirectory) = @_;
    TestString('DirectoryOrFile', $DirectoryOrFile);
    TestString('BaseDirectory',   $BaseDirectory);

    # remove back links
    $DirectoryOrFile = RealPath(AbsolutePath($DirectoryOrFile));
    $BaseDirectory   = RealPath(AbsolutePath($BaseDirectory));

    return 1 if $DirectoryOrFile eq $BaseDirectory;
    return $DirectoryOrFile =~ /^\Q$BaseDirectory\E[\/\\]/;
}

#========================================================================================
# function     GetMountPoint
#----------------------------------------------------------------------------------------
# syntax       $MountPoint = GetMountPoint($Path);
#----------------------------------------------------------------------------------------
# description  returns mount point for $Path (if found in /etc/mtab).
#----------------------------------------------------------------------------------------
# input        $Path | path | string
# return       $MountPoint | mount point of $Path | string
#========================================================================================
sub GetMountPoint ($) {
    my ($Path) = @_;

    return if $Config{osname} eq 'MSWin32';

    # mount information in /etc/mtab not found
    return unless -f '/etc/mtab';

    $Path = AbsolutePath($Path);
    my $MountPoint;

    # mount information is stored in /etc/mtab
    foreach (@{GetFileContentLines('/etc/mtab')}) {
        my (undef, $Mount) = split;
        # get the longest mount point
        $MountPoint = $Mount if index($Path, $Mount) == 0
            and length($Mount) > length($MountPoint);
    }
    return $MountPoint;
}

#========================================================================================
# function     SameDevice
#----------------------------------------------------------------------------------------
# syntax       $OnSameDevice = SameDevice($Path1, $Path2);
#----------------------------------------------------------------------------------------
# description  Tests if $Path1 and $Path2 exist on same device.
#               Returns 0 if $Path1 and/or $Path2 do not exist.
#               Returns 0 if $Path1 and/or $Path2 are UNC paths.
#               Returns 0 if $Path1 and/or $Path2 exist on different mount points.
#----------------------------------------------------------------------------------------
# input          $Path1 | path to compare | string
# input          $Path2 | path to compare | string
# return  $OnSameDevice | 1 - $Path1 and $Path2 exist on same device; 0 -otherwise | boolean
#========================================================================================
sub SameDevice ($$) {
    my ($Path1, $Path2) = @_;

    return 0 unless -e $Path1;
    return 0 unless -e $Path2;

    if ($Config{osname} ne 'MSWin32') {
        $Path1 .= '/' if -d $Path1 && $Path1 !~ m,/$,;
        $Path2 .= '/' if -d $Path2 && $Path2 !~ m,/$,;
        # $Path1 and $Path2 exist on different devices
        return 0 if (stat $Path1)[0] != (stat $Path2)[0];

        # find for both paths their mount point
        my $Mount1 = GetMountPoint($Path1);
        my $Mount2 = GetMountPoint($Path2);

        # mount point not found for $Path1 or $Path2
        return 0 unless $Mount1 and $Mount2;
        # $Path1 and $Path2 exist on same mount point
        return 1 if $Mount1 eq $Mount2;
        # $Path1 and $Path2 exist on different mount points
        return 0;
    }

    my $Device1 = lc(AbsolutePath($Path1));
    return 0 unless $Device1 =~ s/^([a-z]):.*/$1/;
    my $Device2 = lc(AbsolutePath($Path2));
    return 0 unless $Device2 =~ s/^([a-z]):.*/$1/;

    return $Device1 eq $Device2;
}

#========================================================================================
# function     GetFileInfo
#----------------------------------------------------------------------------------------
# syntax       $hFileInfo = GetFileInfo($FileName);
#----------------------------------------------------------------------------------------
# description  Returns file size, modification and access time of a file.
#               Returns an empty has if the file does not exist.
#----------------------------------------------------------------------------------------
# input        $FileName | file name to get informations from | string
# return       $hFileInfo | flags
#                   <ul>
#                       <li>Name            - filename with Path - String
#                       <li>FileName        - filename without Path - String
#                       <li>Extension       - extension of file - String
#                       <li>Path            - path of file - String
#                       <li>Size            - file size in byte - Integer
#                       <li>LastAccess      - DateTime - Object
#                       <li>LastChange      - DateTime - Object
#                       <li>Mimetype        - Mimetype - String
#                   </ul>
#               | ref.hash
#========================================================================================
sub GetFileInfo ($) {
    my ($FileName) = @_;
    my %FileInfo = ();

    my $time1 = time;
    if( my ($size, $atime, $mtime, $ctime) = (stat $FileName) [7..10] ) {

        $FileInfo{'Name'} = $FileName;
        $FileInfo{'Size'} = $size;

        my $TimeZone = GetLocalTimeZone();
        $FileInfo{'LastAccess'} = _EpochToDateTime( $atime, $TimeZone );
        $FileInfo{'LastChange'} = _EpochToDateTime( $mtime, $TimeZone );
        $FileInfo{'InodeChange'} = _EpochToDateTime( $ctime, $TimeZone );

        ($FileInfo{'Path'}, $FileInfo{'FileName'}, $FileInfo{'Extension'}) = fsplit($FileName);
        $FileInfo{'Path'} = '' unless defined $FileInfo{'Path'};
        $FileInfo{'FileName'} = '' unless defined $FileInfo{'FileName'};
        $FileInfo{'Extension'} = '' unless defined $FileInfo{'Extension'};
        $FileInfo{'FileName'} .= $FileInfo{'Extension'};
        $FileInfo{'Mimetype'} = GetMimeTypeByExtension($FileInfo{'Extension'}) if $FileInfo{'Extension'} ne '';
    }
    my $time2 = time;
    AddStatistic('GetFileInfo', $time2 - $time1);
    return \%FileInfo;
}

# converts an Epoch value to a DateTime object in the server time zone
sub _EpochToDateTime {
    my ($Epoch, $TimeZone) = @_;

    my $DateTime = DateTime->from_epoch( epoch => $Epoch );
    $DateTime->set_time_zone( $TimeZone );
    return $DateTime;
}

#========================================================================================
# function     CompareDirectories
#----------------------------------------------------------------------------------------
# syntax       $Compare = CompareDirectories($Path1, $Path2);
#----------------------------------------------------------------------------------------
# description  Compares contents (file names) of two directories.
#----------------------------------------------------------------------------------------
# input        $Path1    | relative or absolute directory name | string
# input        $Path2    | relative or absolute directory name | string
# return       $Compare  | 1 - $Path2 contains all file names of $Path1, 0 - $Path1
#                           contains file names not included in $Path2.
#========================================================================================
sub CompareDirectories ($$) {
    my ($Path1, $Path2) = @_;

    return 1 unless -d $Path1;
    return 0 unless -d $Path2;

    foreach my $File (sort @{GetFileNames( $Path1, {'recursive' => 1} )}) {
        $File =~ s/\Q$Path1\E/$Path2/;
        if (! -f $File) {
            GetLog->debug("file not found '$File' at CompareDirectories\n");
            return 0;
        }
    }
    return 1;
}

#========================================================================================
# function     CreateDirectory
#----------------------------------------------------------------------------------------
# syntax       $CountCreated = CreateDirectory($Path, $hFlags);
#----------------------------------------------------------------------------------------
# example      $CountCreated = CreateDirectory('C:\\Test', {'recursive'=>0,'ExecBit'=>1});
#----------------------------------------------------------------------------------------
# description  Creates the directory path. Does nothing if the directory already exists.
#----------------------------------------------------------------------------------------
# input        $Path    | relative or absolute directory name | string
# input        $hFlags  | hash (see SetAccessRights, optional), 'recursive' => bool:
#               if true (default), directory will be created recursively (mkpath) otherwise only
#               the top directory will be created (mkdir), if its parent path exists | hash
# return       $CountCreated | count of created directories, 0 directory did already exists
#                          before | boolean
#========================================================================================
sub CreateDirectory ($;$) {
    my ($Path, $hFlags) = @_;
    $hFlags->{'recursive'} = 1 unless defined $hFlags->{'recursive'};

    return 0 if -d $Path;

    umask(BASE_MASK);
    my @DirsCreated;
    if ($hFlags->{'recursive'}) {
        # creates the full path
        my $aErrors;
        @DirsCreated = mkpath( $Path, {
            mode => BASE_RWX,
            error => \$aErrors,
        });
        if( scalar @$aErrors ) {
            my $Message = join '; ', map { sprintf '%s: %s', each %$_ } @$aErrors;
            Error('DIR_CREATE_FAILED', {'Directory'=> $Path, 'Code' => $Message});
        }
    } else {
        if (! mkdir ($Path, BASE_RWX)) {
            Error('DIR_CREATE_FAILED', {'Directory'=> $Path, 'Code' => $!});
        }
        @DirsCreated = ($Path);
    }

    SetAccessRights( $_, $hFlags ) foreach @DirsCreated;
    return scalar @DirsCreated;
}

#========================================================================================
# function     _MoveToTrash
# state        private
#----------------------------------------------------------------------------------------
# syntax       $TrashPath = _MoveToTrash($Path, $TrashFor);
#----------------------------------------------------------------------------------------
# example      my $TrashPath = _MoveToTrash($Path, $ENV{EPAGES_STATIC});
#----------------------------------------------------------------------------------------
# description  Moves the directory to the _TRASH_ directory of the given directory, so
#               an scheduler can remove them offline.
#----------------------------------------------------------------------------------------
# input        $Path | directory | string
# input        $TrashFor |  directory | string
# return       $TrashPath | directory | string
#========================================================================================
sub _MoveToTrash ($$) {
    my ($Path, $TrashFor) = @_;

    my $Trash = $TrashFor.'/_TRASH_';
    CreateDirectory($Trash);
    my $GUID = GenerateGUID();
    RenameDirectory($Path, "$Trash/$GUID");
    return "$Trash/$GUID";
}

#========================================================================================
# function     GetTrashRoot
#----------------------------------------------------------------------------------------
# syntax       $TrashRoot = GetTrashRoot($Path);
#----------------------------------------------------------------------------------------
# example      $TrashRoot = GetTrashRoot($Path);
#----------------------------------------------------------------------------------------
# description  Get the EPAGES_STATIC EPAGES_STORES EPAGES_WEBROOT Environment
#               for creating a trash root for this path
#----------------------------------------------------------------------------------------
# input        $Path | directory | string
# return       $TrashRoot | directory | string
#========================================================================================
sub GetTrashRoot($) {
    my ($Path) = @_;

    # remove trailing slashes
    foreach (qw(EPAGES_STATIC EPAGES_STORES EPAGES_WEBROOT EPAGES_WEBSECURE)) {
        $ENV{$_} =~ s,([^/])//+,$1/,g;
        $ENV{$_} =~ s,([^/])/+$,$1,g;
    }

    return $ENV{EPAGES_STATIC} if _UnixPathStartsWith($Path, $ENV{EPAGES_STATIC});
    return $ENV{EPAGES_STORES} if _UnixPathStartsWith($Path, $ENV{EPAGES_STORES});
    return $ENV{EPAGES_WEBROOT} if _UnixPathStartsWith($Path, $ENV{EPAGES_WEBROOT});
    return $ENV{EPAGES_WEBSECURE} if _UnixPathStartsWith($Path, $ENV{EPAGES_WEBSECURE});
    Error('NO_TRASH_DIR', { 'Path' => $Path });
}

sub DeleteDirectory ($;$$); ##no critic qw(RequireCommentForExportedFunctions)
#========================================================================================
# function     DeleteDirectory
#----------------------------------------------------------------------------------------
# syntax       $Deleted = DeleteDirectory($Path, $hFlags);
#----------------------------------------------------------------------------------------
# example      $Deleted = DeleteDirectory('C:\\Test',{'recursive' => 1});
#----------------------------------------------------------------------------------------
# description  Removes the directory. Throws an error if the directory cannot be
#               deleted, for example because it contains files or sub directories and the
#               recursive flag is not set. If the directory did not exist before,
#               then the function does nothing.
#----------------------------------------------------------------------------------------
# input        $Path | directory | string
# input        $hFlags | flags to modify the behaviour of DeleteDirectory.
#                         The following flag can be used:
#                       <ul>
#                           <li>recursive - delete directory recursive with
#                               all contents like rm -rf (default = 0 ) </li>
#                           <li>move - instead of deleting the directory files, move it
#                               to the _TRASH_ folder in EPAGES_STATIC, EPAGES_STORES
#                               or EPAGES_WEBROOT (default=0, implies recursive)</li>
#                       </ul>
#                       | hash ref
# input        $TrashRoot | trash directory | string (optional)
# return       $Deleted | 1 directory was removed , 0 directory doesnt exists | boolean
#========================================================================================
sub DeleteDirectory ($;$$) {
    my ($Path,$hFlags,$TrashRoot) = @_;
    return 0 unless -e $Path;

    # workaround for EPG-4777: delete directory containing files with special characters
    $Path = encode('iso-8859-1', $Path);

    eval {
        if( $hFlags->{'move'} ) {
            $TrashRoot = defined $TrashRoot ? $TrashRoot : GetTrashRoot($Path) ;

            # attempt to move the dir to trash...
            my $TrashPath;
            eval {
                $TrashPath = _MoveToTrash( $Path, $TrashRoot );
            };
            # if we get an error we delete the dir without moving to trash
            if (ExistsError()) {
                my $Error = GetError();
                warn $Error->message if ExistsError();
                if ($Error->code eq 'DIR_RENAME_FAILED') {
                    $hFlags->{'move'} = 0;
                    $hFlags->{'recursive'} = 1;
                    DeleteDirectory($Path,$hFlags);
                }
            }
            elsif ($TrashPath and $Config{'osname'} ne 'MSWin32' ) {
                # no return value to evaluate
                # ignore errors because cronjob (or someone else)
                # may delete $TrashPath meanwhile
                system( "rm -rf $TrashPath >/dev/null 2>&1 &" );
            }
        }
        elsif( -f $Path or -l $Path ) {
            DeleteFile($Path);
        }
        elsif( $hFlags->{'recursive'} ) {
            if ($Config{'osname'} eq 'MSWin32' && $Path =~ /^(\\\\|\/\/)/x) {
                # workaround for UNC paths, see EPG-9636, http://rt.cpan.org/Ticket/Display.html?id=34701
                my $Command = 'rmdir /S /Q "' . WindowsPath($Path) . '"';
                system( $Command );
                Error('SYSTEM_CMD_FAILED', { 'Command' => $Command, 'Code' => $? }) if $?;
            }
            else {
                rmtree $Path or Error('DIR_REMOVE_FAILED', { 'Directory'=> $Path, 'Code'=> $! });
            }
        }
        else {
            rmdir $Path or Error('DIR_REMOVE_FAILED', { 'Directory'=> $Path, 'Code'=> $! });
        }
    };
    if (ExistsError()) {
        # anybody else removes directory meanwhile ?
        if (-d $Path) {
            Error(GetError());
        }
    }
    return 1;
}

#========================================================================================
# function     ExistsDirectory
#----------------------------------------------------------------------------------------
# syntax       $Exists = ExistsDirectory($Path);
#----------------------------------------------------------------------------------------
# example      if( ExistsDirectory('C:\\Test') ) { ... }
#----------------------------------------------------------------------------------------
# description  Returns true if the directory exists.
#----------------------------------------------------------------------------------------
# input        $Path | directory | string
# return       $Exists | true if the directory exists, otherwise false | boolean
#========================================================================================
sub ExistsDirectory ($) {
    my ($Path) = @_;
    return 0 unless -d $Path;
    return 1;
}

#========================================================================================
# function     RenameDirectory
#----------------------------------------------------------------------------------------
# syntax       $Renamed = RenameDirectory($Source, $Target);
# syntax       $Renamed = RenameDirectory($Source, $Target, $hFlags);
#----------------------------------------------------------------------------------------
# example      RenameDirectory('C:\\Test\OldDirName','C:\\Test\NewDirName', { create_directory => 1 });
#----------------------------------------------------------------------------------------
# description  Returns true if the directory was renamed.
#               Returns false if the source directory does not exist.
#               Throws an error if the source directory exists, but could not be renamed.
#----------------------------------------------------------------------------------------
# input        $Source | old directory name | string
# input        $Target | new directory name | string
# input        $Flags     | flags to modify the behaviour of RenameDirectory.
#                            The following flags can be used:
#                          <ul>
#                              <li>create_directory - create directory or not?
#                              <li>stat - defines how to handle UNIX file attributes:
#                                  - epages: set file attributes according to epages rules
#                                    (this is the default behaviour unless flag is defined)
#                                  - preserve: file attributes remain the same
#                                  - ignore: do not set file attributes (makes function faster)
#                               </li>
#                          </ul>
#                          | hash ref
# return       $Renamed | true if the directory was renamed, otherwise false | boolean
#========================================================================================
sub RenameDirectory ($$;$) {
    my ($Source, $Target, $hFlags) = @_;
    TestString('Source', $Source);
    TestString('Target', $Target);

    return 0 unless -d $Source;

    $hFlags->{stat} = 'epages' if !defined $hFlags->{stat};

    if ($hFlags->{'create_directory'} and $hFlags->{'stat'} eq 'preserve') {
        CreateDirectory( dirname($Target), GetAccessRights(dirname($Source)) );
    } elsif ($hFlags->{'create_directory'} and $hFlags->{'stat'} eq 'ignore') {
        CreateDirectory( dirname($Target), {'stat' => 'ignore'} );
    } elsif ($hFlags->{'create_directory'}) {
        CreateDirectory( dirname($Target) );
    }

    if ($Config{'osname'} eq 'MSWin32' &&
        ($Source =~ /^(\\\\|\/\/)/x || $Target =~ /^(\\\\|\/\/)/x)) {
        # workaround for UNC paths
        CopyDirectory($Source, $Target);
        DeleteDirectory($Source, {'recursive' => 1});
    } else {
        my $hAccessRights = $hFlags->{'stat'} eq 'preserve' ? GetAccessRights($Source) : {};
        if (not move($Source, $Target)) {
            Warning('DIR_RENAME_FAILED', {
                'Code'=> _GetOSErrorMessage(),
                'Source'=> $Source,
                'Target'=> $Target });
            CopyDirectory($Source, $Target);
            DeleteDirectory($Source, {'recursive' => 1});
        }
        if ($hFlags->{'stat'} eq 'preserve') {
            SetAccessRights($Target, $hAccessRights);
        } elsif ($hFlags->{'stat'} ne 'ignore') {
            SetAccessRights($Target);
        }
    }

    return 1;
}

#========================================================================================
# function     DeleteStaticAtDirectory (obsolete)
#----------------------------------------------------------------------------------------
# syntax       DeleteStaticAtDirectory($Directory);
#----------------------------------------------------------------------------------------
# description  Removes _STATIC_ directories in given directory.
#----------------------------------------------------------------------------------------
# input        $Directory | directory name | string
#========================================================================================
sub DeleteStaticAtDirectory ($) {
    my ($Directory) = @_;
    warn 'DeleteStaticAtDirectory is obsolete.';

    if (ExistsDirectory($Directory)) {
        find(sub {
                return unless -d $File::Find::name;
                return unless $File::Find::name =~ /_STATIC_$/;
                DeleteDirectory($File::Find::name, {'recursive' => 1});
            },
            $Directory
        );
    }
    return;
}

#========================================================================================
# function     GetDirectorySize
#----------------------------------------------------------------------------------------
# syntax       $Size = GetDirectorySize($Directory);
#----------------------------------------------------------------------------------------
# example      my $Size = GetDirectorySize('/var/log');
#----------------------------------------------------------------------------------------
# description  Returns the size of a directory in bytes.
#               ErrorCodes:<ul>
#                   <li>FILE_NOT_FOUND
#               </ul>
#----------------------------------------------------------------------------------------
# input        $Directory | name of directory | string
# return       $Size | total size of directory, in bytes | integer
#========================================================================================
sub GetDirectorySize ($) {
    my ($Directory) = @_;
    Error('FILE_NOT_FOUND', { FileName=>$Directory } ) unless ExistsDirectory($Directory);

    my $Size = _UnixCommand("du -bs '$Directory'", { Capture => 1});
    $Size =~ s,\s.*,,s;
    return $Size;
}

#========================================================================================
# function     CopyFile
# state        public
#----------------------------------------------------------------------------------------
# syntax       CopyFile( $Source, $Target, $hFlags );
#----------------------------------------------------------------------------------------
# description  Copies a file in binary mode and sets the access rights.
#----------------------------------------------------------------------------------------
# input        $Source | source file name | string
# input        $Target | target file name | string
# input        $Flags  | flags to modify the behaviour of CopyFile.
#                         The following flags can be used:
#                       <ul>
#                           <li>create_directory - CreateDirectory or not</li>
#                           <li>link - If set, create hard link instead of copying.
#                               If set to 'orCopy', copies file if devices differ.
#                               If set to 'noCheck', don't check devices.
#                               If stat is 'epages', set file attributes accordingly;
#                               otherwise do not set file attributes</li>
#                           <li>unlink - If set, remove target before copying</li>
#                           <li>stat - defines how to handle UNIX file attributes:
#                               - epages: set file attributes according to epages rules
#                                 (this is the default behaviour unless flag is defined)
#                               - preserve: file attributes remain the same
#                               - ignore: do not set file attributes (makes function faster)
#                           </li>
#                           <li>md5info - hash ref, keys:
#                               - md5sum: md5sums to compare (scalar or array ref)
#                               - type: text (default) or binary. If text then md5sums are
#                                 compared ignoring EOLs for text files
#                               md5sum defines how to handle a file to be copied:
#                               - skips file if identical in $Source and $Target
#                               - creates new file if $Target does not exist
#                               - overrides $Target if the MD5 sum matches one in the list
#                               - creates $Target with extension .latest if MD5 sum does not match
#                           </li>
#                       </ul>
#                       | hash ref
#========================================================================================
sub CopyFile ($$;$) {
    my ( $FromFile, $ToFile, $hFlags ) = @_;

    $hFlags->{stat} = 'epages' if !defined $hFlags->{stat};
    my $ForceAccessRights = $hFlags->{stat} eq 'epages' ? 1 : 0;

    if ($hFlags->{'create_directory'} and $hFlags->{'stat'} eq 'preserve') {
        CreateDirectory( dirname($ToFile), GetAccessRights(dirname($FromFile)) );
    } elsif ($hFlags->{'create_directory'} and $hFlags->{'stat'} eq 'ignore') {
        CreateDirectory( dirname($ToFile), {'stat' => 'ignore'} );
    } elsif ($hFlags->{'create_directory'}) {
        CreateDirectory( dirname($ToFile) );
    }

    if (defined $hFlags->{md5info} and -f $ToFile) {
        if (compare($FromFile, $ToFile) == 0) {
            GetLog->debug("do not copy($FromFile, $ToFile): they are equal");
            return;
        }
        $ToFile .= '.latest' if IsCustomizedFile($ToFile, $hFlags->{md5info});
    }

    umask(BASE_MASK);

    DeleteFile($ToFile) if $hFlags->{'unlink'} and -f $ToFile;

    if ($hFlags->{'link'} and $hFlags->{'link'} ne 'noCheck') {
        if ( not SameDevice($FromFile, dirname($ToFile)) ) {
            if ($hFlags->{'link'} eq 'orCopy') {
                delete $hFlags->{'link'};
            } else {
                Error('CROSS_DEVICE_FAILURE', {
                    'Code'=> 'Invalid cross-device link',
                    'Source'=> $FromFile,
                    'Target'=> $ToFile
                });
            }
        }
    }

    if ($hFlags->{'link'}) {
        GetLog->debug("link($FromFile, $ToFile)");

        link( $FromFile, $ToFile ) or Error('FILE_LINK_FAILED', {
            'Code'=> _GetOSErrorMessage(),
            'Source'=> $FromFile,
            'Target'=> $ToFile
        });
        SetAccessRights($ToFile, GetExecBit($FromFile)) if $ForceAccessRights;
    } else {
        GetLog->debug("copy($FromFile, $ToFile)");

        copy( $FromFile, $ToFile ) or Error('FILE_COPY_FAILED', {
            'Code'=> _GetOSErrorMessage(),
            'Source'=> $FromFile,
            'Target'=> $ToFile
        });

        if ($hFlags->{'stat'} eq 'preserve') {
            SetAccessRights( $ToFile, GetAccessRights($FromFile) );
        } elsif ($hFlags->{'stat'} ne 'ignore') {
            SetAccessRights( $ToFile );
        }
    }
    return;
}

#========================================================================================
# function     IsCustomizedFile
# state        public
#----------------------------------------------------------------------------------------
# syntax       IsCustomizedFile($File, $hMD5);
#----------------------------------------------------------------------------------------
# description  checks if md5sum in $hMD5 is same as md5sum of $File
#----------------------------------------------------------------------------------------
# input        $File | file name | string
# input        $hMD5 | hash with keys:
#                       <ul>
#                           <li>md5sum - md5sums to compare (scalar or array ref)</li>
#                           <li>type   - text (default) or binary. If text then md5sums
#                                        are compared ignoring EOLs for text files.</li>
#                       </ul>
#                      | hash ref
# return       $IsCustomized | returns 1 if customized, else 0 | boolean
#========================================================================================
sub IsCustomizedFile {
    my ($File, $hMD5) = @_;

    return 0 unless -f $File;

    my @KnownMd5Sums = ref $hMD5->{md5sum} eq 'ARRAY' ?
        @{$hMD5->{md5sum}} : ($hMD5->{md5sum});

    my $rContent = GetFileContent($File);
    my $Md5Sum = Digest::MD5::md5_hex($$rContent);

    return 0 if $Md5Sum ~~ @KnownMd5Sums;

    $hMD5->{type} = 'text' unless defined $hMD5->{type};
    if (-T $File and $hMD5->{type} ne 'binary') {
        # convert all windows newlines to unix ones
        $$rContent =~ s/\r\n/chr(10)/ges if -T $File;
        $Md5Sum = Digest::MD5::md5_hex($$rContent);
        return 0 if $Md5Sum ~~ @KnownMd5Sums;
    }
    return 1;
}

#========================================================================================
# function     MergeIniFiles
#----------------------------------------------------------------------------------------
# syntax       MergeIniFiles($Target, $Source, $EditSections, $hFlags);
#----------------------------------------------------------------------------------------
# example      MergeIniFiles("$ENV{EPAGES_CONFIG}/Scheduler.conf",
#                 "$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Database/Data/Scheduler/Scheduler.conf",
#                 [ 'LOOPS' ], { 'replace' => 'parameters' } );
#----------------------------------------------------------------------------------------
# description  Adds sections from $Source to $Target if not yet exist in $Target.
#               For all $EditSections, do not simply add/replace $Target section by $Source
#               section but add/replace only the $Target parameters that exist in $Source.
#               Removes duplicate comments in $Target.
#----------------------------------------------------------------------------------------
# input        $Target | name of file | string
# input        $Source | name of file | string
# input        $EditSections | (optional) list of section names | ref.array
# input        $hFlags | (optional) flags to modify the behaviour of MergeIniFiles.
#                       <ul>
#                           <li>replace
#                           if set to 'sections' - overwrites target for all but $EditSections
#                           if set to 'parameters' - overwrites target only for $EditSections
#                           if otherwise set - overwrites target for all
#                           if unset - existing sections/parameters in target are not changed</li>
#                           <li>delcomments
#                           if set to ';' - deletes comments starting with ';'
#                           if set to '#' - deletes comments starting with '#'
#                           if otherwise set - deletes all comments
#                           if unset - comments remain
#                           <li>delete
#                           if set delete sections on the fly
#                       </ul>
#                       | hash ref
#========================================================================================
sub MergeIniFiles ($$;$$) {
    my ($Target, $Source, $EditSections, $hFlags) = @_;

    UsePackage( 'DE_EPAGES::Core::API::IniConfig' );
    DE_EPAGES::Core::API::IniConfig::MergeIniFiles($Target, $Source, $EditSections, $hFlags);
    return;
}

#========================================================================================
# function     WriteFile
#----------------------------------------------------------------------------------------
# syntax       WriteFile($FileName, $rContent, $hFlags);
#----------------------------------------------------------------------------------------
# example      WriteFile('/var/log/error.log', \$Content, { binary=>1, create_directory=>1} );
#----------------------------------------------------------------------------------------
# description  Writes a file to disk. Several options are accepted for modifying the
#               write behaviour.
#               ErrorCodes:<ul>
#                   <li>FILE_OPEN_FAILED
#                   <li>FILE_CLOSE_FAILED
#               </ul>
#----------------------------------------------------------------------------------------
# input        $FileName | name of file | string
# input        $rContent | new content of file | ref.string
# input        $hFlags | flags to modify the behaviour of WriteFile. The following flags
#                       can be used:
#                       <ul>
#                           <li>create_directory - automatically create the directory if
#                               it doen't exist</li>
#                           <li>append - append to the end of the file instead of
#                               deleting the previous content</li>
#                           <li>binary - don't convert newlines</li>
#                           <li>flock - block access to the file during the write operation</li>
#                           <li>syswrite - uses 'syswrite' instead of 'print'</li>
#                           <li>utf8 - writes the stream using perl's internal encoding</li>
#                           <li>encoding - character encoding, default: 'iso-8859-1'</li>
#                           <li>usebom - write a byte order mark for unicode files if 'utf8' or
#                                       encoding =~ /^utf/i</li>
#                           <li>ignore_error - don't die if write or create directory fails</li>
#                       </ul>
#                       <ul>
#                           (additionally all flags as defd in SetAccessRights)
#                       </ul>
#                       | hash ref
#========================================================================================
sub WriteFile ($$;$) {
    my ($FileName, $rContent, $hFlags) = @_;

    if ($hFlags->{'ignore_error'}) {
        eval { _WriteFile($FileName, $rContent, $hFlags); };
        warn GetError()->message if ExistsError();
    }
    else {
        _WriteFile($FileName, $rContent, $hFlags);
    }
    return;
}

sub _WriteFile ($$;$) {
    my ($FileName, $rContent, $hFlags) = @_;

    my $ExecBit = $hFlags->{ExecBit};
    $hFlags->{ExecBit} = 1;

    if ($hFlags->{'create_directory'}) {
        my $DirName = dirname ($FileName);
        CreateDirectory( $DirName, $hFlags );
        if (not ExistsDirectory($DirName)) {
            GetLog('BUGTRACKER::2538::Direct')->debug("directory '$DirName' not exists after create");
            sleep(1);
            if (not ExistsDirectory($DirName)) {
                GetLog('BUGTRACKER::2538::Waited')->debug("directory '$DirName' not exists after create and wait");
                CreateDirectory( $DirName, $hFlags );
                sleep(1);
                if (not ExistsDirectory($DirName)) {
                    GetLog('BUGTRACKER::2538::Waited2')->debug("directory '$DirName' not exists after create");
                }
            }
        }
    }

    my $WriteBOM = $hFlags->{'usebom'}
                    && ($hFlags->{'utf8'} || $hFlags->{'encoding'} =~ /^utf/i)
                    && !$hFlags->{'append'};

    my $OpenMode = $hFlags->{'append'} ? '>>' : '>';
    $OpenMode .= ':' . _EncodingByFlags($hFlags) unless $hFlags->{'binary'};

    my $time1 = time;
    umask(BASE_MASK);
    open my $FILE_DESCR, $OpenMode, $FileName or Error('FILE_OPEN_FAILED',{'Mode'=>$OpenMode, 'FileName'=>$FileName, 'Code'=>$!});
    _LockFile($FILE_DESCR, 'w') if $hFlags->{'flock'};
    binmode ($FILE_DESCR)    if $hFlags->{'binary'};
    if (defined $hFlags->{'syswrite'}) {
        syswrite( $FILE_DESCR, BOM, 1 ) if $WriteBOM;
        syswrite( $FILE_DESCR, $$rContent, length( $$rContent ) );
    } else {
        print $FILE_DESCR BOM if $WriteBOM;
        print $FILE_DESCR $$rContent;
    }
    unless( close $FILE_DESCR ) {
        my $OSErrorCode = $!;
        rmtree $FileName unless $hFlags->{'append'};
        Error('FILE_CLOSE_FAILED',{'FileName'=>$FileName, 'Code'=>$OSErrorCode});
    }

    $hFlags->{ExecBit} = $ExecBit;
    SetAccessRights($FileName, $hFlags) if $OpenMode !~ /^>>/;

    my $time2 = time;
    AddStatistic('WriteFile', $time2 - $time1);
    return;
}

#========================================================================================
# function     OpenFile
# state        public
#----------------------------------------------------------------------------------------
# syntax       OpenFile( $Mode, $FileName, $Sub );
# example      OpenFile( '>>', 'LogFile.txt', sub {
#                   my( $FileHandle ) = @_;
#                   print $FileHandle $Message;
#               });
#----------------------------------------------------------------------------------------
# description  Opens a file, executes a code reference and closes the file after the
#               code reference is finished.
#               The handle to the open file is passed as first parameter to the code.
#----------------------------------------------------------------------------------------
# input        $Mode | file open mode, see "perldoc opn", for example: '<', '>', '>>' | string
# input        $FileName | file name | string
# input        $Sub | code to execute while the file is open | code ref
#========================================================================================
sub OpenFile($$$) {
    my ($Mode, $FileName, $Sub) = @_;

    my $KnownBefore = -f $FileName;
    open my $FileHandle, $Mode, $FileName or Error('FILE_OPEN_FAILED', {
        'Mode'=>$Mode,
        'FileName'=>$FileName,
        'Code'=>$!
    });
    eval { $Sub->( $FileHandle ); };
    HandleError( Finally => sub {
        close $FileHandle or Error ('FILE_CLOSE_FAILED', {'FileName'=> $FileName, 'Code'=>$!});
    });
    SetAccessRights($FileName) unless $KnownBefore;
}

#========================================================================================
# function     LockFile
# state        public
#----------------------------------------------------------------------------------------
# syntax       LockFile( $FileHandle, $Access, $cCode );
# example      OpenFile( '>>', 'test.txt', sub {
#                   my ($FileHandle) = @_;
#                   LockFile( $FileHandle, 'w', sub {
#                       # seek to the end of the file in case someone appended
#                       # while we were waiting...
#                       seek( $FileHandle, 0, 2 ) or die $!;
#                       print $FileHandle, "...\n";
#                   });
#               });
#----------------------------------------------------------------------------------------
# description  Locks a file and unlocks it at the end of the code block.
#               There are 2 locking modes awailable:
#               <ul>
#                 <li>w - write or exclusive lock. Once this lock has been obtained by one
#                   process, no other process can read from or write to the file. Use
#                   this mode when writing to a file to prevent other processes from
#                   reading while writing is in progress and prevent multiple processes
#                   from writing simultanously to the same file</li>
#                 <li>r - read or shared lock. Prevents other processes from writing to
#                   the file. However, multiple processes can obtain read locks and read
#                   concurrently from the same file.</li>
#               </ul>
#               Locking is implemented using flock() on Windows.
#               On Unix systems fcntl() is used, because flock() does not work over NFS.
#               Both methods are defined as advisory locks, meaning that processes that
#               do not aquire locks can still access the file regardless of any locks.
#----------------------------------------------------------------------------------------
# input        $FileHandle | open file handle | handle
# input        $Access | access mode ('w' or 'r') | string
# input        $cCode | code to execute while the file is locked | code ref
#========================================================================================
sub LockFile($$$) {
    my ($FileHandle, $Access, $cCode) = @_;
    Error('InvalidLockMode', { Access => $Access }) unless $Access =~ /^[rw]$/;

    _LockFile( $FileHandle, $Access );

    # execute the code
    &$cCode;

    _UnlockFile( $FileHandle );
    return;
}

sub _LockFile($$) {
    my ($FileHandle, $Access) = @_;

    if( $Config{osname} eq 'MSWin32' ) {
        flock( $FileHandle, $Access eq 'r' ? LOCK_SH : LOCK_EX ) or Error( 'FILE_LOCK_FAILED', {
            Method => 'flock',
            Code => $!,
        });
    }
    else {
        my $params = pack('ssx32', ($Access eq 'r' ? F_RDLCK : F_WRLCK), 0);
        fcntl( $FileHandle, F_SETLKW, $params ) or Error( 'FILE_LOCK_FAILED', {
            Method => 'fcntl',
            Code => $!,
        });
    }
    return;
}

sub _UnlockFile($) {
    my ($FileHandle) = @_;

    if( $Config{osname} eq 'MSWin32' ) {
        flock($FileHandle, LOCK_UN) or Error( 'FILE_UNLOCK_FAILED', {
            Method => 'flock',
            Code => $!,
        });
    }
    else {
        my $params = pack('ssx32', F_UNLCK, 0);
        fcntl($FileHandle, F_SETLKW, $params) or Error( 'FILE_UNLOCK_FAILED', {
            Method => 'fcntl',
            Code => $!,
        });
    }
    return;
}

#========================================================================================
# function     MoveFile
#----------------------------------------------------------------------------------------
# syntax       MoveFile($SourceFile, $TargetFile, $hFlags);
#----------------------------------------------------------------------------------------
# example      MoveFile('/var/log/error.log', '/var/logger/error.b1');
#----------------------------------------------------------------------------------------
# description  Moves a file on disk if file on same device rename was used. We don't
#               use perlfunc rename because it may not work across file system boundaries.
#----------------------------------------------------------------------------------------
# input        $SourceFile | name of source file | string
# input        $TargetFile | name of target file | string
# input        $Flags  | flags to modify the behaviour of MoveFile.
#                         The following flags can be used:
#                       <ul>
#                           <li>create_directory - CreateDirectory or not</li>
#                           <li>stat - defines how to handle UNIX file attributes:
#                               - epages: set file attributes according to epages rules
#                                 (this is the default behaviour unless flag is defined)
#                               - preserve: file attributes remain the same
#                               - ignore: do not set file attributes (makes function faster)
#                            </li>
#                       </ul>
#                       | hash ref
#========================================================================================
sub MoveFile ($$;$) {
    my ($Source, $Target, $hFlags) = @_;

    $hFlags->{stat} = 'epages' if !defined $hFlags->{stat};

    if ($hFlags->{'create_directory'} and $hFlags->{'stat'} eq 'preserve') {
        CreateDirectory( dirname($Target), GetAccessRights(dirname($Source)) );
    } elsif ($hFlags->{'create_directory'} and $hFlags->{'stat'} eq 'ignore') {
        CreateDirectory( dirname($Target), {'stat' => 'ignore'} );
    } elsif ($hFlags->{'create_directory'}) {
        CreateDirectory( dirname($Target) );
    }

    my $hAccessRights = $hFlags->{'stat'} eq 'preserve' ? GetAccessRights($Source) : {};
    umask(BASE_MASK);

    move($Source, $Target) or Error('FILE_MOVE_FAILED', {
        'Code'=> _GetOSErrorMessage(),
        'Source'=> $Source,
        'Target'=> $Target
    });

    if ($hFlags->{'stat'} eq 'preserve') {
        SetAccessRights( $Target, $hAccessRights );
    } elsif ($hFlags->{'stat'} ne 'ignore') {
        SetAccessRights( $Target );
    }
    return;
}

#========================================================================================
# function     DeleteFile
#----------------------------------------------------------------------------------------
# syntax       $Deleted = DeleteFile($FileName);
#----------------------------------------------------------------------------------------
# example      DeleteFile('/var/log/error.log');
#----------------------------------------------------------------------------------------
# description  Removes a file from disk. The function throws an error if the file
#               cannot be deleted. If the file did not exist before, the function does
#               nothing.
#               ErrorCodes:<ul>
#                   <li>FILE_REMOVE_FAILED
#               </ul>
#----------------------------------------------------------------------------------------
# input        $FileName | name of file | string
# return       $Deleted | true if the file existed and was deleted, false if the file
#                   did not exist. | boolean
#========================================================================================
sub DeleteFile ($) {
    my ($FileName) = @_;
    return 0 unless -f $FileName;
    unlink($FileName) or Error('FILE_REMOVE_FAILED', {'Code'=>$!, 'FileName'=>$FileName});
    return 1;
}

sub CleanPath ($$;$); ##no critic qw(equireCommentForExportedFunctions)
#========================================================================================
# function     CleanPath
#----------------------------------------------------------------------------------------
# syntax       CleanPath($ShopPath, $FileAlias);
#----------------------------------------------------------------------------------------
# example      CleanPath('/srv/epages/eproot/Shared/WebRoot/Store/Shops/DemoShop','54C0/BC5B/BBD1/C0DF/A091/AC18/3245/E842');
#----------------------------------------------------------------------------------------
# description  Deletes old directories, if there is something. This function recursly walk
#               through the folder starting with $ShopPath."/".$FileDebug. It walk until it
#               comes to $ShopPath.
#----------------------------------------------------------------------------------------
# input        $ShopPath | PublicPath of the Shop. | string
# input        $Path | The path of the directory which should be cleaned. | string
# input        $hFlags  | flags to modify the behaviour of CleanPath.
#                         The following flags can be used:
#                       <ul>
#                           <li>force - If set > 0 the function delete folders with data
#                                       in there. The force attribute is a counter to show
#                                       how deep you want to ignore files in folders.
#                                       E.g.:
#                                           force: 2
#                                           folder: /ABC/DEF/GHI
#                                           Delete folder AND files in directory /ABC/DEF/GHI
#                                               and /ABC/DEF
#                           </li>
#                       </ul>
#                       | hash ref
# return       $DeletSomeThing | true if a folder was deleted, false if not | boolean
#========================================================================================
sub CleanPath ($$;$) {
    my ($ShopPath, $Path, $hFlags) = @_;
    my $Force = (defined $hFlags->{force}) ? $hFlags->{force} : 0;

    my $FullPath = $ShopPath . '/' . $Path;

    # quit this routine if you are on ShopPath
    return if ($Path eq '');

    # if folder exists
    if (opendir(my $DirHandler, $FullPath)) {
        # if directory is empty or you forced a deletion
        unless ((grep { not /^\.\.?$/ } readdir $DirHandler) && ($Force <= 0)) {
            rmtree $FullPath or GetLog->error("Cant delete the directory '$FullPath'.");

            my @Directory = split m/[\/]/, $Path;
            pop @Directory;
            my $NewDir = join('/', @Directory);

            $Force = ($Force == 0) ? 0 : $Force - 1;

            CleanPath($ShopPath, $NewDir, {'force' => $Force});

            return 1;
        }
    }

    return;
}

#========================================================================================
# function     DeleteListedFiles
#----------------------------------------------------------------------------------------
# syntax       $Deleted = DeleteListedFiles($Path, $FileList);
#----------------------------------------------------------------------------------------
# example      $Deleted = DeleteListedFiles('C:\\Test','Test.inexistent');
#----------------------------------------------------------------------------------------
# description  Removes listed files from patch. If an entry ends in .inexistent assume
#               this as a file file to be handled by DeleteListedFiles. The function throws
#               an error if a file cannot be deleted. If the file did not exist before, the
#               function does nothing.
#----------------------------------------------------------------------------------------
# input        $Path | directory | string
# input        $FileList | list with file names relative to directory | string
# return       $Deleted | true if at least one file was deleted, false if not | boolean
#========================================================================================
sub DeleteListedFiles {
    my ($Path, $FileList) = @_;

    my $Result = 0;
    return $Result if not -f $FileList or not -d $Path;

    OpenFile( '<', $FileList, sub {
        my ($FILE_LIST) = @_;

        while (<$FILE_LIST>) {
            chomp;
            next unless /\S/;
            if (/(.+)\.inexistent$/) {
                my $RelPath = $1;
                $FileList =~ /(.+)\.inexistent$/;
                $Result = DeleteListedFiles("$Path/$RelPath", "$1/$_");
            } elsif ( -d "$Path/$_" && $Config{'osname'} eq 'MSWin32' && $Path =~ /^(\\\\|\/\/)/x ) {
                $Result = 1;
                # workaround for UNC paths, see EPG-9636, http://rt.cpan.org/Ticket/Display.html?id=34701
                my $Command = 'rmdir /S /Q "' . WindowsPath("$Path/$_") . '"';
                system( $Command );
                Error('SYSTEM_CMD_FAILED', { 'Command' => $Command, 'Code' => $? }) if $?;
            } elsif ( -d "$Path/$_" ) {
                $Result = 1;
                rmtree "$Path/$_" or Error('DIR_REMOVE_FAILED', {'Code'=> $!, 'Directory'=> "$Path/$_"});
            } elsif ( -f "$Path/$_") {
                $Result = 1;
                unlink "$Path/$_" or Error('FILE_REMOVE_FAILED', {'Code'=>$!, 'FileName'=>"$Path/$_"});
            }
        }
    });

    unlink $FileList;
    return $Result;
}

#========================================================================================
# function     GetFileContent
#----------------------------------------------------------------------------------------
# syntax       $sContent = GetFileContent($FileName, $hFlags);
#----------------------------------------------------------------------------------------
# example      my $sContent = GetFileContent('/var/log/error.log');
#----------------------------------------------------------------------------------------
# description  Returns the content of a file as a scalar reference.
#               ErrorCodes:<ul>
#                   <li>FILE_NOT_FOUND
#                   <li>FILE_OPEN_FAILED
#                   <li>FILE_CLOSE_FAILED
#               </ul>
#----------------------------------------------------------------------------------------
# input        $FileName | name of file | string
# input        $hFlags | flags to modify the behaviour of GetFileContent. The following
#                       flags can be used:
#                       <ul>
#                           <li>binary - don't convert newlines</li>
#                           <li>flock - block access to the file during the read operation</li>
#                           <li>utf8 - reads the stream using perl's internal encoding,
#                               otherwise the content is marked as binary</li>
#                           <li>encoding - character encoding, default: 'iso-8859-1'</li>
#                           <li>usebom - use the BOM to detect Unicode files.
#                                   If the files has a byte order mark, then the flags
#                                   'utf8' and 'encoding' are ignored.</li>
#                       </ul>
#                       | hash ref
# return       $sContent | content of full file | ref.string
#========================================================================================
sub GetFileContent ($;$){
    my ( $FileName, $hFlags ) = @_;

    Error('FILE_NOT_FOUND', {'FileName'=>$FileName}) unless -f $FileName;

    my $time1 = time;

    my $HasBOM = 0;
    local $hFlags->{'encoding'} = $hFlags->{'encoding'};
    if( $hFlags->{'usebom'} ) {
        my $DetectedEncoding = DetectFileEncoding( $FileName );
        if (defined $DetectedEncoding) {
            $hFlags->{'encoding'} = $DetectedEncoding;
            $HasBOM = 1 if $DetectedEncoding =~ /^utf/i;
        }
    }

    my $OpenMode = '<';
    $OpenMode .= ':' . _EncodingByFlags($hFlags) unless $hFlags->{'binary'};

    my $Content;
    OpenFile( $OpenMode, $FileName, sub {
        my ($FILE_DESCR) = @_;

        _LockFile($FILE_DESCR, 'r') if defined $hFlags->{'flock'};
        binmode $FILE_DESCR if defined $hFlags->{'binary'};
        my $BOM;
        read $FILE_DESCR, $BOM, 1 if $HasBOM;
        $Content =  do { local $/ = undef; <$FILE_DESCR> };
    });

    my $time2 = time;
    AddStatistic('GetFileContent', $time2 - $time1);
    return \$Content;
}

#========================================================================================
# function     GetFileContentLines
#----------------------------------------------------------------------------------------
# syntax       $aContent = GetFileContentLines($FileName, $hFlags, $LineSep);
#----------------------------------------------------------------------------------------
# example      my $aContent = GetFileContentLines('/var/log/error.log');
#----------------------------------------------------------------------------------------
# description  Returns the content of a file as an array of lines.
#               ErrorCodes:<ul>
#                   <li>FILE_NOT_FOUND
#                   <li>FILE_OPEN_FAILED
#                   <li>FILE_CLOSE_FAILED
#               </ul>
#----------------------------------------------------------------------------------------
# input        $FileName | name of file | string
# input        $hFlags | flags to modify the behaviour of GetFileContent. The following
#                       flags can be used:
#                       <ul>
#                           <li>binary - don't convert newlines</li>
#                           <li>flock - block access to the file during the read operation</li>
#                           <li>utf8 - reads the stream using perl's internal encoding</li>
#                           <li>encoding - character encoding, default: 'iso-8859-1'</li>
#                           <li>usebom - use the BOM to detect Unicode files.
#                                   If the files has a byte order mark, then the flags
#                                   'utf8' and 'encoding' are ignored.</li>
#                       </ul>
# input        $LineSep | line seperator (default is "\n") | string
# return       $aContent | content of full file | ref.array.string
#========================================================================================
sub GetFileContentLines ($;$$) {
    my ( $FileName, $hFlags, $LineSep) = @_;
    TestString( 'FileName', $FileName );
    local $/ = defined $LineSep ? $LineSep : "\n"; #save global $/ and set default

    Error('FILE_NOT_FOUND', {'FileName'=>$FileName}) unless -f $FileName;

    my $HasBOM = 0;
    local $hFlags->{'encoding'} = $hFlags->{'encoding'};
    if( $hFlags->{'usebom'} ) {
        my $DetectedEncoding = DetectFileEncoding( $FileName );
        if (defined $DetectedEncoding) {
            $hFlags->{'encoding'} = $DetectedEncoding;
            $HasBOM = 1 if $DetectedEncoding =~ /^utf/i;
        }
    }

    my $OpenMode = '<';
    $OpenMode .= ':' . _EncodingByFlags($hFlags) unless $hFlags->{'binary'};

    my @Content;
    OpenFile( $OpenMode, $FileName, sub {
        my ($FILE_DESCR) = @_;

        _LockFile( $FILE_DESCR, 'r') if $hFlags->{'flock'};
        binmode $FILE_DESCR if $hFlags->{'binary'};
        my $BOM;
        read $FILE_DESCR, $BOM, 1 if $HasBOM;
        @Content = <$FILE_DESCR>;
    });

    return \@Content;
}

#========================================================================================
# function     GetFileSize
#----------------------------------------------------------------------------------------
# syntax       $Size = GetFileSize($FileName);
#----------------------------------------------------------------------------------------
# example      my $Size = GetFileSize('/var/log/error.log');
#----------------------------------------------------------------------------------------
# description  Returns the size of a file in bytes.
#               ErrorCodes:<ul>
#                   <li>FILE_NOT_FOUND
#                   <li>FILE_STATISTIC_FAILED
#               </ul>
#----------------------------------------------------------------------------------------
# input        $FileName | name of file | string
# return       $Size | total size of file, in bytes | integer
#========================================================================================
sub GetFileSize ($) {
    my ( $FileName ) = @_;
    TestString( 'FileName', $FileName );

    my $FileSize = -s $FileName;
    Error('FILE_NOT_FOUND', {'FileName'=>$FileName}) unless defined $FileSize;

    return $FileSize;
}

#========================================================================================
# function     GetFileSHA1Digest
#----------------------------------------------------------------------------------------
# syntax       $HexDigest = GetFileSHA1Digest($FileName);
#----------------------------------------------------------------------------------------
# example      my $HexDigest = GetFileSHA1Digest('/var/log/error.log');
#----------------------------------------------------------------------------------------
# description  Returns the sha1 hex digest of the content of a file
#----------------------------------------------------------------------------------------
# input        $FileName | name of file | string
# return       $HexDigest | sha1 hex digest ( see perldoc Digest::SHA1) | string
#========================================================================================
sub GetFileSHA1Digest {
    my ($FileName) = @_;

    my $HexDigest;
    OpenFile( '<', $FileName, sub {
        my ($FILE_DESCR) = @_;

        binmode $FILE_DESCR;
        $HexDigest = Digest::SHA1->new->addfile($FILE_DESCR)->hexdigest;
    });

    return $HexDigest;
}

#========================================================================================
# function     GetFileMD5Digest
#----------------------------------------------------------------------------------------
# syntax       $HexDigest = GetFileMD5Digest($FileName);
#----------------------------------------------------------------------------------------
# example      my $HexDigest = GetFileMD5Digest('/var/log/error.log');
#----------------------------------------------------------------------------------------
# description  Returns the md5 hex digest of the content of a file
#----------------------------------------------------------------------------------------
# input        $FileName | name of file | string
# return       $HexDigest | md5 hex digest ( see perldoc Digest::MD5) | string
#========================================================================================
sub GetFileMD5Digest {
    my ($FileName) = @_;

    my $HexDigest;
    OpenFile( '<', $FileName, sub {
        my ($FILE_DESCR) = @_;

        binmode $FILE_DESCR;
        $HexDigest = Digest::MD5->new->addfile($FILE_DESCR)->hexdigest;
    });

    return $HexDigest;
}

#========================================================================================
# function     CompareMD5TextFile
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsEqual = CompareMD5TextFile($File, $MD5Digest);
#----------------------------------------------------------------------------------------
# example      $IsEqual = CompareMD5TextFile('testtemp.txt','4c832c34a88de899dab3d00b7ccca632');
#----------------------------------------------------------------------------------------
# description  returns 1 if the MD5Hash of the file content,
#               or the MD5Hash of the dos2unix converted file content
#               or the MD5Hash of the unix2dos converted file content
#               is equal $MD5Digest
#----------------------------------------------------------------------------------------
# input        $File | filename | string
# input        $MD5Digest | md5 hex digest  | string
# return       $IsEqual | true if the file content matches the MD5Digest | boolean
#========================================================================================
sub CompareMD5TextFile {
    my ($File, $MD5Digest ) = @_ ;

    my $rContent = GetFileContent($File,{'binary'=>1});

    my $UnixContent = $$rContent;
    #dos2unix
    $UnixContent  =~ s/\r?\n/chr(10)/ges;

    my $WinContent = $$rContent;
    #unix2dos
    $WinContent  =~ s/\r?\n/chr(13).chr(10)/ges;

    if ( GetMD5Digest($rContent)  eq $MD5Digest or
         GetMD5Digest(\$UnixContent) eq $MD5Digest or
         GetMD5Digest(\$WinContent) eq $MD5Digest
       ) {
        return 1;
    }    else {
        return 0;
    }
}

#========================================================================================
# function     CompareMD5File
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsEqual = CompareMD5File($File, $MD5Digest);
#----------------------------------------------------------------------------------------
# example      $IsEqual = CompareMD5File('testtemp.txt','4c832c34a88de899dab3d00b7ccca632');
#----------------------------------------------------------------------------------------
# description  if it is a TextFile <function IsTextFile>
#               returns <function CompareMD5TextFile>($File, $MD5Digest);
#               else returns true if <function GetFileMD5Digest> eq $MD5Digest
#----------------------------------------------------------------------------------------
# input        $File | filename | string
# input        $MD5Digest | md5 hex digest  | string
# return       $IsEqual | true if the file content matches the MD5Digest | boolean
#========================================================================================
sub CompareMD5File {
    my ($FileName, $MD5Digest ) = @_ ;
    TestString('FileName', $FileName);
    TestString('MD5Digest', $MD5Digest);

    if (IsTextFile($FileName)) {
        return CompareMD5TextFile($FileName, $MD5Digest );
    }
    else {
        my $BinMD5Digest = GetFileMD5Digest($FileName);
        return $BinMD5Digest eq $MD5Digest ? 1: 0;
    }
}

#========================================================================================
# function     CompareFileWithMD5File
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsEqual = CompareFileWithMD5File($File, $MD5File);
#----------------------------------------------------------------------------------------
# example      $IsEqual = CompareFileWithMD5File('testtemp.txt','testtemp.txt.md5');
#----------------------------------------------------------------------------------------
# description  compares md5sum(testtemp.txt) with contents of testtemp.txt.md5
#----------------------------------------------------------------------------------------
# input        $File | filename | string
# input        $MD5File | filename | string
# return       $IsEqual | true if the file content matches the MD5Digest | boolean
#========================================================================================
sub CompareFileWithMD5File {
    my ($FileName, $MD5File ) = @_ ;
    TestString('FileName', $FileName);
    TestString('MD5File', $MD5File);

    return 0 unless ExistsFile($FileName);
    return 0 unless ExistsFile($MD5File);

    my $MD5Digest = GetFileContent($MD5File);
    chomp($$MD5Digest);

    return CompareMD5File($FileName, $$MD5Digest);
}

#========================================================================================
# function     CompareDirectoryWithFilelist
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsEqual = CompareDirectoryWithFilelist($Dir, $Filelist, $hFlags);
#----------------------------------------------------------------------------------------
# example      $IsEqual = CompareDirectoryWithFilelist('dir/','dir.txt');
#----------------------------------------------------------------------------------------
# description  Checks if all files in $Filelist are stored in dirname($Dir) with size as
#               defined in $Filelist. Each line of $Filelist has the format '<size> <path>'.
#----------------------------------------------------------------------------------------
# input        $Dir | dirname | string
# input        $Filelist | filename | string
# input        $hFlags | (optional) possible flags are:
#                         <ul>
#                           <li>verbose - show differences</li>
#                           <li>depth   - cut $Dir by depth, default: 0;
#                               use this flag to compare parts of $Dir, e.g.:
#                               if $Dir eq '/a/b/c' and $Filelist contains c/x/y
#                               then with depth == 0 file /a/b/c/c/x/y must exist
#                               and with depth == 1 file /a/b/c/x/y must exist.
#                           </li>
#                         </ul> | ref.hash.string
# return       $IsEqual | true if $Filelist matches $Dir | boolean
#========================================================================================
sub CompareDirectoryWithFilelist {
    my ($Dir, $Filelist, $hFlags) = @_ ;
    $hFlags->{depth} //= 0;
    TestInteger('hFlags->{depth}', $hFlags->{depth});
    TestString('Dir', $Dir);
    TestString('Filelist', $Filelist);

    if (not ExistsDirectory($Dir)) {
        print "directory missing: $Dir\n" if $hFlags->{verbose};
        return 0;
    }
    if (not ExistsFile($Filelist)) {
        print "filelist missing: $Filelist\n" if $hFlags->{verbose};
        return 0;
    }

    my $CheckDir = $Dir;
    $CheckDir = dirname($CheckDir) foreach 1 .. $hFlags->{depth};

    my $Result = 1;
    foreach my $ContentLine (@{GetFileContentLines($Filelist)}) {
        chomp($ContentLine);
        # each $ContentLine has format '<$ExpectedSize> <$CheckFile>'
        my ($ExpectedSize, $CheckFile);
        if ($ContentLine =~ /^(\S+)\s+(.+)/) {
            $ExpectedSize = $1;
            $CheckFile = $2;
        }
        next unless $CheckFile;

        my $FoundSize = -s "$CheckDir/$CheckFile";

        if (not defined $FoundSize) {
            # failure if file is missing
            $Result = 0;
            last unless $hFlags->{verbose};
            print "file missing: $CheckDir/$CheckFile\n";
        } elsif ($FoundSize != $ExpectedSize) {
            # failure if file doesn't has expected size
            $Result = 0;
            last unless $hFlags->{verbose};
            print "file corrupt: $CheckDir/$CheckFile\n";
        }
    }
    return $Result;
}


#========================================================================================
# function     ExistsFile
#----------------------------------------------------------------------------------------
# syntax       $Exists = ExistsFile($FileName);
#----------------------------------------------------------------------------------------
# example      if( ExistsDirectory('C:\\Test\\test.txt') ) { ... }
#----------------------------------------------------------------------------------------
# description  Returns true if the file exits.
#----------------------------------------------------------------------------------------
# input        $FileName | filename | string
# return       $Exists | true if the file exists, otherwise false | boolean
#========================================================================================
sub ExistsFile ($) {
    my ($FileName) = @_;
    TestString( 'FileName', $FileName );

    return 0 unless -f $FileName;
    return 1;
}

#========================================================================================
# function     ExistsFileCaseSensitive
#----------------------------------------------------------------------------------------
# syntax       $Exists = ExistsFileCaseSensitive($FileName);
#----------------------------------------------------------------------------------------
# example      if( ExistsFileCaseSensitive('C:\\Test\\test.txt') ) { ... }
#----------------------------------------------------------------------------------------
# description  Returns true if the file exits. Under windows filenames are not
#               case sensitive, this function is needed to guarantee equal behavior
#               for all operating systems. Now these file 'tEsT.XmL' is no longer
#               found under windows.
#----------------------------------------------------------------------------------------
# input        $FileName | filename | string
# return       $Exists | true if the file exists, otherwise false | boolean
#========================================================================================
sub ExistsFileCaseSensitive{
    my ($FilePath) = @_;
    my $result;
    my($Path, $FileName, $Extension) = fsplit($FilePath);
    opendir DIR, $Path or return 0;
        $result = scalar grep { $_ eq $FileName . $Extension } readdir DIR;
    closedir DIR;
    return $result;
}

my %CACHED_FILES = (); # includes for each cachedfile two values [valid until, NORMALIZER value], the composed key includes file name an function reference string
#========================================================================================
# function     GetCachedFile
#----------------------------------------------------------------------------------------
# syntax       $Content = GetCachedFile($FileName, $Sub);
#----------------------------------------------------------------------------------------
# example      my $IniConfig = GetCachedFile( $CONFIG_FILENAME, sub {
#                    my ($FileName) = @_;
#                    return DE_EPAGES::Core::API::IniConfig->new( FileName => $FileName );
#                });
#----------------------------------------------------------------------------------------
# description  Return the content of an file. If you call this function twice and the
#               file was not changed the file wasnt read again. The up-to-dateness of the
#               file will be tested each second.
#----------------------------------------------------------------------------------------
# input        $FileName | filename | string
# input        $Sub | reference to a function (default <function GetFileContent> | ref.code
# return       $Content | returned content depend on the $Sub function | *
#========================================================================================
sub GetCachedFile($;$) {
    my ($FileName, $Sub) = @_;
    $Sub //= \&GetFileContent;

    my $Caller = join(':', (caller())[1,2]);
    my $Key = $FileName.':'.$Caller;
    my $aFileNode = $CACHED_FILES{$Key};
    my $CurrentTime = time;
    my $LastChanged;
    if( defined $aFileNode ) {
        my ($LastAccess, $OldLastChanged, $Content) = @$aFileNode;
        # avoid multiple calls to stat() within the same second
        return $Content if $CurrentTime == $LastAccess;

        $LastChanged = (stat($FileName))[9] // 0;
        if( $LastChanged == $OldLastChanged ) {
            $aFileNode->[0] = $CurrentTime;
            return $Content;
        }
    }

    $LastChanged //= (stat($FileName))[9] // 0;
    my $Content = $Sub->( $FileName );
    $CACHED_FILES{$Key} = [ $CurrentTime, $LastChanged, $Content ];

    return $Content;
}

#========================================================================================
# function     IsFileNewer
#----------------------------------------------------------------------------------------
# syntax       IsFileNewer($Source, $Target);
#----------------------------------------------------------------------------------------
# description  Checks if the target file is newer than the source file. If the source
#               does not exist it throws the error 'FILE_NOT_FOUND'. If the target file
#               does not exist or is older than this function it returns 'false' (0).
#----------------------------------------------------------------------------------------
# input        $Source | source file | string
# input        $Target | target file | string
# return       $isNewer | 0 or 1 | boolean
#========================================================================================
sub IsFileNewer ($$) {
    my ($Source, $Target) = @_;

    my $time1 = time;

    # -M (modified is age)
    my $SourceModified = -M $Source;
    my $TargetModified = -M $Target;

    Error('FILE_NOT_FOUND', {'FileName'=>$Source}) unless defined $SourceModified;
    my $isNewer = (defined $TargetModified && $TargetModified < $SourceModified) ? 1 : 0;

    my $time2 = time;
    AddStatistic('IsFileNewer', $time2 - $time1);
    return $isNewer;
}

#========================================================================================
# function     Text2Unix
#----------------------------------------------------------------------------------------
# syntax       Text2Unix($FileName, $Force);
#----------------------------------------------------------------------------------------
# description  Checks if text and moves newlines to UNIX style newlines.
#----------------------------------------------------------------------------------------
# input        $FileName | file to be modified | string
# input        $Force    | file is changed even if is is a binary (opt.) | integer
#========================================================================================
sub Text2Unix ($;$) {

    my ($FileName, $Force) = @_;

    if (-T $FileName) {
        GetLog->debug("File '$FileName' is text.");
    } elsif ($Force) {
        GetLog->debug("File '$FileName' is binary.");
    } else {
        GetLog->warn("File '$FileName' is binary.");
        return;
    }

    # read file
    my $rContent = GetFileContent($FileName);

    # edit file
    $$rContent =~ s/\r?\n/chr(10)/ges;

    # write file
    my $hFlags = GetAccessRights($FileName);
    delete $hFlags->{aTime};
    delete $hFlags->{mTime};
    $hFlags->{syswrite} = 1;
    $hFlags->{binary} = 1;

    WriteFile($FileName, $rContent, $hFlags);
    return;
}

#========================================================================================
# function     IsTextFile
#----------------------------------------------------------------------------------------
# syntax       $IsTextFile = IsTextFile($FileName);
#----------------------------------------------------------------------------------------
# description  Checks if the file is a textfile using the file extensions see <package DE_EPAGES::Core::API::Constants>
#               constant TEXT_FILE_TYPES.
#----------------------------------------------------------------------------------------
# input        $FileName | file to be modified | string
# return       $IsTextFile | true if the file has a known text file extension | boolean
#========================================================================================
sub IsTextFile ($) {
    my ( $FileName ) = @_;
    TestString('FileName', $FileName);

    my (undef, $Name, $Extension) = fsplit($FileName);

    # file has no extension
    if (!defined $Extension){
        #check if FileName is an element of TEXT_FILE_NAMES, than it is an text-file
        return ElementOf( $Name, TEXT_FILE_NAMES) ? 1 : 0;
    }
    else{
        $Extension =~ s/^\.//;
        return ElementOf(lc $Extension, TEXT_FILE_TYPES) ? 1 : 0;
    }
}

#========================================================================================
# function     IsImageFile
#----------------------------------------------------------------------------------------
# syntax       $IsImageFile = IsImageFile($FileName);
#----------------------------------------------------------------------------------------
# description  Checks if the file is an image file using the file extensions see <package DE_EPAGES::Core::API::Constants>
#               constant IMAGE_FILE_TYPES.
#----------------------------------------------------------------------------------------
# input        $FileName | file to be modified | string
# return       $IsImageFile | true if the file has a known image file extension | boolean
#========================================================================================
sub IsImageFile ($) {
    my ( $FileName ) = @_;

    my (undef, undef, $Extension) = fsplit($FileName);
    # file has no extension
    return 0 unless $Extension;
    $Extension =~ s/^\.//;

    return ElementOf(lc $Extension, IMAGE_FILE_TYPES) ? 1 : 0;
}

#========================================================================================
# function     Text2Dos
#----------------------------------------------------------------------------------------
# syntax       Text2Dos($FileName, $Force);
#----------------------------------------------------------------------------------------
# description  Checks if text and moves newlines to UNIX style newlines.
#----------------------------------------------------------------------------------------
# input        $FileName | file to be modified | string
# input        $Force    | file is changed even if is is a binary (opt.) | integer
#========================================================================================
sub Text2Dos ($;$) {

    my ( $FileName, $Force ) = @_;

    if (-T $FileName) {
        GetLog->debug("File '$FileName' is text.");
    } elsif ($Force) {
        GetLog->debug("File '$FileName' is binary.");
    } else {
        GetLog->warn("File '$FileName' is binary.");
        return;
    }

    # read file
    my $rContent = GetFileContent($FileName );

    # edit file
    $$rContent =~ s/\r?\n/chr(13).chr(10)/ges;

    # write file
    my $hFlags = GetAccessRights($FileName);
    delete $hFlags->{aTime};
    delete $hFlags->{mTime};
    $hFlags->{syswrite} = 1;
    $hFlags->{binary} = 1;

    WriteFile($FileName, $rContent, $hFlags);
    return;
}

#========================================================================================
# function     GetSimilarFileName
#----------------------------------------------------------------------------------------
# syntax       $SimilarFileName = GetSimilarFileName($FileName);
#----------------------------------------------------------------------------------------
# example      $SimilarFileName = GetSimilarFileName('C:\\Photos\\photo.jpg');
#               # $SimilarFileName is 'C:\\Photos\\photo1.jpg', if this does not exist
#----------------------------------------------------------------------------------------
# description  Returns a file name for a base file name, that does not exist in this
#               directory. This is helpful, when a file has to be copied, but shall not
#               overwrite an existing file.
#----------------------------------------------------------------------------------------
# input        $FileName | absolute file path of base file | string
# return       $SimilarFileName | absolute, not existing file path | string
#========================================================================================
sub GetSimilarFileName($) {
    my ($FilePath) = @_;
    my ($Path, $Filename, $Extension) = fsplit($FilePath);
    my $Counter = 1;
    while (-e $FilePath) {
        # add a counter like 'filename1.jpg' if 'filename.jpg' already exists
        $FilePath = $Path . '/' . $Filename . $Counter++ . $Extension;
    }
    return $FilePath;
}

#========================================================================================
# function     DetectFileEncoding
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Encoding = DetectFileEncoding( $FileName );
#----------------------------------------------------------------------------------------
# description  Checks if the file has a BOM marker (Unicode Byte Order Mark) and retruns
#               the corresponding encoding, for example 'utf-8' or 'utf-7'.
#               Returns undef if the file does not have a BOM.
#----------------------------------------------------------------------------------------
# input        $FileName | file name | string
# return       $Encoding | character set name | string
#========================================================================================
sub DetectFileEncoding {
    my ($FileName) = @_;

    Error('FILE_NOT_FOUND', {'FileName'=>$FileName}) unless -f $FileName;

    my $Header;
    OpenFile( '<', $FileName, sub {
        my ($FILE_DESCR) = @_;

        binmode $FILE_DESCR;
        read $FILE_DESCR, $Header, 5;
    });

    use bytes;
    return 'utf-7' if $Header eq GetBOM( 'utf-7' );
    return 'utf-8' if substr($Header, 0, 3) eq GetBOM( 'utf-8' );
    return 'utf16le' if substr($Header, 0, 2) eq GetBOM( 'utf16le' );
    return 'utf16be' if substr($Header, 0, 2) eq GetBOM( 'utf16be' );
    return undef;
}

#========================================================================================
# function     GetBOM
# state        public
#----------------------------------------------------------------------------------------
# syntax       $BOM = GetBOM( $Encoding );
# example      $BOM = GetBOM( 'utf16be' );
#----------------------------------------------------------------------------------------
# description  Returns the Byte Order Mark for the given encoding.
#----------------------------------------------------------------------------------------
# input        $Encoding | 'utf-7' or 'utf-8' or 'utf16le' or 'utf16be' | string
# return       $BOM | Byte Order Mark | binary
#========================================================================================
sub GetBOM {
    my ($Encoding) = @_;
    return encode( $Encoding, BOM );
}

#========================================================================================
# function     GetSubDirectories
#----------------------------------------------------------------------------------------
# syntax       $aSubDirs = GetSubDirectories($Directory, $hFlags);
#----------------------------------------------------------------------------------------
# example      $aSubDirs = GetSubDirectories('C:\\', {'recursive' => 1});
#----------------------------------------------------------------------------------------
# description  Returns subdirectories of the given directory in an array reference.
#               Please note that the array doesn't contain the '.' and '..' directories.
#----------------------------------------------------------------------------------------
# input        $DirPath | directory, for which the subdirectories are wanted (default '.') | string
# input        $hFlags | (optional) see <function GetInfoDirectory> | ref.hash.string
# return       $aSubDirs | array of subdirectories names | ref.array.string
#========================================================================================
sub GetSubDirectories ($;$) {
    my ($Directory, $hFlags) = @_;
    my $ret = GetInfoDirectory($Directory, $hFlags);
    return $ret->{'Directories'};
}

#========================================================================================
# function     GetFileNames
#----------------------------------------------------------------------------------------
# syntax       $aFiles = GetFileNames( $DirPath, $hFlags );
#----------------------------------------------------------------------------------------
# example      $aFiles = GetFileNames( 'C:\\', {'recursive' => 1} );
#----------------------------------------------------------------------------------------
# description  Returns filenames which are included in the given directory
#               in an array reference.
#----------------------------------------------------------------------------------------
# input        $DirPath | directory, for which the filenames are wanted (default '.') | string
# input        $hFlags | (optional) see <function GetInfoDirectory> | ref.hash.string
# return       $aFiles | array of file names | ref.array.string
#========================================================================================
sub GetFileNames ($;$) {
    my ($DirPath, $hFlags) = @_;
    my $ret = GetInfoDirectory($DirPath, $hFlags);
    return $ret->{'Files'};
}

sub GetInfoDirectory ($;$); ##no critic qw(equireCommentForExportedFunctions)
#========================================================================================
# function     GetInfoDirectory
#----------------------------------------------------------------------------------------
# syntax       $hInfo = GetInfoDirectory( $DirPath , $hFlags );
#----------------------------------------------------------------------------------------
# example      $hInfo = GetInfoDirectory( 'C:\\', { recursive => 1, ignore => $raIgnore } );
#----------------------------------------------------------------------------------------
# description  Returns all items of a directory like files, subdirectories
#               or others (e.g. links).
#----------------------------------------------------------------------------------------
# input        $DirPath | directory, for which the filenames are wanted (default '.') | string
# input        $hFlags | (optional) possible flags are:
#                         <ul>
#                           <li>recursive - process sub directories</li>
#                           <li>ignore - Ignore files with path names (quoted) relative to $hFlags->{basedir}</li>
#                           <li>basedir - see above, default is $DirPath</li>
#                           <li>relative - dirs and files are relative to $DirPath</li>
#                         </ul> | ref.hash.string
# return       $hInfo | directory items keys are:
#               <ul>
#                   <li>Directories - contains directories - ref.array.string
#                   <li>Files - contains files - ref.array.string
#                   <li>Other - contains links - ref.array.string
#               </ul>
#               | ref.hash.array.string
#========================================================================================
sub GetInfoDirectory ($;$) {
    my ($DirPath,  $rhFlags) = @_;
    $DirPath = '.' unless defined $DirPath;
    $rhFlags->{basedir} = $DirPath unless defined $rhFlags->{basedir};

    opendir (DIR, $DirPath) or Error('DIR_OPEN_FAILED', {'Directory'=> $DirPath , 'Code'=> $!, 'CurrentDir' => getcwd()});
    my @SubItems = readdir DIR;
    closedir DIR or Error('DIR_CLOSE_FAILED', {'Directory'=> $DirPath , 'Code'=> $!});
    my (@Directories, @Files, @Others);
    SUBITEM: foreach my $SubItem (@SubItems) {
        next if $SubItem eq '.' or $SubItem eq '..';
        my $FullName = "$DirPath/$SubItem";
        if ( defined $rhFlags->{ignore} ) {
            my $AbsName = AbsolutePath($FullName);
            foreach my $Ignore (@{$rhFlags->{ignore}}) {
                $Ignore = AbsolutePath($Ignore, $rhFlags->{basedir});
                next SUBITEM if $Config{osname} eq 'MSWin32' and lc($AbsName) eq lc($Ignore);
                next SUBITEM if $Config{osname} ne 'MSWin32' and $AbsName eq $Ignore;
            }
        }

        if (-d $FullName) {
            push @Directories, $rhFlags->{'relative'} ? $SubItem : $FullName;
        } elsif (-f $FullName) {
            push @Files, $rhFlags->{'relative'} ? $SubItem : $FullName;
        } else {
            push @Others, $rhFlags->{'relative'} ? $SubItem : $FullName;
        }
    }

    if ($rhFlags->{'recursive'}) {
        my @SubDirs = @Directories;
        foreach my $Directory (@SubDirs) {
            if( $rhFlags->{'relative'} ) {
                my $subs = GetInfoDirectory("$DirPath/$Directory", $rhFlags);
                push @Directories, map { "$Directory/$_" } @{$subs->{'Directories'}};
                push @Files,       map { "$Directory/$_" } @{$subs->{'Files'}};
                push @Others,      map { "$Directory/$_" } @{$subs->{'Others'}};
            }
            else {
                my $subs = GetInfoDirectory($Directory, $rhFlags);
                push @Directories, @{$subs->{'Directories'}};
                push @Files,       @{$subs->{'Files'}};
                push @Others,      @{$subs->{'Others'}};
            }
        }
    }

    return {
        'Directories'   => \@Directories,
        'Files'         => \@Files,
        'Others'        => \@Others
    };
}

#========================================================================================
# function     MoveToDateDir
# state        public
#----------------------------------------------------------------------------------------
# syntax       MoveToDateDir($Dir, $FilePrefix, $Days, $Now, $Format, $Verbose);
#----------------------------------------------------------------------------------------
# example      MoveToDateDir($Dir, '[a-z]', '%Y/%m', 7, time());
#----------------------------------------------------------------------------------------
# description  Moves files in $Dir/$FilePrefix* older than $Days (starting from
#               $Now) to $Dir/$Format/$FilePrefix*.
#----------------------------------------------------------------------------------------
# input        $Dir | move files in $Dir
# input        $FilePrefix | move files starting with $FilePrefix
# input        $Days | move files older than $Days
# input        $Now | $Now is now (seconds since epoch) (default: time())
# input        $Format | move files to $Format (default: %Y/%m)
# input        $Verbose | send output to stdout  (default: 0)
# ========================================================================================
sub MoveToDateDir {
    my ($Dir, $FilePrefix, $Days, $Now, $Format, $Verbose) = @_;

    $Now = time() unless defined $Now;
    $Format = '%Y/%m' unless defined $Format;

    my @OldFiles = grep {$_ =~ /\b$FilePrefix/} @{GetFileNames($Dir,{'recursive' => 0})};

    foreach my $DumpFile ( @OldFiles ) {
        my $DumpFileDays = -M $DumpFile;
        if ( $DumpFileDays >=  $Days ) {
            my $dt = DateTime->from_epoch(epoch => $Now - $DumpFileDays * SECONDS_PER_DAY);
            my $FormatDir = $dt->strftime( $Format );
            my $FileName = basename($DumpFile);
            print "Moving $FileName to $Dir/$FormatDir/$FileName\n" if $Verbose;
            CreateDirectory("$Dir/$FormatDir");
            rename($DumpFile, "$Dir/$FormatDir/$FileName");
        }
    }
}

#========================================================================================
# function     BackupFile
#----------------------------------------------------------------------------------------
# syntax       BackupFile( $SourceFile, $TargetFile, $BackupFile );
#----------------------------------------------------------------------------------------
# description  If BackupFile doesn't exist and SourceFile and TargetFile are not equal
#               this function copies TargetFile to BackupFile.
#----------------------------------------------------------------------------------------
# input        $SourceFile | source file | string
# input        $TargetFile | target file | string
# input        $BackupFile | backup file | string
# input        $Flags      | flags to modify the behaviour of BackupFile.
#                             The following flags can be used:
#                           <ul>
#                               <li>stat - defines how to handle UNIX file attributes:
#                                   - epages: set file attributes according to epages rules
#                                     (this is the default behaviour unless flag is defined)
#                                   - preserve: file attributes remain the same
#                                   - ignore: do not set file attributes (makes function faster)
#                                </li>
#                           </ul>
#                           | hash ref
# return       $BackupDone | 0 or 1 | boolean
#========================================================================================
sub BackupFile ($$$;$) {
    my ($Source, $Target, $Backup, $Flags) = @_;
    return 0 if -f $Backup;
    return 0 unless -f $Target;
    return 0 if compare($Source, $Target) == 0;

    $Flags->{create_directory} = 1;
    CopyFile( $Target, $Backup, $Flags );
    return 1;
}

#========================================================================================
# function     BackupDirectory
#----------------------------------------------------------------------------------------
# syntax       BackupDirectory( $SourceDir, $TargetDir, $BackupDir, $Flags );
#----------------------------------------------------------------------------------------
# description  Copies files which exist in both, source directory and target directory,
#               but not in the backup directory from the target directory to the backup
#               directory.
#----------------------------------------------------------------------------------------
# input        $SourceDir | source directory | string
# input        $TargetDir | target directory | string
# input        $BackupDir | backup directory | string
# input        $Flags     | flags to modify the behaviour of BackupDirectory.
#                            The following flags can be used:
#                          <ul>
#                              <li>ignore - Ignore files with path names (quoted) relative to $SourceDir</li>
#                              <li>stat - defines how to handle UNIX file attributes:
#                                  - epages: set file attributes according to epages rules
#                                    (this is the default behaviour unless flag is defined)
#                                  - preserve: file attributes remain the same
#                                  - ignore: do not set file attributes (makes function faster)
#                               </li>
#                          </ul>
#                          | hash ref
#========================================================================================
sub BackupDirectory ($$$;$) {
    my ( $FromDir, $ToDir, $BackupDir, $Flags) = @_;

    GetLog->debug("BackupDirectory $FromDir, $ToDir, $BackupDir");

    # copy files
    my $DirInfo = GetInfoDirectory($FromDir, $Flags);

    my $Content;
    foreach my $Source ( @{$DirInfo->{'Directories'}} ) {
        my $RelPath = $Source;
        $RelPath =~ s,^\Q$FromDir\E[\\/]*,,;
        my $Target = "$ToDir/$RelPath";
        $Content .= "$RelPath\n" unless -d $Target;
    }

    foreach my $Source ( @{$DirInfo->{'Files'}} ) {
        my $RelPath = $Source;
        $RelPath =~ s,^\Q$FromDir\E[\\/]*,,;
        my $Target = "$ToDir/$RelPath";
        my $Backup = "$BackupDir/$RelPath";
        $Content .= "$RelPath\n" unless -f $Target;
        BackupFile($Source, $Target, $Backup, $Flags);
    }
    WriteFile( "$BackupDir.inexistent", \$Content, {append=>1, create_directory=>1, stat=>$Flags->{stat}} ) if $Content;

    return;
}

#========================================================================================
# function     RebuildFile
#----------------------------------------------------------------------------------------
# syntax       RebuildFile( $SourceFile, $TargetFile, $BackupFile, $Flags );
#----------------------------------------------------------------------------------------
# description  Deletes TargetFile if TargetFile and SourceFile exist and are equal.
#               Moves BackupFile to TargetFile if BackupFile exists.
#----------------------------------------------------------------------------------------
# input        $SourceFile | source file | string
# input        $TargetFile | target file | string
# input        $BackupFile | backup file | string
# input        $Flags  | flags to modify the behaviour of RebuildFile.
#                         The following flag can be used:
#                       <ul>
#                           <li>stat - defines how to handle UNIX file attributes:
#                               - epages: set file attributes according to epages rules
#                                 (this is the default behaviour unless flag is defined)
#                               - preserve: file attributes remain the same
#                               - ignore: do not set file attributes (makes function faster)
#                            </li>
#                       </ul>
#                       | hash ref
#========================================================================================
sub RebuildFile ($$$;$) {
    my ($Source, $Target, $Backup, $Flags) = @_;

    $Flags->{'create_directory'} = 1;

    if (compare($Source, $Target) == 0 and -f $Backup) {
        DeleteFile($Target);
        MoveFile( $Backup, $Target, $Flags );
    }
    return;
}

#========================================================================================
# function     RebuildDirectory
#----------------------------------------------------------------------------------------
# syntax       RebuildDirectory( $SourceDir, $TargetDir, $BackupDir, $Flags );
#----------------------------------------------------------------------------------------
# description  Copies files which exist in the source and backup directory and have the
#               same modification time in source and target from the backup
#               to the target directory. Deletes the backup file if it was copied.
#----------------------------------------------------------------------------------------
# input        $SourceDir | source directory | string
# input        $TargetDir | target directory | string
# input        $BackupDir | backup directory | string
# input        $Flags     | flags to modify the behaviour of RebuildDirectory.
#                            The following flags can be used:
#                          <ul>
#                              <li>ignore - Ignore files with path names (quoted) relative to $SourceDir</li>
#                              <li>recursive - rebuilds directory recursively</li>
#                              <li>stat - defines how to handle UNIX file attributes:
#                                  - epages: set file attributes according to epages rules
#                                    (this is the default behaviour unless flag is defined)
#                                  - preserve: file attributes remain the same
#                                  - ignore: do not set file attributes (makes function faster)
#                               </li>
#                          </ul>
#                          | hash ref
#========================================================================================
sub RebuildDirectory ($$$;$) {
    my ( $FromDir, $ToDir, $BackupDir, $Flags ) = @_;
    $Flags->{recursive} = 1 unless defined $Flags->{recursive};

    DeleteListedFiles($ToDir, "$BackupDir.inexistent");

    # copy files
    my $aFiles = GetFileNames( $FromDir, $Flags );
    foreach my $Source ( @$aFiles ) {
        my $Backup = $Source;
        $Backup =~ s/^\Q$FromDir\E/$BackupDir/e;
        my $Target = $Source;
        $Target =~ s/^\Q$FromDir\E/$ToDir/e;

        RebuildFile($Source, $Target, $Backup, $Flags);
    }
    return;
}

#========================================================================================
# function     PathGlob2TypeGlob
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Path = PathGlob2TypeGlob($Path)
#----------------------------------------------------------------------------------------
# description  Changes $Path from path glob (as a/*/z/??) to type glob (as a\/.+\/z\/..)
#----------------------------------------------------------------------------------------
# input        $Path | directory | string
# return       $Path | directory | string
#========================================================================================
sub PathGlob2TypeGlob {
    my ($Path) = @_;

    chomp($Path);
    $Path =~ s/\\/\//go if $Config{osname} eq 'MSWin32';
    $Path = "\Q$Path\E";
    $Path =~ s/\\\*/[^\/]+/go;
    $Path =~ s/\\\?/[^\/]/go;
    $Path = "(?i)$Path" if $Config{osname} eq 'MSWin32';
    return $Path;
}

#========================================================================================
# function     FilesOnlyInDir1
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aFileList =  FilesOnlyInDir1($Dir1, $Dir2);
#----------------------------------------------------------------------------------------
# description  Returns a list of files that exist in $Dir1, but not in $Dir2. This
#               works with symbolic links.
#----------------------------------------------------------------------------------------
# input        $Dir1 | directory | string
# input        $Dir2 | directory | string
#----------------------------------------------------------------------------------------
# return       $aFileList | list of files | array ref
#========================================================================================
sub FilesOnlyInDir1 ($$) {
    my ( $Dir1, $Dir2 ) = @_;

    if ($Config{'osname'} eq 'MSWin32') {
        $Dir1 = WindowsPath($Dir1);
        $Dir1 =~ s,(\s|[\\/])+$,,x;
        $Dir2 = WindowsPath($Dir2);
        $Dir2 =~ s,(\s|[\\/])+$,,x;
    }

    my @FileList = ();
    return \@FileList unless -e $Dir1;

    foreach my $File ((@{GetSubDirectories( $Dir1, {'recursive' => 1} )},
                      @{GetFileNames( $Dir1, {'recursive' => 1} )})) {
        $File =~ s,^\Q$Dir1\E,$Dir2,;
        if (! -e $File) {
            $File =~ s,^\Q$Dir2\E[\\/]*,,x;
            push @FileList, $File;
        }
    }
    return \@FileList;
}

#========================================================================================
# function     CopyDirectory
# state        public
#----------------------------------------------------------------------------------------
# syntax       CopyDirectory( $SourceDir, $TargetDir, $BackupDir, $Flags );
# example      CopyDirectory( 'c:\\Temp','d:\\Temp', 'e:\\backup', { recursive => 1 } );
#----------------------------------------------------------------------------------------
# description  Copies all files from one directory to another. This function works
#               recursively and creates the necessary sub directories.
#               If a file already exists in the target directory, then it will be saved
#               in the backup directory unless there is already a backup file.
#               In this case the target file is overwritten by the source file.
#               The source directory must exist. The target and backup directories will
#               be created if they don't exist.
#----------------------------------------------------------------------------------------
# input        $SourceDir | source directoy | string
# input        $TargetDir | target directory | string
# input        $BackupDir | backup directory (optional) | string
# input        $Flags     | flags to modify the behaviour of CopyDirectory.
#                            The following flags can be used:
#                          <ul>
#                              <li>recursive - copies directory recursively</li>
#                              <li>link - If set, create hard link instead of copying.
#                                  If set to 'orCopy', copies file if devices differ.
#                                  If set to 'noCheck', don't check devices.</li>
#                              <li>unlink - If set, remove target before copying</li>
#                              <li>ignore - ignore files with path names (quoted) relative to $SourceDir</li>
#                              <li>stat - defines how to handle UNIX file attributes for $TargetDir:
#                                  - epages: set file attributes according to epages rules
#                                  - preserve: file attributes remain the same
#                                  - ignore: do not set file attributes (makes function faster)
#                                  If stat is not set, default values are either 'epages' or 'preserve':
#                                  stat is set to 'epages' if: a) 'by' is perl
#                                    b) target is located inside epages, but source not
#                                    c) source and target are located in different epages parts
#                                  stat is set to 'preserve' if: a) target is located outside epages
#                                    b) source and target are located in the same epages part
#                              </li>
#                              <li>bakstat - defines how to handle UNIX file attributes for $BackupDir:
#                                  - epages: set file attributes according to epages rules
#                                    (this is the default behaviour unless flag is defined)
#                                  - preserve: file attributes remain the same
#                                  - ignore: do not set file attributes (makes function faster)
#                              </li>
#                              <li>md5sum - name of an ini file
#                                  - sections: file names relative to $SourceDir
#                                  - keys:
#                                  - - md5sum: md5sums to compare
#                                  - - type: text (default) or binary. If text then md5sums
#                                      are compared ignoring EOLs for text files
#                                  - example:
#                                    [License.xml]
#                                    type=text
#                                    md5sum=1f99100011f69ed8d81500facad25da3 # 6.11.0 mysql
#                                  - the ini file defines how to handle a file to be copied:
#                                  - - skips file if identical in $TargetDir and $SourceDir
#                                  - - creates new file if does not exist in $TargetDir
#                                  - - overrides file in $TargetDir if the MD5 sum matches one in the list
#                                  - - creates a file with extension .latest if MD5 sum does not match
#                              </li>
#                              <li>by - defines the tool that is used for copying:
#                                  - perl: use File::Copy
#                                  - exe: use cp/copy executable (default for UNIX)
#                                  Default value for 'by' is 'perl' if:
#                                  - Windows
#                                  - $Flags->{stat} is set to 'epages'
#                                  - $Flags->{recursive} is unset
#                                  - $Flags->{md5sum} is set
#                                  - $Flags->{ignore} is set
#                              </li>
#                              <li>Print - print the command line (only used with by => exe) - boolean</li>
#                          </ul>
#                          | hash ref
#========================================================================================
sub CopyDirectory ($$;$$) {
    my ( $SourceDir, $TargetDir, $BackupDir, $Flags ) = @_;
    TestString('SourceDir', $SourceDir);
    TestString('TargetDir', $TargetDir);
    TestString('BackupDir', $BackupDir) if defined $BackupDir;

    $Flags->{recursive} = 1 unless defined $Flags->{recursive};
    $Flags->{basedir} = $SourceDir;
    $Flags->{stat} = _SetStat($SourceDir, $TargetDir, $Flags);
    $Flags->{by} //= 'perl';
    $Flags->{by} = 'perl' if $Flags->{stat} eq 'epages';
    $Flags->{by} = 'perl' unless $Flags->{recursive};
    $Flags->{by} = 'perl' if $Flags->{md5sum};
    $Flags->{by} = 'perl' if $Flags->{ignore};
    $Flags->{by} = 'perl' if $Config{osname} eq 'MSWin32';

    if ($Flags->{'link'} and $Flags->{'link'} ne 'noCheck') {
        CreateDirectory($TargetDir, $Flags);

        if ( SameDevice($SourceDir, $TargetDir) ) {
            $Flags->{'link'} = 'noCheck';
        } else {
            if ($Flags->{'link'} eq 'orCopy') {
                delete $Flags->{'link'};
            } else {
                Error('CROSS_DEVICE_FAILURE', {
                    'Code'=> 'Invalid cross-device link',
                    'Source'=> $SourceDir,
                    'Target'=> $TargetDir,
                });
            }
        }
    }

    my $BakFlags = {};
    do { $BakFlags->{$_} = $Flags->{$_}; } foreach keys %$Flags;
    $BakFlags->{stat} = $Flags->{bakstat};
    delete $BakFlags->{link};
    delete $BakFlags->{unlink};

    BackupDirectory($SourceDir, $TargetDir, $BackupDir, $BakFlags) if defined $BackupDir;
    GetLog->debug("CopyDirectory $SourceDir, $TargetDir");

    if ( defined $Flags->{md5sum} and -s $Flags->{md5sum} ) {
        my %Md5Hash = ();
        tie %Md5Hash, 'Config::IniFiles', ( -file => $Flags->{md5sum} )
            or Error('FILE_OPEN_FAILED',
                     {'Mode'=>'<', 'FileName'=>$Flags->{md5sum}, 'Code'=>$!});
        $Flags->{hMd5Sum} = \%Md5Hash;
    }

    _CopyDirectory( $SourceDir, $TargetDir, $Flags );
    delete $Flags->{hMd5Sum};

    return;
}

# copy directory recursively without backup
sub _CopyDirectory ($$;$); # forward declaration
sub _CopyDirectory ($$;$) {
    my ( $SourceDir, $TargetDir, $Flags ) = @_;

    my $CopyFast = $Flags->{recursive} && $Flags->{by} eq 'exe' ? 1 : 0;

    # don't use GetInfoDirectory recursively, because it would too much RAM
    my $DirInfo;
    unless ($CopyFast) {
        $DirInfo = GetInfoDirectory($SourceDir, { %$Flags, 'recursive'=> 0 });
        return unless $DirInfo->{'Directories'} or $DirInfo->{'Files'};
    }

    my $DirRights = {};
    $DirRights = GetAccessRights($SourceDir)
        if defined $Flags->{'stat'} and $Flags->{'stat'} eq 'preserve';

    CreateDirectory($TargetDir, { %$Flags, %$DirRights, 'recursive'=> 1 });

    if( $Flags->{recursive} ) {
        return _cp($SourceDir, $TargetDir, $Flags) if $Flags->{by} eq 'exe';
        foreach my $Source ( @{$DirInfo->{'Directories'}} ) {
            my $Target = $Source;
            $Target =~ s/^\Q$SourceDir\E/$TargetDir/e;
            _CopyDirectory( $Source, $Target, $Flags );
        }
    }

    my $fileFlags = { %$Flags, 'create_directory' => 1 };
    delete $fileFlags->{md5sum};
    delete $fileFlags->{hMd5Sum};

    foreach my $Source ( @{$DirInfo->{'Files'}} ) {
        my $Target = $Source;
        $Target =~ s/^\Q$SourceDir\E/$TargetDir/e;

        if (defined $Flags->{hMd5Sum}) {
            my $File = $Source;
            $File =~ s/^\Q$Flags->{basedir}\E\/?//x;
            $fileFlags->{md5info} = $Flags->{hMd5Sum}{$File} if
                defined $Flags->{hMd5Sum}{$File};
        }
        CopyFile( $Source, $Target, $fileFlags );
    }

    return;
}

sub _SetStat ($$$) {
    my ($SourceDir, $TargetDir, $Flags) = @_;

    # if stat is unset and by is perl, permissions are set by perl as well
    if (defined $Flags->{by} and $Flags->{by} eq 'perl') {
        return 'epages' unless $Flags->{stat};
    } elsif (defined $Flags->{by} and $Flags->{by} eq 'exe') {
        return $Flags->{stat} unless $Flags->{stat} eq 'epages';
    }

    # if stat is unset set it to default value
    my $Stat = $Flags->{stat} // 'epages';

    # return stat, windows doesn't matter
    return $Stat if $Config{'osname'} eq 'MSWin32';
    # return stat if set to preserve/ignore
    return $Stat if $Stat ne 'epages';

    $SourceDir = AbsolutePath($SourceDir);
    $TargetDir = AbsolutePath($TargetDir);

    # set stat to preserve if target is located outside epages
    return 'preserve' unless _UnixPathStartsWith($TargetDir, $ENV{EPAGES});
    # set stat to epages if target is located inside epages, but source not
    return 'epages' unless _UnixPathStartsWith($SourceDir, $ENV{EPAGES});

    foreach my $Env (qw(APACHE CONFIG LOG STATIC STORES WEBROOT WEBSECURE CARTRIDGES PERL PATCHES)) {
        my $SourceStart = _UnixPathStartsWith($SourceDir, $ENV{"EPAGES_$Env"});
        my $TargetStart = _UnixPathStartsWith($TargetDir, $ENV{"EPAGES_$Env"});
        # set stat to preserve if source and target are located in same epages part
        return 'preserve' if $SourceStart and $TargetStart;
        # set stat to epages if source and target are located in different epages parts
        return 'epages' if $SourceStart or $TargetStart;
    }

    my $SharedSource = _UnixPathStartsWith($SourceDir, $ENV{EPAGES_SHARED});
    my $SharedTarget = _UnixPathStartsWith($TargetDir, $ENV{EPAGES_SHARED});
    if ($SharedSource and $SharedTarget) {
        $SourceDir =~ s,$ENV{EPAGES_SHARED}/+(.+?)($|/).*,$1,;
        $TargetDir =~ s,$ENV{EPAGES_SHARED}/+(.+?)($|/).*,$1,;
        return $SourceDir eq $TargetDir ? 'preserve' : 'epages';
    } elsif ($SharedSource or $SharedTarget) {
        return 'epages';
    }

    $SourceDir =~ s,$ENV{EPAGES}/+(.+?)($|/).*,$1,;
    $TargetDir =~ s,$ENV{EPAGES}/+(.+?)($|/).*,$1,;
    return $SourceDir eq $TargetDir ? 'preserve' : 'epages';
}

sub _cp ($$$) {
    my ( $SourceDir, $TargetDir, $Flags ) = @_;

    Error('SYSTEM_CMD_MISSING', { 'Command' => '/bin/cp', 'OS' => $^O })
        unless -x '/bin/cp';
    Error('DIR_OPEN_FAILED', { 'Directory' => $SourceDir, 'Code' => 'no such directory' })
        unless -d $SourceDir;
    Error('DIR_OPEN_FAILED', { 'Directory' => $TargetDir, 'Code' => 'no such directory' })
        unless -d $TargetDir;

    my $Options = '-r';
    $Options  .= ' -l' if $Flags->{link};
    $Options  .= ' --remove-destination' if $Flags->{unlink};
    $Options  .= ' -f' unless $Flags->{unlink};
    $Options  .= ' -p' if $Flags->{stat} eq 'preserve';

    _UnixCommand("cd \"$SourceDir\" && /bin/cp $Options . \"$TargetDir\"", $Flags);

    return;
}

sub _EncodingByFlags {
    my ($hFlags) = @_;

    return 'utf8' if $hFlags->{'utf8'};
    my $Encoding = defined $hFlags->{'encoding'} ? $hFlags->{'encoding'} : DEFAULT_ENCODING;
    return "raw:encoding($Encoding):crlf:utf8" if $Encoding =~ /^utf16/i;
    return "encoding($Encoding)";
}

sub  _UnixCommand {
    my ($Command, $Flags) = @_;

    # we can't use SystemCommand here because
    # Core::API::SystemCommand and Core::API::File require each other

    print "$Command\n" if $Flags->{Print};
    my $Output = `$Command 2>&1`;
    my $Result = $?;

    my $ErrorParams = { Command => $Command };
    my $ErrorCode;
    ## no critic qw(ProhibitMagicNumbers)
    if ($Result == -1) {
        $ErrorCode = 'SYSTEM_CMD_ERROR';
        $ErrorParams->{Message} = $!;
        $ErrorParams->{Code} = -1;
    } elsif ($Result & 127) {
        $ErrorCode = 'SYSTEM_CMD_TERMINATED';
        $ErrorParams->{Signal} = ($Result & 127);
    }
    elsif ($Result != 0) {
        $ErrorCode = 'SYSTEM_CMD_FAILED';
        $ErrorParams->{Message} = $Output;
        $ErrorParams->{Code} = ($Result >> 8);
    }
    Error($ErrorCode, $ErrorParams) if $ErrorCode;
    ## use critic

    return $Output if $Flags->{Capture};
    return;
}

#========================================================================================
# function     _UnixPathStartsWith
# state        private
#----------------------------------------------------------------------------------------
# syntax       _UnixPathStartsWith($Path, @PathStart)
#----------------------------------------------------------------------------------------
# description  returns 1 if $Path starts with one of @PathStart, else 0
#----------------------------------------------------------------------------------------
# input        $Path | path | string
# input        @PathStart | array of paths | string
#----------------------------------------------------------------------------------------
# return       0 or 1
#========================================================================================
sub _UnixPathStartsWith {
    my ($Path, @PathStart) = @_;
    foreach ( @PathStart ) {
        next unless defined $_ and /\S/;
        return 1 if $Path =~ /^\Q$_\E[\/\\]?/x;
    }
    return 0;
}

#========================================================================================
# function     _GET_EPAGES_APACHE_PERMS
# state        private
#----------------------------------------------------------------------------------------
# syntax       _GET_EPAGES_APACHE_PERMS($Path, $hFlags, $User, $Group)
#----------------------------------------------------------------------------------------
# description  returns permissions for EPAGES_APACHE
#----------------------------------------------------------------------------------------
# input        $Path   | path                                | string
# input        $hFlags | hash w/ flags Mode, UserID, GroupID | hash ref
# input        $User   | user name                           | string
# input        $Group  | group name                          | string
#----------------------------------------------------------------------------------------
# return       $href   | hash w/ flags Mode, UserID, GroupID | hash ref
#========================================================================================
sub _GET_EPAGES_APACHE_PERMS {
    my ($Path, $hFlags, $User, $Group) = @_;
    my ($Mode, $UserID, $GroupID) = ($hFlags->{Mode}, (_DefaultUserGroup($hFlags)));

    if (not defined $Mode or $Mode !~ /\S/) {
        $Mode = GROUP_R;
        $Mode = GROUP_RX if $Path =~ m,^$ENV{EPAGES_APACHE}/(bin|lib|modules|build)/,x;
    }
    $UserID = $User->{ep_web} if not defined $UserID or $UserID !~ /\S/;
    $GroupID = $Group->{ep_web} if not defined $GroupID or $GroupID !~ /\S/;

    return { Mode => $Mode, UserID => $UserID, GroupID => $GroupID };
}

#========================================================================================
# function     _GET_EPAGES_CONFIG_PERMS
# state        private
#----------------------------------------------------------------------------------------
# syntax       _GET_EPAGES_CONFIG_PERMS($Path, $hFlags, $User, $Group)
#----------------------------------------------------------------------------------------
# description  returns permissions for EPAGES_CONFIG
#----------------------------------------------------------------------------------------
# input        $Path   | path                                | string
# input        $hFlags | hash w/ flags Mode, UserID, GroupID | hash ref
# input        $User   | user name                           | string
# input        $Group  | group name                          | string
#----------------------------------------------------------------------------------------
# return       $href   | hash w/ flags Mode, UserID, GroupID | hash ref
#========================================================================================
sub _GET_EPAGES_CONFIG_PERMS {
    my ($Path, $hFlags, $User, $Group) = @_;
    my ($Mode, $UserID, $GroupID) = ($hFlags->{Mode}, (_DefaultUserGroup($hFlags)));

    if ($Path =~ /\/Database.d/x) {
        $Mode = OWNER_RW if not defined $Mode or $Mode !~ /\S/;
        $UserID = $User->{eprunapp} if not defined $UserID or $UserID !~ /\S/;
    } elsif ($Path =~ /((Data(base|Cache)\.conf|\.db)$|\/Cartridges.d)/x or $Path eq $ENV{EPAGES_SHARED}) {
        $Mode = OWNER_RW if not defined $Mode or $Mode !~ /\S/;
        $UserID = $User->{eprunapp} if not defined $UserID or $UserID !~ /\S/;
    } else {
        $Mode = GROUP_R if not defined $Mode or $Mode !~ /\S/;
        $UserID = $User->{ep_appl} if not defined $UserID or $UserID !~ /\S/;
    }
    $GroupID = $Group->{epages} if not defined $GroupID or $GroupID !~ /\S/;
    $GroupID = $Group->{ep_web} if $Path =~ /(log4perl\.conf|WebInterface\.conf|ServerConfig\.xml)$/x;

    return { Mode => $Mode, UserID => $UserID, GroupID => $GroupID };
}

#========================================================================================
# function     _GET_EPAGES_LOG_PERMS
# state        private
#----------------------------------------------------------------------------------------
# syntax       _GET_EPAGES_LOG_PERMS($Path, $hFlags, $User, $Group)
#----------------------------------------------------------------------------------------
# description  returns permissions for EPAGES_LOG
#----------------------------------------------------------------------------------------
# input        $Path   | path                                | string
# input        $hFlags | hash w/ flags Mode, UserID, GroupID | hash ref
# input        $User   | user name                           | string
# input        $Group  | group name                          | string
#----------------------------------------------------------------------------------------
# return       $href   | hash w/ flags Mode, UserID, GroupID | hash ref
#========================================================================================
sub _GET_EPAGES_LOG_PERMS {
    my ($Path, $hFlags, $User, $Group) = @_;
    my ($Mode, $UserID, $GroupID) = ($hFlags->{Mode}, (_DefaultUserGroup($hFlags)));

    $Mode = OWNER_RW if not defined $Mode or $Mode !~ /\S/;
    $UserID = $User->{eprunapp} if not defined $UserID or $UserID !~ /\S/;
    $GroupID = $Group->{ep_web} if not defined $GroupID or $GroupID !~ /\S/;

    return { Mode => $Mode, UserID => $UserID, GroupID => $GroupID };
}

#========================================================================================
# function     _GET_EPAGES_STATIC_PERMS
# state        private
#----------------------------------------------------------------------------------------
# syntax       _GET_EPAGES_STATIC_PERMS($Path, $hFlags, $User, $Group)
#----------------------------------------------------------------------------------------
# description  returns permissions for EPAGES_STATIC
#----------------------------------------------------------------------------------------
# input        $Path   | path                                | string
# input        $hFlags | hash w/ flags Mode, UserID, GroupID | hash ref
# input        $User   | user name                           | string
# input        $Group  | group name                          | string
#----------------------------------------------------------------------------------------
# return       $href   | hash w/ flags Mode, UserID, GroupID | hash ref
#========================================================================================
sub _GET_EPAGES_STATIC_PERMS {
    my ($Path, $hFlags, $User, $Group) = @_;
    my ($Mode, $UserID, $GroupID) = ($hFlags->{Mode}, (_DefaultUserGroup($hFlags)));

    $Mode = GROUP_R if not defined $Mode or $Mode !~ /\S/;
    $UserID = $User->{eprunapp} if not defined $UserID or $UserID !~ /\S/;
    $GroupID = $Group->{epages} if not defined $GroupID or $GroupID !~ /\S/;

    return { Mode => $Mode, UserID => $UserID, GroupID => $GroupID };
}

#========================================================================================
# function     _GET_EPAGES_STORES_PERMS
# state        private
#----------------------------------------------------------------------------------------
# syntax       _GET_EPAGES_STORES_PERMS($Path, $hFlags, $User, $Group)
#----------------------------------------------------------------------------------------
# description  returns permissions for EPAGES_STORES
#----------------------------------------------------------------------------------------
# input        $Path   | path                                | string
# input        $hFlags | hash w/ flags Mode, UserID, GroupID | hash ref
# input        $User   | user name                           | string
# input        $Group  | group name                          | string
#----------------------------------------------------------------------------------------
# return       $href   | hash w/ flags Mode, UserID, GroupID | hash ref
#========================================================================================
sub _GET_EPAGES_STORES_PERMS {
    my ($Path, $hFlags, $User, $Group) = @_;
    my ($Mode, $UserID, $GroupID) = ($hFlags->{Mode}, (_DefaultUserGroup($hFlags)));

    $Mode = GROUP_R if not defined $Mode or $Mode !~ /\S/;
    if (not defined $Mode or $Mode !~ /\S/)  {
        $Mode = $Path =~ /\/_MONITOR_/ ? OWNER_RW : GROUP_R;
    }

    $UserID = $User->{eprunapp} if not defined $UserID or $UserID !~ /\S/;
    $GroupID = $Group->{epages} if not defined $GroupID or $GroupID !~ /\S/;

    return { Mode => $Mode, UserID => $UserID, GroupID => $GroupID };
}

#========================================================================================
# function     _GET_EPAGES_WEBSECURE_PERMS
# state        private
#----------------------------------------------------------------------------------------
# syntax       _GET_EPAGES_WEBSECURE_PERMS($Path, $hFlags, $User, $Group)
#----------------------------------------------------------------------------------------
# description  returns permissions for EPAGES_WEBSECURE
#----------------------------------------------------------------------------------------
# input        $Path   | path                                | string
# input        $hFlags | hash w/ flags Mode, UserID, GroupID | hash ref
# input        $User   | user name                           | string
# input        $Group  | group name                          | string
#----------------------------------------------------------------------------------------
# return       $href   | hash w/ flags Mode, UserID, GroupID | hash ref
#========================================================================================
sub _GET_EPAGES_WEBSECURE_PERMS {
    my ($Path, $hFlags, $User, $Group) = @_;
    return _GET_EPAGES_WEBROOT_PERMS($Path, $hFlags, $User, $Group);
}

#========================================================================================
# function     _GET_EPAGES_WEBROOT_PERMS
# state        private
#----------------------------------------------------------------------------------------
# syntax       _GET_EPAGES_WEBROOT_PERMS($Path, $hFlags, $User, $Group)
#----------------------------------------------------------------------------------------
# description  returns permissions for EPAGES_WEBROOT
#----------------------------------------------------------------------------------------
# input        $Path   | path                                | string
# input        $hFlags | hash w/ flags Mode, UserID, GroupID | hash ref
# input        $User   | user name                           | string
# input        $Group  | group name                          | string
#----------------------------------------------------------------------------------------
# return       $href   | hash w/ flags Mode, UserID, GroupID | hash ref
#========================================================================================
sub _GET_EPAGES_WEBROOT_PERMS {
    my ($Path, $hFlags, $User, $Group) = @_;
    my ($Mode, $UserID, $GroupID) = ($hFlags->{Mode}, (_DefaultUserGroup($hFlags)));

    $Mode = GROUP_R if not defined $Mode or $Mode !~ /\S/;
    if (not defined $Mode or $Mode !~ /\S/)  {
        $Mode = $Path =~ /\/_MONITOR_/ ? OWNER_RW : GROUP_R;
    }

    $UserID = $User->{eprunapp} if not defined $UserID or $UserID !~ /\S/;
    $GroupID = $Group->{ep_web} if not defined $GroupID or $GroupID !~ /\S/;

    return { Mode => $Mode, UserID => $UserID, GroupID => $GroupID };
}

#========================================================================================
# function     _GET_EPAGES_CARTRIDGES_PERMS
# state        private
#----------------------------------------------------------------------------------------
# syntax       _GET_EPAGES_CARTRIDGES_PERMS($Path, $hFlags, $User, $Group)
#----------------------------------------------------------------------------------------
# description  returns permissions for EPAGES_CARTRIDGES
#----------------------------------------------------------------------------------------
# input        $Path   | path                                | string
# input        $hFlags | hash w/ flags Mode, UserID, GroupID | hash ref
# input        $User   | user name                           | string
# input        $Group  | group name                          | string
#----------------------------------------------------------------------------------------
# return       $href   | hash w/ flags Mode, UserID, GroupID | hash ref
#========================================================================================
sub _GET_EPAGES_CARTRIDGES_PERMS {
    my ($Path, $hFlags, $User, $Group) = @_;
    my ($Mode, $UserID, $GroupID) = ($hFlags->{Mode}, (_DefaultUserGroup($hFlags)));

    if (not defined $Mode or $Mode !~ /\S/) {
        $Mode = BASE_RW;
        $Mode = BASE_RWX if $Path =~ m,^$ENV{EPAGES_CARTRIDGES}/[^/]+/[^/]+/bin/,x;
        $Mode = BASE_RWX if $Path =~ m,\.sh$,x;
    }

    $UserID = $User->{ep_appl} if not defined $UserID or $UserID !~ /\S/;
    $GroupID = $Group->{epages} if not defined $GroupID or $GroupID !~ /\S/;

    return { Mode => $Mode, UserID => $UserID, GroupID => $GroupID };
}

#========================================================================================
# function     _GET_EPAGES_PERL_PERMS
# state        private
#----------------------------------------------------------------------------------------
# syntax       _GET_EPAGES_PERL_PERMS($Path, $hFlags, $User, $Group)
#----------------------------------------------------------------------------------------
# description  returns permissions for EPAGES_PERL
#----------------------------------------------------------------------------------------
# input        $Path   | path                                | string
# input        $hFlags | hash w/ flags Mode, UserID, GroupID | hash ref
# input        $User   | user name                           | string
# input        $Group  | group name                          | string
#----------------------------------------------------------------------------------------
# return       $href   | hash w/ flags Mode, UserID, GroupID | hash ref
#========================================================================================
sub _GET_EPAGES_PERL_PERMS {
    my ($Path, $hFlags, $User, $Group) = @_;
    my ($Mode, $UserID, $GroupID) = ($hFlags->{Mode}, (_DefaultUserGroup($hFlags)));

    if (not defined $Mode or $Mode !~ /\S/) {
        $Mode = BASE_RW;
        $Mode = BASE_RWX if _UnixPathStartsWith($Path, "$ENV{EPAGES_PERL}/bin");
        $Mode = BASE_RWX if _UnixPathStartsWith($Path, "$ENV{EPAGES_PERL}/lib32");
        $Mode = BASE_RWX if $Path =~ /\.so$/;
    }
    $UserID = $User->{ep_appl} if not defined $UserID or $UserID !~ /\S/;
    $GroupID = $Group->{epages} if not defined $GroupID or $GroupID !~ /\S/;

    return { Mode => $Mode, UserID => $UserID, GroupID => $GroupID };
}

#========================================================================================
# function     _GET_EPAGES_SHARED_PERMS
# state        private
#----------------------------------------------------------------------------------------
# syntax       _GET_EPAGES_SHARED_PERMS($Path, $hFlags, $User, $Group)
#----------------------------------------------------------------------------------------
# description  returns permissions for EPAGES_SHARED
#----------------------------------------------------------------------------------------
# input        $Path   | path                                | string
# input        $hFlags | hash w/ flags Mode, UserID, GroupID | hash ref
# input        $User   | user name                           | string
# input        $Group  | group name                          | string
#----------------------------------------------------------------------------------------
# return       $href   | hash w/ flags Mode, UserID, GroupID | hash ref
#========================================================================================
sub _GET_EPAGES_SHARED_PERMS {
    my ($Path, $hFlags, $User, $Group) = @_;
    my ($Mode, $UserID, $GroupID) = ($hFlags->{Mode}, (_DefaultUserGroup($hFlags)));

    if (not defined $Mode or $Mode !~ /\S/) {
        $Mode = GROUP_R;
        $Mode = GROUP_RX if _UnixPathStartsWith($Path, "$ENV{EPAGES_SHARED}/Monitor");
        $Mode = BASE_RW if $Path eq $ENV{EPAGES_SHARED} or $Path eq $ENV{EPAGES_PATCHES};
        $Mode = BASE_RW if $Path eq "$ENV{EPAGES_SHARED}/Patches";
        $Mode = BASE_RW if $Path =~ m,$ENV{EPAGES_PATCHES}/[^/]*$,;
        $Mode = BASE_RW if $Path =~ m,$ENV{EPAGES_SHARED}/Patches/[^/]*$,;
    }

    $UserID = $User->{ep_appl} if not defined $UserID or $UserID !~ /\S/;
    if (not defined $GroupID or $GroupID !~ /\S/) {
        $GroupID = $Group->{epages};
        $GroupID = $Group->{ep_web} if _UnixPathStartsWith($Path, "$ENV{EPAGES_SHARED}/Monitor");
    }
    return { Mode => $Mode, UserID => $UserID, GroupID => $GroupID };
}

#========================================================================================
# function     _GET_EPAGES_PERMS
# state        private
#----------------------------------------------------------------------------------------
# syntax       _GET_EPAGES_PERMS($Path, $hFlags, $User, $Group)
#----------------------------------------------------------------------------------------
# description  returns permissions for EPAGES
#----------------------------------------------------------------------------------------
# input        $Path   | path                                | string
# input        $hFlags | hash w/ flags Mode, UserID, GroupID | hash ref
# input        $User   | user name                           | string
# input        $Group  | group name                          | string
#----------------------------------------------------------------------------------------
# return       $href   | hash w/ flags Mode, UserID, GroupID | hash ref
#========================================================================================
sub _GET_EPAGES_PERMS {
    my ($Path, $hFlags, $User, $Group) = @_;
    my ($Mode, $UserID, $GroupID) = ($hFlags->{Mode}, (_DefaultUserGroup($hFlags)));

    if (not defined $Mode or $Mode !~ /\S/) {
        $Mode = GROUP_R;
        $Mode = BASE_RW if $Path eq $ENV{EPAGES};
        $Mode = BASE_RWX if _UnixPathStartsWith($Path, "$ENV{EPAGES}/bin");
        foreach (qw(ASPoolDBCacheServer MySQLDumps RequestRouter Service Setup WebAdapter)) {
            $Mode = GROUP_RX if _UnixPathStartsWith($Path, "$ENV{EPAGES}/$_");
        }
    }

    if (not defined $UserID or $UserID !~ /\S/) {
        $UserID = $User->{ep_appl};
        $UserID = $User->{root} if _UnixPathStartsWith($Path, "$ENV{EPAGES}/Setup");
    }
    if (not defined $GroupID or $GroupID !~ /\S/) {
        $GroupID = $Group->{epages};
        $GroupID = $Group->{ep_web} if _UnixPathStartsWith($Path, "$ENV{EPAGES}/WebAdapter");
    }

    return { Mode => $Mode, UserID => $UserID, GroupID => $GroupID };
}

#========================================================================================
# function     _DefaultUserGroup
# state        private
#----------------------------------------------------------------------------------------
# syntax       _DefaultUserGroup($hFlags)
#----------------------------------------------------------------------------------------
# description  returns uid/gid for user as defined in $hFlags
#----------------------------------------------------------------------------------------
# input        $hFlags  | hash w/ flags UserID, GroupID, User, Group | hash ref
#----------------------------------------------------------------------------------------
# return       $UserID  |  uid | integer
# return       $GroupID |  gid | integer
#========================================================================================
sub _DefaultUserGroup {
    my ($hFlags) = @_;

    my ($UserID, $GroupID) = ($hFlags->{UserID},$hFlags->{GroupID});

    $UserID = (getpwnam $hFlags->{User})[2]   if (not defined $hFlags->{UserID})  and $hFlags->{User};
    $GroupID = (getgrnam $hFlags->{Group})[2] if (not defined $hFlags->{GroupID}) and $hFlags->{Group};
    $GroupID = (getpwuid $UserID)[3] if (not defined $GroupID or $GroupID !~ /\S/) and (defined $hFlags->{UserID} or $hFlags->{User});

    return ($UserID, $GroupID);
}

#========================================================================================
# function     ConvertPathChars
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DirectoryName = ConvertPathChars( $Alias );
# example      $DirectoryName = ConvertPathChars( "IBN-08/25.1234" );
#----------------------------------------------------------------------------------------
# description  Encodes a normal string to a string, which can be used as direcory name.
#               The following characters are converted to a 4-digit hex code surrounded
#               by underline: \ / : * ? " < > | & ' % # . \x{0000}-\x{0020} \x{0080}-\x{ffff}
#               For example: '.' -> '_002E_'
#               Note on dots in file names:
#                 Windows silently removes a trailing dot from directry or file names.
#                 UrlScan for IIS rejects all directory names containing dots.
#----------------------------------------------------------------------------------------
# input        $Alias | alias of object | string
# return       $DirectoryName | directory name | string
#========================================================================================
sub ConvertPathChars {
    local $_ = shift;

    s/[\\
      \/
      \:
      \*
      \?
      \"
      \<
      \>
      \|
      \&
      \'
      \%
      \#
      \.
      \x{0000}-\x{0020}
      \x{0080}-\x{ffff}
      ]/sprintf("_%04X_",ord($&))/xge;  #" help syntax highlighting

    return $_;
}

#========================================================================================
# function     RelativePath
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Path = RelativePath( $BasePath, $RelatingPath );
# example      $Path = RelativePath( 'C:\Windows\System32', 'C:\Windows\Fonts');
#               # $Path == '../Fonts';
#----------------------------------------------------------------------------------------
# description  Returns the relative path to get from $BasePath to $RelatingPath.
#               All output slashes are converted to /.
#----------------------------------------------------------------------------------------
# input        $BasePath | absolute base path | string
# input        $RelatingPath | absolute path to use get the relative path to base path
#               | string
# return       $RelativePath | the relative path | string
#========================================================================================
sub RelativePath {
    my ($BasePath, $RelatingPath) = @_;

    # unify slashes
    $BasePath =~ s/[\\\/]+/\//g;
    $RelatingPath =~ s/[\\\/]+/\//g;

    my @BasePathSplitted = split /\//, $BasePath;
    my @RelatingPathSplitted = split /\//, $RelatingPath;

    # check different drives - if so then return the $RelatingPath
    if ($BasePath =~ /^\// && $RelatingPath =~ /^\// &&
        $BasePathSplitted[1] ne $RelatingPathSplitted[1]) {
        # *nix style, starting /
        return $RelatingPath;
    } elsif ($BasePath =~ /^[a-zA-Z]:/ && $RelatingPath =~ /^[a-zA-Z]:/ &&
        $BasePathSplitted[0] ne $RelatingPathSplitted[0]) {
        # Windows style, starting with [Letter]:
        return $RelatingPath;
    }

    # remove equal parts
    while (scalar @BasePathSplitted) {
        my $BasePathPart = shift @BasePathSplitted;
        my $RelatingPathPart = shift @RelatingPathSplitted;
        if (not defined $RelatingPathPart) {
            unshift @BasePathSplitted, $BasePathPart;
            last;
        } elsif ($BasePathPart ne $RelatingPathPart) {
            unshift @BasePathSplitted, $BasePathPart;
            unshift @RelatingPathSplitted, $RelatingPathPart;
            last;
        }
    }
    return (scalar (@BasePathSplitted) ? '../' x scalar(@BasePathSplitted) : '') . join('/', @RelatingPathSplitted);
}

#========================================================================================
# function     _GetOSErrorMessage
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Message = _GetOSErrorMessage();
# example      say _GetOSErrorMessage();
#----------------------------------------------------------------------------------------
# description  Returns the operating-system specific error message of the last I/O
#               operation. Uses Win32::FormatMessage() for Windows and $! otherwise.
#----------------------------------------------------------------------------------------
# return       $Message | the error message | string
#========================================================================================
sub _GetOSErrorMessage {

    return $! unless $Config{'osname'} eq 'MSWin32';

    UsePackage('Win32');
    my $Message = Win32::FormatMessage(Win32::GetLastError());
    $Message =~ s/\s+$//;
    return $Message;
}

#========================================================================================
# function     RunOnDirectory
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Result = RunOnDirectory($FolderName, $Sub);
# example      my $aOldFiles = RunOnDirectory($OldFolder,
#                   sub{
#                       my $aFiles = GetFileNames( undef , {'recursive' => 1} );
#                       return $aFiles;
#                   }
#               );
#----------------------------------------------------------------------------------------
# description  execute the given function on specific folder and returns the result of
#               given function.
#----------------------------------------------------------------------------------------
# input        $Directory | name of folder to run | string
# input        $Function | reference to function that would be executed | coderef
# return       $Return | return the value of given function | array/scalar
#========================================================================================
sub RunOnDirectory {
    my ($Directory, $Function) = @_;
    TestString( 'Directory', $Directory );

    #get actual directory
    my $Cwd = getcwd();

    #change to target directory
    unless( chdir $Directory ) {
        Error('CHANGE_DIR_FAILED', {
            'Directory' => $Directory,
            'Code' => $!,
        });
    }

    #placeholder for return values of given function
    my @Result;
    my $Result;

    #decision of scalar context or array context
    if (wantarray()) {
        @Result = $Function->();
    }
    else {
        $Result = $Function->();
    }

    #change to original directory
    unless($Cwd and chdir $Cwd){
        Error('CHANGE_DIR_FAILED', {
            'Directory' => $Cwd,
            'Code' => $!,
        });
    }

    #return array or scalar result
    return wantarray ? @Result : $Result;
}

#========================================================================================
# function     RunOnGlob
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Result = RunOnGlob($Directory, $GlobPattern, $Sub, $hFlags);
# example      RunOnGlob($XmlDir, 'Hooks*.xml', sub { DeleteHooks($_) }, {reverse => 1});
#----------------------------------------------------------------------------------------
# description  execute function $Sub for files matching $GlobPattern in $Directory
#----------------------------------------------------------------------------------------
# input        $Directory | name of folder to run | string
# input        $GlobPattern | glob pattern | string
# input        $Sub | reference to function that would be executed | coderef
# input        $hFlags | flags
#                   <ul>
#                       <li>reverse   - execute function for each file in reverse order
#                       <li>force     - ignore errors (just warn)
#                   </ul>
#               | ref.hash
# return       $Return | return the value of given function | array/scalar
#========================================================================================
sub RunOnGlob {
    my ($Directory, $Pattern, $Sub, $hFlags) = @_;

    return unless -d $Directory;

    my $Result;
    eval {
        $Result = RunOnDirectory( $Directory, sub {
            if ($hFlags->{reverse}) {
                $Sub->() foreach reverse sort glob $Pattern;
            } else {
                $Sub->() foreach sort glob $Pattern;
            }
                                  });
    };

    if( ExistsError() ) {
        if( $hFlags->{force} ) {
            Warning( GetError() );
        } else {
            Error( GetError() );
        }
    }

    return $Result;
}

#========================================================================================
# function     CaptureStream
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Variable = CaptureStream($FileHandle, $Sub);
# example      my $Variable = CaptureStream( *STDERR, sub {
#                                  print STDERR 11111;
#                              });
#----------------------------------------------------------------------------------------
# description  redirects $FileHandle output of function $Sub to $Variable
#----------------------------------------------------------------------------------------
# input        $FileHandle | name of the file handle | file handle
# input        $Sub | reference to function that would be executed | coderef
# return       $Variable | string that otherwise would be sent to $FileHandle | string
#========================================================================================
sub CaptureStream {
    my ($FileHandle, $Sub) = @_;

    local $| = 1;

    open my $SaveHandle, '>&', $FileHandle or ##no critic (RequireBriefOpen)
        Error('FILE_OPEN_FAILED', {'FileName'=>$FileHandle, 'Code'=>$!});
    close $FileHandle or
        Warning('FILE_CLOSE_FAILED', {'FileName'=>$FileHandle, 'Code'=>$!});

    my $Variable;
    eval {
        open $FileHandle, '>', \$Variable or ## no critic qw(RequireBriefOpen)
            Error('FILE_OPEN_FAILED', {'FileName'=>$FileHandle, 'Code'=>$!});
        $Sub->();
        close $FileHandle or
            Error('FILE_CLOSE_FAILED', {'FileName'=>$FileHandle, 'Code'=>$!});
    };
    HandleError( Finally => sub {
        # restore old STDERR even if an error occured
        open $FileHandle, '>&', $SaveHandle or ##no critic (RequireBriefOpen)
            Warning('FILE_OPEN_FAILED', {'FileName'=>$FileHandle, 'Code'=>$!});
    });

    return $Variable;
}

#========================================================================================
# function     _define_users
# state        private
#----------------------------------------------------------------------------------------
# syntax       _define_users($hFlags, $hUser, $hGroup);
# example      _define_users($hFlags, $hUser, $hGroup);
#----------------------------------------------------------------------------------------
# description  sets UserID, GroupID in $hFlags; then fills $hUser, $hGroup accordingly
#----------------------------------------------------------------------------------------
# input        $hFlags | flags
#                   <ul>
#                       <li>UserID    - user ID  - integer
#                       <li>GroupID   - group ID - integer
#                       <li>User      - user name  - string
#                       <li>Group     - group name - string
#                   </ul>
#               | ref.hash
# input        $hUser  | key: user name,  value: user ID  | ref.hash
# input        $hGroup | key: group name, value: group ID | ref.hash
#========================================================================================
sub _define_users ($$$) {
    my ($hFlags, $hUser, $hGroup) = @_;

    $hFlags->{UserID} = (getpwnam $hFlags->{User})[2]
        if !defined $hFlags->{UserID} && defined $hFlags->{User};

    $hFlags->{GroupID} = (getgrnam $hFlags->{Group})[2]
        if !defined $hFlags->{GroupID} && defined $hFlags->{Group};

    $hFlags->{GroupID} = (getpwuid $hFlags->{UserID})[3]
        if !defined $hFlags->{GroupID} && defined $hFlags->{UserID};

    $hUser->{ep_appl} = (getpwnam 'ep_appl')[2];
    local $ENV{'EPAGES_APPUSER'} = 'eprunapp' unless $ENV{'EPAGES_APPUSER'};
    $hUser->{eprunapp} = (getpwnam $ENV{'EPAGES_APPUSER'})[2];
    $hUser->{ep_db} = (getpwnam 'ep_db')[2];
    $hUser->{ep_web} = (getpwnam 'ep_web')[2];

    $hGroup->{epages} = (getgrnam 'epages')[2];
    if ($ENV{'EPAGES_WEBGROUP'} =~ /.*:([0-9]+)$/) {
        $hGroup->{ep_web} = $1;
    }
    return;
}

#========================================================================================
# function     _define_mode
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $Mode = _define_mode($Path, $hFlags, $epStat);
# example      my $Mode = _define_mode($Path, $hFlags, $epStat);
#----------------------------------------------------------------------------------------
# description  returns $Mode for $Path depending on $hFlags->{Mode} and $hFlags->{ExecBit}
#               if $hFlags->{Mode} is empty, sets $hFlags->{Mode} to $epStat->{Mode}
#----------------------------------------------------------------------------------------
# input        $Path | file path | string
# input        $hFlags | user-defined flags
#                   <ul>
#                       <li>Mode    - constant as defd in DE_EPAGES::Core::API::Constants
#                       <li>ExecBit - 0: remove EXECUTE_BIT
#                                   - 1: add EXECUTE_BIT
#                                   - 2: add EXECUTE_BIT (even if digit == 0)
#                   </ul>
#               | ref.hash
# input        $epStat | default mode, depending on file location
#                   <ul>
#                       <li>Mode    - constant as defd in DE_EPAGES::Core::API::Constants
#                   </ul>
#               | ref.hash
# return       $Mode | resulting file mode | constant as defd in DE_EPAGES::Core::API::Constants
#========================================================================================
sub _define_mode ($$$) {
    my ($Path, $hFlags, $epStat) = @_;

    my %ExecBit = qw(del 0 add 1 force 2);

    if (!defined $hFlags->{Mode}) {
        $hFlags->{'ExecBit'} = $ExecBit{add} if -d $Path && !defined $hFlags->{'ExecBit'};
        $hFlags->{Mode} = defined $epStat->{Mode} ? $epStat->{Mode} : BASE_RW;
    }

    my $Mode = $hFlags->{Mode};
    $Mode &= ALL_RWS; # shorten mode to 4 octets

    return $Mode unless defined $hFlags->{'ExecBit'};
    return $Mode | ALL_X if $hFlags->{'ExecBit'} == $ExecBit{force};
    return $Mode & (ALL_RW|SET_EFFECTIVE) if $hFlags->{'ExecBit'} == $ExecBit{del};
    return $Mode unless $hFlags->{'ExecBit'} == $ExecBit{add};

    foreach my $Bit (0, 3, 6) {
        $Mode |= ($Mode & 7 << $Bit) ? (1 << $Bit) : 0;
    }
    return $Mode;
}

sub _chmod {
    my ($Path, $MyUserID, $Mode, $DoWarn, @EntryStat) = @_;

    if ( defined $Mode and $Mode != ( $EntryStat[2] & ALL_RWS ) ) {
        if ( $MyUserID == $EntryStat[4] or $MyUserID == 0 ) {    # i\'m the owner or root
            chmod( $Mode, $Path )
            or Error('FILE_CHMOD_FAILED', { 'FileName' => $Path, 'Code' => $!});
        } elsif ($DoWarn) {
            GetLog->warn(sprintf('user %s can\'t chmod: File=%s: 0%o -> 0%o',
                                 $MyUserID, $Path, $EntryStat[2] & ALL_RWS, $Mode));
        }
    }
}

sub _chown {
    my ($Path, $MyUserID, $UserID, $DoWarn, @EntryStat) = @_;

    if ( defined $UserID and $UserID != $EntryStat[4] ) {
        if ( $MyUserID == 0 ) {    # i\'m root
            chown( $UserID, -1, $Path )
            or Error('FILE_CHOWN_FAILED', { 'FileName' => $Path, 'Code' => $!});
        } elsif ($DoWarn) {
            GetLog->warn(sprintf('user %s can\'t chown: File=%s: %s -> %s',
                                 $MyUserID, $Path, $EntryStat[4], $UserID));
        }
    }
}

sub _chgrp {
    my ($Path, $MyUserID, $DoWarn, $GroupID, @EntryStat) = @_;

    if ( defined $GroupID and $GroupID != $EntryStat[5] ) {
        my $MyGroup = $MyUserID == $EntryStat[4] && ElementOf( $GroupID, split( /\s/, $( ) );
        if ( $MyGroup or $MyUserID == 0 ) {    # i\'m the owner and a group member or root
            chown( -1, $GroupID, $Path )
            or Error('FILE_CHOWN_FAILED', { 'FileName' => $Path, 'Code' => $!});
        } elsif ($DoWarn) {
            GetLog->warn(sprintf('user %s (groups: %s) can\'t chgrp: File=%s: %s -> %s',
                                 $MyUserID, $(, $Path, $EntryStat[5], $GroupID));
        }
    }
}

sub _utime {
    my ($Path, $MyUserID, $Mode, $mTime, $UserID, $aTime, $DoWarn, $GroupID, @EntryStat) = @_;

    if ( $aTime != $EntryStat[8] or $mTime != $EntryStat[9] ) {
        if ( $MyUserID == $EntryStat[4] or $MyUserID == 0 ) {    # i\'m the owner or root
            utime( $aTime, $mTime, $Path )
            or Error('FILE_UTIME_FAILED', { 'FileName' => $Path, 'Code' => $!});
        } elsif ($DoWarn) {
            GetLog->warn(sprintf('user %s can\'t utime: File=%s, Mode=0%o, UserID=%s, GroupID=%s, aTime=%s, mTime=%s',
                $MyUserID, $Path, $Mode, $UserID, $GroupID, $aTime, $mTime));
        }
    }
    return ();
}

#========================================================================================
# function     CreateSymLink
# state        public
#----------------------------------------------------------------------------------------
# syntax       CreateSymLink($Source, $Target);
#----------------------------------------------------------------------------------------
# example      CreateSymLink('/temp/sub1/sub2', '/temp/linked_sub2');
#----------------------------------------------------------------------------------------
# description  Creates a symbolic link from one file or directory to another.
#----------------------------------------------------------------------------------------
# input        $Source | absolute path of source file or directory (existing directory or file) | string
# input        $Target | absolute path of target file or directory (link name) | string
#========================================================================================
sub CreateSymLink {
    my ($Source, $Target) = @_;
    TestString('Source', $Source);
    TestString('Target', $Target);

    # If $Target links to non-existent object, -e returns 0.
    return if -e $Target or IsSymLink($Target);

    my $ParentDir = dirname($Target);

    if (not -e $ParentDir) {
        CreateDirectory( $ParentDir );
    }

    if ($Config{'osname'} eq 'MSWin32') {
        my @Args = (WindowsPath($Target), WindowsPath($Source), '>nul');
        unshift @Args, '/D' if -d $Source;
        unshift @Args, 'mklink';
        my $Result = system(@Args);
        Error('SYSTEM_CMD_FAILED', { 'Command' => join(' ', @Args), 'Code' => $Result })
            if $Result;
    }
    else {
        symlink($Source, $Target)
            or Error('SYMLINK_CREATE_FAILED', {'Code'=>$!, 'FileName'=>$Target});
    }

    return;
}

#========================================================================================
# function     DeleteSymLink
#----------------------------------------------------------------------------------------
# syntax       $Deleted = DeleteSymLink($FileName);
#----------------------------------------------------------------------------------------
# example      DeleteSymLink('/temp/symlink');
#----------------------------------------------------------------------------------------
# description  Removes a symlink from disk. The function throws an error if the link
#               cannot be deleted. If the link did not exist before, the function does
#               nothing.
#               ErrorCodes:<ul>
#                   <li>SYMLINK_REMOVE_FAILED
#               </ul>
#----------------------------------------------------------------------------------------
# input        $FileName | name of symlink | string
# return       $Deleted | true if the symlink existed and was deleted, false if it
#                   did not exist. | boolean
#========================================================================================
sub DeleteSymLink ($) {
    my ($FileName) = @_;

    TestString('FileName', $FileName);
    return 0 unless IsSymLink($FileName);

    if ($Config{'osname'} eq 'MSWin32') {
        DeleteDirectory($FileName);
    }
    else {
        unlink($FileName)
            or Error('SYMLINK_REMOVE_FAILED', {'Code'=>$!, 'FileName'=>$FileName});
    }

    return 1;
}

#========================================================================================
# function     IsSymLink
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsSymLink = IsSymLink($FileName);
#----------------------------------------------------------------------------------------
# example      if( IsSymLink('/temp/symlink') ) { ... }
#----------------------------------------------------------------------------------------
# description  Returns true if the file name is a symlink.
#----------------------------------------------------------------------------------------
# input        $FileName | target file or directory link | string
# return       $IsSymLink | true if the target is a symlink, otherwise false | boolean
#========================================================================================
sub IsSymLink {
    my ($FileName) = @_;

    TestString('FileName', $FileName);
    return -l $FileName ? 1 : 0 if $Config{'osname'} ne 'MSWin32';

    my @PathParts = fsplit($FileName);
    my $ParentDir = WindowsPath($PathParts[0]);
    $FileName = $PathParts[2] ? join('.', @PathParts[1..2]) : $PathParts[1];

    return scalar qx(dir $ParentDir /a:l | find "$FileName") ? 1 : 0;
}

#========================================================================================
# function     GetLocalFSTypes
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aLocalFSTypes = GetLocalFSTypes();
#----------------------------------------------------------------------------------------
# description  Returns a list of local file systems (UNIX-only)
#----------------------------------------------------------------------------------------
# return       $aLocalFSTypes | list of local file systems | ref.array
#========================================================================================
sub GetLocalFSTypes {
    return if $^O eq 'MSWin32';
    my @LocalFSTypes = qw(rootfs ramfs initramfs tmpfs ecryptfs);

    push @LocalFSTypes,
    grep {s,.*mkfs\.,,;} map {glob($_);} '/usr/sbin/mkfs.*', '/sbin/mkfs.*'; ##no critic (ProhibitMutatingListFunctions)

    return \@LocalFSTypes;
}

#========================================================================================
# function     GetFSType
# state        public
#----------------------------------------------------------------------------------------
# syntax       $FSType = GetFSType($Path);
#----------------------------------------------------------------------------------------
# description  Returns the file system type the $Path is stored on (or undef unless exists)
#----------------------------------------------------------------------------------------
# example      print 'is local' if GetFSType($Path) ~~ @{GetLocalFSTypes()};
#----------------------------------------------------------------------------------------
# input        $Path | file path | string
# return       $FSType | file system type for $Path | string
#========================================================================================
sub GetFSType {
    my ($Path) = @_;
    return if $^O eq 'MSWin32';
    return unless -e $Path;

    my $FSType = `find "$Path" -maxdepth 0 -printf "%F\n"`;
    chomp($FSType);
    return $FSType;
}

1;
