use strict;

# load libs at compile time to enable other cartridges to implement policies beyond ./Scripts/Perl/Critic
BEGIN {
    my $DE_EPAGES = "$ENV{EPAGES_CARTRIDGES}/DE_EPAGES";
    opendir( my $DIR, $DE_EPAGES ) || die "Cannot open directory $DE_EPAGES: $!";
    my @cartridgeDirectories = readdir($DIR);
    closedir($DIR);

    foreach my $cartridgeDirectory (@cartridgeDirectories) {
        if ( -d "$DE_EPAGES/$cartridgeDirectory/Scripts/Perl/Critic" ) {
            push @INC, "$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/$cartridgeDirectory/Scripts";
        }
    }
}

use threads;
use DateTime;
use DE_EPAGES::Core::API::File qw (
  CorrectPathSlashes
  CreateDirectory
  DeleteFile
  GetFileContentLines
  GetFileInfo
  GetFileNames
);
use DE_EPAGES::Core::API::PerlTools qw ( GetLocalTimeZone );
use DE_EPAGES::Core::API::Script qw ( RunScript );
use DE_EPAGES::XML::API::UseProblems qw ( WriteCartridgeDependencies WriteExportsCache );
use Getopt::Long;
use Perl::Critic;
use Test::More;
use Time::HiRes qw (time);

my $DEFAULT_PROFILE = "$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Core/Scripts/perlcritic.conf";
my $VERBOSE         = 0;
my %FORMATS = (
    'default' => "%m (%p)\n  at %f line %l.\n  %e.\n-\n",
    'compact' => "%m at %f line %l. %p: %e.\n",
    'review'  => "%f line %l: %m (%p)\n",
    'sonar'   => "%f~|~%s~|~%l~|~%c~|~%m~|~%e~|~%p~||~%n",
);

sub Main {
    my $Help;
    my $Profile = $DEFAULT_PROFILE;
    my $Format = 'default';
    my $Blacklist;
    my $newerTimestamp;
    my $Threads = 1;
    my $refreshCaches = 0;
    my $PPICacheDir;

    GetOptions(
        'help'          => \$Help,
        'verbose'       => \$VERBOSE,
        'profile=s'     => \$Profile,
        'format=s'      => \$Format,
        'blacklist=s'   => \$Blacklist,
        'cnewer=s'      => \$newerTimestamp,
        'threads=s'     => \$Threads,
        'refreshcaches' => \$refreshCaches,
        'ppicachedir=s' => \$PPICacheDir,
    );

    usage() if $Help or scalar @ARGV == 0;
    diag("Using profile: $Profile") if $Profile ne $DEFAULT_PROFILE and $VERBOSE;

    # set rc file
    local $ENV{PERLCRITIC} = $Profile;
    my $time1 = time;

    # set output format
    die "Unknown format: $Format" unless exists $FORMATS{$Format};
    Perl::Critic::Violation::set_format( $FORMATS{$Format} );

    # refresh exports/dependencies caches
    _refreshCaches($refreshCaches);

    # setup PPI::Cache
    _setupPPICache( $PPICacheDir, $Threads > 1 ) if ($PPICacheDir);

    # get Perl files from directories
    my $aFilteredFiles = _collectAndFilterFiles( \@ARGV, $newerTimestamp, $Blacklist );

    if ( $Threads > 1 ) {
        plan tests => $Threads;

        my %Worker; # worker processes
        my $aaFiles = _distributeWorkload( $aFilteredFiles, $Threads );
        foreach my $aFiles ( @$aaFiles ) {

            my $pid = fork();
            die "fork() failed: $!" unless defined $pid;
            if ($pid) {
                $Worker{$pid} = 1;
            }
            else {
                my $failures = critique($aFiles);
                exit $failures == 0 ? 0 : 1;
            }
        }

        while( scalar keys %Worker ) {
            printf "waiting for %d worker...\n", scalar keys %Worker if $VERBOSE;
            my $wpid = wait;
            print "worker $wpid has finished with status $?\n" if $VERBOSE;
            ok( $? == 0, sprintf "Perl::Critic worker process $wpid" ) or diag "status=$?";
            delete $Worker{$wpid};
            last if $wpid == -1;
        }
    }
    else {
        my $result = critique($aFilteredFiles);
        print $result == 0 ? "ok\n" : "failed\n";
        exit $result  == 0 ? 0 : 1;
    }
    diag( sprintf( "need %.1f seconds\n", time - $time1 ) ) if $VERBOSE;
}

sub _refreshCaches {
    my ($refreshCaches) = @_;

    if($refreshCaches) {
        DeleteFile($DE_EPAGES::XML::API::UseProblems::DEPENDENCIES_DATAFILE);
        DeleteFile($DE_EPAGES::XML::API::UseProblems::EXPORTS_DATAFILE);
    }
    if (! -e $DE_EPAGES::XML::API::UseProblems::DEPENDENCIES_DATAFILE) {
        diag( $$ . ' writing dependencies cache file ') if ($VERBOSE);
        WriteCartridgeDependencies();
    }
    if (! -e $DE_EPAGES::XML::API::UseProblems::EXPORTS_DATAFILE) {
        diag( $$ . ' writing exports cache file ') if ($VERBOSE);
        WriteExportsCache();
    }
}

sub _collectAndFilterFiles {
    my ( $aInput, $newerTimestamp, $Blacklist ) = @_;

    if ($newerTimestamp) {
        $newerTimestamp = DateTime->from_epoch( epoch => $newerTimestamp );
        $newerTimestamp->set_time_zone( GetLocalTimeZone() );
    }

    my @Files = grep { !$newerTimestamp || GetFileInfo($_)->{LastChange} >= $newerTimestamp }
      map {
        -d $_
          ? grep {/\.pm$|\.pl$|\.t$/x} @{ GetFileNames( $_, { 'recursive' => 1 } ) }
          : $_
      }
      map { m|(.+)[\\/]$| ? $1 : $_ }    # remove trailing slash
      @$aInput;
    return $Blacklist ? _filterOutBlacklistedFiles( \@Files, $Blacklist ) : \@Files;
}

sub _setupPPICache {
    my ( $PPICacheDir, $PreCreateCacheDirs ) = @_;

    if (! -d $PPICacheDir) {
        CreateDirectory($PPICacheDir);
    }

    # creating ppi cache dirs, as this is not crashing in thread-mode
    require PPI::Cache;
    PPI::Cache->import( path => $PPICacheDir );
    if ($PreCreateCacheDirs) {
        _precreateCacheDirs($PPICacheDir);
    }
    return;
}

sub _precreateCacheDirs {
    my ($PPICacheDir) = @_;
    foreach my $AA ( ( 0 .. 9 ), ( 'A' .. 'F' ) ) {
        CreateDirectory( $PPICacheDir . '/' . $AA );
        foreach my $BB ( ( 0 .. 9 ), ( 'A' .. 'F' ) ) {
            CreateDirectory( $PPICacheDir . '/' . $AA . '/' . $AA . $BB );
        }
    }
}

sub _distributeWorkload {
    my ( $aFilteredFiles, $Threads ) = @_;
    my $aWorkload = [];
    my $i         = 0;
    my @TempFiles = @$aFilteredFiles;
    while ( scalar @TempFiles ) {
        push @{ $aWorkload->[ $i % $Threads ] }, pop(@TempFiles);
        ++$i;
    }
    return $aWorkload;
}

sub _filterOutBlacklistedFiles {
    my ($aFiles, $BlacklistFile) = @_;

    #Initialze Blacklisted Files.
    my @BlacklistedFiles = map { s|\\|\/|g; $_; }
                           map { /(.*?)\s*$/; }
                           @{GetFileContentLines( $BlacklistFile )};

    my $isBlackListed = sub {
        my ($File) = @_;
        $File =~ s|\\|\/|g;

        # Blacklist entry has to match end of filename
        return scalar grep { $File =~ /^\Q$_\E/ } @BlacklistedFiles;
    };

    return [ grep { !$isBlackListed->($_) } @$aFiles ];
}

sub critique {
    my ($aFiles) = @_;

    my $critic          = Perl::Critic->new();
    my $foundViolations = 0;
    foreach my $file (@$aFiles) {
        $file = CorrectPathSlashes($file);

        diag( $$ . ' testing ' . $file ) if ($VERBOSE);
        my @violations = $critic->critique($file);
        foreach my $violation (@violations) {
            diag($violation);
            $foundViolations++;
        }
    }

    if( $foundViolations > 0 ) {
        diag( sprintf 'found %d failure(s) perlcrtitic-tesing %d files(s)',  $foundViolations, scalar @$aFiles );
    }

    return $foundViolations;
}

RunScript( Sub => \&Main );

sub usage {
    print <<END_USAGE;
Description:
    Criticises given files or directories regarding the active policies of the rc file.
    Different policies are applied sequentially to the module like common
    test cases. If one policy fails, a 'not_ok' message is displayed,
    'ok' otherwise.

Usage:
    perl $0 [-profile perlcritic.conf] files or directories

Flags:
    -help          show the command line options
    -verbose       gives some verbose information during run
    -profile       name of the policy definition file
                   default: $DEFAULT_PROFILE
    -format        output format (optional, one of: default|compact|review)
    -blacklist     name of a file containing a list of blacklisted files or
                   directories
    -cnewer        timestamp: only check files newer than given timestamp
    -threads       number of threads to run in parallel  (optional, default=1)
    -ppicachedir   folder to store PPI::Cache files to speed up repeated calls
                   of $0 (optional)
    -refreshcaches If set, critic.pl will update the exports and dependencies caches
                   in $ENV{EPAGES_STATIC} (Must enable this in a CI environment)

Examples:
    perl $0 Test.pm
    perl $0 $ENV{EPAGES_CARTRIDGES}/DE_EPAGES
    perl $0 -profile /home/user/.perlcritic_file Test1.pm Test2.pm
    perl $0 -profile /home/user/.perlcritic_file -blacklist bl.txt /path/to/directory
    perl $0 -cnewer 1321344784 .
END_USAGE
    exit 2;
}
