#==============================================================================
# package      DE_EPAGES::Design::API::TLE::StyleHandler
# state        public
#==============================================================================
package DE_EPAGES::Design::API::TLE::StyleHandler;

use strict;
use DE_EPAGES::Core::API::Error qw ( ExistsError GetError );
use DE_EPAGES::Core::API::Log qw( GetLog LogError );
use DE_EPAGES::Core::API::TestParameter qw ( TestObject );
use DE_EPAGES::Object::API::Factory qw( LoadObject );
use DE_EPAGES::Shop::API::Object::Shop qw( getUnityApiShopContent );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );

sub RegisterHandlerProc {
    my ($Params) = @_;
    GetLog->debug( 'StyleHandler.RegisterHandlerProc' );

    __PACKAGE__->new()->register( $Params->{'Processor'} );
}

sub new { bless {}, shift; }

#========================================================================================
# function     register
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler->register( $Processor );
#----------------------------------------------------------------------------------------
# description  Registers this handle on the processor
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
#========================================================================================
sub register {
    my $self = shift;
    my ($Processor) = @_;
    GetLog->debug( 'StyleHandler.register' );

    $Processor->registerHandler('VariableHandler', $self, 'Style');
    $Processor->registerHandler('FunctionHandler', $self, 'STYLEGROUPBYID');
    $Processor->registerHandler('FunctionHandler', $self, 'STYLETEMPLATEBYID');
    $Processor->registerHandler('FunctionHandler', $self, 'GetStyleURL');
    $Processor->registerHandler('FunctionHandler', $self, 'GetUnityThemeStyleURL');

    return;
}

#========================================================================================
# function     tle
# state        public
#----------------------------------------------------------------------------------------
# syntax       #Style
#----------------------------------------------------------------------------------------
# description  Returns the currently active style.
#========================================================================================
sub tle {
    my $self = shift;
    my ($Processor, $TLEName) = @_;
    GetLog->debug( "StyleHandler.tle($TLEName)" );

    if( $TLEName eq 'Style' ) {
        my $ServletVars = $Processor->tle('INPUT');
        my $USE_STYLE = $ServletVars->{'TLEPROCESSOR_STYLE'};
        return $USE_STYLE if defined $USE_STYLE;

        my $ViewObject = LoadObject($ServletVars->{'ObjectID'});
        my $Style = undef;
        if ($ViewObject->instanceOf('Style')) {
            $Style = $ViewObject ;
        } else {
            my $PreviewStyleAlias = $ServletVars->{'PreviewStyle'};
            my $Shop = $Processor->tle('Shop')->{'Object'};
            my $PreviewStyleID = defined $PreviewStyleAlias ? $Shop->child('Styles')->findChild($PreviewStyleAlias) : undef;
            $Style = defined $PreviewStyleID ? LoadObject($PreviewStyleID) : $Shop->get('DefaultStyle');
        }
        $USE_STYLE = $Style->tleHash;
        $ServletVars->{'TLEPROCESSOR_STYLE'} = $USE_STYLE;
        return $USE_STYLE;
    }
    return undef;
}

#========================================================================================
# function     STYLEGROUPBYID
# state        public
#----------------------------------------------------------------------------------------
# syntax       #STYLEGROUPBYID(#StyleGroupID)
#----------------------------------------------------------------------------------------
# description  Returns the StyleGroup by the given ID.
#----------------------------------------------------------------------------------------
# input        #StyleGroupID | style group id | integer
# return       #StyleGroup | style group for the given id | object
#========================================================================================
sub STYLEGROUPBYID {
    my $self = shift;
    my ($Processor, $aParams, $cTemplate) = @_;
    my ($StyleGroupID) = @$aParams;

    my $StyleGroup = LoadObject($StyleGroupID);
    TestObject('StyleGroup', $StyleGroup, 'DE_EPAGES::Design::API::Object::StyleGroup');
    return defined $StyleGroup ? $StyleGroup->tleHash : undef;
}

#========================================================================================
# function     STYLETEMPLATEBYID
# state        public
#----------------------------------------------------------------------------------------
# syntax       #STYLETEMPLATEBYID(#StyleTemplateID)
#----------------------------------------------------------------------------------------
# description  Returns the StyleTemplate by the given ID.
#----------------------------------------------------------------------------------------
# input        #StyleTemplateID | style template id | integer
# return       #StyleTemplate | style template for the given id | object
#========================================================================================
sub STYLETEMPLATEBYID {
    my $self = shift;
    my ($Processor, $aParams, $cTemplate) = @_;
    my ($StyleTemplateID) = @$aParams;

    my $StyleTemplate = LoadObject($StyleTemplateID);
    TestObject('StyleTemplate', $StyleTemplate, 'DE_EPAGES::Design::API::Object::StyleTemplate');
    return defined $StyleTemplate ? $StyleTemplate->tleHash : undef;
}

#========================================================================================
sub GetStyleURL {
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Style, $hInput) = @$aParams;

    my $hValues = { 'Input' => $hInput, 'StyleURL' => undef };

    TriggerHook('API_GetStyleURL', { 'Object' => $Style, 'Values' => $hValues });

    return defined $hValues->{'StyleURL'} ? $hValues->{'StyleURL'} : $Style->get('URL');
}

#========================================================================================
# function     GetUnityThemeStyleURL
# state        public
#----------------------------------------------------------------------------------------
# syntax       #GetUnityThemeStyleURL(#Shop, #INPUT.ChangeAction)
#----------------------------------------------------------------------------------------
# description  Saves the path to the Unity checkout CSS stylesheet to the shop object,
#               using a GET call to the Unity REST API in case the ChangeAction is
#               PickupBasket or loads the path from the shop object,
#               returns undef in case of an error.
#               This function triggers the hook 'OBJ_LoadCheckoutCSS'.
# example      #LOCAL("UnityCheckoutStylesheetURL", #GetUnityThemeStyleURL(#Object, #INPUT.ChangeAction))
#                 #IF(#UnityCheckoutStylesheet)
#                   <link href="#UnityCheckoutStylesheetURL" rel="stylesheet" type="text/css" />
#                 #ENDIF
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# input        #Shop | the current shop object | object
# input        #INPUT.ChangeAction | ChangeAction of the current servlet | string
# return       #UnityStylesheetURL | URL to the Unity checkout stylesheet | string
#========================================================================================
sub GetUnityThemeStyleURL {
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Shop, $ChangeAction) = @$aParams;
    my $ThemeStylesheetPath;

    eval {
        TestObject( 'Shop', $Shop, 'DE_EPAGES::Shop::API::Object::Shop' );
    };
    if (ExistsError()) {
        # Log the error if any, but do not throw it to the frontend
        my $Error = GetError();
        LogError($Error);
    } else {
        $ThemeStylesheetPath = $Shop->get('UnityThemeStylePath');
        # only add Unity chechout stylesheet we have the change action 'PickupBasket'

        # check if any change action wants to request css from unity
        my $hValues = {
            'RequestCSSPathFromUnity' => 0,
            'ChangeAction' => $ChangeAction
        };
        TriggerHook('OBJ_LoadCheckoutCSS', { 'Object' => $Shop, 'Values' => $hValues });

        if ($hValues->{'RequestCSSPathFromUnity'}) {
            my $ThemeContent;
            eval {
                $ThemeContent = $Shop->getUnityApiShopContent->{'theme'};
            };
            if (ExistsError()) {
                # Log the error if any, but do not throw it to the frontend
                my $Error = GetError();
                LogError($Error);
                $ThemeStylesheetPath = undef;
            }
            elsif (defined $ThemeContent) {
                # concat theme path and add '.css' file ending
                my $ThemeStylesheetPathUnity = '/themes/' . $ThemeContent->{'name'}
                . '@' . $ThemeContent->{'version'} . '/stylesheets/'
                . $ThemeContent->{'themeSettings'}->{'settings'}->{'style'}
                . '-checkout.css';
                if ($ThemeStylesheetPathUnity ne $ThemeStylesheetPath) {
                    # set path to shop object
                    $Shop->set({ 'UnityThemeStylePath' => $ThemeStylesheetPathUnity });
                    $ThemeStylesheetPath = $ThemeStylesheetPathUnity;
                }
            }
        }

        # no Unity checkout stylesheet saved --> use the default one
        return undef unless defined $ThemeStylesheetPath;

        my $UnityStylesheetURL= $Shop->get('WebUrlUnitySSL');
        # insert path into the Unity URL
        $UnityStylesheetURL =~ s:(\?|$):$ThemeStylesheetPath$1:;

        return $UnityStylesheetURL;
    }

    return undef;
}

1;
