#========================================================================================
# package      DE_EPAGES::GoogleAdWords::API::GoogleAdWords
# state        public
#----------------------------------------------------------------------------------------
# description  Google AdWords
#========================================================================================
package DE_EPAGES::GoogleAdWords::API::GoogleAdWords;
use base Exporter;

use strict;

use DE_EPAGES::Core::API::File qw (GetFileContentLines);
use DE_EPAGES::GoogleAdWords::API::Table::GoogleAdWordsCode qw (
    GetPKeyGoogleAdWordsCodesByIsUsed
    UpdateGoogleAdWordsCode
    InsertGoogleAdWordsCode
    GetAllInfoGoogleAdWordsCodes
);
use DE_EPAGES::Core::API::Warning qw( Warning );

our @EXPORT_OK = qw (
    ImportPromotionCodes
    ApplyPromotionCodeToShop
    FormatAdwordsTrackingCode
);

#========================================================================================
# function     ImportPromotionCodes
# state        public
#----------------------------------------------------------------------------------------
# syntax       ImportPromotionCodes( $Filename );
#----------------------------------------------------------------------------------------
# description  Import promotional codes from Google txt file
#----------------------------------------------------------------------------------------
# input        $Filename | file with promotion codes | string
#========================================================================================
sub ImportPromotionCodes {
    my ($Filename) = @_;

    my $NewCodes = GetFileContentLines($Filename);
    return unless (scalar @$NewCodes);

    my $ahCurrentCodes = GetAllInfoGoogleAdWordsCodes();

    # filter duplicates.
    chomp(@$NewCodes);
    SKIPCODE: foreach my $NewCode (@$NewCodes){
        foreach my $ExistingCode (@$ahCurrentCodes){
            # coupon code is already existing in our database.
            next SKIPCODE if ($NewCode eq $ExistingCode->{'CouponCode'});
        }
        my $hInfoNewCode = { 'CouponCode' => $NewCode, 'IsUsed' => 0 };
        InsertGoogleAdWordsCode($hInfoNewCode);
        push @$ahCurrentCodes, $hInfoNewCode;
    }
}

#========================================================================================
# function     ApplyPromotionCodeToShop
# state        public
#----------------------------------------------------------------------------------------
# syntax       ApplyPromotionCodeToShop( $Shop );
#----------------------------------------------------------------------------------------
# description  Get promotion code and apply to shop
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
# return       $Code | promotion code | string
#========================================================================================
sub ApplyPromotionCodeToShop {
    my ($Shop) = @_;
    my $OldCode = $Shop->get('GoogleAdWordsPromotionCode');
    return $OldCode if ($OldCode);

    my $aPKeyAdWordsCodes = GetPKeyGoogleAdWordsCodesByIsUsed(0);
    return unless (scalar @$aPKeyAdWordsCodes);

    my $UseCode = pop @$aPKeyAdWordsCodes;
    UpdateGoogleAdWordsCode($UseCode, {'ShopID' => $Shop->id, 'IsUsed' => 1 });
    if (scalar @$aPKeyAdWordsCodes < 100) {
        Warning('Only '.(scalar @$aPKeyAdWordsCodes - 1). ' google codes left.');
    }

    return;
}

#========================================================================================
# function     FormatAdwordsTrackingCode
# state        public
#----------------------------------------------------------------------------------------
# syntax       FormatAdwordsTrackingCode( $OriginalTrackingString );
#----------------------------------------------------------------------------------------
# description  Format adwords conversion tracking code.
#----------------------------------------------------------------------------------------
# input        $OriginalTrackingString | tracking string | string
# return       $FormattedString | formatted string | string
#========================================================================================
sub FormatAdwordsTrackingCode {
    my ($OriginalTrackingString) = @_;

    my $FormattedString = $OriginalTrackingString;
    $FormattedString =~ s/http:(\/\/www)/$1/sgx;
    # EPG-23599: Replace default conversation value with total order amount.
    $FormattedString =~ s/var\sgoogle_conversion_value.*?=.*?;/var google_conversion_value = \"#OrderAmount\";/sx;
    $FormattedString =~ s/\?value=.*?&amp;/\?value=#OrderAmount&amp;/sx;

    return $FormattedString;
}

1;
