#=================================================================================
# package      DE_EPAGES::Installer::API::EpagesJ
#---------------------------------------------------------------------------------
# description  Functions for handling epagesj
#---------------------------------------------------------------------------------
# state        public
#=================================================================================
package DE_EPAGES::Installer::API::EpagesJ;
use strict;

use base qw (Exporter);

our @EXPORT_OK = qw (
    GetEpagesJBuild
    GetEpagesJVersion
    GetEpagesJZipUrl
    GetEpagesJEtcUrl
    UpdateEpagesJEtc
    UpdateEpagesJ
    UnpackEpagesJ
    AddEpagesJService
    AddMongoDBService
);

use File::Basename qw(basename);
use File::Find qw(find);
use LWP::UserAgent;
use HTTP::Status;
use LWP::Simple;
use Time::HiRes qw(time);

use DE_EPAGES::Core::API::Constants qw(BASE_RWX USER_RWX GROUP_R GROUP_RX);
use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Core::API::File qw(GetFileContentLines CopyFile WindowsPath WriteFile
                                  RenameDirectory DeleteDirectory CreateDirectory MoveFile
                                  CorrectPathSlashes RunOnGlob RunOnDirectory DeleteFile
                                  CopyDirectory);
use DE_EPAGES::Core::API::IniConfig;
use DE_EPAGES::Core::API::PerlTools qw(UsePackage);
use DE_EPAGES::Core::API::Program::Tools qw ( FindProgramFile Has64BitOS );
use DE_EPAGES::Core::API::SystemCommand qw ( SystemCommand );
use DE_EPAGES::Installer::API::Service qw(StartService StopService ExistsService);
use DE_EPAGES::Installer::API::Constants qw (
    INTERN_PUB_CARTRIDGE_PACK_SERVERS
    PUBLIC_CARTRIDGE_PACK_SERVERS
);

BEGIN {
    if( $^O eq 'MSWin32' ) {
        UsePackage( 'Win32' );
        UsePackage( 'Win32::Service' );
    }
}

#ERRORDEFINITION
sub ErrorDefinition () { return {
    'SOURCE_NOT_FOUND'    => 'epagesj.zip#Version not found',
    'DOWNLOAD_FAILED'     => 'Download #Url to #File failed (return code: #Result).',
    'JDK_7_NOT_FOUND'     => 'No directory Java/jdk1.7* found.',
    'FILE_NOT_FOUND'      => 'File not found (#FileName).',
    'DIR_NOT_FOUND'       => 'Directory not found (#Dir).',
    'NO_EPAGESJ_USER'     => 'No epages java user "#User" found.',
    'NO_EPAGESJ_GROUP'    => 'No epages java group "#Group" found.',
    'FILE_CHOWN_FAILED'   => 'User "#User" cannot chown "#Dir"',
    'FILE_CHMOD_FAILED'   => 'User "#User" cannot chmod "#Dir"',
    'SYMLINK_FAILED'      => 'User "#User" cannot symlink #File to #Link',
                         }
}
#ERRORDEFINITION END

#========================================================================================
# function     GetEpagesJBuild
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Build = GetEpagesJBuild($JDir);
#----------------------------------------------------------------------------------------
# description  Get build nubmer as defd. in $JDir/webapps/epagesj.version
#               Returns 0 unless exists.
#----------------------------------------------------------------------------------------
# input        $JDir  | path to j/ (default: $ENV{EPAGESJ}) | string
#----------------------------------------------------------------------------------------
# return       $Build | version of installed epagesj | string
#========================================================================================
sub GetEpagesJBuild (;$) {
    my ($JDir) = @_;
    $JDir //= $ENV{EPAGESJ} =~ /\S/ ? $ENV{EPAGESJ} : "$ENV{EPAGES}/j";

    # $JDir either contains j/ or is j/
    my $VersionFile = -f "$JDir/webapps/epagesj.version" ?
        "$JDir/webapps/epagesj.version" : "$JDir/j/webapps/epagesj.version";
    return 0 unless -f $VersionFile;

    my $Build = (split /:/,
                 (grep { /build:/ }
                  @{GetFileContentLines($VersionFile)})[0])[1];

    $Build =~ s/[^0-9]//g;
    return 0 unless $Build;
    return $Build;
}

#========================================================================================
# function     GetEpagesJVersion
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Version = GetEpagesJVersion($JDir);
#----------------------------------------------------------------------------------------
# description  Get version as defd. in $JDir/webapps/epagesj.version
#               Returns 0 unless exists.
#----------------------------------------------------------------------------------------
# input        $JDir    | path to j/ (default: $ENV{EPAGESJ}) | string
#----------------------------------------------------------------------------------------
# return       $Version | version of installed epagesj | string
#========================================================================================
sub GetEpagesJVersion (;$) {
    my ($JDir) = @_;
    $JDir //= $ENV{EPAGESJ} =~ /\S/ ? $ENV{EPAGESJ} : "$ENV{EPAGES}/j";

    # $JDir either contains j/ or is j/
    my $VersionFile = -f "$JDir/webapps/epagesj.version" ?
        "$JDir/webapps/epagesj.version" : "$JDir/j/webapps/epagesj.version";
    return 0 unless -f $VersionFile;

    my $Version = (split /:/,
                   (grep { /epagesj-version:/ }
                    @{GetFileContentLines($VersionFile)})[0])[1];

    return 0 unless $Version;
    $Version =~ s/\s+//gs;
    return $Version;
}

#========================================================================================
# function     GetEpagesJZipUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DownloadUrl = GetEpagesJZipUrl($Version);
#----------------------------------------------------------------------------------------
# description  Returns URL for epagesj.zip on the download site.
#----------------------------------------------------------------------------------------
# input        $Version | version is either empty, then get the URL for the latest
#                          version, or epages version string as 6.15.0, or snapshot
#                          version as 6.15.0-SNAPSHOT (only epages internal use) | string
#----------------------------------------------------------------------------------------
# return       $DownloadUrl | download URL | string
#========================================================================================
sub GetEpagesJZipUrl (;$) {
    my ($Version) = @_;
    $Version //= '';
    my %Speed;

    my @UrlList = qw(
      http://epages7:8081/artifactory/simple/epages-release-local/com/epages/epagesj
      http://epages7:8081/artifactory/epages-snapshot-local/com/epages/epagesj
    );
    unshift @UrlList, map { "$_/windows/3rd-party/32" }
    (INTERN_PUB_CARTRIDGE_PACK_SERVERS, PUBLIC_CARTRIDGE_PACK_SERVERS);

    if ($Version) {
        unshift @UrlList, map { "$_/pub/epages/build/repo/$Version/latest/windows/3rd-party/32" }
        qw(ftp://ftp.epages.com);
    }
    unshift @UrlList, map { "$_/windows/3rd-party/32" }
    split(/\s+/,$ENV{EPAGES_REPO}) if $ENV{EPAGES_REPO};

    my $Latest;
    # first internal resources because they might contain newer versions
    foreach my $Url (@UrlList) {
        # stop if EPAGES_REPO found
        last if $ENV{EPAGES_REPO} and scalar keys %Speed;
        # stop if internal resource found
        last if $Url =~ /epages-software.de/ and scalar keys %Speed;
        # epagesj.zip on epages7 always have version
        next if not $Version and $Url =~ /epages7/;
        # snapshots always have version
        next if not $Version and $Url =~ /snapshot/;
        # get snapshots from snapshot download
        next if $Version =~ /SNAPSHOT/ and $Url !~ /snapshot/;
        # ignore snapshot if release version already found
        next if $Url =~ /snapshot/ and $Version !~ /SNAPSHOT/ and
            scalar grep { !/SNAPSHOT/ } keys %Speed;

        my $EpagesJZip = $Version ? "epagesj-$Version.zip" : 'epagesj.zip';
        $EpagesJZip = "$Version/$EpagesJZip" if $Url =~ /epages7/;

        if ($Url =~ /epages-snapshot-local/) {
            $EpagesJZip = "$Version-SNAPSHOT/epagesj-$Version-SNAPSHOT.zip" if $Version !~ /SNAPSHOT/;
        } elsif ($Url =~ s/\/snapshot$//) {
            $EpagesJZip = "epagesj-$Version-SNAPSHOT.zip" if $Version !~ /SNAPSHOT/;
        }

        $Latest //= "$Url/$EpagesJZip";

        # we must use max_redirect => 0 because http pages may redirect
        # to an error page with RC == 200 if original page does not exist
        my $UserAgent = LWP::UserAgent->new(max_redirect => 0);
        my $time1 = time;
        my $Response = $UserAgent->head("$Url/$EpagesJZip");
        next unless $Response->is_success;
        $Speed{"$Url/$EpagesJZip"} = time - $time1;
    }

    return $Latest unless scalar keys %Speed;

    my $Fastest = (sort { $Speed{$a} <=> $Speed{$b} } keys %Speed)[0];
    return $Fastest;
}

#========================================================================================
# function     GetEpagesJEtcUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DownloadUrl = GetEpagesJEtcUrl($Version);
#----------------------------------------------------------------------------------------
# description  Returns URL for epagesj start/stop scripts (Linux) on the download site.
#----------------------------------------------------------------------------------------
# input        $Version | version is either empty, then get the URL for the latest
#                          version, or epages version string as 6.15.0, or snapshot
#                          version as 6.15.0-SNAPSHOT (only epages internal use) | string
#----------------------------------------------------------------------------------------
# return       $DownloadUrl | download URL | string
#========================================================================================
sub GetEpagesJEtcUrl (;$) {
    my ($Version) = @_;
    $Version //= 'latest';
    $Version =~ s/-SNAPSHOT//gs;

    my @UrlList = map { "$_/windows/epagesj" }
    (INTERN_PUB_CARTRIDGE_PACK_SERVERS, PUBLIC_CARTRIDGE_PACK_SERVERS);

    unshift @UrlList, map { "$_/windows/epagesj" }
    split(/\s+/,$ENV{EPAGES_REPO}) if $ENV{EPAGES_REPO};

    my $Latest;
    foreach my $Url (@UrlList) {
        # we must use max_redirect => 0 because http pages may redirect
        # to an error page with RC == 200 if original page does not exist
        my $UserAgent = LWP::UserAgent->new(max_redirect => 0);
        my $Response = $UserAgent->head("$Url/$Version/etc/init.d/epagesj");
        return "$Url/$Version/etc/init.d" if $Response->is_success;
        $Latest //= "$Url/latest/etc/init.d";
    }
    return $Latest;
}

#========================================================================================
# function     UpdateEpagesJEtc
# state        public
#----------------------------------------------------------------------------------------
# syntax       UpdateEpagesJEtc($OldVersion, $NewVersion, $Source);
#----------------------------------------------------------------------------------------
# example      UpdateEpagesJEtc('6.14.4', '6.14.6', 'Data/j/etc');
# example      UpdateEpagesJEtc(undef, '6.14.6');
# example      UpdateEpagesJEtc()
#----------------------------------------------------------------------------------------
# description  Updates epagesj start/stop scripts (Linux).
#----------------------------------------------------------------------------------------
# input        $OldVersion | installed epagesj version | string
# input        $NewVersion | epagesj version to be installed | string
# input        $Source     | etc/ directory that contains init.d/epagesj etc. (opt) | string
#========================================================================================
sub UpdateEpagesJEtc (;$$$) {
    my ($OldVersion, $NewVersion, $Source) = @_;
    $OldVersion //= 'old';
    return if $^O eq 'MSWin32';

    my (undef, $EpagesJGid) = @{_GetEpagesJUser()};
    my $DownloadUrl = GetEpagesJEtcUrl($NewVersion);
    $Source //= $DownloadUrl;

    my $InitDir = '/srv/epages/etc/init.d';
    foreach (qw(epagesj epages-solr epages-solr-slave)) {
        my $Content;
        if (not -d $Source) {
            print "getstore($DownloadUrl/$_, $InitDir/$_)\n";
            $Content = get("$DownloadUrl/$_");
            Error('DOWNLOAD_FAILED', {Url=>"$DownloadUrl/$_", File=>"$InitDir/$_",
                                      Result=>'undef'}) unless defined $Content;
            $Content  =~ s/\r?\n/chr(10)/ges;
        }
        DeleteFile("/etc/init.d/$_") if -e "/etc/init.d/$_";
        CreateDirectory($InitDir, {Mode=>BASE_RWX}) unless -d $InitDir;
        MoveFile( "$InitDir/$_","$InitDir/$_-$OldVersion") if -f "$InitDir/$_";

        if (-d $Source) {
            my $InitFile = "$Source/init.d/$_";
            Error('FILE_NOT_FOUND', {FileName=>$InitFile}) unless -f $InitFile;
            CopyFile($InitFile, "$InitDir/$_");
        } else {
            WriteFile("$InitDir/$_", \$Content);
        }

        chown(0, $EpagesJGid, "$InitDir/$_") or
            Error('FILE_CHOWN_FAILED', {Dir => "$InitDir/$_", User => getpwuid($<)});
        chmod(GROUP_R | USER_RWX, "$InitDir/$_") or
            Error('FILE_CHMOD_FAILED', {Dir => "$InitDir/$_", User => getpwuid($<)});
        symlink("$InitDir/$_","/etc/init.d/$_")  or
            Error('SYMLINK_FAILED', {File => "$InitDir/$_", Link => "/etc/init.d/$_", User => getpwuid($<)});
    }
}

#========================================================================================
# function     UpdateEpagesJ
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Update = UpdateEpagesJ($Version, $Source, $hFlags);
#----------------------------------------------------------------------------------------
# example      UpdateEpagesJ('6.14.6', 'Data/j', { start => 1 });
# example      UpdateEpagesJ('6.14.6', 'Data/epagesj.zip');
# example      UpdateEpagesJ('6.14.6');
# example      UpdateEpagesJ();
#----------------------------------------------------------------------------------------
# description  If $Version is set, upgrades/downgrades/installs epagesj.
#               If $Version is unset, installs epagesj unless installed.
#               Does nothing if installed and new version are same.
#               Returns error if upgrade/downgrade/install fails.
#----------------------------------------------------------------------------------------
# input        $Version | version is one of:
#                          - empty : update epagesj to the latest version
#                          - epages version (6.15.0) : upgrade/downgrade to specific
#                            version if exists, else use snapshot version (6.15.0-SNAPSHOT)
#                          - snapshot version (6.15.0-SNAPSHOT) : upgrade/downgrade to
#                            specific snapshot version (only epages internal use) | string
# input        $Source  | download epagesj from source (dir/zip/URL) | string
# input        $hFlags | flags
#                   <ul>
#                       <li>start - 1: start epagesj after update (default); 0 - not
#                   </ul>
#               | ref.hash
#----------------------------------------------------------------------------------------
# return       $Update | true if epagesj was updated, else false | boolean
#========================================================================================
sub UpdateEpagesJ (;$$$) {
    my ($Version, $Source, $hFlags) = @_;

    my $StartEpagesJ = defined $hFlags->{start} ? $hFlags->{start} : 1;

    # on linux throw error unless epagesj user exists
    _GetEpagesJUser();

    my $Installed = GetEpagesJVersion();
    my $DownloadUrl = GetEpagesJZipUrl($Version);
    my $EpagesJZip = basename($DownloadUrl);
    $EpagesJZip = $Version ? "epagesj-$Version.zip" : 'epagesj.zip' if $Source;
    $Source //= $DownloadUrl;

    # skip update if epagesj is installed and new version is unknown
    my $SkipUpdate = $Installed && ! $Version ? 1 : 0;

    if (not $SkipUpdate) {
        Error('SOURCE_NOT_FOUND', {Version=>" for version $Version"}) unless $Source;

        my $Package;
        if (-f "$Source/webapps/epagesj.version") {
            $Package = $Source;
        } else {
            $Package = "$ENV{EPAGES}/$EpagesJZip";
            if (-f $Source) {
                CopyFile($Source, $Package);
            } else {
                print "getstore($Source, $Package)\n";
                my $Result = getstore($Source, $Package);
                Error('DOWNLOAD_FAILED', {Url=>$Source, File=>$EpagesJZip,
                                          Result=>$Result}) if $Result != RC_OK;
            }
        }

        UnpackEpagesJ($Package, $Installed, $Version, {dont_move => 1});

        if ($Installed) {
            my $InstalledBase = $Installed;
            $InstalledBase =~ s/-SNAPSHOT//gs;

            my $JDir = $ENV{EPAGESJ} =~ /\S/ ? $ENV{EPAGESJ} : "$ENV{EPAGES}/j";
            my $InstalledBuild = GetEpagesJBuild();
            my $Build = GetEpagesJBuild("$JDir-$Version");

            # skip update if installed and new version are same
            $SkipUpdate = 1 if $InstalledBuild >= $Build and
                ($InstalledBase eq $Version or $Installed eq $Version);
        }
    }

    if ($SkipUpdate) {
        # on windows, epages-j might be installed, but services not (yet)
        return 0 unless $^O eq 'MSWin32';

        my %ExistingServices;
        foreach (qw(solr-server epagesj-server)) {
            $ExistingServices{$_} = 1 if ExistsService($_);
        }

        # return if both services are installed
        return 0 if scalar keys %ExistingServices == 2;

        # stop existing services
        StopService($_) foreach keys %ExistingServices;

        # install services
        AddEpagesJService();

        # start services
        StartService('epagesj', {verbose => 1}) if $StartEpagesJ;

        return 0;
    }

    my $ServiceExists = ExistsService('epagesj');
    StopService('epagesj', {verbose => 1}) if $ServiceExists;
    _MoveEpagesJ({ old => $Installed, new => $Version});

    my $Etc = -f "$Source/etc/init.d/epagesj" ? "$Source/etc" : undef;
    UpdateEpagesJEtc($Installed, $Version, $Etc);
    AddEpagesJService() unless $ServiceExists;
    StartService('epagesj', {verbose => 1}) if $StartEpagesJ;

    return 1;
}

#========================================================================================
# function     UnpackEpagesJ
# state        public
#----------------------------------------------------------------------------------------
# syntax       UnpackEpagesJ($Package, $OldVersion, $NewVersion, $hFlags);
#----------------------------------------------------------------------------------------
# description  Updates $ENV{EPAGESJ} with package $Package.
#----------------------------------------------------------------------------------------
# input        $Package    | path to epagesj zip file or j/ directory | string
# input        $OldVersion | installed epagesj version | string
# input        $NewVersion | epagesj version to be installed | string
# input        $hFlags | flags
#                   <ul>
#                       <li>dont_move  - if set to 1, old j/ is not overwritten by new version
#                   </ul>
#               | ref.hash
#========================================================================================
sub UnpackEpagesJ ($;$$$) {
    my ($Package, $OldVersion, $NewVersion, $hFlags) = @_;

    my $JDir = $ENV{EPAGESJ} =~ /\S/ ? $ENV{EPAGESJ} : "$ENV{EPAGES}/j";

    my $New = $NewVersion ? $NewVersion : 'new';
    DeleteDirectory("$JDir-$New", { recursive => 1});
    CreateDirectory("$JDir-$New");

    $Package = CorrectPathSlashes($Package);
    my $UnpackDir = CorrectPathSlashes("$JDir-$New");
    if (-d $Package) {
        CopyDirectory($Package, "$JDir-$New/j")
    } else {
        SystemCommand("unzip -qq $Package -d \"$UnpackDir\"", { 'Print' => 1 });
    }

    RunOnGlob("$JDir-$New",'epagesj-*', sub { RenameDirectory($_,'j') });

    Error('DIR_NOT_FOUND', {Dir=>CorrectPathSlashes("$JDir-$New/j")}) unless -d "$JDir-$New/j";

    _MoveEpagesJ({old => $OldVersion, new => $NewVersion}) unless $hFlags->{dont_move};
}

# function moves j to j-old and j-new to j, e.g.: move j j-6.14.2 ; move j-6.14.3 j
sub _MoveEpagesJ ($) {
    my ($hFlags) = @_;

    my ($EpagesJUid, $EpagesJGid) = @{_GetEpagesJUser()};
    my $JDir = $ENV{EPAGESJ} =~ /\S/ ? $ENV{EPAGESJ} : "$ENV{EPAGES}/j";

    my $Old = $hFlags->{old} // 'old';
    my $New = $hFlags->{new} // 'new';

    # note that $Old might be $New (same version, different build)
    DeleteDirectory("$JDir-$Old.tmp", { recursive => 1});
    RenameDirectory($JDir,"$JDir-$Old.tmp",{create_directory => 1});
    RenameDirectory("$JDir-$New/j",$JDir,{create_directory=>1, stat=>'epages'});
    DeleteDirectory("$JDir-$New", { recursive => 1});
    RenameDirectory("$JDir-$Old.tmp","$JDir-$Old", {create_directory => 1});

    _UpdateEpagesJConf("$JDir-$Old", $JDir);
    _SetEpagesJAccessRights($JDir, $EpagesJUid, $EpagesJGid);
}

#========================================================================================
# function     AddEpagesJService
# state        public
#----------------------------------------------------------------------------------------
# syntax       AddEpagesJService();
#----------------------------------------------------------------------------------------
# description  Adds windows epagesj services (does nothing under Linux)
#========================================================================================
sub AddEpagesJService {
    return unless $^O eq 'MSWin32';

    #get Info
    my $CurrentEpagesJava = $ENV{EPAGES_JAVA};
    my $EpagesJava7 = FindProgramFile('Java/jdk1.7*');
    Error('JDK_7_NOT_FOUND') unless -d $EpagesJava7;

    local $ENV{EPAGES_JAVA} = $ENV{EPAGES_JAVA};
    if ($ENV{EPAGES_JAVA} ne $EpagesJava7) {

        #logging
        GetLog()->info("EPAGES_JAVA is not correct, set value temporary, from '$CurrentEpagesJava' to correct value '$EpagesJava7'");

        #update envioronment variable
        SystemCommand(['setx', 'EPAGES_JAVA', $EpagesJava7, '-m'], { 'Print' => 1 });
        $ENV{EPAGES_JAVA} = $EpagesJava7; ## no critic qw(RequireLocalizedPunctuationVars)
    }

    my $JVM = "$ENV{EPAGES_JAVA}/jre/bin/server/jvm.dll";
    Error('FILE_NOT_FOUND', {FileName=>$JVM}) unless -f $JVM;

    #run on ePagesJ folder
    RunOnDirectory($ENV{EPAGESJ} =~ /\S/ ? $ENV{EPAGESJ} : "$ENV{EPAGES}/j", sub {
        local $ENV{PROCESSOR_ARCHITECTURE} = 'AMD64' if Has64BitOS();

        foreach my $Service (qw(solr-server solr-server-slave epagesj-server)) {

            #get status
            my %Status;
            Win32::Service::GetStatus('', $Service, \%Status);

            my $Script = WindowsPath("bin/windows/installWindowsService_$Service.bat");
            Error('FILE_NOT_FOUND', {FileName=>$Script}) unless -f $Script;

            if (defined $Status{CurrentState}){
                #logging
                GetLog()->info("skip script '$Script' to install service '$Service', because it is already installed, status '$Status{CurrentState}'");

                #skip
                next;
            };

            #logging
            GetLog()->info("call script '$Script' to install service '$Service'");

            #call script
            SystemCommand($Script, { 'Print' => 1 });
        }
    });
}

#========================================================================================
# function     AddMongoDBService
# state        public
#----------------------------------------------------------------------------------------
# syntax       AddMongoDBService($MongoBinDir, $MongoDataDir, $hFlags);
#----------------------------------------------------------------------------------------
# description  Adds windows mongodb service (does nothing under Linux)
#----------------------------------------------------------------------------------------
# input        $MongoBinDir | directory that contains the mongodb executables | string
# input        $MongoDataDir | directory that contains the mongodb data | string
# input        $hFlags | flags
#                   <ul>
#                       <li>remove    - uninstall service
#                       <li>reinstall - uninstall service and install again
#                   </ul>
#               | ref.hash
#========================================================================================
sub AddMongoDBService (;$$$) {
    my ($MongoBinDir, $MongoDataDir, $hFlags) = @_;

    return unless $^O eq 'MSWin32';

    $MongoDataDir //= Win32::GetFolderPath(Win32::CSIDL_COMMON_APPDATA()). '\\mongodb';
    $MongoDataDir = WindowsPath($MongoDataDir);

    my $Service = 'mongodb';
    my $MongodCfgFile = WindowsPath("$MongoDataDir/mongod.cfg");
    my (@MongodCfgLines, %hMongodCfg, $MongodExe);

    # read mongod.cfg
    if (-s $MongodCfgFile) {
        foreach (@{GetFileContentLines($MongodCfgFile)}) {
            my ($Key, $Value);
            if (/^\s*(\w+)\s*=\s*(.+?)\s*$/) {
                $Key = $1; $Value = $2;
                $hMongodCfg{$Key} = $Value;
            } elsif (/^\s*#MongodExe\s*=\s*(.+?)\s*$/) {
                $MongodExe = $1;
            } elsif (/\S/) {
                $Value = $_;
            }
            chomp $Value;
            push @MongodCfgLines, { key => $Key, value => $Value};
        }
        if (not defined $hMongodCfg{smallfiles}) {
            $hMongodCfg{smallfiles} = 'true';
            push @MongodCfgLines, { key => 'smallfiles', value => 'true'};
        }
    }

    $MongoBinDir //= FindProgramFile('mongodb-*/bin');
    $MongodExe //= WindowsPath("$MongoBinDir/mongod.exe");

    # define values for mongod.cfg
    if (not -s $MongodCfgFile) {
        push @MongodCfgLines, {value => '# see http://docs.mongodb.org/manual/reference/configuration-options/'};
        push @MongodCfgLines, {value => '#MongodExe: path to mongod.exe (ePages specific)'};

        %hMongodCfg = ( dbpath => WindowsPath("$MongoDataDir/data"),
                        pidfilepath => $MongoDataDir,
                        logpath => WindowsPath("$MongoDataDir/log/$Service.log"),
                        logappend => 'true',
                        directoryperdb => 'true',
                        smallfiles => 'true',
            );

        push @MongodCfgLines, {key => $_, value => $hMongodCfg{$_}} foreach
            qw(dbpath pidfilepath logpath logappend directoryperdb smallfiles);
    }

    return if not -f $MongodExe and $hFlags->{remove};

    Error('FILE_NOT_FOUND', { FileName => $MongodExe}) unless -f $MongodExe;

    my %Status;
    Win32::Service::GetStatus('', $Service, \%Status);

    if (defined $Status{CurrentState}) {
        # do nothing unless remove/reinstall
        return unless scalar keys %$hFlags;
        SystemCommand([$MongodExe, '--remove'], { 'Print' => 1 });
    }
    return if $hFlags->{remove};

    # if mongod was updated, new exe is stored in different (versioned) directory
    $MongodExe = WindowsPath("$MongoBinDir/mongod.exe") if -f "$MongoBinDir/mongod.exe";

    my $Content = '';
    CreateDirectory($hMongodCfg{dbpath}) unless -d $hMongodCfg{dbpath};
    WriteFile($hMongodCfg{logpath}, \$Content, {create_directory=>1}) unless -f $hMongodCfg{logpath};

    # write mongod.cfg
    foreach my $Line (@MongodCfgLines) {
        if (defined $Line->{key}) {
            $Content .= "$Line->{key} = $hMongodCfg{$Line->{key}}\n";
        } else {
            $Content .= "$Line->{value}\n";
        }
    }
    $Content =~ s/\n+$/\n/;
    # add meta comment for path of mongod.exe
    $Content .=  "#MongodExe=$MongodExe\n";

    WriteFile($MongodCfgFile, \$Content, {create_directory=>1});
    SystemCommand([$MongodExe, '--config',  $MongodCfgFile, '--install'], { 'Print' => 1 } );
}

#========================================================================================
# state        internal
#========================================================================================

sub _UpdateEpagesJConf ($$) {
    my ($OldJ, $NewJ) = @_;

    # on win32, do nothing unless $OldJ/conf/epagesj.conf exists
    return if $^O eq 'MSWin32' and not -f "$OldJ/conf/epagesj.conf";

    CopyFile("$NewJ/conf/epagesj.sample.conf","$NewJ/conf/epagesj.conf")
        if -f "$NewJ/conf/epagesj.sample.conf" and not -f "$NewJ/conf/epagesj.conf";
    return unless -f "$NewJ/conf/epagesj.conf";
    return unless -f "$OldJ/conf/epagesj.conf";

    my $OldConf = DE_EPAGES::Core::API::IniConfig->new(
        FileName => "$OldJ/conf/epagesj.conf", CaseSensitive => 1);
    my $NewConf = DE_EPAGES::Core::API::IniConfig->new(
        FileName => "$NewJ/conf/epagesj.conf", CaseSensitive => 1);

    foreach my $Section (@{$OldConf->sectionNames}) {
        my $NewSection = $Section eq 'Server' ? 'EPagesJServer' : $Section;
        foreach my $Key (keys %{$OldConf->section($Section)}) {
            $NewConf->set($NewSection, $Key, $OldConf->get($Section, $Key));
        }
    }
    $NewConf->rewrite();
}

sub _GetEpagesJUser {
    return [] if $^O eq 'MSWin32';

    Error('FILE_CHOWN_FAILED', {Dir => '/etc/init.d/epagesj', User => getpwuid($<)})
        unless $< == 0; # i'm root

    my $EpagesJUid = getpwnam($ENV{EPAGESJ_USER} // 'epagesj');
    Error('NO_EPAGESJ_USER', {User=>$ENV{EPAGESJ_USER} // 'epagesj'})
        unless defined $EpagesJUid;
    my $EpagesJGid = getgrnam('epages');
    Error('NO_EPAGESJ_GROUP', {Group=>'epages'}) unless defined $EpagesJGid;
    return [$EpagesJUid, $EpagesJGid];
}

sub _SetEpagesJAccessRights ($$$) {
    my ($JDir, $EpagesJUid, $EpagesJGid) = @_;
    return if $^O eq 'MSWin32';

    find(sub {
        chown($EpagesJUid, $EpagesJGid, $_) or
            Error('FILE_CHOWN_FAILED', {Dir => $File::Find::name, User => getpwuid($<)});
        chmod(-d $_ ? GROUP_RX : GROUP_R, $_) or
            Error('FILE_CHMOD_FAILED', {Dir => $File::Find::name, User => getpwuid($<)});
         },
         $JDir
        );
}

1;
