#=================================================================================
# package      DE_EPAGES::Installer::API::NodeJS
#---------------------------------------------------------------------------------
# description  Functions for handling epages-nodejs
#---------------------------------------------------------------------------------
# state        public
#=================================================================================
package DE_EPAGES::Installer::API::NodeJS;
use strict;

use base qw (Exporter);

our @EXPORT_OK = qw (
    GetNodeJSVersion
    GetNodeJSZipUrl
    UpdateNodeJS
    UnpackNodeJS
);

use File::Basename qw(basename);
use File::Find qw(find);
use LWP::UserAgent;
use HTTP::Status;
use LWP::Simple;
use Time::HiRes qw(time);

use DE_EPAGES::Core::API::Constants qw(BASE_RW BASE_RWX);
use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::File qw(GetFileContent CopyFile DeleteDirectory CreateDirectory
                                  CorrectPathSlashes CopyDirectory);
use DE_EPAGES::Core::API::SystemCommand qw ( SystemCommand );
use DE_EPAGES::Installer::API::Constants qw (
    INTERN_PUB_CARTRIDGE_PACK_SERVERS
    PUBLIC_CARTRIDGE_PACK_SERVERS
);

#ERRORDEFINITION
sub ErrorDefinition () { return {
    'SOURCE_NOT_FOUND'    => 'epages-nodejs.zip#Version not found',
    'DOWNLOAD_FAILED'     => 'Download #Url to #File failed (return code: #Result).',
    'DIR_NOT_FOUND'       => 'Directory not found (#Dir).',
    'NO_EPAGES_USER'     => 'No epages user "#User" found.',
    'NO_EPAGES_GROUP'    => 'No epages group "#Group" found.',
    'FILE_CHOWN_FAILED'   => 'User "#User" cannot chown "#Dir"',
    'FILE_CHMOD_FAILED'   => 'User "#User" cannot chmod "#Dir"',
                         }
}
#ERRORDEFINITION END

#========================================================================================
# function     GetNodeJSVersion
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Version = GetNodeJSVersion($NodeDir);
#----------------------------------------------------------------------------------------
# description  Get version as defd. in Perl/bin/nodejs.d/epages.version.
#               Returns 0 unless exists.
#----------------------------------------------------------------------------------------
# input        $NodeDir | path to nodejs.d/ (default: $ENV{EPAGES_PERL}/bin/nodejs.d) | string
#----------------------------------------------------------------------------------------
# return       $Version | version of installed epages-nodejs | string
#========================================================================================
sub GetNodeJSVersion (;$) {
    my ($NodeDir) = @_;
    $NodeDir //=  "$ENV{EPAGES_PERL}/bin/nodejs.d";
    return 0 unless -f "$NodeDir/epages.version";

    my $Version = GetFileContent("$NodeDir/epages.version");
    chomp($$Version);
    return $$Version ? $$Version : 0;
}

#========================================================================================
# function     GetNodeJSZipUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DownloadUrl = GetNodeJSZipUrl($Version);
#----------------------------------------------------------------------------------------
# description  Returns URL for epages-nodejs.zip on the download site.
#----------------------------------------------------------------------------------------
# input        $Version | version is either empty, then get the URL for the latest
#                          version, or epages version string as 6.17.0 | string
#----------------------------------------------------------------------------------------
# return       $DownloadUrl | download URL | string
#========================================================================================
sub GetNodeJSZipUrl (;$) {
    my ($Version) = @_;
    $Version //= '';
    my %Speed;

    my @UrlList = (INTERN_PUB_CARTRIDGE_PACK_SERVERS, PUBLIC_CARTRIDGE_PACK_SERVERS);

    if ($Version) {
        unshift @UrlList, map { "$_/pub/epages/build/repo/$Version/latest" }
        qw(ftp://ftp.epages.com);
    }
    unshift @UrlList, split(/\s+/,$ENV{EPAGES_REPO}) if $ENV{EPAGES_REPO};

    my $Latest;
    # first internal resources because they might contain newer versions
    foreach my $Url (@UrlList) {
        # stop if EPAGES_REPO found
        last if $ENV{EPAGES_REPO} and scalar keys %Speed;
        # stop if internal resource found
        last if $Url ~~ PUBLIC_CARTRIDGE_PACK_SERVERS and scalar keys %Speed;

        my $NodeJSZip = $Version ? "epages-nodejs-$Version.zip" : 'epages-nodejs.zip';
        $Latest //= "$Url/windows/3rd-party/32/$NodeJSZip";

        # we must use max_redirect => 0 because http pages may redirect
        # to an error page with RC == 200 if original page does not exist
        my $UserAgent = LWP::UserAgent->new(max_redirect => 0);
        my $time1 = time;
        my $Response = $UserAgent->head("$Url/windows/3rd-party/32/$NodeJSZip");
        next unless $Response->is_success;
        $Speed{"$Url/windows/3rd-party/32/$NodeJSZip"} = time - $time1;
    }
    return $Latest unless scalar keys %Speed;

    my $Fastest = (sort { $Speed{$a} <=> $Speed{$b} } keys %Speed)[0];
    return $Fastest;
}

#========================================================================================
# function     UpdateNodeJS
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Update = UpdateNodeJS($Version, $Source);
#----------------------------------------------------------------------------------------
# example      UpdateNodeJS('6.17.6', 'epages-nodejs.zip');
# example      UpdateNodeJS('6.17.6');
# example      UpdateNodeJS();
#----------------------------------------------------------------------------------------
# description  If $Version is set, upgrades/downgrades/installs epages-nodejs.
#               If $Version is unset, installs epages-nodejs unless installed.
#               Does nothing if installed and new version are same.
#               Returns error if upgrade/downgrade/install fails.
#----------------------------------------------------------------------------------------
# input        $Version | version is one of:
#                          - empty : update epages-nodejs to the latest version
#                          - epages version (6.17.0) : upgrade/downgrade to specific
#                            version if exists, else use snapshot version (6.17.0-SNAPSHOT)
# input        $Source  | download epages-nodejs from source (zip/URL) | string
#----------------------------------------------------------------------------------------
# return       $Update | true if epages-nodejs was updated, else false | boolean
#========================================================================================
sub UpdateNodeJS (;$$) {
    my ($Version, $Source) = @_;

    my $Installed = GetNodeJSVersion();
    my $DownloadUrl = GetNodeJSZipUrl($Version);
    my $NodeJSZip = basename($DownloadUrl);
    $NodeJSZip = $Version ? "epages-nodejs-$Version.zip" : 'epages-nodejs.zip' if $Source;
    $Source //= $DownloadUrl;

    # do nothing if epages-nodejs is installed and new version is unknown
    return 0 if $Installed && ! $Version ? 1 : 0;

    Error('SOURCE_NOT_FOUND', {Version=>" for version $Version"}) unless $Source;

    my $Package = "$ENV{EPAGES_PERL}/bin/$NodeJSZip";;
    if (-f $Source) {
        CopyFile($Source, $Package);
    } else {
        print "getstore($Source, $Package)\n";
        my $Result = getstore($Source, $Package);
        Error('DOWNLOAD_FAILED', {Url=>$Source, File=>$NodeJSZip,
                                  Result=>$Result}) if $Result != RC_OK;
    }

    my $UnpackDir = UnpackNodeJS($Package, $Version, {dont_move => 1});
    return 0 if GetNodeJSVersion("$UnpackDir/bin/nodejs.d") eq $Installed;

    _InstallNodeJS($UnpackDir);
    return 1;
}

#========================================================================================
# function     UnpackNodeJS
# state        public
#----------------------------------------------------------------------------------------
# syntax       UnpackNodeJS($Package, $Version, $hFlags);
#----------------------------------------------------------------------------------------
# description  Updates $ENV{EPAGES_PERL}/bin/nodejs.d with package $Package.
#----------------------------------------------------------------------------------------
# input        $Package | path to epages-nodejs.zip | string
# input        $Version | epages-nodejs version to be installed | string
# input        $hFlags  | flags
#                   <ul>
#                       <li>dont_move  - if set to 1, old bin/ is not overwritten by new version
#                   </ul>
#               | ref.hash
#----------------------------------------------------------------------------------------
# return       $UnpackDir | package was unpacked into $UnpackDir | string
#========================================================================================
sub UnpackNodeJS ($;$$) {
    my ($Package, $Version, $hFlags) = @_;

    my $New = $Version ? $Version : 'new';
    my $UnpackDir = "$ENV{EPAGES_PERL}/bin/nodejs-$New";

    DeleteDirectory($UnpackDir, { recursive => 1});
    CreateDirectory($UnpackDir);

    $Package = CorrectPathSlashes($Package);
    $UnpackDir = CorrectPathSlashes($UnpackDir);
    SystemCommand("unzip -qq $Package -d \"$UnpackDir\"", { 'Print' => 1 });

    Error('DIR_NOT_FOUND', {Dir=>CorrectPathSlashes("$UnpackDir/bin/nodejs.d")}) unless -d "$UnpackDir/bin/nodejs.d";

    _InstallNodeJS($UnpackDir) unless $hFlags->{dont_move};

    return $UnpackDir;
}

#========================================================================================
# state        internal
#========================================================================================

sub _InstallNodeJS ($) {
    my ($SourceDir) = @_;
    _SetNodeJSAccessRights($SourceDir);

    foreach my $NodeJSDir (glob "$SourceDir/bin/nodejs.d/lib/node_modules/*") {
        next unless -d $NodeJSDir;
        $NodeJSDir = "$ENV{EPAGES_PERL}/bin/nodejs.d/lib/node_modules/". basename($NodeJSDir);
        DeleteDirectory($NodeJSDir, {recursive => 1});
    }
    print "CopyDirectory($SourceDir, $ENV{EPAGES_PERL})\n";
    CopyDirectory($SourceDir, $ENV{EPAGES_PERL}, undef, {stat => 'preserve'})
}

sub _GetNodeJSUser {
    return [] if $^O eq 'MSWin32';

    Error('FILE_CHOWN_FAILED', { Dir => "$ENV{EPAGES_PERL}/bin/nodejs.d/",
                                 User => getpwuid($<)}
        ) unless $< == 0; # i'm root

    my $NodeJSUid = getpwnam('ep_appl');
    Error('NO_EPAGES_USER', {User => 'ep_appl'}) unless defined $NodeJSUid;

    my $NodeJSGid = getgrnam('epages');
    Error('NO_EPAGES_GROUP', {Group => 'epages'}) unless defined $NodeJSGid;
    return [$NodeJSUid, $NodeJSGid];
}

sub _SetNodeJSAccessRights ($) {
    my ($NodeDir) = @_;
    return if $^O eq 'MSWin32';

    my ($NodeJSUid, $NodeJSGid) = @{_GetNodeJSUser()};

    find(sub {
        chown($NodeJSUid, $NodeJSGid, $_) or
            Error('FILE_CHOWN_FAILED', {Dir => $File::Find::name, User => getpwuid($<)});
        chmod(-d $_ ? BASE_RWX : BASE_RW, $_) or
            Error('FILE_CHMOD_FAILED', {Dir => $File::Find::name, User => getpwuid($<)});
         },
         $NodeDir
        );
}

1;
