use strict;
use Getopt::Long;
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Core::API::GitHub;

# A test for this script is defined in the infra repo:
# /epages-infra/scripts/epages/test/build/test_pull_request_creation.sh

sub Main {

    my $self = shift;
    local $| = 1;

    my $SourceBranch;
    my $TargetBranch;
    my $Account;
    my $UserName;
    my $UserToken;
    my $PropertiesFile;
    my $Help;
    my $UserCrossRepo;
    my $mergePR = 0;
    my $PullRequestId;
    my $MergeId;


    GetOptions(
        'sourceBranch=s'        => \$SourceBranch,
        'targetBranch=s'        => \$TargetBranch,
        'account=s'             => \$Account,
        'userName=s'            => \$UserName,
        'userToken=s'           => \$UserToken,
        'userCrossRepo=s'       => \$UserCrossRepo,
        'propertiesFile=s'      => \$PropertiesFile,
        'help'                  => \$Help,
        'mergePR'               => \$mergePR,
    );

    usage() if $Help;

    unless( $SourceBranch ) {
        print "missing parameter -sourceBranch\n";
        usage();
    }

    unless( $TargetBranch ) {
        print "missing parameter -targetBranch\n";
        usage();
    }

    unless( $Account ) {
        $Account = $UserName;
    }

    unless( $UserName ) {
        print "missing parameter -userName\n";
        usage();
    }

    unless( $UserToken ) {
        print "missing parameter -userToken\n";
        usage();
    }

    $SourceBranch = $UserCrossRepo.':' . $SourceBranch if(defined $UserCrossRepo);
    my $Title = "Merge generated Patch Files from $SourceBranch";
    my $Repository = 'Cartridges';

    print "Create pull request to merge commits from branch '$SourceBranch' into '$TargetBranch' on account '$Account'.\n";

    my $GitHub = DE_EPAGES::Core::API::GitHub->new($UserName, $UserToken, $Account);

    $PullRequestId = $GitHub->CreatePullRequest($Title, $SourceBranch, $TargetBranch, $Repository);
    die 'Could not create pull request.' unless $PullRequestId;

    # Merge pull request.
    if ($mergePR) {
        for (1 .. 6) {
            if($GitHub->IsMergeable($PullRequestId, $Repository)) {
                my $CommitMessage = 'Merge generated patch files';
                $MergeId = $GitHub->MergePullRequest($CommitMessage, $PullRequestId, $Repository);
                last;
            }
            sleep(5);
        }
        if($GitHub->IsMerged($PullRequestId, $Repository)) {
            print "Pull request '$PullRequestId' successfully merged on GitHub.\n";
        } else {
            die "Pull request '$PullRequestId' on GitHub failed.\n";
        }
    }
}

sub usage {
    print <<END_USAGE;
This script creates a pull request from the given source branch to the given target branch.
It will also merge the pull request directly if you set -mergePR.

Usage:
    perl $0 [options] [flags]

options:
    -sourceBranch     required, the name of the branch where your changes are implemented;
    -targetBranch     required, the name of the branch you want your changes pulled into;
                          This needs to be an existing branch of the <account> repository;
    -account          optional, name of the GitHub accout where the <sourceBranch> and
                          <targetBranch> are located. default: <userName>;
    -userName         required, name of the GitHub account of the user who creates the pull request;
    -userToken        required, personal access token for the corresponding <userName>;
    -userCrossRepo    optional, name of a different GitHub account than <userName> where the <sourceBranch> is located;
    -propertiesFile   optional, the path and name of the properties file, e.g. '/tmp/properties.conf'

flags:
    -help             show the command line options.
    -mergePR          merge the pull request directly.

Examples:
    # Cretae pull request from branch "test/PatchBuildGit" into "develop" on the same GitHub repository
    perl $0 -userName jdoe -userToken 6586090af2a56c -sourceBranch test/PatchBuildGit -targetBranch develop
    # Create pull request across different GitHub repositories
    perl $0 -userName jdoe -userToken 6586090af2a56c -sourceBranch test/PatchBuildGit -targetBranch develop -userCrossRepo ttester
    # Create and merge pull request
    perl $0 -userName jdoe -userToken 6586090af2a56c -sourceBranch test/PatchBuildGit -targetBranch develop -mergePR
    # Create and merge pull request on another repo the one from the user
    perl $0 -userName jdoe -userToken 6586090af2a56c -account eCommerce-fr -sourceBranch test/PatchBuildGit -targetBranch develop -mergePR

END_USAGE
    exit 2;
}

RunScript(
    'Debug' => 0,
    'Message' => 'short',
    'Sub' => \&Main
);

1;
