#!/srv/epages/eproot/Perl/bin/perl

use strict;
use Getopt::Long;
use Time::HiRes qw (time);

use DE_EPAGES::Core::API::File qw( DeleteFile
    CompareDirectoryWithFilelist GetFileSHA1Digest );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Core::API::Net qw(GetFastestEpagesMirror GetFile);
use DE_EPAGES::Core::API::SystemCommand qw(UnpackTar);
use DE_EPAGES::Core::API::IniConfig;

sub Main {

    local $| = 1;

    my $time1 = time;

    #-------------------------------------------------------------------
    # check options
    #-------------------------------------------------------------------

    my %Options = ();
    my @GetOpt = qw(help version=s nooutput verbose);
    GetOptions(\%Options, @GetOpt);

    usage() if $Options{help};
    usage() unless $Options{version};
    delete $Options{nooutput} if $Options{verbose};
    my $Quiet = $Options{nooutput};

    my $EpagesPatches = $ENV{EPAGES_PATCHES};
    $EpagesPatches = "$ENV{EPAGES_SHARED}/Patches"
        if not $EpagesPatches and $ENV{EPAGES_SHARED};
    usage('environment variable not set: EPAGES_PATCHES') unless $EpagesPatches;

    #-------------------------------------------------------------------
    # define variables
    #-------------------------------------------------------------------

    my $WebRootPacks = "$EpagesPatches/WebRootPacks";
    my $Patch = 'patch_'. $Options{version};
    $Patch =~ s/\./_/g;
    my $ExtractedWebRootPacks = "$EpagesPatches/$Patch/ExtractedDataPacks".
        "/ExtractedWebRootPacks.v/$Options{version}";

    my @Version = split(/\./, $Options{version});
    $#Version = 2 if $#Version > 2;
    my $Major = join('.', @Version);

    my $StoreTypeChanges = "epages-$Options{version}-storetype-changes";

    #-------------------------------------------------------------------
    # exit if unpacked package found
    #-------------------------------------------------------------------

    _exit(0, 1, "$ExtractedWebRootPacks/$Major", $time1)
        if -d "$ExtractedWebRootPacks/$Major";

    #-------------------------------------------------------------------
    # download package unless exist
    #-------------------------------------------------------------------

    if (not -f "$WebRootPacks/$StoreTypeChanges.tar.gz") {

        DeleteFile("$WebRootPacks/$StoreTypeChanges.info");

        # get epages-$Options{version}-storetype-changes.tar.gz
        my $Url = GetFastestEpagesMirror("patch/$Options{version}/$StoreTypeChanges.tar.gz");
        print "FOUND: $Url\n";
        print "$StoreTypeChanges: download package: " unless $Quiet;
        GetFile($Url, "$WebRootPacks/$StoreTypeChanges.tar.gz");
        print "DONE\n" unless $Quiet;

        # get epages-common.info saved as epages-$Options{version}-storetype-changes.info
        $Url =~ s,patch/$Options{version}/$StoreTypeChanges.tar.gz,yum/origin/epages-common.info,;
        print "FOUND: $Url\n";
        print "$StoreTypeChanges.info: download info file: " unless $Quiet;
        GetFile($Url, "$WebRootPacks/$StoreTypeChanges.info");
        print "DONE\n" unless $Quiet;
    }

    #-------------------------------------------------------------------
    # test correctness of packed package
    #-------------------------------------------------------------------

    if (-f "$WebRootPacks/$StoreTypeChanges.info") {
        print "$StoreTypeChanges: test correctness of packed package: " unless $Quiet;

        my $IniConfig = DE_EPAGES::Core::API::IniConfig->new
            (FileName => "$WebRootPacks/$StoreTypeChanges.info", CaseSensitive => 1);
        my $ExpectedSha1Sum = $IniConfig->get('tgz-sha1sum', 'storetypes');
        my $FoundSha1Sum = GetFileSHA1Digest("$WebRootPacks/$StoreTypeChanges.tar.gz");

        if (not $ExpectedSha1Sum) {
            print "SKIPPING\nFOUND: $WebRootPacks/$StoreTypeChanges.tar.gz\n" unless $Quiet;
        } elsif ($ExpectedSha1Sum eq $FoundSha1Sum) {
            print "OK\nFOUND: $WebRootPacks/$StoreTypeChanges.tar.gz\n" unless $Quiet;
        } else {
            _exit(1, $Quiet, "$WebRootPacks/$StoreTypeChanges.tar.gz", $time1);
        }
    }

    #-------------------------------------------------------------------
    # unpack package
    #-------------------------------------------------------------------

    print "$StoreTypeChanges: unpack package: " unless $Quiet;
    UnpackTar("$WebRootPacks/$StoreTypeChanges.tar.gz", $ExtractedWebRootPacks);
    print "DONE\n" unless $Quiet;

    #-------------------------------------------------------------------
    # test completeness of unpacked package
    #-------------------------------------------------------------------

    print "$StoreTypeChanges: test completeness of unpacked package: " unless $Quiet;
    # skip test if manifest (file list) does not exist but directory does
    if (not -f "$ExtractedWebRootPacks/$StoreTypeChanges.txt"
        and -d "$ExtractedWebRootPacks/$Major") {
        _exit(2, $Quiet, "$ExtractedWebRootPacks/$Major", $time1);
    } elsif (CompareDirectoryWithFilelist("$ExtractedWebRootPacks/$Major",
                                          "$ExtractedWebRootPacks/$StoreTypeChanges.txt",
                                          { 'depth' => 1, })) {
        _exit(0, $Quiet, "$ExtractedWebRootPacks/$Major", $time1);
    } else {
        _exit(1, $Quiet, "$ExtractedWebRootPacks/$Major", $time1);
    }
}

sub _exit {
    my ($ExitCode, $Quiet, $File, $StartTime) = @_;
    if (not $Quiet) {
        my $Result = $ExitCode == 0 ? 'OK' : 'NO';
        if ( $ExitCode == 2) {
            $Result = 'SKIPPING';
            $ExitCode = 0;
        }
        $File = $ExitCode == 0 ? "FOUND: $File\n" : "FAILURE: $File\n" if $File;
        printf "$Result\n${File}need %.1f seconds\n", time - $StartTime;
    }
    exit $ExitCode;
}

sub usage {
    my ($Msg) = @_;
    print "$Msg\n\n" if $Msg;
    print <<END_USAGE;
Description:
    Downloads epages-*-storetype-changes.tar.gz into \$EPAGES_SHARED/Patches/WebRootPacks/
    and unpacks it into \$EPAGES_SHARED/Patches/patch_*/ExtractedDataPacks/.

Usage:
    perl $0 { -version VERSION | -help }

Options:
    -version       ePages version number which defines the package to
                   download: epages-<version>-storetype-changes.tar.gz
Flags:
    -help          show the command line options
    -nooutput      do not print any output
    -verbose       verbose output

Example:
    \$PERL $0 -version 6.17.17.1
END_USAGE
    exit 2;
}

RunScript(
    'Sub' => \&Main
);

1;
