#========================================================================================
# package      DE_EPAGES::JQuery::API::DictionaryConverter
# state        public
#----------------------------------------------------------------------------------------
# description  This the install handler for all patch install functions.
#========================================================================================
package DE_EPAGES::JQuery::API::DictionaryConverter;
use base qw (Exporter);

use strict;
our @EXPORT_OK = qw (
    ConvertMainDictionary
    ConvertCartridgeDictionary
    ConvertSubDictionary
);
use File::Find;
use Getopt::Long;
use JSON qw( to_json );
use DE_EPAGES::Core::API::File qw ( ExistsDirectory ExistsFile ExistsFileCaseSensitive WriteFile fsplit );
use DE_EPAGES::Dictionary::API::XML::DictionaryImportHandler qw( ImportDictionary );

#========================================================================================
# function     _WantedFiles
# state        private
#----------------------------------------------------------------------------------------
# syntax       _WantedFiles($hFiles, $Language, $Prefix );
# example      _WantedFiles(\%Files, "de", "Dictionary");
#----------------------------------------------------------------------------------------
# description  This function is used by the function File::Find. All dictionary files
#               in the curent directory (and the recursive following directories)
#               will be stored in hFiles.
#----------------------------------------------------------------------------------------
# input        $hFiles | hash to store filenames | ref.hash
# input        $Language | language code | string
# input        $Prefix | file name prefix | string
#========================================================================================
sub _WantedFiles {
    my ($hFiles, $Language, $Prefix) = @_;

    if (     $File::Find::name =~ /$Prefix\.$Language\.xml$/gx
         and $File::Find::name !~ /(?:\/|\\)Patches(?:\/|\\).*\.$Language\.xml$/xg
         and $File::Find::name =~/.\w+(?:-\w+)?\.xml$/) #finds also dialects like 'en-US' or 'pt-BR'
    {
        my $Filename = $File::Find::name;
        $hFiles->{$Filename} = 1;
    }
}

#========================================================================================
# function     _WantedLanguageFiles
# state        private
#----------------------------------------------------------------------------------------
# syntax       _WantedLanguageFiles($hFiles);
# example      _WantedLanguageFiles(\%Files);
#----------------------------------------------------------------------------------------
# description  This function is used by the function File::Find. All xml files
#               in the curent directory (and the recursive following directories)
#               will be stored in hFiles.
#----------------------------------------------------------------------------------------
# input        $hFiles | hash to store filenames | ref.hash
#========================================================================================
sub _WantedLanguageFiles {
    my ($hFiles) = @_;

    if ( $File::Find::name =~ /\.xml$/g)
    {
        my $Filename = $File::Find::name;
        $hFiles->{$Filename} = 1;
    }
}

#========================================================================================
# function     _GetLanguages
# state        private
#----------------------------------------------------------------------------------------
# syntax       _GetLanguages($ahInfo);
# example      _GetLanguages([{'directory' => $Directory1},{'directory' => $Directory1}]);
#----------------------------------------------------------------------------------------
# description  This function looks for language codes at xml files.
#----------------------------------------------------------------------------------------
# input        $ahInfo | hash with directory names | ref.array.hash
# return       $aLanguages | array with language codes | ref.array
#========================================================================================
sub _GetLanguages {
    my ($ahInfo) = @_;

    my %Languages;
    foreach my $hItem (@$ahInfo){
        my $Directory = $hItem->{'directory'};
        if(ExistsDirectory($Directory)){
            my %Files;
            find( sub { _WantedLanguageFiles(\%Files) }, $hItem->{'directory'});
            foreach my $FileName (keys %Files){
                if( $FileName =~ /.+\.(.+)\.xml/ ){
                    $Languages{$1} = 1;
                }
            }
        }
    }

    return [ keys %Languages ];
};

#========================================================================================
# function     _GetDictionary
# state        private
#----------------------------------------------------------------------------------------
# syntax       _GetDictionary($ahInfo);
# example      _GetDictionary([ { 'directory' => $Directory1,
#                                  'prefix'    => 'dictionary',
#                                  'js'        => 0 },
#                                { 'directory' => $Directory2,
#                                  'prefix'    => 'Dictionary',
#                                   'js'        => 1 } ]);
#----------------------------------------------------------------------------------------
# description  This function imports and merges the dictionary xml files and returns them as list of hashes..
#----------------------------------------------------------------------------------------
# input        $ahInfo | array of hashes with directory information
#                           <ul>
#                               <li> directory - directory the dictionary is located (recurses into sub directories)</li>
#                               <li> prefix - prefix for dictionary file name</li>
#                               <li> js - only get translations with attribute JS="1" [true|false]</li>
#                               <li> sub - only get translations for subdictionarys not for Dictionary.xml [true|false]</li>
#                           </ul> | ref.array.hash
# return       $ahDict | dictionary
#                           <ul>
#                               <li> language - language code</li>
#                               <li> filename - name of dictionary file</li>
#                               <li> dict - dictionary hash</li>
#                           </ul> | ref.array.hash.string | ref.array.hash
#========================================================================================
sub _GetDictionary {
    my ($ahInfo) = @_;

    my @Dictionaries;
    my $HasPrefix = 0;

    foreach my $LanguageOrDialect ( @{_GetLanguages($ahInfo)} ){
        my $hDict = undef;

        # If you are using a dialect like 'en-US' or 'pt-BR'
        # we write only the translations of this dialect
        # to the dictionaries of the native language ('en' or 'pt')
        my $Language = $LanguageOrDialect;
        $Language =~ s/^(\w+)(?:-\w+)?$/$1/gx;

        my %Dictionary = ();
        foreach my $hItem (@$ahInfo){
            $HasPrefix = exists $hItem->{'prefix'};
            my %Files;
            find( sub { _WantedFiles(\%Files, $LanguageOrDialect, $hItem->{'prefix'}) }, $hItem->{'directory'});
            foreach my $FileName (keys %Files){
                # do not import the dictionary twice for subdicts
                if (exists $hItem->{'sub'} && $hItem->{'sub'}) {
                    my (undef, $Name) = fsplit ($FileName);
                    next if $Name =~ m/^Dictionary/;
                }
                $hDict = undef unless $HasPrefix;
                $hDict = ExistsFileCaseSensitive($FileName)
                    ? ImportDictionary($FileName, $hDict, $Language, $hItem->{'js'} )
                    : undef;

                $FileName =~ s/.*\/de_epages\/(.+)\.$LanguageOrDialect\.xml$/$1/gx;
                $Dictionary{'filename'} = $FileName if defined $hDict and not $HasPrefix;
                if(scalar keys %$hDict){
                    $Dictionary{'dict'} = $hDict->{$Language};
                    push @Dictionaries, {'language' => $Language, %Dictionary } if not $HasPrefix;
                }
            }
        }
        push @Dictionaries, {'language' => $Language, %Dictionary }  if scalar keys %$hDict and $HasPrefix;
    }

    return \@Dictionaries;
}

#========================================================================================
# function     _SaveToJSON
# state        private
#----------------------------------------------------------------------------------------
# syntax       _SaveToJSON($hInfo);
# example      _SaveToJSON({ 'version'   => $hInfo->{'version'},
#                             'prefix'    => 'dict',
#                             'package'   => $Package,
#                             'cartridgename' => $CartridgeName });
#----------------------------------------------------------------------------------------
# description  This function saves a dictionary as json formatted string to a file.
#----------------------------------------------------------------------------------------
# input       $aDictionaries | dictionary
#                           <ul>
#                               <li> language - language code</li>
#                               <li> filename - name of dictionary file</li>
#                               <li> dict - dictionary hash</li>
#                           </ul> | ref.array.hash.string | ref.array.hash
# input       $hTarget | array of hashes with directory information
#                           <ul>
#                               <li> storetype - type of store</li>
#                               <li> version - version of store type</li>
#                               <li> prefix - prefix for dictionary file name</li>
#                               <li> package - cartrige package (e.g. 'DE_EPAGES')</li>
#                               <li> cartridgename - cartridge name </li>
#                               <li> parentcartridge - cartridge name </li>
#                               <li> parentcartridgeprefix - prefix for dictionary file name </li>
#                               <li> parentpackage - package name </li>
#                               <li> parentpackageprefix - prefix for dictionary file name </li>
#                           </ul> | ref.array.hash
#========================================================================================
sub _SaveToJSON {
    my ($aDictionaries, $hTarget) = @_;

    foreach my $hDict (@$aDictionaries){
        my $Language = $hDict->{'language'};

        my $RootPath = join('/',$ENV{'EPAGES_WEBROOT'},'StoreTypes',$hTarget->{'version'},$hTarget->{'storetype'},'lib');
        my $PathPart = lc($hTarget->{'package'});
        $PathPart .= '/' . lc($hTarget->{'cartridgename'}) if exists $hTarget->{'cartridgename'};
        $PathPart .= '/';
        $PathPart .= lc(exists $hTarget->{'prefix'} ? $hTarget->{'prefix'} : $hDict->{'filename'});
        $PathPart .= ".$Language.json";
        my $Path = "$RootPath/$PathPart";

        # EPG-25401: Build json array with path to the parent dict, current dict content.
        my @DictContent;
        my $ParentDict;
        my $ParentDictCheck;
        # Check if module dict in a specific cartridge.
        if(not exists $hTarget->{'prefix'} and exists $hTarget->{'parentcartridge'} and $hTarget->{'parentcartridgeprefix'}) {
            $ParentDict = lc($hTarget->{'package'}) . '/' . lc($hTarget->{'parentcartridge'}) . "/$hTarget->{'parentcartridgeprefix'}";
            $ParentDictCheck = "$ParentDict.$Language.json";
            # Set ParentDict to undef if the dict is own parent.
            $ParentDict = undef if $ParentDictCheck eq $PathPart;
        }
        # Add global dict as default.
        if(not defined $ParentDict or not ExistsFile($RootPath . '/' . $ParentDictCheck)) {
            if(exists $hTarget->{'parentpackage'} and exists $hTarget->{'parentpackageprefix'}) {
                $ParentDict = lc($hTarget->{'parentpackage'}) . "/$hTarget->{'parentpackageprefix'}";
                $ParentDictCheck = "$ParentDict.$Language.json";
            }
        }
        # Build array.
        push @DictContent, $ParentDict if defined $ParentDict and ExistsFile($RootPath . '/' . $ParentDictCheck);
        push @DictContent, $hDict->{'dict'};
        # Convert perl array to json format.
        my $ToJson = to_json(\@DictContent, { 'pretty' => 1 });
        # Write dict file.
        WriteFile($Path, \$ToJson, { create_directory => 1, 'utf8' => 1, 'encoding' => 'utf-8'} );
    }

    return;
}

#========================================================================================
# function     ConvertMainDictionary
# state        private
#----------------------------------------------------------------------------------------
# syntax       ConvertMainDictionary($hInfo);
# example      ConvertMainDictionary({ 'version'   => '6.11.0',
#                                        'storetype' => 'Store' });
#----------------------------------------------------------------------------------------
# description  This function merges the dictionaries of cartridges JQuery and Dictionary
#               and saves it as json formatted string to a file.
#               Calls the functions _GetDictionary and _SaveToJSON.
#----------------------------------------------------------------------------------------
# input        $hInfo | storetype and version
#                           <ul>
#                               <li> storetype - type of store</li>
#                               <li> version - version of store type</li>
#                           </ul> | ref.array.hash
#========================================================================================
sub ConvertMainDictionary {
    my ($hInfo) = @_;

    my $JQueryDir     = join('/',$ENV{'EPAGES_CARTRIDGES'},'DE_EPAGES/JQuery/Templates');
    my $DictionaryDir = join('/',$ENV{'EPAGES_CARTRIDGES'},'DE_EPAGES/Dictionary/Templates');

    my $aDictionaries = _GetDictionary([ { 'directory' => $JQueryDir,
                                           'prefix'    => 'Dictionary',
                                           'js'        => 0 },
                                         { 'directory' => $DictionaryDir,
                                           'prefix'    => 'Dictionary',
                                           'js'        => 1 } ]);

    #Save: $ENV{'EPAGES_WEBROOT'}/StoreTypes/#Version#/#StoreType#/lib/ep/dict.#Language#.json;
    return _SaveToJSON($aDictionaries, {
        'storetype' => $hInfo->{'storetype'},
        'version'   => $hInfo->{'version'},
        'prefix'    => 'dict',
        'package'   => 'ep',
    });
}

#========================================================================================
# function     ConvertCartridgeDictionary
# state        private
#----------------------------------------------------------------------------------------
# syntax       ConvertCartridgeDictionary($hInfo);
# example      ConvertCartridgeDictionary({ 'version'   => '6.11.0',
#                                             'storetype' => 'Store'
#                                             'cartridge' => 'DE_EPAGES::Design' });
#----------------------------------------------------------------------------------------
# description  This function merges the cartridges' dictionaries files located at
#               'Data/Public/lib/de_epages' and 'Templates'
#               and saves it as json formatted string to a file.
#               Calls the functions _GetDictionary and _SaveToJSON.
#----------------------------------------------------------------------------------------
# input        $hInfo | storetype and version
#                           <ul>
#                               <li> storetype - type of store</li>
#                               <li> version - version of store type</li>
#                               <li> cartridge - cartridge to process</li>
#                           </ul> | ref.array.hash
#========================================================================================
sub ConvertCartridgeDictionary {
    my ($hInfo) = @_;

    my $Cartridge = $hInfo->{'cartridge'};
    my ($Package, $CartridgeName) = split(/::/, $Cartridge);
    my $Directory1 = join('/',$ENV{'EPAGES_CARTRIDGES'},$Package, $CartridgeName,'Data/Public/lib/de_epages');
    my $Directory2 = join('/',$ENV{'EPAGES_CARTRIDGES'},$Package, $CartridgeName,'Templates');

    my $aDictionaries = _GetDictionary([ { 'directory' => $Directory1,
                                           'prefix'    => 'Dictionary',
                                           'js'        => 0 },
                                         { 'directory' => $Directory2,
                                           'prefix'    => 'Dictionary',
                                           'js'        => 1 } ]);

    #Save: $ENV{'EPAGES_WEBROOT'}/StoreTypes/#Version#/#StoreType#/lib/de_epages/#CartridgeName#/dictionary.#Language#.json;
    return _SaveToJSON($aDictionaries, {
        'storetype' => $hInfo->{'storetype'},
        'version'   => $hInfo->{'version'},
        'prefix'    => 'dictionary',
        'package'   => $Package,
        'cartridgename' => $CartridgeName,
        'parentpackage' => 'ep',
        'parentpackageprefix' => 'dict',
    });
}

#========================================================================================
# function     ConvertSubDictionary
# state        private
#----------------------------------------------------------------------------------------
# syntax       ConvertSubDictionary($hInfo);
# example      ConvertSubDictionary({ 'version'   => '6.11.0',
#                                       'storetype' => 'Store'
#                                       'cartridge' => 'DE_EPAGES::Design' });
#----------------------------------------------------------------------------------------
# description  This function saves the cartridges' dictionaries files located at
#               'Data/Public/lib/de_epages' (without prefix)
#               and saves it as json formatted string to a file.
#               Calls the functions _GetDictionary and _SaveToJSON.
#----------------------------------------------------------------------------------------
# input        $hInfo | storetype and version
#                           <ul>
#                               <li> storetype - type of store</li>
#                               <li> version - version of store type</li>
#                               <li> cartridge - cartridge to process</li>
#                           </ul> | ref.array.hash
#========================================================================================
sub ConvertSubDictionary {
    my ($hInfo) = @_;

    my $Cartridge = $hInfo->{'cartridge'};
    my ($Package, $CartridgeName) = split(/::/, $Cartridge);
    my $Directory = join('/',$ENV{'EPAGES_CARTRIDGES'},$Package, $CartridgeName,'Data/Public/lib/'.lc($Package));

    # mark as subdir (sub => 1)
    my $aDictionaries = _GetDictionary([ { 'directory' => $Directory, 'js' => 0, 'sub' => 1 } ]);

    #Save: $ENV{'EPAGES_WEBROOT'}/StoreTypes/#Version#/#StoreType#/lib/de_epages/#PathToCartridgeName#/#FileName#.#Language#.json;
    return _SaveToJSON($aDictionaries, {
        'storetype' => $hInfo->{'storetype'},
        'version'   => $hInfo->{'version'},
        'package'   => $Package,
        'parentcartridge' => $CartridgeName,
        'parentcartridgeprefix' => 'dictionary',
        'parentpackage' => 'ep',
        'parentpackageprefix' => 'dict',
    });
}

1;
