#!/usr/bin/env node
var spawn = require('child_process').spawn,
    path = require('path'),
    phantomjs = require('phantomjs'),
    phantomPath = phantomjs.path,
    testrunner = require('./testrunner'),
    blacklist = require('./blacklist'),

    version,
    storetype,
    vm,
    threads = 4,
    testFilesPaths = [],
    pathToTest,
    runtime,
    timeout = 30,
    debug = false,
    regex = /\..*?\/(ep|de_epages|ckeditor)\//g,
    ssl = false;


//prints instroctions if given parameters are invalid
if(!testrunner.areParamsValid(process.argv)) {
    testrunner.printInstructions();
    process.exit(0);
}

//set arguments for test
version = process.argv[process.argv.indexOf('-v') + 1];
storetype = process.argv[process.argv.indexOf('-storetype') + 1];
vm = process.argv[process.argv.indexOf('-vm') + 1];

if (process.argv.indexOf('-threads') !== -1) {
    threads = process.argv[process.argv.indexOf('-threads') + 1];
}

if (process.argv.indexOf('-timeout') !== -1) {
    timeout = process.argv[process.argv.indexOf('-timeout') + 1];
}

if (process.argv.indexOf('-debug') !== -1) {
    debug = true;
}

if (process.argv.indexOf('-ssl') !== -1) {
    ssl = true;
}

//get main folder holding tests depending on given parameters
pathToTest = testrunner.getPathToTest(process.argv);

//fill array with pathes to all tests
testFilesPaths = testrunner.getTestFiles([pathToTest]);

// filter blacklisted items
testFilesPaths = testFilesPaths.filter(function(path) {
    return !blacklist.some(function(blacklistedTest) {

        return (path.indexOf(blacklistedTest) !== -1);
    });
});

// recursive testrunner
function runTestsRecursively(testArray, concurrency) {
    var failedTests = [],
        outtimedTests = [],
        runningTests = 0,
        currentTest = 0;

    console.log('---------------------------------------------------------------------------');
    console.log('Running ' + testArray.length + ' Tests.');
    console.log('---------------------------------------------------------------------------');

    runtime = new Date().getTime();

    return new Promise((resolve, reject) => {
        function runTest(testFilePath) {

            // Arguments for the phantom script
            var processArgs = [
                    path.join(__dirname, 'testJSphantom.js'),
                    ssl ? 'https://' : 'http://' + vm + '/WebRoot/StoreTypes/' + version + '/' + storetype + '/lib/testsuite/qunit.html?test=' + testFilePath,
                    timeout
                ],
                test;


            currentTest++;
            // spawn additional testprocesses, if max thread number not yet reached
            if (runningTests < concurrency && currentTest < testArray.length) {
                runningTests++;
                runTest(testArray[currentTest]);
            }

            // create test process
            test = spawn(phantomPath, processArgs);

            test.stdout.on('data', (data) => {
                // if no JSON-Object, just print the output
                try {
                    var result = JSON.parse(data.toString().split('\n')[0]);
                } catch (e) {
                    if (debug) {
                        console.log(data.toString().replace('\n', ''));
                    }
                    return;
                }

                // print result
                if (result.timeout) {
                    outtimedTests.push(result.name);
                    console.log('Failed test ' + result.name.replace(regex, '') + ' due to timeout.');
                } else {
                    if (result.success) {
                        console.log('Completed all ' + result.tests + ' assertions for test ' + result.name.replace(regex, '') + '.');
                    } else {
                        failedTests.push(result.name)
                        console.log('Failed ' + result.fails + ' out of ' + result.tests + ' assertions for test ' + result.name.replace(regex, '') + '.');
                    }
                }


                // if all tests are run and no process is left running output result, otherwise start with next test.
                if (currentTest >= testArray.length) {
                    runningTests--;
                    if (runningTests === 0) {
                        if (failedTests.length + outtimedTests.length) {
                            reject([failedTests, outtimedTests]);
                        } else {
                            resolve();
                        }
                    }
                } else {
                    runTest(testArray[currentTest]);
                }
            });

            // print errors
            test.stderr.on('data', (data) => {
                console.log(data.toString().split('\n')[0]);
            });
        }

        // kickoff testing with first test
        runningTests++;
        runTest(testArray[currentTest]);
    });
}

// there are no tests in the array, quit with code -1
if (testFilesPaths.length === 0) {
    process.exit(-1);
}

// start tests;
runTestsRecursively(testFilesPaths, threads)
    .then(() => {
        // on success exit with number of completed tests
        var myRuntime = Math.floor((new Date().getTime() - runtime) / 1000);
        console.log('\n---------------------------------------------------------------------------\n--------------------------------Results------------------------------------\n---------------------------------------------------------------------------\n');
        console.log('Completed ' + testFilesPaths.length + ' tests successfully in ' + myRuntime + ' seconds.');
        process.exit(testFilesPaths.length);
    })
    .catch((data) => {
        console.log('\n---------------------------------------------------------------------------\n--------------------------------Results------------------------------------\n---------------------------------------------------------------------------\n');
        if (Array.isArray(data)) {
            var errors = data[0],
                timeouts = data[1];

            // on error exit with code 0
            var myRuntime = Math.floor((new Date().getTime() - runtime) / 1000);
            console.log('Completed ' + testFilesPaths.length + ' tests in ' + myRuntime + 'seconds with ' + (errors.length + timeouts.length ) + ' failing tests:\n');

            if (errors.length) {
                console.log(errors.join('\n').replace(regex, ''));
            }

            if (timeouts && timeouts.length > 0) {
                console.log('\nThe following tests failed due to timeout:\n');
                console.log(timeouts.join('\n').replace(regex, ''));
            }

            process.exit(0);
        } else {
            console.log('Process crashed :', errors);
            process.exit(-1);
        }
    });
