#========================================================================================
# package      DE_EPAGES::MultiStoreCommon::API::Synchronization
# base         Exporter
# state        public
# description  MultiStoreCommon - Synchronization module
#========================================================================================
package DE_EPAGES::MultiStoreCommon::API::Synchronization;
use base Exporter;

use strict;
use feature 'say';
use feature 'switch';
use Config;
use XML::Simple;

use DE_EPAGES::Core::API::Error qw (
    HandleError
);
use DE_EPAGES::Core::API::File qw (
    ConvertPathChars
    ExistsFile
    WriteFile
    DeleteFile
);
use DE_EPAGES::Core::API::PerlTools qw (
    GetCurrentDateTime
    SubtractDuration
);
use DE_EPAGES::Trigger::API::Trigger qw (
    TriggerHook
);

use DE_EPAGES::MultiStoreCommon::API::MultiStoreLog qw (
    MultiStoreLogDebug
);
use DE_EPAGES::MultiStoreCommon::API::Table::ObjectUpdate qw (
    GetInfoObjectUpdate
    DeleteFromObjectUpdateByShopIDUpdateTime
);

our @EXPORT_OK = qw (
    CleanupTable
    GetCategoryObjXML
    GetCategoryObjXMLType
    GetDefaultValues
    GetLastUpdateID
    GetLastUpdates
    GetObjectPathToIdMap
    GetSyncPath
    RunClocked
    SetLastUpdateID
    SetPeriodicUpdateID
    TransformExportXML
    GetBundleProductMapsFileName
);

#Flag indicating if the current process is the ex/import script.
our $IS_MULTISTORE_EXIMPORT = 0;
our $PRINT_OUTPUT = 0;

#========================================================================================
# function     GetDefaultValues
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $hValues = GetDefaultValues();
#----------------------------------------------------------------------------------------
# description  Triggers hook 'API_MS_GetDefaultValues' and returns the result.
#----------------------------------------------------------------------------------------
# return       $hValues | default values | ref.hash
#========================================================================================
sub GetDefaultValues {

    my %Values;
    TriggerHook('API_MS_GetDefaultValues', { 'Values' => \%Values });
    return \%Values;
}

#========================================================================================
# function     CleanupTable
# state        public
#----------------------------------------------------------------------------------------
# syntax       CleanupTable($Shop)
#----------------------------------------------------------------------------------------
# description  Deletes each record from table ObjectUpdate older than MS_UPDATE_DAYS.
#----------------------------------------------------------------------------------------
# input        $Shop | shop object | DE_EPAGES::Shop::API::Object::Shop
#========================================================================================
sub CleanupTable {
    my ($Shop) = @_;

    say '=> Cleanup table ObjectUpdate by DateTime.' if $PRINT_OUTPUT;
    my $UpdateTime = SubtractDuration(GetCurrentDateTime(), 'days', GetDefaultValues()->{'MS_UPDATE_DAYS'});
    DeleteFromObjectUpdateByShopIDUpdateTime( $Shop->id, $UpdateTime->datetime );
}

#========================================================================================
# function     GetSyncPath
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $SyncPath = GetSyncPath($StoreName,$ShopName);
#----------------------------------------------------------------------------------------
# description  returns the path to the sync folder for the shop.
#----------------------------------------------------------------------------------------
# input        $StoreName | store name | String
# input        $ShopName | shop name | String
# return       $SynPath | sync path | String
#========================================================================================
sub GetSyncPath {
    my ($StoreName,$ShopName) = @_;

    my $SyncPath = GetDefaultValues()->{'MS_SYNC_PATH'};
    $SyncPath =~ s/#StoreName/$StoreName/ig;
    $SyncPath =~ s/#ShopName/$ShopName/ig;
    return $SyncPath;
}

#========================================================================================
# function     GetBundleProductMapsFileName
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $BundleProductMapsFileName = GetBundleProductMapsFileName($StoreName,$ShopName);
#----------------------------------------------------------------------------------------
# description  returns the file name to the bundle product maps file for the shop.
#----------------------------------------------------------------------------------------
# input        $StoreName | store name | String
# input        $ShopName | shop name | String
# return       $BundleProductMapsFileName | bundle product maps file name | String
#========================================================================================
sub GetBundleProductMapsFileName {
    my ($StoreName,$ShopName) = @_;

    my $SyncPath = GetSyncPath($StoreName,$ShopName);

    return $SyncPath . '/BundleProductMaps.json';
}

#========================================================================================
# function     GetObjectPathToIdMap
# state        public
#----------------------------------------------------------------------------------------
# syntax       my %Map = GetObjectPathToIdMap($ahInfoObjects);
# example      my %Map = GetObjectPathToIdMap( LoadClassByAlias('ObjectUpdateID')->infoObjects );
#----------------------------------------------------------------------------------------
# description  Maps objects paths to a list of associate objects ids.
#----------------------------------------------------------------------------------------
# input        $ahInfoObjects | object info | ref.array.hash
# return       %ObjectPathToIdMap | path to id map | hash
#========================================================================================
sub GetObjectPathToIdMap {
    my ($ahInfoObjects) = @_;

    my %ObjectIdToPathMap = map {
        $_->{'ObjectUpdateID'} => $_->{'ObjectPath'}
    } @$ahInfoObjects;
    my %ObjectPathToIdMap;
    foreach my $Id (keys %ObjectIdToPathMap) {
        my $Path = $ObjectIdToPathMap{$Id};
        if( exists $ObjectPathToIdMap{$Path} ){
            push @{ $ObjectPathToIdMap{$Path} }, $Id
        } else {
            $ObjectPathToIdMap{$Path} = [ $Id ];
        }
    }

    return %ObjectPathToIdMap;
}

#========================================================================================
# function     GetLastUpdates
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $ahLastUpdates = GetLastUpdates($Shop, $LastUpdateID_LB, $LastUpdateID_UB);
#----------------------------------------------------------------------------------------
# description  Returns the info list of recently updated objects.
#----------------------------------------------------------------------------------------
# input        $Shop | shop object | DE_EPAGES::Shop::API::Object::Shop
# input        $LastUpdateID_LB | (optional) lower bound - starting number for updates to look for;
#                                if ommited, gets the value from GetLastUpdateID() | Integer
# input        $LastUpdateID_UB | (optional) upper bound - last record to look for
#                                   even if it is not the last one at all | Integer
# return       $ahLastUpdates | List of recently updated objects | ref.array.hash
#========================================================================================
sub GetLastUpdates {
    my ($Shop, $LastUpdateID_LB, $LastUpdateID_UB) = @_;

    $LastUpdateID_LB //= GetLastUpdateID($Shop);
    return GetInfoObjectUpdate( $Shop->id, $LastUpdateID_LB, $LastUpdateID_UB );
}

#========================================================================================
# function     GetLastUpdateID
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $LastUpdateID = GetLastUpdateID($Shop, $ShopAlias);
#----------------------------------------------------------------------------------------
# description  Returns the value of shop attribute 'MS_LastUpdate' or '0' for default.
#----------------------------------------------------------------------------------------
# input        $Shop | shop object | DE_EPAGES::Shop::API::Object::Shop
# input        $ShopAlias | (optional) shop alias; if ommited, $Shop->alias is used. | String
# return       $LastUpdateID | last update id | Integer
#========================================================================================
sub GetLastUpdateID {
    my ($Shop, $ShopAlias) = @_;

    my $Value = 0;
    my $hLastUpdate = $Shop->get('MS_LastUpdate');
    my $Key = $ShopAlias // $Shop->alias;
    if(defined $hLastUpdate && exists $hLastUpdate->{$Key}){
        $Value = $hLastUpdate->{$Key};
    }

    return $Value;
}

#========================================================================================
# function     SetLastUpdateID
# state        public
#----------------------------------------------------------------------------------------
# syntax       SetLastUpdateID($Shop, $MS_LastUpdate, $ShopAlias);
#----------------------------------------------------------------------------------------
# description  Sets a value for shop attribute 'MS_LastUpdate'.
#----------------------------------------------------------------------------------------
# input        $Shop | shop object | DE_EPAGES::Shop::API::Object::Shop
# input        $MS_LastUpdate | value to set | Integer
# input        $ShopAlias | (optional) shop alias; if ommited, $Shop->alias is used. | String
#========================================================================================
sub SetLastUpdateID {
    my ($Shop, $MS_LastUpdate, $ShopAlias) = @_;

    my $hLastUpdate = $Shop->get('MS_LastUpdate');
    $hLastUpdate->{$ShopAlias // $Shop->alias} = $MS_LastUpdate;
    $Shop->set({ 'MS_LastUpdate' => $hLastUpdate });
}

#========================================================================================
# function     SetPeriodicUpdateID
# state        public
#----------------------------------------------------------------------------------------
# syntax       SetPeriodicUpdateID($Shop, $ProcessedUpdatesCount, $ObjectUpdateID, %ScheduledUpdateIDs, $ShopAlias);
#----------------------------------------------------------------------------------------
# description  Calls function 'SetLastUpdateID' depending on DE_EPAGES::MultiStoreCommon::API::Constants::MS_LASTUPDATE_SAVE_PERIOD.
#----------------------------------------------------------------------------------------
# input        $Shop | shop object | DE_EPAGES::Shop::API::Object::Shop
# input        $ProcessedUpdatesCount | current counter of iteration in update process | Integer
# input        $ObjectUpdateID | value of current update id | Integer
# input        $hScheduledUpdateIDs | all update ids that have not been processed yet | Integer
# input        $ShopAlias | (optional) shop alias; if ommited, $Shop->alias is used. | String
#========================================================================================
sub SetPeriodicUpdateID {
    my ( $Shop, $ProcessedUpdatesCount, $ObjectUpdateID, $hScheduledUpdateIDs, $MasterShopAlias) = @_;

    if( scalar keys %$hScheduledUpdateIDs && !( $ProcessedUpdatesCount % GetDefaultValues()->{'MS_LASTUPDATE_SAVE_PERIOD'} ) ){
        my @SortLeftUpdateIDs = sort keys %$hScheduledUpdateIDs;
        SetLastUpdateID( $Shop, $SortLeftUpdateIDs[0], $MasterShopAlias ) ;
    }
}

#========================================================================================
# function     TransformExportXML
# state        public
#----------------------------------------------------------------------------------------
# syntax       TransformExportXML($TagName, $hAttributes)
#----------------------------------------------------------------------------------------
# description  Transformation of some attributes during export of an object.
#----------------------------------------------------------------------------------------
# input        $TagName | tag name | String
# input        $hAttributes | attributes of the tag | ref.hash
#========================================================================================
sub TransformExportXML {
    my ( $TagName, $hAttributes ) = @_;

    my $hDefaultValues = GetDefaultValues();
    delete $hAttributes->{'GUID'} if $hDefaultValues->{'EXPORT_PARAMS'}{'withoutguids'};
    delete $hAttributes->{'CreationDate'} if $hDefaultValues->{'EXPORT_PARAMS'}{'withoutcreationdate'};


    #SubShop objects don't have these attributes.
    delete $hAttributes->{'IsAffiliateVisible'};
    delete $hAttributes->{'HasSelectiveAffiliateVisibility'};

    #Causes errors but is not necessary since objects are imported into their parent objects directly.
    delete $hAttributes->{'Parent'} if exists $hAttributes->{'Parent'};

    TriggerHook('API_MS_TransformExportXML', {
        'TagName'     => $TagName,
        'hAttributes' => $hAttributes,
    });

    return;
}

#========================================================================================
# function     RunClocked
# state        public
#----------------------------------------------------------------------------------------
# syntax       RunClocked($Store,$Shop,$cSub);
#----------------------------------------------------------------------------------------
# description  Runs the given code for this shop in a clocked mode. The process can be startet
#               several times, but ensures that the current shop is processed only once at the same time.
#----------------------------------------------------------------------------------------
# input        $Store | store name | String
# input        $Shop | shop object | DE_EPAGES::Shop::API::Object::Shop
# input        $cSub | code reference | ref.code
#========================================================================================
sub RunClocked {
    my ($Store,$Shop,$cSub) = @_;

    my $Token = GetDefaultValues()->{'MS_SYNC_FOLDER'} .'/'. ConvertPathChars($Store) . '_' . ConvertPathChars($Shop->alias);

    MultiStoreLogDebug("RunClocked(Store: '$Store', Shop: '".$Shop->alias.'\')');
    MultiStoreLogDebug("RunClocked() TOKEN: $Token");

    # Test if sync process is already running for this shop.
    if (not ExistsFile($Token)) {
        # Start synchronization..
        my $Content = $$;
        WriteFile($Token, \$Content, { 'create_directory' => 1, 'flock'=> 1 } );
        eval {
            $cSub->()
        };
        HandleError(
            OnError => sub {
                my ($Error) = @_;
                MultiStoreLogDebug("UNEXPECTED ERROR\n\n".$Error->longMessage()."\n\n");
            },
            Finally => sub {
                # Delete the token in any case.
                DeleteFile($Token);
            }
        );
    } else {
        MultiStoreLogDebug("RunClocked() SKIPPING SYNC: Token file already exists > $Token");
    }

    MultiStoreLogDebug("RunClocked() END");

    return;
}

#========================================================================================
# function     GetCategoryObjXML
# state        public
#----------------------------------------------------------------------------------------
# syntax       GetCategoryObjXML($hXML);
#----------------------------------------------------------------------------------------
# description  Removes the CategoryProductMap from a category object data hash
#----------------------------------------------------------------------------------------
# input        $hXML | Category object data hash including CategoryProductMap; input from XMLin | ref.hash
# return       $hXML | Category object data hash without CategoryProductMap | ref.hash
#========================================================================================
sub GetCategoryObjXML {
    my ($hXML) = @_;

    delete $hXML->{'epages'}{ GetCategoryObjXMLType($hXML) }{'CategoryProductMap'};
    return $hXML;
}

#========================================================================================
# function     GetCategoryObjXMLType
# state        public
#----------------------------------------------------------------------------------------
# syntax       GetCategoryObjXMLType($hXML);
#----------------------------------------------------------------------------------------
# description  Returns the type of a category given by a category object data hash
#----------------------------------------------------------------------------------------
# input        $hXML | Category object data hash; input from XMLin | ref.hash
# return       $hXML | Category type, either 'Category' or 'CategorySpecialOffer' | string
#========================================================================================
sub GetCategoryObjXMLType {
    my ($hXML) = @_;

    return
        exists $hXML->{'epages'}{'CategorySpecialOffer'}
        ? 'CategorySpecialOffer'
        : 'Category';
}

1;
