#========================================================================================
# package      DE_EPAGES::MultiStoreMaster::API::Synchronization
# base         Exporter
# state        public
# description  MultiStoreMaster - Synchronization module
#========================================================================================
package DE_EPAGES::MultiStoreMaster::API::Synchronization;
use base Exporter;

use strict;

use feature 'say';
use File::Temp qw (
    tempdir
    tempfile
);
use JSON::XS qw(
    encode_json
    decode_json
);
use XML::Simple;

use DE_EPAGES::Core::API::Error qw(
    Error
);
use DE_EPAGES::Core::API::File qw (
    fsplit
    WriteFile
    GetFileContent
    ExistsFile
    DeleteFile
);
use DE_EPAGES::Object::API::XML::ObjectExportDriver qw (
    ExportObjects
);
use DE_EPAGES::Object::API::Factory qw (
    JoinObjectPath
    LoadClassByAlias
    LoadRootObject
    SplitObjectPath
);
use DE_EPAGES::Trigger::API::Trigger qw (
    TriggerHook
);
use DE_EPAGES::WebService::API::Client;

use DE_EPAGES::MultiStoreCommon::API::ObjectUpdate qw (
    InsertObjectUpdate
    ObjectPathToFilePath
    TestObjectClass
);
use DE_EPAGES::MultiStoreCommon::API::MultiStoreLog qw (
    MultiStoreLogDebug
);
use DE_EPAGES::MultiStoreCommon::API::Synchronization qw (
    GetBundleProductMapsFileName
    GetSyncPath
    GetDefaultValues
    GetLastUpdates
    GetLastUpdateID
    SetLastUpdateID
    SetPeriodicUpdateID
    TransformExportXML
);
use DE_EPAGES::MultiStoreCommon::API::Table::ObjectUpdate qw (
    DeleteFromObjectUpdateByShopIDObjectUpdateID
);
use DE_EPAGES::Database::API::Config qw ( DatabaseConfig );

our @EXPORT_OK = qw (
    ExportMasterShop
    InitObjectUpdate
);

our $PRINT_OUTPUT = 0;

#========================================================================================
# function     InitObjectUpdate
# state        public
#----------------------------------------------------------------------------------------
# syntax       InitObjectUpdate($StoreName, $Shop)
#----------------------------------------------------------------------------------------
# description  Tags each ContentObject (of class registered in CLASS_FOLDER_MAP) to be exported.
#----------------------------------------------------------------------------------------
# input        $StoreName | store name | String
# input        $Shop | shop object | DE_EPAGES::Shop::API::Object::Shop
#========================================================================================
sub InitObjectUpdate {
    my ($StoreName, $Shop) = @_;

    say '=> Init table ObjectUpdate..' if $PRINT_OUTPUT;

    my $hDefaultValues = GetDefaultValues();
    my $ProductClass = $Shop->get('ProductClass');
    InsertObjectUpdate( $ProductClass ) if TestObjectClass($ProductClass, $hDefaultValues->{'MASTER_CLASSES_LISTENED_TO'});

    my %ClassFolderMap = %{GetDefaultValues()->{'CLASS_FOLDER_MAP'}};

    my $aMasterClassesListenedTo = $hDefaultValues->{'MASTER_CLASSES_LISTENED_TO'};
    foreach my $ClassAlias ( sort { $b cmp $a } keys %ClassFolderMap) {
        foreach my $Class ($ClassFolderMap{$ClassAlias}) {
            foreach my $ClassKey (@$Class) {
                my $iChildren = $Shop->folder( $ClassKey )->childrenIterator;
                while ( my $Child = <$iChildren> ) {
                    InsertObjectUpdate($Child) if TestObjectClass($Child, $aMasterClassesListenedTo);
                    TriggerHook('API_MS_InitObjectUpdate',{ 'Object' => $Child });
                }
            }
        }
    }

}

#========================================================================================
# function     ExportMasterShop
# state        public
#----------------------------------------------------------------------------------------
# syntax       ExportMasterShop($StoreName, $Shop)
#----------------------------------------------------------------------------------------
# description  Exports OverwritableAttributes and ContentObjects from (Master)Shop.
#----------------------------------------------------------------------------------------
# input        $StoreName | store name | String
# input        $Shop | shop object | DE_EPAGES::Shop::API::Object::Shop
#========================================================================================
sub ExportMasterShop {
    my ($StoreName,$Shop) = @_;

    MultiStoreLogDebug('ExportMasterShop() START');

    say '=> Exporting MasterShop..' if $PRINT_OUTPUT;

    _ExportOverwritableAttributes($StoreName,$Shop);
    _ExportContentObjects($StoreName,$Shop);
    _ExportBundleProductMaps($StoreName,$Shop);

    MultiStoreLogDebug('ExportMasterShop() END');
}

sub _ExportOverwritableAttributes {
    my ($StoreName,$Shop) = @_;

    MultiStoreLogDebug('_ExportOverwritableAttributes() START');

    say '-> overwritable attributes' if $PRINT_OUTPUT;
    my $Class = LoadClassByAlias('AffiliateOverwritableMap');
    my $ahObjectInfo = $Class->infoObjects('ByShopID',[ $Shop->id ]);

    my %Map;
    foreach my $hItem (@$ahObjectInfo){
        my $ClassName = $hItem->{'ClassName'};
        my $AttributeName = $hItem->{'AttributeName'};
        if ( exists $Map{$ClassName} and ref $Map{$ClassName} eq 'ARRAY' ) {
            push @{$Map{$ClassName}}, $AttributeName;
        }
        else {
            $Map{$ClassName} = [ $AttributeName ];
        }
    }

    MultiStoreLogDebug('_ExportOverwritableAttributes() DATA:', \%Map);

    my $Content = encode_json(\%Map);
    my $FilePath = join('/', GetSyncPath($StoreName,$Shop->alias), GetDefaultValues()->{'MS_OVERWRITABLES_INFO'});
    MultiStoreLogDebug('_ExportOverwritableAttributes() OutputFile: '.$FilePath);
    WriteFile($FilePath, \$Content, { 'create_directory' => 1, 'flock'=> 1, 'utf8' => 1 } );

    MultiStoreLogDebug('_ExportOverwritableAttributes() END');

    return;
}

sub _WebServiceClient {
    my ($MasterShop, $SiteName, $ServiceUrn) = @_;

    my $hWebServiceConfig = $MasterShop->get( [qw( MS_WebServiceURL MS_WebServiceUser MS_WebServicePassword )] );

    my $WebServiceURL = $hWebServiceConfig->{'MS_WebServiceURL'}{$SiteName};
    my $WebServiceUser = $hWebServiceConfig->{'MS_WebServiceUser'}{$SiteName};
    my $WebServicePassword = $hWebServiceConfig->{'MS_WebServicePassword'}{$SiteName};

    Error('MissingWebServiceConfig', {
        Site => $SiteName,
        WebServiceURL => $WebServiceURL,
        WebServiceUser => $WebServiceUser,
        WebServicePassword => $WebServicePassword ? '******' : undef,
    }) if !$WebServiceURL || !$WebServiceUser || !$WebServicePassword;

    return DE_EPAGES::WebService::API::Client
        ->uri( $ServiceUrn )
        ->proxy( $WebServiceURL )
        ->userinfo( $WebServiceUser . ':' . $WebServicePassword )
        ->xmlschema('2001');
}

sub _MultiStoreConfigService {
    my ($MasterShop) = @_;

    return _WebServiceClient($MasterShop, 'Site', 'urn://epages.de/WebService/MultiStoreConfigService/2011/08');
}

sub _MultiStoreService {
    my ($MasterShop, $SubShopAlias) = @_;

    return _WebServiceClient($MasterShop, $SubShopAlias, 'urn://epages.de/WebService/MultiStoreService/2013/04');
}

sub _GetAssignedSubShops {
    my ($MasterShop) = @_;

    my $ConfigService = _MultiStoreConfigService($MasterShop);
    return $ConfigService->getInfo({
        'ProviderPath'    => LoadRootObject()->get('ProviderPath'),
        'MasterShopAlias' => $MasterShop->alias,
        'FilterActive' => 1
    })->result->{'SubShops'};
}

sub _InformSubShopsAboutUpdates {
    my ($MasterShop, $aObserverUpdates, $aSubShops, $hWebServiceConfig) = @_;

    say '-> Sending updates to sub shops' if $PRINT_OUTPUT;

    # Call web service for each sub shop
    foreach my $SubShopAlias ( @$aSubShops ) {

        say "  $SubShopAlias" if $PRINT_OUTPUT;

        my $MSService = _MultiStoreService($MasterShop, $SubShopAlias);
        $MSService->updateSubShopObjects($SubShopAlias, $aObserverUpdates);
    }
    return;
}

sub _ExportContentObjects {
    my ($StoreName,$Shop) = @_;

    MultiStoreLogDebug('_ExportContentObjects() START');

    say '-> Exporting ContentObjects.' if $PRINT_OUTPUT;
    my $MS_LastUpdate = GetLastUpdateID($Shop);
    my $ahLastUpdates = GetLastUpdates($Shop);

    my $hWebServiceConfig = $Shop->get( [qw( MS_WebServiceURL MS_WebServiceUser MS_WebServicePassword )] );
    my $aSubShops = _GetAssignedSubShops($Shop, $hWebServiceConfig);

    MultiStoreLogDebug('_ExportContentObjects() MS_LastUpdate = '.$MS_LastUpdate);
    MultiStoreLogDebug('_ExportContentObjects() AssignedSubShops', $aSubShops);

    # MS_LastUpdate - get ids of each scheduled update
    my $ProcessedUpdatesCount = 0;
    my %ScheduledUpdateIDs = map { $_->{'ObjectUpdateID'} => undef } @$ahLastUpdates;
    my @ObjectUpdateIDs = keys %ScheduledUpdateIDs;
    my @ObserverUpdates;

    # Filter updates (full updates or single value changes).
    my $haRealUpdates;
    # Loop over all updates.
    foreach my $hInfo ( @$ahLastUpdates ) {
        my $ObjectUpdateID = $MS_LastUpdate = $hInfo->{'ObjectUpdateID'};

        # MS_LastUpdate - set attribute periodically and drop id of currently processed update.
        SetPeriodicUpdateID($Shop, $ProcessedUpdatesCount++, $ObjectUpdateID, \%ScheduledUpdateIDs);
        delete $ScheduledUpdateIDs{$ObjectUpdateID};

        # Get remove flag, value, objectpath.
        my $Values = $hInfo->{'Value'};
        my $Remove = $hInfo->{'Remove'};
        my $ObjectPath = $hInfo->{'ObjectPath'};

        # Init datastructure if not already initialized.
        $haRealUpdates->{$ObjectPath} = {
            'updates' => [],
            'remove' => 0,
            'fullupdate' => 0
        } unless exists $haRealUpdates->{$ObjectPath};

        # If object marked as removed.
        if ($Remove) {
            # Clear all previous updates.
            $haRealUpdates->{$ObjectPath} = {};
            $haRealUpdates->{$ObjectPath} =  {
                remove => 1,
                updates => [],
                fullupdate => 0
            };
        } elsif ($Values) {
            # Check if there are values only for special attributes and no full import necessary.
            if (!$haRealUpdates->{$ObjectPath}->{'fullupdate'}) {
                # Add update
                push @{$haRealUpdates->{$ObjectPath}->{'updates'}}, $Values;
            }
        } else {
            # A fullupdate should be done.
            $haRealUpdates->{$ObjectPath}->{'fullupdate'} = 1;
            $haRealUpdates->{$ObjectPath}->{'updates'} = [];
        }
    }

    foreach my $ObjectPath (keys %$haRealUpdates) {
        my $hInfo = $haRealUpdates->{$ObjectPath};
        MultiStoreLogDebug("_ExportContentObjects() Update Object = $ObjectPath", $hInfo);

        my $hPathInfo = _GetPathInfo($StoreName, $Shop, $ObjectPath);
        my $FileName = $hPathInfo->{'FileName'};

        if ($hInfo->{'remove'}) {
            #Delete the object itself (xml), its AffiliateVisibilityMap (avm) and its CategoryProdutMap (cpm)
            my ($Path, $Name) = fsplit($FileName);
            foreach my $Ext ( qw( xml avm cpm ) ) {
                my $File = "$Path/$Name" . ".$Ext";
                say "- $File" if $PRINT_OUTPUT;
                MultiStoreLogDebug("_ExportContentObjects() Delete file = $File");
                DeleteFile($File) if ExistsFile($File);
            }
            push @ObserverUpdates, { 'ObjectPath' => $ObjectPath, 'Remove' => 1 };
        }

        # Don't care about deleted objects which are not created afterwards.
        unless ($Shop->findPath($ObjectPath)) {
            _InformSubShopsAboutUpdates($Shop, \@ObserverUpdates, $aSubShops, $hWebServiceConfig);
            @ObserverUpdates = ();
            next;
        }

        my %LocalExportParams = %{GetDefaultValues()->{'EXPORT_PARAMS'}};
        my $MainFolder = $hPathInfo->{'MainFolder'};
        $LocalExportParams{'recursive'} = $MainFolder eq 'Categories' ? 0 : 1;

        # Beltway for not working /LockFile( $FH, sub{ ExportObjects(..) })/
        # Export object into temporary file to avoid conflicts with other processes reading from the repository.
        my $TmpDir = tempdir( 'CLEANUP' => 1 );
        my (undef,$TmpFile) = tempfile( 'DIR' => $TmpDir );
        say '+ /' . $Shop->alias . "/$ObjectPath" if $PRINT_OUTPUT;
        MultiStoreLogDebug("_ExportContentObjects() ExportObject = $ObjectPath");
        ExportObjects(
            $hPathInfo->{'ParentPath'},
            $TmpFile,
            \%LocalExportParams,
            [ $hPathInfo->{'ObjectName'} ],
            \&TransformExportXML
        );

        # Remove unwanted control characters, which makes XMLin think the document were not well formed.
        my $TmpFileContent = GetFileContent( $TmpFile, { 'utf8' => 1 } );
        if ($$TmpFileContent =~ s/[\0\a\b\e\f\t\v]//gx) {
            WriteFile($TmpFile, $TmpFileContent, { 'create_directory' => 1, 'utf8' => 1 } );
        }

        my %HookParams = (
            'MainFolder' => $MainFolder,
            'FileName' => $FileName,
            'TmpFile' => $TmpFile,
            'Done' => 0,
        );

        MultiStoreLogDebug('_ExportContentObjects() Trigger Hook > API_MS_ExportContentObjects');
        TriggerHook('API_MS_ExportContentObjects',\%HookParams);

        unless ( $HookParams{'Done'} ) {
            my $ProdObjXML = XMLout(
                _RemoveUnwantedTags( #!FIXME
                    XMLin( $TmpFile, 'KeepRoot' => 1 )
                ),
                'KeepRoot' => 1,
                'XMLDecl' => q(<?xml version='1.0' encoding='utf-8'?>)
            );

            MultiStoreLogDebug("_ExportContentObjects() Write ExportObject.xml file > $FileName");
            WriteFile($FileName, \$ProdObjXML, { 'create_directory' => 1, 'flock'=> 1, 'utf8' => 1 } );
        }

        # AffililateVisibilityMap..
        my $Object = $Shop->path($ObjectPath);
        if ( TestObjectClass($Object, GetDefaultValues()->{'VISIBILITY_LIMITED_CLASS'}) ) {
            my %AVM = (
                $ObjectPath => {
                    'IsVisible' => $Object->get('IsAffiliateVisible')
                }
            );
            if ($Object->get('HasSelectiveAffiliateVisibility')) {
                $AVM{$ObjectPath}->{'VisibleFor'} = $Object->get('AffiliateShopAliases');
            }

            my ($Path, $Name) = fsplit($FileName);
            my $AvmFilename = "$Path/$Name.avm";
            MultiStoreLogDebug("_ExportContentObjects() Save AffililateVisibilityMap > $AvmFilename");
            WriteFile($AvmFilename, \encode_json(\%AVM), { 'create_directory' => 1, 'flock'=> 1, 'utf8' => 1 } );
        }

        # If a fullupdate should be triggered -> do not set value param.
        if ($hInfo->{'fullupdate'}) {
            push @ObserverUpdates, { 'ObjectPath' => $ObjectPath, 'Remove' => 0 };
        } elsif ($hInfo->{'updates'}) {
            # If there are some special updates -> push each update.
            my %Values = ();
            foreach my $Value (@{$hInfo->{'updates'}}) {
                my %TempHash = (%Values, %{decode_json($Value)});
                %Values = %TempHash;
            }
            my $NewJSONString = encode_json(\%Values);
            push @ObserverUpdates, { 'ObjectPath' => $ObjectPath, 'Remove' => 0, 'Value' => $NewJSONString };
        }
        # If observer package size is reached -> inform subshops about updates.
        if (scalar @ObserverUpdates >= GetDefaultValues()->{'MS_OBSERVER_PACKAGE_SIZE'}) {
            # Send remaining update infos.
            _InformSubShopsAboutUpdates($Shop, \@ObserverUpdates, $aSubShops, $hWebServiceConfig);
            @ObserverUpdates = ();
        }
    }

    # Push all remaining updates.
    MultiStoreLogDebug('_ExportContentObjects() _InformSubShopsAboutUpdates');
    _InformSubShopsAboutUpdates($Shop, \@ObserverUpdates, $aSubShops, $hWebServiceConfig) if scalar @ObserverUpdates;

    # MS_LastUpdate - reset the latest value after all updates are finished
    MultiStoreLogDebug('_ExportContentObjects() DeleteFromObjectUpdateByShopIDObjectUpdateID');
    DeleteFromObjectUpdateByShopIDObjectUpdateID($Shop->id, \@ObjectUpdateIDs);
    SetLastUpdateID($Shop, 0);

    MultiStoreLogDebug('_ExportContentObjects() END');

}

sub _RemoveUnwantedTags {
    my ($hXML) = @_;

    # get connection data
    my $Config = DatabaseConfig('System');

    # get config info
    my $EMSStocklevelHandle = $Config->get('MSSynchronizationService','MSSynchronizationServiceHandle-StockLevel');

    # !FIXME - In fact, these attributes are wanted, but not yet supported.
    my @UnwantedTags = qw(
        ScalePrice
        QuantityDiscount
        ValueDiscount
        ProductShippingMethod
    );

    push(@UnwantedTags, qw(StockLevel)) if ($EMSStocklevelHandle);

    if ( exists $hXML->{'epages'}{'Product'} ) {
        my $rProduct = $hXML->{'epages'}{'Product'};
        foreach my $hProduct ( ref $rProduct eq 'ARRAY' ? @$rProduct : $rProduct ) {
            delete $hProduct->{$_} foreach @UnwantedTags;
            $hProduct->{'ProductType'} =
                GetDefaultValues()->{'MS_PRODUCT_CLASS_ALIAS'}
                unless exists $hProduct->{'ProductType'};
            delete $hProduct->{'ProductType'} if exists $hProduct->{'reference'};
        }
    }
    return $hXML;
}

#========================================================================================
# function     _GetPathInfo
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $hPathInfo = _GetPathInfo($MasterStore, $Shop, $Path);
# example      my $hPathInfo = GetPathInfo('Master', $Shop, '/Products/be_40401');
#----------------------------------------------------------------------------------------
# description  Return a couple of information regarding the given path.
#----------------------------------------------------------------------------------------
# input        $MasterStore | master store name | String
# input        $Shop | shop object | DE_EPAGES::Shop::API::Object::Shop
# input        $Path | relative object path | String
# return       $hPathInfo | obejct path info, keys are <ul>
#                    <li>ObjectName | object alias | String</li>
#                    <li>FileName | complete path to object snapshot (file system) | String</li>
#                    <li>ParentPath | parent path on shop (database) | String</li>
#                    <li>MainFolder | folder name of object (e.g. 'Categories', 'Products') | String</li>
#                    </ul> | .ref.hash
#========================================================================================
#!FIXME - cookie monster that needs to be brushed off..
sub _GetPathInfo {
    my ($MasterStore, $Shop, $Path) = @_;

    my $aPathParts = SplitObjectPath($Path);
    my $MainFolder = $aPathParts->[0];
    my $ObjectName = JoinObjectPath([ pop @$aPathParts ]);
    my $ParentRelPath = JoinObjectPath($aPathParts);

    my $FolderName = GetSyncPath($MasterStore, $Shop->alias);
    $FolderName .= '/'. ObjectPathToFilePath($ParentRelPath) if scalar @$aPathParts;
    my $FileName = $FolderName .'/'. ObjectPathToFilePath($ObjectName) . '.xml';
    my $ParentAbsPath = $ParentRelPath
                      ? join('/', $Shop->get('Path'), $ParentRelPath)
                      : $Shop->get('Path');

    return {
        'ObjectName' => $ObjectName,
        'FileName' => $FileName,
        'ParentPath' => $ParentAbsPath,
        'MainFolder' => $MainFolder,
    };
}

sub _ExportBundleProductMaps {
    my ($StoreName,$Shop) = @_;

    # logging, monitoring
    MultiStoreLogDebug('_ExportBundleProductMaps() START');
    say '-> Exporting BundleProductMaps.' if $PRINT_OUTPUT;

    # get info
    my $hBundleProductMaps;
    my $hHandledBundleProducts;
    my $MapClass = LoadClassByAlias('BundleProductMap');
    my $aObjects = $MapClass->getObjects();
    my $aBundleProducts;

    # handle bundle product maps
    foreach my $BundleProductMap (@$aObjects){
        # get info
        my $BundleProduct = $BundleProductMap->get('BundleProduct');
        my $aFokusedBundleProductMaps;

        # check if bundle product is already handled
        next if $hHandledBundleProducts->{$BundleProduct->id};

        # check for siteid
        next unless ($BundleProduct->getSite()->id eq $Shop->id);

        foreach my $FokusedBundleProductMap (@{$MapClass->getObjects('ByBundleProductID',[$BundleProduct->id])}){

            # get info
            my $BundledProduct = $FokusedBundleProductMap->get('BundledProduct');
            my $Quantity = $FokusedBundleProductMap->get('BundledQuantity');
            my $Position = $FokusedBundleProductMap->get('Position');

            # build result
            push(
                @$aFokusedBundleProductMaps,
                {
                    'alias' => $BundledProduct->alias,
                    'path' => $BundledProduct->pathString,
                    'quantity' => $Quantity,
                    'position' => $Position
                }
            );
        }

        push(
            @$aBundleProducts,
            {
                'alias' => $BundleProduct->alias,
                'path' => $BundleProduct->pathString,
                'bundle_products' => $aFokusedBundleProductMaps
            }
        );

        # mark bundle product as already handled
        $hHandledBundleProducts->{$BundleProduct->id} = 1;
    }

    # finalize result
    $hBundleProductMaps->{'shops'}->{$Shop->alias}->{'bundle_product_maps'} = $aBundleProducts;

    # write json content to file
    my $FileName = GetBundleProductMapsFileName($StoreName,$Shop->alias);
    open(my $FileHandle, '>', $FileName) or die "Could not open file '$FileName' $!";
    print $FileHandle encode_json($hBundleProductMaps);
    close($FileHandle) or die "error while closing $FileName";
}

1;
