#========================================================================================
# package      DE_EPAGES::Order::API::Basket
# state        public
#----------------------------------------------------------------------------------------
# description  This package provides the main basket functions.
#========================================================================================
package DE_EPAGES::Order::API::Basket;
use base Exporter;
use strict;

our @EXPORT_OK = qw (
    CheckBasketAddress
    CreateBasket
    GetExistingBasket
    GetBasket
    IsFiscalCodeRequired
);

use DE_EPAGES::Address::API::Address qw( CheckZipcode );
use DE_EPAGES::Core::API::Error qw( ExistsError Error GetError );
use DE_EPAGES::Database::API::Connection qw( GetCurrentDBHandle );
use DE_EPAGES::DataCache::API::Cache qw( UpdateCache );
use DE_EPAGES::Object::API::Factory qw( LoadClassByAlias );
use DE_EPAGES::Object::API::Table::Country qw( GetInfoCountry );
use DE_EPAGES::Order::API::Constants qw( COUNTRYID_ITALY FISCALCODE_LIMIT );
use DE_EPAGES::Shop::API::Address qw( CheckStringVATID CheckVATID UpdateVATID CheckShopCountry );

#========================================================================================
# function     CreateBasket
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Basket = CreateBasket($hInfo);
#----------------------------------------------------------------------------------------
# description  Creates a basket for a user and shop.
#----------------------------------------------------------------------------------------
# input        $hInfo | keys
#               <ul>
#                   <li>User - user - object
#                   <li>Shop - shop - object
#                   <li>LocaleID - Locale - char(5)
#                   <li>Parent - baskets folder (optional) - object
#               </ul> |  ref.hash
# return       $Basket| basket | object
#========================================================================================
sub CreateBasket {
    my ($hInfo) = @_;
    my $Parent = defined $hInfo->{'Parent'} ? $hInfo->{'Parent'} : $hInfo->{'User'}->folder('Baskets');
    return $Parent->insertChild('Basket', $hInfo);
}

#========================================================================================
# function     GetExistingBasket
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Basket = GetExistingBasket($User, $Shop, $hInfo);
#----------------------------------------------------------------------------------------
# description  Gets a basket for a user in a shop if the basket exists else it returns
#               'undef'.
#----------------------------------------------------------------------------------------
# input        $User | user | object
# input        $Shop | shop | object
# input        $hInfo | keys
#               <ul>
#                   <li>Parent - baskets folder (optional) - object
#               </ul> |  ref.hash
# return       $Basket| basket (optional) | object
#========================================================================================
sub GetExistingBasket($$;$) {
    my ($User, $Shop, $hInfo) = @_;
    my $Parent = defined $hInfo->{'Parent'} ? $hInfo->{'Parent'} : $User->folder('Baskets');
    return $Parent->child($Shop->id) if $Parent->existsChild($Shop->id);
    return undef;
}

#========================================================================================
# function     GetBasket
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Basket = GetBasket($User, $Shop, $hInfo);
#----------------------------------------------------------------------------------------
# description  Creates a basket for a user if the basket does not exist<br>
#               else it returns the existing basket.
#----------------------------------------------------------------------------------------
# input        $User | user | object
# input        $Shop | shop | object
# input        $hInfo | basket infos keys (optional):
#               <ul>
#                   <li>LocaleID | Locale (optional) | char(5)
#                   <li>CurrencyID | currency (optional) | char(3)
#                   <li>Parent - baskets folder (optional) - object
#               </ul> | ref.array.hash
# return       $Basket| basket | object
#========================================================================================
sub GetBasket($$;$) {
    my ($User, $Shop, $hInfo) = @_;
    $hInfo = {} unless defined $hInfo;

    my $Basket = GetExistingBasket($User, $Shop, $hInfo);
    return $Basket if defined $Basket;
    my %BasketInfo = (
        %$hInfo,
        'User' => $User,
        'Shop' => $Shop,
    );
    eval { $Basket = CreateBasket(\%BasketInfo); };
    if (ExistsError()) {
        my $Error = GetError();
        Error($Error) unless $Error->code eq 'DBIExecute' and $Error->vars->{'code'} == 2601 and $Error->vars->{'message'} =~ /u_object_alias/;
        my $dbi = GetCurrentDBHandle();
        UpdateCache($dbi->cacheKey,'ObjectByParentIDAlias');
        $Basket = GetExistingBasket($User, $Shop, $hInfo);
    }
    return $Basket;
}

#========================================================================================
# function     CheckBasketAddress
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $hAddress = CheckBasketAddress($Form, $Basket, $AddressType, $UsesShipping);
#----------------------------------------------------------------------------------------
# description  Checks if all required address data has been entered according to the
#               particular registration set (non-member or shipping address).
#               Adds a form error for each input field that is not correctly filled in.
#               Used at storefront and from webservice.
#----------------------------------------------------------------------------------------
# input        $Form | form object | object
# input        $Basket | basket object | object
# input        $AddressType | address type | object
# input        $UsesShipping | true if shipping address is given | boolean
# return       $hAddress | shipping or billing address values | ref.hash
#========================================================================================
sub CheckBasketAddress {
    my ($Form, $Basket, $AddressType, $UsesShipping) = @_;

    # get registration set
    my $Shop = $Basket->getSite();
    my $aRegistrationSet = $Shop->get('RegistrationSet');

    # test registration set
    my $ClassAddress = LoadClassByAlias('Address');

    my %Address;
    # if a field is mandatory for shipping it's also mandatory for billing (reason for use of UsesShipping)
    my $MandRegistrationSetName    = 'Is'.($UsesShipping ? $AddressType : 'ShippingAndBilling'). 'AddressMandatory';
    my $VisibleRegistrationSetName = 'Is'.($UsesShipping ? $AddressType : 'ShippingAndBilling'). 'AddressVisible';

    # Is it a country with a mandatory county/state/subdivision?
    my $SubdivsMandatory = 0;
    my $CountryField = $AddressType eq 'Shipping' ? 'ShippingCountryID' : 'CountryID';
    if ( $Form->value($CountryField) and $Form->value($CountryField) =~ /\d/ ) {
        $SubdivsMandatory = 1 if GetInfoCountry($Form->value($CountryField))->{'SubdivsMandatory'};
    }

    # special exception for Italy EPG-21131
    if ( $AddressType eq 'Billing' && IsFiscalCodeRequired( $Basket, $AddressType.'Address', $Form->value($CountryField) ) ) {
        my $FiscalCode = $Form->value( 'FiscalCode' );
        $Address{'FiscalCode'} = $Form->field( $Shop, $FiscalCode, {
            Name => 'FiscalCode',
            Form => $AddressType . 'Address',
            Type => 'string',
            Mandatory => 1
        });
    }

    foreach my $Set (@$aRegistrationSet) {
        my $Attribute = $Set->{'Attribute'};
        if ($Attribute->alias eq 'State') {
            # If the country has a mandatory state/county/subdivision we have to set the visible and mandatory flag.
            # But if the new country has no longer a state/county/subdivision this field have to be deleted, so
            # we set the visible flag only!
            $Set->{$VisibleRegistrationSetName} = 1;
            $Set->{$MandRegistrationSetName} = 1 if $SubdivsMandatory;
        }
        next unless $Set->{$VisibleRegistrationSetName};

        # remove these lines when attributes are removed from all non-member registration sets
        my $AttributeName = $Attribute->alias;
        next unless $ClassAddress->existsAttribute( $AttributeName );

        # get form value eg: 'ShippingCountry' | 'BillingCountry' + 'Country'
        my $FieldName = defined $Form->value( $AddressType . $AttributeName ) ? $AddressType . $AttributeName : $AttributeName;
        my $hFieldDef = { Name => $FieldName, Form => $AddressType . 'Address' };
        $hFieldDef->{Mandatory} = 1 if $Set->{$MandRegistrationSetName};

        my $Type = $Attribute->get('Type');
        if( $Type eq 'String' ) {
            $hFieldDef->{Type} = 'string';
            $hFieldDef->{MaxValue} = $Attribute->get('Length');
        }
        elsif( $Type eq 'File' ) {
            $hFieldDef->{Type} = 'file';
        }
        elsif( $Type eq 'EMail' ) {
            $hFieldDef->{Type} = 'email_address';
        }
        elsif( $Attribute->get('IsObject') ) {
            $hFieldDef->{Type} = 'integer';
            $hFieldDef->{MinValue} = 1;
        }
        else { # Integer, Float, DateTime, Money
            $hFieldDef->{Type} = 'reg_'. lc $Type;
        }
        my $OrgValue = $Form->value( $FieldName );
        $Address{$AttributeName} = $Form->field( $Shop, $OrgValue, $hFieldDef );
    }
    if ($Address{'VATID'}) {
        if (!CheckStringVATID($Address{'VATID'})) {
            $Form->addFormError({
                'Reason'=> 'VATIDIsNotValid',
                'Form'  => $AddressType . 'Address',
                'Name'  => 'VATID',
            });
        }
        else {
            $Address{'VATID'} = UpdateVATID($Address{'VATID'}, $Address{'CountryID'}, $Shop);
            my $check = CheckVATID($Address{'VATID'}, $Address{'CountryID'}, $Shop);
            $Form->addFormError({
                'Reason'=> $check ? 'VATIDIsNotFromSameCountry' : 'VATIDIsNotValid',
                'Form'  => $AddressType . 'Address',
                'Name'  => $AddressType eq 'Billing' ? 'VATID' : 'ShippingVATID',
            }) unless $check==1;
        }
    }
    if (!CheckZipcode($Address{'Zipcode'}, $Address{'CountryID'})) {
        $Form->addFormError({
            'Reason'=> 'ZIPCODE_NOT_VALID',
            'Form'  => $AddressType . 'Address',
            'Name'  => $AddressType eq 'Billing' ? 'Zipcode' : 'ShippingZipcode',
        });
    }
    if (!CheckShopCountry($Shop, $Address{'CountryID'})) {
        $Form->addFormError({
            'Reason'=> 'COUNTRYID_NOT_VALID',
            'Form'  => $AddressType . 'Address',
            'Name'  => 'CountryID',
        });
    }

    return \%Address;
}

#========================================================================================
# function     IsFiscalCodeRequired
# state        public
#----------------------------------------------------------------------------------------
# syntax       $required = IsFiscalCodeRequired( $Basket, $AddressType, $BasketCountryID );
#----------------------------------------------------------------------------------------
# description  returns true if exception for Italy EPG-21131 is true
#----------------------------------------------------------------------------------------
# input        $Basket | basket object | object
# input        $AddressType | address type | string
# input        $BasketCountryID | country id from the basket address | int
# return       $required | true if required | int
#========================================================================================
sub IsFiscalCodeRequired {
    my ($Basket, $AddressType, $BasketCountryID) = @_;

    my $Shop = $Basket->getSite;
    my $ShopAddress = $Shop->get('Address');
    return 0 unless defined $ShopAddress;

    my $ShopCountryID = $ShopAddress->get('CountryID');
    return 0 unless defined $ShopCountryID;
    return 0 unless $ShopCountryID == COUNTRYID_ITALY;

    # use country from basketaddress if no form fields available
    if (!$BasketCountryID) {
        my $BasketAddress = $Basket->get($AddressType);
        $BasketAddress = $Basket->get('BillingAddress') if not defined $BasketAddress and $AddressType eq 'ShippingAddress';
        $BasketCountryID = $BasketAddress->get('CountryID') if defined $BasketAddress;

        # set basket country to shop country if we have no informations about the basket country
        $BasketCountryID ||= $ShopCountryID;
    }

    return 0 if $BasketCountryID != COUNTRYID_ITALY;

    return $Basket->container->get('GrandTotal') >= FISCALCODE_LIMIT ? 1 : 0;
}

1;
