#========================================================================================
# package      DE_EPAGES::PayPalExpress::API::Payment
# state        public
#----------------------------------------------------------------------------------------
# description  Payment functions for PayPal Express.
#========================================================================================
package DE_EPAGES::PayPalExpress::API::Payment;
use base Exporter;
use strict;

our @EXPORT_OK = qw (
    CheckoutRequest
    CheckoutRequestDetails
    DoCheckoutPayment
    RefundTransaction
    SendPayPalRequest
    TestPPSettings
    GetInstantUpdateParams
    ResetBasket
);

use DE_EPAGES::Core::API::Constants qw ( PERL516 );
use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::Object::NumberFormat;
use DE_EPAGES::Core::API::PerlTools qw ( fcmp isValid trim UsePackage );
use DE_EPAGES::Core::API::TestParameter qw ( TestFloat );
use DE_EPAGES::Core::API::Url qw ( ParseQueryString );
use DE_EPAGES::Database::API::Connection qw ( GetCurrentDBHandle );
use DE_EPAGES::ExternalPayment::API::Log qw ( LogPayment );
use DE_EPAGES::Object::API::Factory qw ( LoadObject LoadRootObject );
use DE_EPAGES::Object::API::Language qw ( GetCodeByLanguageID );
use DE_EPAGES::Object::API::Table::Country qw ( ExistsCountry GetInfoCountry );
use DE_EPAGES::Object::API::Template qw ( ProcessFileTemplate );
use DE_EPAGES::Order::API::LineItemContainer qw ( SavePayment );
use DE_EPAGES::PayPalExpress::API::Constants qw (
    API_VERSION
    ITEM_CATEGORY_DIGITAL
    ITEM_CATEGORY_PHYSICAL
    PARAMS_SKIPPED_AT_LOGGING
    REQUEST_ENCODING
);
use DE_EPAGES::Shop::API::Constants qw ( TAX_MODEL_GROSS );
use DE_EPAGES::Shop::API::Url qw ( BuildShopUrl );
use DE_EPAGES::ThirdPartyConfig::API::ThirdPartyConfig qw ( GetDefaultTimeOut );
use DE_EPAGES::XML::API::DateTimeFormatXML;
use Encode qw ( encode );
use HTTP::Request::Common qw ( POST );
use LWP::UserAgent;
use Readonly;
use Time::HiRes;

BEGIN {
    if ($] < PERL516) {
        # use old packages for old perl
        UsePackage('Net::INET6Glue');
    }
}

#========================================================================================
# function     SendPayPalRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $hResult = SendPayPalRequest($PaymentMethod, \%Params);
#----------------------------------------------------------------------------------------
# description  sends a request with name/value pairs as parameters to PayPal
#----------------------------------------------------------------------------------------
# input        $PaymentMethod | the payment method | object
# input        $hParams | name/value input parameter hash for PayPal request | ref.hash
# return       $hResponse | response hash from PayPal; undef if an error occurs | ref.hash or undef
#========================================================================================
sub SendPayPalRequest {
    my ($PaymentMethod, $hParams) = @_;

    my $PaymentTypeAlias = $PaymentMethod->get('PaymentType')->alias; # EPG-14718, payment type can also be PayPalPro
    my $hPM = $PaymentMethod->get([qw (APIUsername APIPassword APISignatur IsInTestMode)]);
    $hParams->{'VERSION'}   = API_VERSION;
    $hParams->{'USER'}      //= $hPM->{'APIUsername'};
    $hParams->{'PWD'}       //= $hPM->{'APIPassword'};
    $hParams->{'SIGNATURE'} //= $hPM->{'APISignatur'};


    # override URL for testing
    if (defined $hParams->{'IsInTestMode'}) {
        $hPM->{'IsInTestMode'} = $hParams->{'IsInTestMode'};
        delete $hParams->{'IsInTestMode'};
    }

    my $pmlog = $PaymentMethod->id . ' (' . $PaymentMethod->pathString . ')';
    my $url = LoadRootObject()->get($hPM->{'IsInTestMode'}  ?
        'PayPalExpress_SandboxURL_APISignatur'  :  'PayPalExpress_ProductionURL_APISignatur');

    foreach my $Key (keys %$hParams){
        $hParams->{$Key} = encode('utf8', $hParams->{$Key});
    }

    LogPayment($PaymentTypeAlias, "parameters to send to $url for payment method $pmlog", _ClearParamsForLogging($hParams));

    my $Request = POST($url, $hParams);
    $Request->content_type('application/x-www-form-urlencoded; charset=utf-8');
    my $UserAgent = LWP::UserAgent->new;
    $UserAgent->timeout(GetDefaultTimeOut);
    $UserAgent->agent('ePages/6 ' . $UserAgent->agent());

    my $Response = $UserAgent->request($Request);
    if ($Response->is_success) {
        my $Content = $Response->content;
        LogPayment($PaymentTypeAlias, "response for payment method $pmlog", $Content);
        my $hResult = ParseQueryString($Content, REQUEST_ENCODING);
        # convert arrays to scalars
        my %Result = ();
        if (defined $hResult) {
            %Result = map { $_ => $hResult->{$_}->[0] } keys %$hResult;
        }
        LogPayment($PaymentTypeAlias, "response parameters for payment method $pmlog", \%Result);
        return \%Result;
    }
    else {
        LogPayment($PaymentTypeAlias, "request for payment method $pmlog failed", $Response->as_string);
        return undef;
    }
}

sub _ClearParamsForLogging {
    my ($hParams) = @_;

    my %ParamsForLogging = %$hParams;

    foreach my $SkippedParam (PARAMS_SKIPPED_AT_LOGGING) {
        $ParamsForLogging{$SkippedParam} = '****' if defined $ParamsForLogging{$SkippedParam};
    }

    return \%ParamsForLogging;
}

#========================================================================================
# function     TestPPSettings
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $hResult = TestPPSettings($PaymentMethod);
#----------------------------------------------------------------------------------------
# description  tests the settings of the payment method by calling SetExpressCheckout
#----------------------------------------------------------------------------------------
# input        $PaymentMethod | the payment method | object
# input        $hParams | parameter hash | ref.hash (optional)
# return       $hResult | result from PayPal, may be undef on failure | ref.hash
#========================================================================================
sub TestPPSettings {
    my ($PaymentMethod, $hParams) = @_;
    $hParams //= {};

    my $returnUrl   = LoadRootObject()->get('WebServerScriptUrlSSL');
    return SendPayPalRequest($PaymentMethod, {
        %$hParams,
        'METHOD'                        => 'SetExpressCheckout',
        'PAYMENTREQUEST_0_AMT'          => 1,
        'PAYMENTREQUEST_0_CURRENCYCODE' => 'EUR',
        'RETURNURL'                     => $returnUrl,
        'CANCELURL'                     => $returnUrl
    });
}

#========================================================================================
# function     CheckoutRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       CheckoutRequest($LineItem, \%Params);
#----------------------------------------------------------------------------------------
# description  calls SetExpressCheckout
#----------------------------------------------------------------------------------------
# input        $LineItemPaymentPayPalExpress | the payment lineitem | object
# input        $hInfo | data for SetExpressCheckout request:
#               <ul>
#               <li>ObjectID - basket id - int</li>
#               <li>LocaleCode - locale code (2 char country code) - string
#               <li>BuyerEmail - email of buyer - string</li>
#               <li>ReturnURL - return url - string</li>
#               <li>CancelURL - cancel url - string</li>
#               <li>BUTTONSOURCE - build notation id followed by 'S' (express) or 'M' (normal) - string</li>
#               </ul>
#               | ref.hash
# return       $redirectUrl | redirect url with token from response | string
#========================================================================================
sub CheckoutRequest {
    my ($pli, $hInfo) = @_;

    # check payment method
    Error('NO_PAYPAL_EXPRESS') unless ($pli->instanceOf('LineItemPaymentPayPalExpress'));

    my $PaymentMethod = $pli->get('PaymentMethod');
    my $hValues  = $PaymentMethod->get([qw(
        APIUsername
        APIPassword
        APISignatur
        IsInTestMode
        CPPLogoImage
        CPPHeaderImage
        CPPHeaderBorderColor
        CPPHeaderBackColor
        CPPPayflowColor
        PageStyle
        AllowGuestCheckout
        TransactionType
        HidePayPalExpress
    )]);

    my $CanDoInstandUpdate = 0;
    my $IsPayPalExpressVisible = $hValues->{'HidePayPalExpress'} ? 0 : 1; # invert logic
    my $DoInstandUpdate = $hValues->{'TransactionType'};
    if( $DoInstandUpdate && $IsPayPalExpressVisible ){
        $CanDoInstandUpdate = 1;
    }

    # provide a registered address for PayPal
    my $PaymentObject = LoadObject($hInfo->{'ObjectID'});
    my $Address  = $PaymentObject->get('ShippingAddress')
                || $PaymentObject->get('BillingAddress')
                || $PaymentObject->get('User')->get('ShippingAddress')
                || $PaymentObject->get('User')->get('BillingAddress');

    # killing # at the beginning of colour codes
    foreach my $Code (qw ( CPPHeaderBorderColor CPPHeaderBackColor CPPPayflowColor )) {
        $hValues->{$Code} ||= '';
        $hValues->{$Code} =~ s/^#//;
        $hValues->{$Code} =~ s/transparent//;
    }

    # split-up container items
    my $Container = $pli->container;
    my $hInstantUpdateParams = GetInstantUpdateParams($Container);

    my $Formatter = DE_EPAGES::Core::API::Object::NumberFormat->new;
    my $Amount = $Container->get('GrandTotal');
    my $TaxAmount = $hInstantUpdateParams->{'L_TAXAMT0'} // 0;
    my $NetShippingAmount = $hInstantUpdateParams->{'L_SHIPPINGOPTIONAMOUNT0'} // 0;
    my $NetItemAmount = $Formatter->round($Amount - $TaxAmount - $NetShippingAmount, 2);

    # calculate MAXAMT
    # This value is required in case the instant update option is used.
    # It has to be greater than the total amount.
    my $MaxAmount = 0;
    my $MaxShippingAmount = _CalculateMaxShippingAmount($pli);

    if ($MaxShippingAmount != 0) {
        # add a factor to avoid rounding errors
        my $RoundErrorFactor = 1.15; ## no critic (ProhibitMagicNumbers)
        $MaxAmount = $MaxShippingAmount * $RoundErrorFactor + $TaxAmount + $NetItemAmount;

        if ($MaxAmount < $Amount) {
            $MaxAmount = $Amount * $RoundErrorFactor;
        }
    } else {
        # no shipping method has implemented its price function
        # or only free shipping possible?
        $MaxAmount = $Amount * 3;
    }

    # we round the max amount to 0 decimals because some currencies
    # (e.g. Japanese Yen) are not supporting decimals.
    $MaxAmount = $Formatter->round($MaxAmount, 0);

    # build params hash
    my %RequestValues = (
        'METHOD'                         => 'SetExpressCheckout',

        'PAYMENTREQUEST_0_PAYMENTACTION' => 'Sale',
        'PAYMENTREQUEST_0_AMT'           => $Amount,
        'PAYMENTREQUEST_0_ITEMAMT'       => $NetItemAmount,
        'PAYMENTREQUEST_0_SHIPPINGAMT'   => $NetShippingAmount,
        'PAYMENTREQUEST_0_TAXAMT'        => $TaxAmount,
        'PAYMENTREQUEST_0_CURRENCYCODE'  => $pli->get('CurrencyID'),
        'PAYMENTREQUEST_0_INVNUM'        => '',
        'PAYMENTREQUEST_0_DESC'          => '',
        'MAXAMT'                         => $MaxAmount,

        'EMAIL'                          => $hInfo->{'BuyerEmail'},
        'LOCALECODE'                     => $hInfo->{'LocaleCode'},

        'RETURNURL'                      => $hInfo->{'ReturnURL'},
        'CANCELURL'                      => $hInfo->{'CancelURL'},
        'BUTTONSOURCE'                   => $hInfo->{'BUTTONSOURCE'},
        'LOGOIMG'                        => $hValues->{'CPPLogoImage'},
        'HDRIMG'                         => $hValues->{'CPPHeaderImage'},
        'HDRBORDERCOLOR'                 => $hValues->{'CPPHeaderBorderColor'},
        'HDRBACKCOLOR'                   => $hValues->{'CPPHeaderBackColor'},
        'PAYFLOWCOLOR'                   => $hValues->{'CPPPayflowColor'},
        'PAGESTYLE'                      => $hValues->{'PageStyle'},

        'NOSHIPPING'                     => 0,
        'ALLOWNOTE'                      => 0
    );

    # add Instant Update API params (if shipping is needed)
    if ( $Container->get('NeedShipping') && $CanDoInstandUpdate && $hInfo->{'IsExpressCheckout'} ) {
        %RequestValues = (
            %RequestValues,
            %$hInstantUpdateParams,
            'CALLBACK' => $hInfo->{'CallbackURL'},
            'CALLBACKTIMEOUT' => 5
        );
    }

    my $HasGuestCheckoutOption = LoadRootObject()->get('PayPalExpress_EnableGuestCheckoutOption');

    if ($HasGuestCheckoutOption) {
        $RequestValues{'LANDINGPAGE'} = 'Billing' if ($hValues->{'AllowGuestCheckout'});
        $RequestValues{'SOLUTIONTYPE'} = $hValues->{'AllowGuestCheckout'} ? 'Sole' : 'Mark';
    }
    # use shop name instead of email address (EPG-23532), only if it's set in the current language (EPG-32351)
    my $ShopName = $pli->getSite()->get('Name', $Container->get('LanguageID'));
    $RequestValues{'BRANDNAME'} = $ShopName if $ShopName;

    # GiroPay params
    if ($hInfo->{'GiroPaySuccessURL'}) {
        $RequestValues{'GIROPAYSUCCESSURL'} = $hInfo->{'GiroPaySuccessURL'};
        $RequestValues{'BANKTXNPENDINGURL'} = $hInfo->{'GiroPaySuccessURL'};
    }
    $RequestValues{'GIROPAYCANCELURL'} = $hInfo->{'GiroPayCancelURL'} if (defined $hInfo->{'GiroPayCancelURL'});

    # Address params
    if (defined ($Address)) {
        my $AddressName = join (' ', grep { defined $_ } ($Address->get('FirstName'), $Address->get('MiddleName'), $Address->get('LastName')));
        my $hAddressData = $Address->get([qw (Street Street2 City State CountryID Zipcode Company)]);
        my $CompanyAndStreet2 = _buildCompanyAndStreet2AddressDataForPaypalRequest($hAddressData);
        my $AddressCountryCode2 = ($hAddressData->{'CountryID'} and ExistsCountry($hAddressData->{'CountryID'}))
            ? GetInfoCountry($hAddressData->{'CountryID'})->{'Code2'}
            : '';

        %RequestValues = (%RequestValues,
            'PAYMENTREQUEST_0_SHIPTONAME'        => $AddressName,
            'PAYMENTREQUEST_0_SHIPTOSTREET'      => $hAddressData->{'Street'},
            'PAYMENTREQUEST_0_SHIPTOSTREET2'     => $CompanyAndStreet2,
            'PAYMENTREQUEST_0_SHIPTOCITY'        => $hAddressData->{'City'},
            'PAYMENTREQUEST_0_SHIPTOSTATE'       => $hAddressData->{'State'},
            'PAYMENTREQUEST_0_SHIPTOZIP'         => $hAddressData->{'Zipcode'},
            'PAYMENTREQUEST_0_SHIPTOCOUNTRYCODE' => $AddressCountryCode2,
        );
    }
    $RequestValues{'ADDROVERRIDE'} = $hInfo->{'ADDROVERRIDE'}  ?  1 : 0;

    # Basket as a single line item (EPG-18538)
    my ($BasketName, $BasketDescr) = _getBasketTexts($Container->get('LanguageID'));
    $RequestValues{'L_PAYMENTREQUEST_0_NAME0'}         = $BasketDescr if($CanDoInstandUpdate);
    $RequestValues{'L_PAYMENTREQUEST_0_AMT0'}          = $NetItemAmount,
    $RequestValues{'L_PAYMENTREQUEST_0_TAXAMT0'}       = $TaxAmount,
    $RequestValues{'L_PAYMENTREQUEST_0_NUMBER0'}       = 1;
    $RequestValues{'L_PAYMENTREQUEST_0_QTY0'}          = 1;
    $RequestValues{'L_PAYMENTREQUEST_0_ITEMCATEGORY0'} = $Container->get('NeedShipping') ? ITEM_CATEGORY_PHYSICAL : ITEM_CATEGORY_DIGITAL,
    $RequestValues{'L_PAYMENTREQUEST_0_DESC0'}         = $BasketName if($CanDoInstandUpdate);

    # execute and evaluate request
    my $hResult = SendPayPalRequest($PaymentMethod, \%RequestValues);
    _TestSuccess($hResult);

    $pli->set({
        'Token'  => $hResult->{'TOKEN'},
        'Field1' => 'SetExpressCheckout'
    });

    my $redirectUrl = LoadRootObject()->get($hValues->{'IsInTestMode'}
        ? 'PayPalExpress_SandboxURL_Redirect'
        : 'PayPalExpress_ProductionURL_Redirect'
    );

    return $redirectUrl . $hResult->{'TOKEN'};
}

#========================================================================================
# function     _CalculateMaxShippingAmount
# state        private
#----------------------------------------------------------------------------------------
# syntax       _CalculateMaxShippingAmount( $PaymentLineItem );
#----------------------------------------------------------------------------------------
# description  get the maximum shipping and tax amount based on the current items.
#               go thru all usable shipping method of the shop
#               we can, however, only use shipping methods which have implemented the "price" function
#----------------------------------------------------------------------------------------
# input        $PaymentLineItem | payment line item | DE_EPAGES::PayPalExpress::API::Object::LineItemPaymentPayPalExpress
# return       $MaxShippingAmount | Max shipping amount | float
#========================================================================================
sub _CalculateMaxShippingAmount {
    my ( $PaymentLineItem ) = @_;

    my $aLineItemProducts = $PaymentLineItem->container()->childrenByClass('LineItemProduct');
    my @ShippingElements = ();
    foreach my $LineItemProduct ( @{$aLineItemProducts} ) {
        push @ShippingElements, {
            'Element'  => $LineItemProduct->get('Product'),
            'Quantity' => $LineItemProduct->get('Quantity')
        };
    }

    my $aShopShippingMethods = $PaymentLineItem->getSite->existsChild('ShippingMethods')
                             ? $PaymentLineItem->getSite->child('ShippingMethods')->children()
                             : [];
    my $CurrencyID = $PaymentLineItem->get('CurrencyID');
    my $MaxShippingAmount = 0;

    foreach my $ShopShippingMethod ( @{$aShopShippingMethods} ) {
        my $ShippingPrice = 0;
        eval {
            $ShippingPrice = $ShopShippingMethod->price({
                'CurrencyID' => $CurrencyID,
                'TaxModel'   => TAX_MODEL_GROSS,
                'ShippingElements' => \@ShippingElements
            });
            TestFloat('ShippingPrice', $ShippingPrice);
        };
        if (not ExistsError() and $ShippingPrice > $MaxShippingAmount) {
            $MaxShippingAmount = $ShippingPrice;
        }
    }

    return $MaxShippingAmount;
}
#========================================================================================
# function     GetInstantUpdateParams
# state        public
#----------------------------------------------------------------------------------------
# syntax       GetInstantUpdateParams($Container);
#----------------------------------------------------------------------------------------
# description  fetches the params for the Instant Update API
#               used in the SetExpressCheckout request and CallbackResponse,
#               see PayPalExpress::API::Payment::CheckoutRequest
#               and PayPalExpress::UI::Basket::PayPalExpressCallback
#----------------------------------------------------------------------------------------
# input        $Container | the container object | object
# return       $hParams | the response details | ref.hash
#========================================================================================
sub GetInstantUpdateParams {
    my ($Container) = @_;

    my $ShippingLineItem = $Container->get('Shipping');
    my $ShippingMethod = defined $ShippingLineItem ? $ShippingLineItem->get('ShippingMethod') : undef;

    #--- check if it's PayPalExpress and if this method is allowed for the given billing address
    #--- it may not be PayPalExpress, e.g. in the call from PayPalExpress::UI::Basket::PayPalExpressCallback
    my $PaymentMethod = $Container->get('Payment')->get('PaymentMethod');

    LogPayment('PayPalExpress', 'PaymentMethod', $PaymentMethod);

    my $PaymentMethodRegion = $PaymentMethod->get('Region');
    my $BillingCountryID = $Container->get('BillingCountryID');

    my $IsPPEAllowed = (
        $PaymentMethod->instanceOf('PaymentMethodPayPalExpress')
        and not (
            defined $PaymentMethodRegion
            and defined $BillingCountryID
            and not $PaymentMethodRegion->existsCountry($BillingCountryID)
        )
    ) ? 1 : 0;

    LogPayment('PayPalExpress', 'IsPPEAllowed', $IsPPEAllowed);

    #--- build and format the return hash
    my %Params = ();
    my $Formatter = DE_EPAGES::Core::API::Object::NumberFormat->new;

    if ($IsPPEAllowed and defined $ShippingMethod) {
        my $LanguageID = $Container->get('LanguageID');

        my $ShippingLineItemTax = 0;
        if ($Container->get('TaxModel') == TAX_MODEL_GROSS) {
            $ShippingLineItemTax += $_->{'TaxAmount'} foreach (@{$ShippingLineItem->get('Taxes') // []});
        }

        my $ShippingLineItemNetPrice = $Formatter->round($ShippingLineItem->get('LineItemPrice') - $ShippingLineItemTax, 2);

        my $FileNameNetTranslation = $ENV{'EPAGES_CARTRIDGES'} . '/DE_EPAGES/PayPalExpress/Templates/NetTranslation.txt';
        my $NetTranslation = ProcessFileTemplate($FileNameNetTranslation, GetCodeByLanguageID($LanguageID), {});
        $NetTranslation =~ s/[\n\r]//g;

        %Params = (
            'L_SHIPPINGOPTIONNAME0'      => $ShippingMethod->get('NameOrAlias', $LanguageID),
            'L_SHIPPINGOPTIONLABEL0'     => $NetTranslation,
            'L_SHIPPINGOPTIONAMOUNT0'    => $ShippingLineItemNetPrice,
            'L_SHIPPINGOPTIONISDEFAULT0' => 'true',
            'L_TAXAMT0'                  => $Container->get('TotalTax')
        );
    }
    else {
        %Params = (
            'NO_SHIPPING_OPTION_DETAILS' => 1
        );
    }

    LogPayment('PayPalExpress', 'GetInstantUpdateParams', \%Params);

    return \%Params;
}

#========================================================================================
# function     CheckoutRequestDetails
# state        public
#----------------------------------------------------------------------------------------
# syntax       CheckoutRequestDetails($LineItem, \%Params);
#----------------------------------------------------------------------------------------
# description  calls GetExpressCheckoutDetails
#----------------------------------------------------------------------------------------
# input        $LineItemPaymentPayPalExpress | the payment lineitem | object
# input        $hInfo | data for GetExpressCheckoutDetails request:
#               <ul>
#               <li>ObjectID - basket/order id - int</li>
#               </ul>
#               | ref.hash
# return       $details | the response details  | ref.hash
#========================================================================================
sub CheckoutRequestDetails {
    my ($pli, $hInfo) = @_;

    #--- check payment method
    Error('NO_PAYPAL_EXPRESS') unless($pli->instanceOf('LineItemPaymentPayPalExpress'));

    my $Token = $pli->get('Token');
    Error('PayPalExpressTokenNotAvailabe', $hInfo->{'ObjectID'}) unless defined $Token;

    my $hResult = SendPayPalRequest($pli->get('PaymentMethod'), {
        'METHOD' => 'GetExpressCheckoutDetails',
        'TOKEN' => $Token
    });

    # did it work out?
    _TestSuccess($hResult);
    $pli->set({'Field1' => 'CheckoutRequestDetails'});
    $pli->set({'PayerID' => $hResult->{'PAYERID'}}) if $hResult->{'PAYERID'};
    return $hResult;
}

#========================================================================================
# function     DoCheckoutPayment
# state        public
#----------------------------------------------------------------------------------------
# syntax       DoCheckoutPayment($LineItem, \%Params);
#----------------------------------------------------------------------------------------
# description  calls DoExpressCheckoutPayment
#----------------------------------------------------------------------------------------
# input        $LineItemPaymentPayPalExpress | the payment lineitem | object
# input        $hInfo | data for DoExpressCheckoutPayment request:
#               <ul>
#               <li>ObjectID - order id - int</li>
#               </ul>
#               | ref.hash
# return       $payinfo | the response payment info  | ref.hash
#========================================================================================
sub DoCheckoutPayment {
    my ($pli, $hInfo) = @_;
    my $Order     = LoadObject($hInfo->{'ObjectID'});
    my $Container = $Order->container;
    my $Shop      = $Order->getSite();

    #--- check payment method
    Error('NO_PAYPAL_EXPRESS') unless ($pli->instanceOf('LineItemPaymentPayPalExpress'));

    my $PayerID = $pli->get('PayerID');
    if (!defined $PayerID) {
        my %response = (
            'Ack'           => 'PayPalExpressPayerIDNotAvailable',
            'ErrorCode'     => 'PayPalExpressPayerIDNotAvailable',
            'LongMessage'   => 'PayPalExpressPayerIDNotAvailable',
            'CorrelationID' => 0,
            'ObjectID'      => $hInfo->{'ObjectID'},
        );
        $response{'Response'} = {%response};
        Error('PayPalExpressFailed', \%response);
    }

    my $NotifyURL = BuildShopUrl($Shop, {
        'ChangeAction' => ['BlockMonitor', 'PayPalExpressIPN'],
        'PaymentGUID'  => $Container->get('Payment')->get('GUID')
    }, {
        'UseSSL'        => 1,
        'UseObjectPath' => 1
    });

    # update amount and currency; may change due to address update
    $pli->set({
        'Amount'     => $Container->get('GrandTotal'),
        'CurrencyID' => $Container->get('CurrencyID')
    });

    my $BuildNotation = $pli->get('IsExpressCheckout') ? 'PayPalExpress_BuildNotationID' : 'PayPalExpress_BuildNotationID_ST';
    my $ButtonSource = LoadRootObject()->get($BuildNotation);

    my %RequestValues = (
        'METHOD'                         => 'DoExpressCheckoutPayment',
        'BUTTONSOURCE'                   => $ButtonSource,
        'TOKEN'                          => $pli->get('Token'),
        'PAYERID'                        => $PayerID,
        'PAYMENTREQUEST_0_PAYMENTACTION' => 'Sale',
        'PAYMENTREQUEST_0_CURRENCYCODE'  => $pli->get('CurrencyID'),
        'PAYMENTREQUEST_0_AMT'           => $pli->get('Amount'),
        'PAYMENTREQUEST_0_DESC'          => $Order->getSite()->get('NameOrAlias') . ' Order# ' . $Order->get('NameOrAlias'),
        'PAYMENTREQUEST_0_NOTIFYURL'     => $NotifyURL,
    );

    if ($pli->get('ShippingAddressChanged')) {
        my $ShippingAddress = $Order->get('ShippingAddress');
        my $ShippingName = join (' ', grep { defined $_ } ($ShippingAddress->get('FirstName'), $ShippingAddress->get('MiddleName'), $ShippingAddress->get('LastName')));
        my $hAddressData = $ShippingAddress->get([qw (Street Street2 City State CountryID Zipcode Company)]);
        my $CompanyAndStreet2 = _buildCompanyAndStreet2AddressDataForPaypalRequest($hAddressData);
        %RequestValues   = (%RequestValues,
            'PAYMENTREQUEST_0_SHIPTONAME'        => $ShippingName,
            'PAYMENTREQUEST_0_SHIPTOSTREET'      => $hAddressData->{'Street'},
            'PAYMENTREQUEST_0_SHIPTOSTREET2'     => $CompanyAndStreet2,
            'PAYMENTREQUEST_0_SHIPTOCITY'        => $hAddressData->{'City'},
            'PAYMENTREQUEST_0_SHIPTOSTATE'       => $hAddressData->{'State'},
            'PAYMENTREQUEST_0_SHIPTOCOUNTRYCODE' => ($hAddressData->{'CountryID'} &&
                                    ExistsCountry($hAddressData->{'CountryID'})) ?
                                    GetInfoCountry($hAddressData->{'CountryID'})->{'Code2'} : '',
            'PAYMENTREQUEST_0_SHIPTOZIP'         => $hAddressData->{'Zipcode'},
        );
    }

    my $hResult = SendPayPalRequest($pli->get('PaymentMethod'), \%RequestValues);

    my $Now = GetCurrentDBHandle()->currentDateTime();

    # did it work out?
    eval {
        _TestSuccess($hResult);
    };
    if (ExistsError()) {
        my $Error = GetError();
        if ($Error->code eq 'PayPalExpressFailed') {
            $Order->set({'CancelledOn' => $Now});
        }
        Error($Error);
    }

    # set payment line item values
    my %Values = (
        'TransTime' => $Now,
        'Field1'    => 'DoExpressCheckoutPayment'
    );
    $Values{'TransID'}       = $hResult->{'PAYMENTINFO_0_TRANSACTIONID'} if $hResult->{'PAYMENTINFO_0_TRANSACTIONID'};
    $Values{'TransStatus'}   = $hResult->{'PAYMENTINFO_0_PAYMENTSTATUS'} if $hResult->{'PAYMENTINFO_0_PAYMENTSTATUS'};
    $Values{'PendingReason'} = $hResult->{'PAYMENTINFO_0_PENDINGREASON'} if $hResult->{'PAYMENTINFO_0_PENDINGREASON'};

    $pli->set(\%Values);

    # set order status
    _SetTransStatus($Order, lc($hResult->{'PAYMENTINFO_0_PAYMENTSTATUS'}), $Now);

    return $hResult;
}

#========================================================================================
# function     RefundTransaction
# state        public
#----------------------------------------------------------------------------------------
# syntax       RefundTransaction($LineItem, \%Params);
#----------------------------------------------------------------------------------------
# description  calls RefundTransaction
#----------------------------------------------------------------------------------------
# input        $LineItemPaymentPayPalExpress | the payment lineitem | object
# input        $amount | the amount to refund (optional); if omitted refund whole amount | float
# input        $note | note to send to PayPal (optional) | string
# return       $success | successful? | boolean
# return       $hResult | response of RefundTransaction call | ref.hash
#========================================================================================
sub RefundTransaction {
    my ($pli, $amount, $note) = @_;

    #--- check payment method
    Error('NO_PAYPAL_EXPRESS') unless($pli->instanceOf('LineItemPaymentPayPalExpress'));

    my %RequestValues = (
        'METHOD'        => 'RefundTransaction',
        'TRANSACTIONID' => $pli->get('TransID')
    );

    my $remainingAmount = $pli->get('Amount');
    $remainingAmount -= $pli->get('Field2') // 0;
    $remainingAmount -= $pli->get('Field3') // 0;
    if (!defined($amount) || fcmp($remainingAmount, $amount) == 0) { # refund full amount
        $RequestValues{'REFUNDTYPE'} = 'Full';
    }
    else {
        $RequestValues{'REFUNDTYPE'}   = 'Partial';
        $RequestValues{'AMT'}          = $amount;
        $RequestValues{'CURRENCYCODE'} = $pli->get('CurrencyID');
    }

    $RequestValues{'NOTE'} = $note if(isValid($note));

    my $hResult = SendPayPalRequest($pli->get('PaymentMethod'), \%RequestValues);
    if (defined($hResult->{'ACK'}) && $hResult->{'ACK'}  =~ /^Success/i) {
        # update payment lineitem
        # TransID2: refund transaction id
        # TransStatus: Refunded or PartiallyRefunded
        # TransTimeMod: timestamp
        # Field2: net refund amount
        # Field3: refund fee
        # Field2 + Field3 = gross refund amount
        # if more than 1 refund take place the new values will be added to the saved ones
        my $DateParser = DE_EPAGES::XML::API::DateTimeFormatXML->new;
        $pli->set({
            'TransID2'     => $hResult->{'REFUNDTRANSACTIONID'},
            'TransStatus'  => $RequestValues{'REFUNDTYPE'} eq 'Full'  ?  'Refunded' : 'PartiallyRefunded',
            'TransTimeMod' => $DateParser->parse_datetime($hResult->{'TIMESTAMP'}),
            'Field2'       => ($pli->get('Field2') // 0) + $hResult->{'NETREFUNDAMT'},
            'Field3'       => ($pli->get('Field3') // 0) + $hResult->{'FEEREFUNDAMT'},
        });
        return (1, $hResult);
    }
    else {
        return (0, $hResult);
    }
}

#========================================================================================
sub _SetTransStatus {
    my ($Order, $TransStatus, $DateTime) = @_;

    $Order->set({'PendingOn',   $DateTime}) if $TransStatus eq 'pending';
    $Order->set({'PaidOn',      $DateTime}) if $TransStatus eq 'completed';
    return;
}

#========================================================================================
# function     _TestSuccess
# state        private
#----------------------------------------------------------------------------------------
# syntax       eval { _TestSuccess($response); };
#----------------------------------------------------------------------------------------
# description  throws Error in case of not succeeded request
#----------------------------------------------------------------------------------------
# input        $hResult | the result hash | ref.hash
#========================================================================================
sub _TestSuccess {
    my ($hResult) = @_;
    if (!defined($hResult)) {
        Error('PayPalExpressFailed', {});
    }
    unless (defined($hResult->{'ACK'}) && $hResult->{'ACK'}  =~ /^Success/i) {
        LogPayment('PayPalExpress', '_TestSuccess throws error, hResult=', $hResult);
        Error('PayPalExpressFailed', {
            'Ack'           => $hResult->{'ACK'},
            'ErrorCode'     => $hResult->{'L_ERRORCODE0'},
            'LongMessage'   => $hResult->{'L_LONGMESSAGE0'},
            'CorrelationID' => $hResult->{'CORRELATIONID'},
            'Response'      => $hResult
        });
    }
}

#========================================================================================
# function     _getBasketTexts
# state        private
#----------------------------------------------------------------------------------------
# syntax       eval { _getBasketTexts($LanguageID); };
#----------------------------------------------------------------------------------------
# description  process the templates for the basket texts and return the localized texts
#----------------------------------------------------------------------------------------
# input        $LanguageID | language id | int
# return       $BasketName | name of basket | string
# return       $BasketDescr | description of basket | string
#========================================================================================
sub _getBasketTexts {
    my ($LanguageID) = @_;

    my $FileNameBasketName  = $ENV{'EPAGES_CARTRIDGES'} . '/DE_EPAGES/PayPalExpress/Templates/BasketName.txt';
    my $FileNameBasketDescr = $ENV{'EPAGES_CARTRIDGES'} . '/DE_EPAGES/PayPalExpress/Templates/BasketDescription.txt';
    my $LanguageCode = GetCodeByLanguageID($LanguageID);
    my $BasketName  = ProcessFileTemplate($FileNameBasketName, $LanguageCode, {});
    my $BasketDescr = ProcessFileTemplate($FileNameBasketDescr, $LanguageCode, {});
    return trim($BasketName, $BasketDescr);
}

#========================================================================================
# function     ResetBasket
# state        public
#----------------------------------------------------------------------------------------
# syntax       ResetBasket($Basket);
#----------------------------------------------------------------------------------------
# description  Deletes any address of a anonymous basket user,
#               set the payment line item to the default payment method
#----------------------------------------------------------------------------------------
# input        $Basket | basket object | object
#========================================================================================
sub ResetBasket {
    my ($Basket) = @_;

    my $BasketUser = $Basket->get('User');
    my $Container = $Basket->container;

    return unless defined $Container;

    #--- delete all address information for a anonymous user
    if (defined $BasketUser and $BasketUser->get('IsAnonymous')) {
        foreach my $Address (@{$BasketUser->get('Addresses')}) {
            $Address->delete;
        }

        $Basket->set({'BillingAddress' => undef, 'ShippingAddress' => undef});
        $Basket->child('BillingAddress')->delete if ($Basket->existsChild('BillingAddress'));
        $Basket->child('ShippingAddress')->delete if ($Basket->existsChild('ShippingAddress'));
    }

    #--- use default payment method
    my $aPaymentMethods = $Container->get('PaymentMethods');
    my ($NewPaymentMethod) = grep { $_->get('IsDefault') } @$aPaymentMethods;
    SavePayment($Container, $NewPaymentMethod);
    $Container->recalculate;

    return;
}

#========================================================================================
# function     _buildCompanyAndStreet2AddressDataForPaypalRequest
# state        private
#----------------------------------------------------------------------------------------
# syntax       _buildCompanyAndStreet2AddressDataForPaypalRequest($hAddressData);
#----------------------------------------------------------------------------------------
# description  This method build a string with the company and street2 fields from the address data
#----------------------------------------------------------------------------------------
# input        $hAddressData | hash with Address data | refhash
# return       The concatenation of company and street2 of the address data| string
#========================================================================================
sub _buildCompanyAndStreet2AddressDataForPaypalRequest {
    my ($hAddressData) = @_;
    my $CompanyName = $hAddressData->{'Company'};
    my $Street2 = $hAddressData->{'Street2'};

    if (isValid($CompanyName) && isValid($Street2)){
        return sprintf("%s %s", $CompanyName, $Street2);
    } elsif(isValid($CompanyName)){
        return $CompanyName;
    } else{
        return $Street2;
    }
}

1;
