#========================================================================================
# package      DE_EPAGES::PayPalPlus::API::PaymentWallDataBuilder
# state        public
#----------------------------------------------------------------------------------------
# description  Collects Data from LineItemPaymentPayPalPlus for PaymentWall
#========================================================================================
package DE_EPAGES::PayPalPlus::API::PaymentWallDataBuilder;

use DE_EPAGES::Core::API::TestParameter qw (TestObject TestUrl TestInteger);
use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );

use strict;

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       new( );
#----------------------------------------------------------------------------------------
# description  constructor
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::PayPalPlus::API::PaymentWallDataBuilder
#========================================================================================
sub new {
    my $class = shift;
    my $self  = bless {}, $class;
    return $self;
}

#========================================================================================
# function     extractData
# state        public
#----------------------------------------------------------------------------------------
# syntax       extractData( $LineItemPaymentPayPalPlus );
#----------------------------------------------------------------------------------------
# description  Extracts Data from LineItemPaymentPayPalPlus for PaymentWall
#----------------------------------------------------------------------------------------
# input        $LineItemPaymentPayPalPlus | LineItem Payment PayPalPlus | DE_EPAGES::PayPalPlus::API::Object::LineItemPaymentPayPalPlus
# return       1: if successful
#               0: else          | boolean
#========================================================================================
sub extractData {
    my $self = shift;
    my ($LineItemPaymentPayPalPlus) = @_;

    eval {
        $self->_extractApprovalURL($LineItemPaymentPayPalPlus);
        $self->_extractMode($LineItemPaymentPayPalPlus);
        $self->_extractCountryCode($LineItemPaymentPayPalPlus);
        $self->_extractLanguage($LineItemPaymentPayPalPlus);
    };
    if ( ExistsError() ) {
        my $Error = GetError();
        $self->_handleError($Error);
        return 0;
    }

    return 1;
}

#========================================================================================
# function     getData
# state        public
#----------------------------------------------------------------------------------------
# syntax       getData(  );
#----------------------------------------------------------------------------------------
# description  returns Data for Paymentwall
#----------------------------------------------------------------------------------------
# return       $hPaymentWallData | PaymentWallData | refhash
#========================================================================================
sub getData {
    my $self = shift;

    my $hPaymentWallData = {
            'wasSuccessful' => 1,
            'approvalUrl' => $self->{'ApprovalURL'},
            'mode' => $self->{'Mode'},
            'buttonLocation' => 'outside',
            'language' => $self->{'Language'},
            'country' => $self->{'CountryCode'},
        };

    return $hPaymentWallData;
}

#========================================================================================
# function     getError
# state        public
#----------------------------------------------------------------------------------------
# syntax       getError(  );
#----------------------------------------------------------------------------------------
# description  return Data in error case
#----------------------------------------------------------------------------------------
# return       $hError | Error hash | refhash
#========================================================================================
sub getError {
    my $self = shift;

    my $hError = {
        'wasSuccessful' => 0,
        'errorCode' => $self->{'ErrorCode'},
        'errorMessage' => $self->{'ErrorMessage'},
    };

    return $hError;
}

#----------------------------------------------------------------------------------------
# private
#----------------------------------------------------------------------------------------
sub _extractApprovalURL {
    my $self = shift;
    my ($LineItemPaymentPayPalPlus) = @_;

    TestObject('LineItemPaymentPayPalPlus',$LineItemPaymentPayPalPlus,'DE_EPAGES::PayPalPlus::API::Object::LineItemPaymentPayPalPlus');

    my $ApprovalURL = $LineItemPaymentPayPalPlus->get('ApprovalURL');
    TestUrl('ApprovalURL', $ApprovalURL);

    $self->{'ApprovalURL'} = $ApprovalURL;

    return;
}

#----------------------------------------------------------------------------------------
sub _extractMode {
    my $self = shift;
    my ($LineItemPaymentPayPalPlus) = @_;

    TestObject('LineItemPaymentPayPalPlus',$LineItemPaymentPayPalPlus,'DE_EPAGES::PayPalPlus::API::Object::LineItemPaymentPayPalPlus');

    my $PaymentMethodPayPalPlus = $LineItemPaymentPayPalPlus->get('PaymentMethod');
    TestObject('PaymentMethodPayPalPlus',$PaymentMethodPayPalPlus,'DE_EPAGES::PayPalPlus::API::Object::PaymentMethodPayPalPlus');

    my $IsInTestMode = $PaymentMethodPayPalPlus->isInTestMode();
    TestInteger('IsInTestMode', $IsInTestMode);

    $self->{'Mode'} = $IsInTestMode?'sandbox':'live'; # Transform to paypal values;

    return;
}

#----------------------------------------------------------------------------------------
sub _extractCountryCode {
    my $self = shift;
    my ($LineItemPaymentPayPalPlus) = @_;

    #TODO: AD-3106 use TestObject to make sure we have a valid LineItemContainer
    my $LineItemContainer = $LineItemPaymentPayPalPlus->container();
    my $Basket = $LineItemContainer->get('BasketOrOrder');
    my $Address = $self->_getBasketShippingAddress($Basket);
    if(defined($Address)) {
        my $Country = $Address->get('Country');
        $self->{'CountryCode'} = $Country->{'Code2'};
    } else {
        $self->{'CountryCode'} = 'DE';
    }

    return;
}

#----------------------------------------------------------------------------------------
sub _extractLanguage {
    my $self = shift;
    my ($LineItemPaymentPayPalPlus) = @_;

    #TODO: AD-3106 use TestObject to make sure we have a valid LineItemContainer
    my $LineItemContainer = $LineItemPaymentPayPalPlus->container();
    $self->{'Language'} = $LineItemContainer->get('LocaleID');

    return;
}

#----------------------------------------------------------------------------------------
sub _getBasketShippingAddress {
    my $self = shift;
    my ($Basket) = @_;
    my $Address = $Basket->get('ShippingAddress');
    if(not isValid($Address)){
        $Address = $Basket->get('BillingAddress');
    }
    return $Address;
}

#----------------------------------------------------------------------------------------
sub _handleError {
    my $self = shift;
    my ($Error) = @_;

    #TODO: AD-3106 use ErrorTypes and log the error!!
    if ( $Error->code() ~~ [ qw ( PARAMETER_MISSING PARAMETER_NO_OBJECT WRONG_CLASS FORMAT_NOT_URL FORMAT_NOT_INTEGER) ]) {
        $self->{'ErrorCode'} = $Error->code();
        $self->{'ErrorMessage'} = $Error->message();
    } else {
        Error($Error); #forward all other Errors
    }

    return;
}

1;