#!/srv/epages/eproot/Perl/bin/perl

use strict;
use Getopt::Long;
use Time::HiRes qw (time);
use File::Basename qw(dirname);

use DE_EPAGES::Core::API::File qw( DeleteFile
    CompareDirectoryWithFilelist GetFileSHA1Digest );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Core::API::Net qw(GetFastestEpagesMirror GetFile);
use DE_EPAGES::Core::API::SystemCommand qw(UnpackTar);
use DE_EPAGES::Installer::API::LanguagePack qw(GetUsedLanguages);

sub Main {

    local $| = 1;

    my $time1 = time;

    #-------------------------------------------------------------------
    # check options
    #-------------------------------------------------------------------

    my %Options = ();
    my @GetOpt = qw(help storename=s version=s lang=s nooutput verbose);
    GetOptions(\%Options, @GetOpt);

    usage() if $Options{help};
    usage() unless $Options{version};
    delete $Options{nooutput} if $Options{verbose};
    my $Quiet = $Options{nooutput};

    my $EpagesPatches = $ENV{EPAGES_PATCHES};
    $EpagesPatches = "$ENV{EPAGES_SHARED}/Patches"
        if not $EpagesPatches and $ENV{EPAGES_SHARED};
    usage('environment variable not set: EPAGES_PATCHES') unless $EpagesPatches;

    #-------------------------------------------------------------------
    # define variables
    #-------------------------------------------------------------------

    my $StoreLanguages;
    if ($Options{lang}) {
        $StoreLanguages = [split(/\s+/, $Options{lang})];
    } else {
        (undef, $StoreLanguages) = GetUsedLanguages([$Options{storename}]);
        $StoreLanguages = [sort keys %$StoreLanguages];
    }

    my $CtmplPacks = "$EpagesPatches/CtmplPacks";
    my $Patch = 'patch_'. $Options{version};
    $Patch =~ s/\./_/g;
    my $ExtractedCtmplPacks = "$EpagesPatches/$Patch/ExtractedDataPacks".
        "/ExtractedCtmplPacks.v/$Options{version}";

    #-------------------------------------------------------------------
    # loop over languages
    #-------------------------------------------------------------------

    foreach my $Lang ( @$StoreLanguages ) {

        my $Ctmpls = "epages-$Options{version}-$Lang-compiled-templates";

        #-------------------------------------------------------------------
        # next if unpacked package found
        #-------------------------------------------------------------------

        next if -d "$ExtractedCtmplPacks/$Lang";

        #-------------------------------------------------------------------
        # download package unless exist
        #-------------------------------------------------------------------

        if (not -f "$CtmplPacks/$Ctmpls.tar.gz") {

            DeleteFile("$CtmplPacks/$Ctmpls.info");

            # get epages-$Options{version}-$Lang-compiled-templates.tar.gz
            my $Url = GetFastestEpagesMirror(
                "patch/$Options{version}/$Ctmpls.tar.gz");
            print "FOUND: $Url\n";
            print "$Ctmpls: download package: " unless $Quiet;
            GetFile($Url, "$CtmplPacks/$Ctmpls.tar.gz");
            print "DONE\n" unless $Quiet;

            # get epages-common.info saved as
            # epages-$Options{version}-$Lang-compiled-templates.info
            $Url =~ s,patch/$Options{version}/$Ctmpls.tar.gz,yum/origin/epages-common.info,;
            print "FOUND: $Url\n";
            print "$CtmplPacks/$Ctmpls.info: download info file: " unless $Quiet;
            GetFile($Url, "$CtmplPacks/$Ctmpls.info");
            print "DONE\n" unless $Quiet;
        }

        #-------------------------------------------------------------------
        # test correctness of packed package
        #-------------------------------------------------------------------

#        if (-f "$CtmplPacks/$Ctmpls.info") {
            print "$Ctmpls.info: test correctness of packed package: " unless $Quiet;

            my $IniConfig = DE_EPAGES::Core::API::IniConfig->new
                (FileName => "$CtmplPacks/$Ctmpls.info", CaseSensitive => 1);
            my $ExpectedSha1Sum = $IniConfig->get('tgz-sha1sum', "ctmpl-$Lang");
            my $FoundSha1Sum = GetFileSHA1Digest("$CtmplPacks/$Ctmpls.tar.gz");

            if (not $ExpectedSha1Sum) {
                _showResult('SKIPPING', "$CtmplPacks/$Ctmpls.tar.gz") unless $Quiet;
            } elsif ($ExpectedSha1Sum eq $FoundSha1Sum) {
                _showResult('OK', "$CtmplPacks/$Ctmpls.tar.gz") unless $Quiet;
            } else {
                _showResult('FAILURE', "$CtmplPacks/$Ctmpls.tar.gz");
                next;
            }
#        }

        #-------------------------------------------------------------------
        # unpack package
        #-------------------------------------------------------------------

        print "$Ctmpls: unpack package: " unless $Quiet;
        UnpackTar("$CtmplPacks/$Ctmpls.tar.gz", dirname($ExtractedCtmplPacks));
        print "DONE\n" unless $Quiet;

        #-------------------------------------------------------------------
        # test completeness of unpacked package
        #-------------------------------------------------------------------

        print "$Ctmpls: test completeness of unpacked package: " unless $Quiet;
        print "\n" if $Options{verbose};

        # skip test if manifest (file list) does not exist but directory does
        if (not -f "$ExtractedCtmplPacks/$Ctmpls.txt"
            and -d "$ExtractedCtmplPacks/$Lang") {
            _showResult('SKIPPING', "$ExtractedCtmplPacks/$Lang") unless $Quiet;
        } elsif (CompareDirectoryWithFilelist("$ExtractedCtmplPacks/$Lang",
                                              "$ExtractedCtmplPacks/$Ctmpls.txt",
                                              { 'verbose' => $Options{verbose},
                                                'depth' => 2 })) {
            _showResult('OK', "$ExtractedCtmplPacks/$Lang") unless $Quiet;
        } else {
            _showResult('FAILURE', "$ExtractedCtmplPacks/$Ctmpls.txt $ExtractedCtmplPacks/$Lang") unless $Quiet;
        }
    }
    printf "need %.1f seconds\n", time - $time1 unless $Quiet;
}

sub _showResult {
    my ($Result, $File) = @_;
    my $Found = $Result eq 'FAILURE' ? 'FAILURE' : 'FOUND';
    print "$Result\n$Found: $File\n";
}

sub usage {
    my ($Msg) = @_;
    print "$Msg\n\n" if $Msg;
    print <<END_USAGE;
Description:
    Downloads epages-VERSION-LANG-compiled-templates.tar.gz into
    \$EPAGES_SHARED/Patches/CtmplPacks/ and unpacks it into
    \$EPAGES_SHARED/Patches/patch_*/ExtractedDataPacks/.
Usage:
    perl $0 { -storename STORE -version VERSION [-lang 'LIST']| -help }

Options:
    -storename  name of store
    -version    ePages version number which defines the package to
                download: epages-<version>-LANG-compiled-templates.tar.gz
    -lang       consider only (space-sep. list of) languages (default: all store langs)

Flags:
    -help          show the command line options
    -nooutput      do not print any output
    -verbose       verbose output

Example:
    \$PERL $0 -storename Store -version 6.17.19
    \$PERL $0 -storename Store -version 6.17.19 -lang 'de en_US'
END_USAGE
    exit 2;
}

RunScript(
    'Sub' => \&Main
);

1;
