#========================================================================================
# package      DE_EPAGES::Product::API::CSV::ProductExportDriver;
# base         DE_EPAGES::ShopCSVExportImport::API::CSV::ShopObjectExportDriver
# state        public
#----------------------------------------------------------------------------------------
# description  This module implements a handler to export shop products via csv
# example      use DE_EPAGES::Product::API::CSV::ProductExportDriver;
#               DE_EPAGES::Product::API::CSV::ProductExportDriver->exportCSV($FileName, $hFileFormat, $aColumns, $Shop, $aObjects );
#========================================================================================
package DE_EPAGES::Product::API::CSV::ProductExportDriver;
use base qw ( DE_EPAGES::ShopCSVExportImport::API::CSV::ShopObjectExportDriver );

use strict;
use DE_EPAGES::Core::API::CSV qw( CombineCSV );
use DE_EPAGES::Product::API::Constants qw( @UNITY_CSV_ATTRIBUTEVALUES_NOT_FOR_SUBPRODUCTS );

#========================================================================================
# function     getObjectAttributeValue
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $Value = $self->getObjectAttributeValue($Product, $hAttribute);
#----------------------------------------------------------------------------------------
# input        $Product    | Product object | object
# input        $hAttribute | properties of object attribute to get:
#               <ul>
#                   <li>Alias     - attribute alias, e.g. Alias, Color
#                   <li>Type      - Attribute type, e.g. LocalizedString
#                   <li>IsObject  - 0 or 1, set to 1 if attribute IsObject
#                   <li>LanguageID - only set if Type =~ /Localized/
#                   <li>CurrencyID - only set if Type eq 'Price'
#               </ul> | ref.hash
# return       $Value | value of attribute for object | scalar
#========================================================================================
sub getObjectAttributeValue  {
    my $self = shift;
    my ($Object, $hAttribute ) = @_;

    my $Alias = $hAttribute->{'Alias'};
    my $Class = $Object->class;
    return undef unless $Class->existsAttribute($Alias);

    my $Attribute = $Class->attribute($Alias);
    return undef unless $hAttribute->{'Attribute'}->id == $Attribute->id;


    my $IsUnityRegistered = $Object->getSite()->get('IsUnityRegistered');

    # return empty string (attribute values) for unity and sub products and some aliases
    # we will not have values here because the master value is relevant and sub product values are ignored on import
    if ($IsUnityRegistered and defined $Object->superProduct and $Alias ~~ [@UNITY_CSV_ATTRIBUTEVALUES_NOT_FOR_SUBPRODUCTS] ) {
        return '';
    }elsif($Alias eq 'Class') {
         my $ClassAlias = $Class->alias;
         return $ClassAlias eq 'ProductClass' ? '' : $ClassAlias;
    }elsif ($Alias eq 'SelectedVariations') {
        my $SuperProduct = $Object->superProduct;
        my $Value = '';
        if (defined $SuperProduct ) {
            my @VariationTypes = map {$_->{'Attribute'}->alias} @{$SuperProduct->get('VariationTypeValues')};
            if (scalar (@VariationTypes ) ) {
                $Value= CombineCSV( \@VariationTypes, {'sep_char' => '~'} );
            }
        }
        return $Value;
    #add shippingmethods
    }elsif ($Alias eq 'ShippingMethods') {
        my $ahShippingMethods = $Object->get('ShippingMethods');
        my $Value = '';
        if (scalar @$ahShippingMethods ) {
            my @ShippingAttributesAliases = map {$_->alias} @{$ahShippingMethods};
            if (scalar (@ShippingAttributesAliases ) ) {
                $Value= CombineCSV( \@ShippingAttributesAliases, {'sep_char' => '~'} );
            }
        }
        return $Value;
    }elsif ($Alias eq 'ListPrices') {
        return $Object->getListPrice($hAttribute->{'CurrencyID'},$self->topID('TaxModel'),1);
    }elsif ($Alias eq 'SuperProduct') {
        my $SuperProduct = $Object->superProduct;
        return defined $SuperProduct ? $SuperProduct->alias : '';
    }elsif ($Alias ~~ [qw( TaxClass EcoParticipationCategory TaxClassDigital )] or
        $hAttribute->{'Type'} eq 'UnitOfMeasurement' or
        $hAttribute->{'Type'} =~ /^PreDef(Localized)?String$/) {
        my $Item = $Object->getWithParent($Alias);
        return defined $Item ? $Item->alias : '';
    }elsif ($hAttribute->{'Type'} =~ /^PreDefMulti(Localized)?String$/ )  {
        my $aValues = $Object->getWithParent($Alias);
        my @Aliases = map { $_->alias } @$aValues;
        return CombineCSV( \@Aliases, {'sep_char' => '~'} );
    }elsif ($Alias eq 'ManufacturerPrices') {
        return $Object->getManufacturerPrice($hAttribute->{'CurrencyID'},$self->topID('TaxModel'),1);
    } elsif ($Alias eq 'EcoParticipations') {
        return $Object->getEcoParticipation($hAttribute->{'CurrencyID'},$self->topID('TaxModel'));
    } elsif ($Alias eq 'Deposits') {
        return $Object->getDeposit($hAttribute->{'CurrencyID'},$self->topID('TaxModel'));
    } elsif ($Alias eq 'Position') {
        my $Position = $Object->get('Position');
        return $Position ? $Position : '';
    } elsif ($Alias eq 'URI') {
        return defined $Object->superProduct ? undef : $Object->get('URI', $hAttribute->{'LanguageID'});
    } else {
        return $Object->getWithParent($Alias,$hAttribute->{'LanguageID'});
    }
}

1;
