#========================================================================================
# package      DE_EPAGES::Product::API::Resource::ProductSlideshow
# state        public
# base         DE_EPAGES::WebService::API::Resource::BaseObjectResource
# description  implements the REST resource for getting slideshow images of a product.
#========================================================================================
package DE_EPAGES::Product::API::Resource::ProductSlideshow;
use base DE_EPAGES::WebService::API::Resource::BaseObjectResource;

use DE_EPAGES::Object::API::Factory qw ( LoadObjectByPath );
use DE_EPAGES::Core::API::Constants qw ( IMAGE_FILE_TYPES );
use DE_EPAGES::Core::API::PerlTools qw ( ElementOf );
use DE_EPAGES::Core::API::File qw ( fsplit );
use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::Url qw ( ParseQueryString );
use DE_EPAGES::Shop::API::Resource::ShopRequestContext;
use DE_EPAGES::SEO::API::SEO qw ( PerformSEOCheckAndUpdateObject );
use DE_EPAGES::Product::API::Constants qw ( PRODUCT_IMAGE_MAX_COUNT );
use DE_EPAGES::Product::API::ProductImageService;
use DE_EPAGES::Product::API::ProductImageValidation;
use DE_EPAGES::Product::API::SlideshowService;
use DE_EPAGES::Core::API::Unicode qw ( Unicode2Filename );
use feature 'switch';

use strict;



sub service {
    return LoadObjectByPath('/RestServices/ProductSlideshow');
}



sub class {
    return 'Product';
}



sub producedContentTypes {
    shift;
    my ($RequestMethod) = @_;

    return {
        GET => [
            ['get', 'application/vnd.epages.v1+json'],
            ['get', 'application/json'],
        ],
        POST => [
            [ 'post', 'application/vnd.epages.v1+json' ],
            [ 'post', 'application/json' ],
        ],
    }->{$RequestMethod};
}



sub match {
    my $self = shift;
    my ($Request) = @_;

    return unless $Request->uri->path =~ m|^/api/products/([^/]+)/slideshow/?$|;
    return $1;
}



sub href {
    my $self = shift;
    my ($APIBaseURL, $ProductGUID) = @_;
    return "$APIBaseURL/api/products/$ProductGUID/slideshow";
}



sub params {
    my $self = shift;
    my ($ActionName, $Object, $Request) = @_;

    given ($ActionName) {
        when ('post') {
            my $hQuery = ParseQueryString( $Request->uri->query, 'utf-8' );
            my $Imagename = $hQuery->{'imagename'}->[0];
            return [ $Object, $Request, $Imagename ];
        }
        default {
            return [$Object];
        }
    }
}



sub get {
    my $self = shift;
    my ($Product) = @_;

    # if slideshow is empty, we do not throw an error but return an empty result instead (AD-3471).

    my $Context = DE_EPAGES::Shop::API::Resource::ShopRequestContext->new($Product->getSite());
    my $SlideshowService = DE_EPAGES::Product::API::SlideshowService->new($Product, $Context->imageBaseUrl());
    my $hSlideshow = $SlideshowService->getSlideshow();

    return $hSlideshow;
}



sub post {
    my $self = shift;
    my ( $Product, $Request, $Imagename ) = @_;

    my (undef, $FileName, $FileExtension) = fsplit($Imagename);
    # remove white spaces and special charactors from imagename
    $FileName = Unicode2Filename($FileName);
    $Imagename = $FileName.$FileExtension;

    if (!($FileExtension && ElementOf(substr(lc($FileExtension),1), IMAGE_FILE_TYPES()))) {
        Error('UNSUPPORTED_IMAGE_FORMAT');
    }

    $self->_checkIfImageExistsWithSameName($Product, $Imagename);
    $self->_checkIfMaxImageCountIsAlreadyReached($Product);

    my $ProductImageService = $self->_createProductImageService($Product);

    my $NewFileName = $FileName.$FileExtension;
    $ProductImageService->uploadAndScaleImage($NewFileName, $Request->content());
    $ProductImageService->updateSlideShowString($NewFileName);

    $Product->updatePageCache();
    $self->_PerformSEOCheckAndUpdateObject($Product);

    my $Context = DE_EPAGES::Shop::API::Resource::ShopRequestContext->new($Product->getSite());
    my $SlideshowService = DE_EPAGES::Product::API::SlideshowService->new($Product, $Context->imageBaseUrl());
    my $hSlideshowImage = $SlideshowService->getSlideshowImage($NewFileName);

    return $hSlideshowImage;
}



sub _checkIfMaxImageCountIsAlreadyReached {
    my $self = shift;
    my ($Product) = @_;

    my $ProductImageValidation = $self->_createProductImageValidation($Product);
    my $ImagesCount = $ProductImageValidation->getImageCount();

    if($ImagesCount >= PRODUCT_IMAGE_MAX_COUNT){
        Error('PRODUCT_IMAGE_MAX_COUNT');
    }

    return;
}



sub _checkIfImageExistsWithSameName {
    my $self = shift;
    my ($Product, $Imagename) = @_;

    # check if name exists in PublicFiles.
    my $aPublicFiles = $Product->get('PublicFiles');
    my @PublicFilesNames = map {$_->{'FileName'}} @$aPublicFiles;
    my %hPublicFilesNames = map { $_ => 1 } @PublicFilesNames;

    if(exists($hPublicFilesNames{$Imagename})){
        Error('BadRequest', {
            Class => class(),
            GUID => $Product->guid(),
            Reason => 'ImageNameIsTaken',
        });
    }
    return;
}



sub _PerformSEOCheckAndUpdateObject {
    my $self = shift;
    my ($Product) = @_;
    PerformSEOCheckAndUpdateObject($Product);
    return;
}



sub _createProductImageService {
    my $self = shift;
    my ($Product) = @_;
    return DE_EPAGES::Product::API::ProductImageService->new($Product);
}



sub _createProductImageValidation {
    my $self = shift;
    my ($Product) = @_;
    return DE_EPAGES::Product::API::ProductImageValidation->new($Product);
}

1;
