#========================================================================================
# package      DE_EPAGES::Product::t::unit::ProductImageService
# base         DE_EPAGES::Core::API::Test
# state        public
#----------------------------------------------------------------------------------------
# description  Unit test for DE_EPAGES::Product::API::ProductImageService
#========================================================================================
package DE_EPAGES::Product::t::unit::ProductImageService;
use base DE_EPAGES::Core::API::Test;

use DE_EPAGES::Product::API::ProductImageService;
use DE_EPAGES::Test::Mock::MockifyEpagesObject;
use DE_EPAGES::Test::Mock::EpagesObject;
use DE_EPAGES::Test::Mock::Mockify qw ( WasCalled GetParametersFromMockifyCall );
use DE_EPAGES::Test::Mock::Mockery;
use Test::Exception;
use Test::More;

use strict;



sub test {
    my $self = shift;

    $self->test_setSlideShowImageAsProductImage_invisibleProduct();
    $self->test_setSlideShowImageAsProductImage_imageDoesNotExist();
    $self->test_setSlideShowImageAsProductImage();

    $self->test_deleteSlideShowImage_invisibleProduct();
    $self->test_deleteSlideShowImage_imageDoesNotExist();
    $self->test_deleteSlideShowImage();

    done_testing();
}



sub test_setSlideShowImageAsProductImage_invisibleProduct {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Mockery = $self->_createMocks('InvisibleProduct');
    my $ProductImageService = $Mockery->get('ProductImageService');

    $ProductImageService->setSlideShowImageAsProductImage('TestImage.png');

    my $Product = $Mockery->get('InvisibleProduct');
    is($Product->get('ImageLarge'), 'TestImage.png', "$SubTestName - test if selected image was set taken as product image.");
    ok(WasCalled($Product, 'updatePageCache'), "$SubTestName - test if updatePageCache was called.");
}



sub test_setSlideShowImageAsProductImage_imageDoesNotExist {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Mockery = $self->_createMocks('ProductWithoutImages');
    my $ProductImageService = $Mockery->get('ProductImageService');

    throws_ok(sub {
        $ProductImageService->setSlideShowImageAsProductImage('TestImage.png');
    }, qr/ResourceNotFound/,
    "$SubTestName - test if error ResourceNotFound not was thrown if the image does not exist");
}



sub test_setSlideShowImageAsProductImage {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Mockery = $self->_createMocks('Product');
    my $ProductImageService = $Mockery->get('ProductImageService');

    $ProductImageService->setSlideShowImageAsProductImage('TestImage.png');

    my $Product = $Mockery->get('Product');
    is($Product->get('ImageLarge'), 'TestImage.png', "$SubTestName - test if selected image was set taken as product image.");
    ok(WasCalled($Product, 'updatePageCache'), "$SubTestName - test if updatePageCache was called.");
}



sub test_deleteSlideShowImage_invisibleProduct {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Mockery = $self->_createMocks('InvisibleProduct');
    my $ProductImageService = $Mockery->get('ProductImageService');

    lives_ok(sub {
        $ProductImageService->deleteSlideShowImage('TestImage.png');
    }, "$SubTestName - test if deleteSlideShowImage can be called without error");

    my $aArguments = GetParametersFromMockifyCall($ProductImageService, '_DeleteProductImage');
    is($aArguments->[0]->alias(), 'InvisibleTestProduct', "$SubTestName - test if _DeleteProductImage was called with shop as 1st argument.");
    is($aArguments->[1], 'TestImage.png', "$SubTestName - test if _DeleteProductImage was called with image name as 2nd argument.");
}



sub test_deleteSlideShowImage_imageDoesNotExist {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Mockery = $self->_createMocks('ProductWithoutImages');
    my $ProductImageService = $Mockery->get('ProductImageService');

    throws_ok(sub {
        $ProductImageService->deleteSlideShowImage('TestImage.png');
    }, qr/ResourceNotFound/,
    "$SubTestName - test if error ResourceNotFound not was thrown if the image does not exist");
}



sub test_deleteSlideShowImage {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $Mockery = $self->_createMocks('Product');
    my $ProductImageService = $Mockery->get('ProductImageService');

    lives_ok(sub {
        $ProductImageService->deleteSlideShowImage('TestImage.png');
    }, "$SubTestName - test if deleteSlideShowImage can be called without error");

    my $aArguments = GetParametersFromMockifyCall($ProductImageService, '_DeleteProductImage');
    is($aArguments->[0]->alias(), 'TestProduct', "$SubTestName - test if _DeleteProductImage was called with shop as 1st argument.");
    is($aArguments->[1], 'TestImage.png', "$SubTestName - test if _DeleteProductImage was called with image name as 2nd argument.");
}



sub _createMocks {
    my $self = shift;
    my ($NameOfMockedProduct) = @_;

    my $Mockery = DE_EPAGES::Test::Mock::Mockery->new();

    $Mockery->register('ProductImageService', [$NameOfMockedProduct], sub {
        my ($Product) = @_;
        my $Mockify = DE_EPAGES::Test::Mock::Mockify->new('DE_EPAGES::Product::API::ProductImageService', [$Product]);
        $Mockify->addMockWithReturnValueAndParameterCheck('_PerformSEOCheckAndUpdateObject', undef, [{'object' => 'DE_EPAGES::Product::API::Object::Product'}]);
        $Mockify->addMockWithReturnValueAndParameterCheck('_DeleteProductImage', undef, ['object', 'string']);
        return $Mockify->getMockObject();
    });

    $Mockery->register('Product', ['Shop'], sub {
        my ($Shop) = @_;
        my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Product::API::Object::Product', {
            'Alias' => 'TestProduct',
            'Site' => $Shop,
            'IsVisible' => 1,
            'ImagesSlideShow' => $self->_createImageSlideShow(),
        });
        $Mockify->addMockWithReturnValue('updatePageCache', undef);
        return $Mockify->getMockObject();
    });

    $Mockery->register('ProductWithoutImages', ['Shop'], sub {
        my ($Shop) = @_;
        return DE_EPAGES::Test::Mock::EpagesObject->new('DE_EPAGES::Product::API::Object::Product', {
            'Site' => $Shop,
            'IsVisible' => 1,
            'ImagesSlideShow' => [],
        });
    });

    $Mockery->register('InvisibleProduct', ['Shop'], sub {
        my ($Shop) = @_;
        my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Product::API::Object::Product', {
            'Alias' => 'InvisibleTestProduct',
            'Site' => $Shop,
            'IsVisible' => 0,
            'ImagesSlideShow' => $self->_createImageSlideShow(),
        });
        $Mockify->addMockWithReturnValue('updatePageCache', undef);
        return $Mockify->getMockObject();
        
    });

    $Mockery->register('Shop', [], sub {
        return DE_EPAGES::Test::Mock::EpagesObject->new('DE_EPAGES::Shop::API::Object::Shop', {
            'Alias' => 'TestShop',
        });
    });

    return $Mockery->assemble();
}



sub _createImageSlideShow {
    return [
        {'ImageName' => 'test123.jpg'},
        {'ImageName' => 'TestImage.png'},
    ];
}

__PACKAGE__->runScript();