#========================================================================================
# package      DE_EPAGES::Product::t::unit::ProductLowestPriceResource
# base         DE_EPAGES::Core::API::Test
# state        public
#----------------------------------------------------------------------------------------
# description  Unit test for DE_EPAGES::Product::API::Resource::ProductLowesPrice
#========================================================================================
package DE_EPAGES::Product::t::unit::ProductLowestPriceResource;
use base DE_EPAGES::Core::API::Test;

use DE_EPAGES::Product::API::Resource::ProductLowestPrice;
use DE_EPAGES::Test::Mock::EpagesObject;
use DE_EPAGES::Test::Mock::Mockify;
use Test::Exception;
use Test::More;

use strict;

sub test {
    my $self = shift;

    $self->test_get_errorOnNoSubProductPricesFound();
    $self->test_get_errorOnNoSubProductFound();
    $self->test_get_FromPriceIfAllSubProductPricesEquals();
    $self->test_get_FromPriceIfAllSubProductPricesNotEquals();
    
    done_testing();
}

sub test_get_errorOnNoSubProductPricesFound {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $ProductLowestPrice = $self->_createProductLowestPriceResource();
    my $Product = $self->_createProductWithoutSubproductPrices();

    throws_ok{
        $ProductLowestPrice->get($Product, 'de_DE', 'EUR');
    } qr/ResourceNotFound: Product "123-456" does not have a lowest price. \(ShopAlias: TestShop\)/, ## no critic (ProhibitComplexRegexes)
    "$SubTestName - test if error is thrown when no sub product prices are found";
}

sub test_get_errorOnNoSubProductFound {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $ProductLowestPrice = $self->_createProductLowestPriceResource();
    my $Product = $self->_createProduct();

    throws_ok{
        $ProductLowestPrice->get($Product, 'de_DE', 'EUR');
    } qr/ResourceNotFound: Product "123-456" does not have a lowest price. \(ShopAlias: TestShop\)/, ## no critic (ProhibitComplexRegexes)
    "$SubTestName - test if error is thrown if no sub product was found";
}

# test if for subproducts with same prices the function ->price was called
sub test_get_FromPriceIfAllSubProductPricesEquals {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $ProductLowestPrice = $self->_createProductLowestPriceResource();
    my $Product = $self->_createProductWithSubproductPrices(1);

    throws_ok{
        $ProductLowestPrice->get($Product, 'de_DE', 'EUR');
    } qr/NativeDie: Can't call method "price".*/, ## no critic (ProhibitComplexRegexes)
    "$SubTestName - test if the product prices are equals and we have no from prices";
}

# test if for subproducts with different prices the function ->fromPrice was called
sub test_get_FromPriceIfAllSubProductPricesNotEquals {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $ProductLowestPrice = $self->_createProductLowestPriceResource();
    my $Product = $self->_createProductWithSubproductPrices(0);

    throws_ok{
        $ProductLowestPrice->get($Product, 'de_DE', 'EUR');
    } qr/NativeDie: Can't call method "fromPrice".*/, ## no critic (ProhibitComplexRegexes)
    "$SubTestName - test if the product prices are not equals and we have from prices";
}

sub _createProductLowestPriceResource {
    my $self = shift;
    my $Mockify = DE_EPAGES::Test::Mock::Mockify->new('DE_EPAGES::Product::API::Resource::ProductLowestPrice', []);
    $Mockify->addMock('_createShopRequestContext', sub {
        my $MockedSelf = shift;
        my ($Shop) = @_;
        my $ShopRequestContextMockify = DE_EPAGES::Test::Mock::Mockify->new('DE_EPAGES::Shop::API::Resource::ShopRequestContext', [$Shop]);
        $ShopRequestContextMockify->addMockWithReturnValueAndParameterCheck('validLocale', 'de_DE', [{'string' => 'de_DE'}]);
        $ShopRequestContextMockify->addMockWithReturnValueAndParameterCheck('validCurrency', 'EUR', [{'string' => 'EUR'}]);
        $ShopRequestContextMockify->addMockWithReturnValue('apiBaseUrl', undef);
        return $ShopRequestContextMockify->getMockObject();
    });
    $Mockify->addMockWithReturnValueAndParameterCheck('_createFromPriceFormatter', undef, ['object', 'string']);

    return $Mockify->getMockObject();
}



sub _createProductWithoutSubproductPrices {
    my $self = shift;
    my $Shop = DE_EPAGES::Test::Mock::EpagesObject->new('DE_EPAGES::Product::API::Object::Product', {
        'Site' => $self->_createShop(),
        'GUID' => '123-456',
    });
    $Shop->mock('existSubProductPrices', sub {
        return 0;
    });
    return $Shop;
}

sub _createProductWithSubproductPrices{
    my $self = shift;
    my ($UseSamePrices) = @_;
    my $Shop = DE_EPAGES::Test::Mock::EpagesObject->new('DE_EPAGES::Product::API::Object::Product', {
        'Site' => $self->_createShop(),
        'GUID' => '123-456',
        'MinOrder' => 5,
    });
    $Shop->mock('existSubProductPrices', sub {
        return 1;
    });
    $Shop->mock('subProductWithLowestPrice', sub {
        return $self->_createSubProduct();
    });
    $Shop->mock('subProducts', sub {
        my $aSubs;
        for (1..5) {
            push @$aSubs , $self->_createSubProduct($UseSamePrices ? '12.99' : $_ + 10 );            
        }
        return $aSubs;
    });
    return $Shop;
}

sub _createSubProduct{
    my $self = shift;
    my ($Price) = @_;

    my $Shop = DE_EPAGES::Test::Mock::EpagesObject->new('DE_EPAGES::Product::API::Object::Product', {
        'Site' => $self->_createShop(),
        'GUID' => '123-456',
        'MinOrder'    => 5,
        'IsAvailable' => 1,
        'IsVisible'   => 1,
        'HasSubOwnPrices' => 1

    });
    $Shop->mock('getListPrice', sub {
        return $Price ? $Price : '12.99';
    });
    $Shop->mock('displayPrice', sub {
        return $Price ? $Price : '12.99';
    });
    return $Shop;
}

sub _createProduct{
    my $self = shift;
    my $Shop = DE_EPAGES::Test::Mock::EpagesObject->new('DE_EPAGES::Product::API::Object::Product', {
        'Site' => $self->_createShop(),
        'GUID' => '123-456',
        'MinOrder' => 5,
    });
    $Shop->mock('existSubProductPrices', sub {
        return 1;
    });
    $Shop->mock('subProductWithLowestPrice', sub {
        return undef;
    });
    return $Shop;
}

sub _createShop {
    my $self = shift;
    return DE_EPAGES::Test::Mock::EpagesObject->new('DE_EPAGES::Shop::API::Object::Shop', {
        'Alias' => 'TestShop',
        'TaxModel' => 1,
    });
}

__PACKAGE__->runScript();