#========================================================================================
# package      DE_EPAGES::SPSBillingOnline::API::Cartridge
# base         DE_EPAGES::Design::API::DesignInstaller
# state        public
#----------------------------------------------------------------------------------------
# description  This is the main cartridge class for install/patch/uninstall.
#========================================================================================
package DE_EPAGES::SPSBillingOnline::API::Cartridge;
use base DE_EPAGES::Payment::API::PaymentInstaller;

use strict;
use Config;
use DE_EPAGES::Core::API::File qw ( ExistsFileCaseSensitive );
use DE_EPAGES::Core::API::Error qw (GetError ExistsError Error);
use DE_EPAGES::Core::API::EpagesConfig qw (EpagesConfig);

#========================================================================================
# function     new
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer = DE_EPAGES::SPSBillingOnline::API::Cartridge->new(%options);
#----------------------------------------------------------------------------------------
# description  Creates a new installer object of DE_EPAGES::SPSBillingOnline.
#----------------------------------------------------------------------------------------
# input        %options | options for un/install
#               <ul>
#                 <li>IsRecursive - recursive un/installation of required or dependent
#                     cartridges
#               </ul>
#               | hash
# return       $Installer | cartridge installer | object
#========================================================================================
sub new {
    my ($class, %options) = @_;

    my $self = __PACKAGE__->SUPER::new(
        %options,
        'CartridgeDirectory' => 'DE_EPAGES/SPSBillingOnline',
        'Version'            => '7.5.0',
        'Patches'            => ['6.15.1','6.15.2','6.15.3','6.15.4','6.16.0','6.16.1','6.16.2','6.16.3','6.17.0','6.17.1','6.17.2','6.17.3','6.17.4','6.17.5','6.17.6','6.17.7','6.17.8','6.17.9','6.17.9.1','6.17.9.2','6.17.9.3','6.17.10','6.17.11','6.17.12','6.17.12.1','6.17.12.2','6.17.13','6.17.13.1','6.17.13.2','6.17.14','6.17.14.1','6.17.15','6.17.15.1','6.17.16','6.17.17','6.17.17.1','6.17.18','6.17.19','6.17.20','6.17.20.1','6.17.21','6.17.21.1','6.17.22','6.17.23','6.17.24','6.17.24.1','6.17.25','6.17.26','6.17.26.1','6.17.27','6.17.27.1','6.17.28','6.17.28.1','6.17.29','6.17.29.1','6.17.30','6.17.31','6.17.32','6.17.33','6.17.33.1','6.17.34','6.17.34.1','6.17.35','6.17.36','6.17.36.1','6.17.37','6.17.38','6.17.38.1','6.17.39','6.17.39.1','6.17.40','6.17.40.1','6.17.41','6.17.41.1','6.17.42','6.17.43','6.17.43.1','6.17.44','6.17.45','6.17.45.1','6.17.46','6.17.47','6.17.47.1','6.17.48','6.17.48.1','6.17.49','6.17.49.1','6.17.50','6.17.51','6.17.51.1','6.17.52','6.17.53','7.0.0','7.1.0','7.2.0','7.3.0','7.4.0'],
    );
    return bless $self, $class;
}

sub testInstall {
    my $self = shift;

    my $Checks = [
        # no windows
        sub { Error('UNSUPPORTED_OS') if ( $Config{osname} eq 'MSWin32'); return; },
        # xmlsec
        sub { Error('MISSING_XML_SEC') unless `which xmlsec1 2> /dev/null`; return; },
        # file must exist
        sub {
            my $CertPath = EpagesConfig()->get('DE_EPAGES::SPSBillingOnline', 'CertificatePath', $ENV{'EPAGES_CONFIG'} . '/yellowworld.epages.com');
            Error('MISSING_PFX') unless ExistsFileCaseSensitive("$CertPath.pfx");
            return;
        }
    ];

    foreach my $Check (@$Checks) {
        eval {
            $Check->();
        };
        if (ExistsError()) {
            my $Text = GetError()->code;
            $self->addReason({
                'Reason'    => 'REQIREMENT_MISSING',
                'Package'   => $self->cartridgePackage,
                'Version'   => $self->version,
                'Text'      => $Text,
            });
        }
    }

    return $self->SUPER::testInstall;
}

#========================================================================================
# function     install
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer->install;
#----------------------------------------------------------------------------------------
# description  Installs the cartridge.
#========================================================================================
sub install {
    my $self = shift;

    $self->startInstall;
    $self->SUPER::install;
    $self->finishInstall;
    return;
}

#========================================================================================
# function     uninstall
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer->uninstall;
#----------------------------------------------------------------------------------------
# description  Uninstalls the cartridge.
#========================================================================================
sub uninstall {
    my $self = shift;

    $self->startUninstall;
    $self->SUPER::uninstall;
    $self->finishUninstall;
    return;
}

1;
