#========================================================================================
# package      DE_EPAGES::ShopCSVExportImport::API::CSV::ShopExportImport;
# state        public
#----------------------------------------------------------------------------------------
# description  Some functions to export / import shop objects.
#========================================================================================
package DE_EPAGES::ShopCSVExportImport::API::CSV::ShopExportImport;
use base qw( Exporter );
use strict;

our @EXPORT_OK = qw (
    ExportCSV
    ImportCSV
);

use DE_EPAGES::Core::API::Error qw ( GetError ExistsError Error);
use DE_EPAGES::Core::API::PerlTools qw (ElementOf UsePackage );
use DE_EPAGES::Core::API::TestParameter qw ( TestObject TestString );
use DE_EPAGES::Core::API::Warning qw ( BlockWarnings Warning );

use DE_EPAGES::Core::API::Log qw (GetLog);
use DE_EPAGES::Trigger::API::BatchProcessor qw ( RunBatchProcess ExistsBatchHook );

use constant MAX_WARNINGS => 100;

sub ErrorDefinition {
    return {
        'GenerateBatchName' => 'Can\'t generate batch action name from driver package: #Package'
    };
}

#========================================================================================
# function     ExportCSV
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $aWarnings = ExportCSV($Shop, $FileName, $Driver, $hOptions, $aObjects)
#----------------------------------------------------------------------------------------
# description  Exports objects in csv file.
#----------------------------------------------------------------------------------------
# input        $Shop | shop object | object
# input        $FileName | export file | string
# input        $DriverPackage | csv export driver package | string
# input        $hOptions | (optional) format options.
#                   <ul>
#                       <li>LocaleID - locale (default: localeid of shop)</li>
#                       <li>LanguageCode2s - array of languages with code2, e.g. ['de', 'en']
#                   </ul>
#                   The following parameters are optional. Their default values are
#                   determined by the given locale id:
#                   <ul>
#                       <li>eol - row separator, e.g. "\n" - string</li>
#                       <li>sep_char - quote character, e.g. ',' - string</li>
#                       <li>quote_char - quote character, e.g. '"' - string</li>
#                       <li>escape_char - escape character, e.g. '"' - string</li>
#                       <li>DecimalSep - decimal point, e.g. '.' - string</li>
#                       <li>ThousandSep - thousands separator, e.g. ',' - string</li>
#                       <li>encoding - character encoding, e.g. 'iso-8859-1' - string</li>
#                   </ul> | ref.hash
# input        $aObjects | list of objects to export | ref.array.object
# return       $aWarnings | Warnings which have occured | ref.array
#========================================================================================
sub ExportCSV {
    my ($Shop, $FileName, $DriverPackage, $hOptions, $aObjects) = @_;
    TestObject( 'Shop', $Shop );
    TestString( 'FileName', $FileName );
    TestString( 'DriverPackage', $DriverPackage );

    $hOptions->{'LocaleID'} = $Shop->get('LocaleID') unless defined $hOptions->{'LocaleID'};
    $hOptions->{'LanguageCode2s'} = [] unless defined $hOptions->{'LanguageCode2s'}; # all shop languages
    UsePackage($DriverPackage);
    my $Driver = $DriverPackage->new;
    my $aWarnings = BlockWarnings( sub {
        eval {
            $Driver->csvExport( $Shop, $FileName, $hOptions, $aObjects );
        };

        if (ExistsError()) {
            my $Error =GetError();
            if( ElementOf($Error->code,  @{$Driver->warningCodes}) ) {
                Warning($Error);
            } else {
                Error($Error);
            }
        }
    }, $Driver->warningCodes );

    return $aWarnings;
}

#========================================================================================
# function     ImportCSV
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $aWarnings = ImportCSV($Shop, $FileName, $Driver, $hOptions)
#----------------------------------------------------------------------------------------
# description  Imports objects from csv file.
#----------------------------------------------------------------------------------------
# input        $Shop | shop object | object
# input        $FileName | import file | string
# input        $DriverPackage | csv import driver package | string
# input        $hOptions | (optional) format options.
#                   <ul>
#                       <li>LocaleID - locale (default: localeid of shop)</li>
#                       <li>LanguageCode2s - array of languages with code2, e.g. ['de', 'en']
#                   </ul>
#                   The following parameters are optional. Their default values are
#                   determined by the given locale id:
#                   <ul>
#                       <li>eol - row separator, e.g. "\n" - string</li>
#                       <li>sep_char - quote character, e.g. ',' - string</li>
#                       <li>quote_char - quote character, e.g. '"' - string</li>
#                       <li>escape_char - escape character, e.g. '"' - string</li>
#                       <li>DecimalSep - decimal point, e.g. '.' - string</li>
#                       <li>ThousandSep - thousands separator, e.g. ',' - string</li>
#                       <li>encoding - character encoding, e.g. 'iso-8859-1' - string</li>
#                   </ul> | ref.hash
# return       $aWarnings | Warnings which have occured | ref.array
# return       $TooManyWarnings | to many warnings occured (see: MAX_WARNINGS) | boolean
#========================================================================================
sub ImportCSV {
    my ($Shop, $FileName, $DriverPackage, $hOptions) = @_;
    TestObject( 'Shop', $Shop );
    TestString( 'FileName', $FileName );
    TestString( 'DriverPackage', $DriverPackage );

    my $BatchRun;
    if($DriverPackage =~ m/^.+::(.+?)ExportImport$/x){
        $BatchRun = 'CSV' . $1 . 'Import';
    }
    else{
        Error( 'GenerateBatchName', {Package => $DriverPackage} );
    }

    GetLog->debug('Generate name for batch process: '. $DriverPackage . ' => ' . $BatchRun);

    $hOptions->{'LocaleID'} = $Shop->get('LocaleID') unless defined $hOptions->{'LocaleID'};
    $hOptions->{'LanguageCode2s'} = []; # all shop languages
    UsePackage($DriverPackage);
    my $Driver = $DriverPackage->new;

    my ($aWarnings, $TooManyWarnings) = BlockWarnings( sub {
        eval {
            my $cSub = sub { $Driver->csvImport( $Shop, $FileName, $hOptions) };
            if (ExistsBatchHook($BatchRun)) {
                RunBatchProcess($BatchRun, $cSub);
            } else {
                GetLog->warn('BatchHook for '.$BatchRun . ' does not exist. Running import non-batched.');
                $cSub->();
            }
        };
        if (ExistsError()) {
            my $Error = GetError();
            if ( ElementOf($Error->code, @{$Driver->warningCodes}) ) {
                Warning($Error);
            } else {
                Error($Error);
            }
        }
    }, $Driver->warningCodes, MAX_WARNINGS);


    return ($aWarnings, $TooManyWarnings);
}

1;
