use strict;
use feature 'say';
use feature 'switch';

use Data::Dumper;
use Time::HiRes qw (time);
use Getopt::Long;
use DE_EPAGES::Core::API::Error qw( GetError ExistsError Error );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::Connection qw( RunOnStore );
use DE_EPAGES::Object::API::Factory qw( LoadRootObject );
use DE_EPAGES::WebService::API::Client;

sub Main {

    my $Proxy;
    my $User;
    my $Password;
    my $ShopRefAlias;
    my $ShopTypeAlias;
    my $StoreName;
    my $ShopAlias;
    my $DomainName;
    my $HasSSLCertificate;
    my $WebServerScriptNamePart;
    my $IsTrialShop        = 0;
    my $IsInternalTestShop = 0;
    my $IsClosed;
    my $MerchantLogin;
    my $MerchantPassword;
    my $MerchantEMail;

    my $PublicFiles;
    my $Help;

    # validate command-line parameters
    GetOptions(
        'proxy=s'            => \$Proxy,
        'wsuser=s'           => \$User,
        'passwd=s'           => \$Password,
        'alias=s'            => \$ShopRefAlias,
        'shoptype=s'         => \$ShopTypeAlias,
        'storename=s'        => \$StoreName,
        'shopalias=s'        => \$ShopAlias,
        'domainname=s'       => \$DomainName,
        'ssl'                => \$HasSSLCertificate,
        'trialshop'          => \$IsTrialShop,
        'testshop'           => \$IsInternalTestShop,
        'closed'             => \$IsClosed,
        'scriptname=s'       => \$WebServerScriptNamePart,
        'publicfiles=s'      => \$PublicFiles,
        'merchantlogin=s'    => \$MerchantLogin,
        'merchantpassword=s' => \$MerchantPassword,
        'merchantemail=s'    => \$MerchantEMail,
        'help'               => \$Help,
    );

    my @ImportFiles = @ARGV;

    Usage() if $Help;
    Usage('Missing required parameter -wsuser')    unless defined $User;
    Usage('Missing required parameter -passwd')    unless defined $Password;
    Usage('Missing required parameter -alias')     unless defined $ShopRefAlias;
    Usage('Missing required parameter -shoptype')  unless defined $ShopTypeAlias;
    Usage('Missing required parameter -storename') unless defined $StoreName;
    $ShopAlias = $ShopRefAlias unless defined $ShopAlias;

    if ( !defined $Proxy ) {
        RunOnStore(
            'Store' => 'Site',
            'Sub'   => sub {
                $Proxy = LoadRootObject()->folder('Stores')->child('Site')->get('WebServiceProxy');
            }
        );
    }

    # create a web service client
    my $ShopConfigService = DE_EPAGES::WebService::API::Client->uri(
        'urn://epages.de/WebService/ShopConfigService/2016/02')->proxy( $Proxy, timeout => 900 );
    $ShopConfigService->userinfo("$User:$Password");

    my $time1 = time;

    # create a shop
    my $StoreObjectID;

    eval {
        $StoreObjectID = $ShopConfigService->create({
            Database                => $StoreName,
            ShopType                => $ShopTypeAlias,
            ShopAlias               => $ShopAlias,
            Alias                   => $ShopRefAlias,
            DomainName              => $DomainName,
            HasSSLCertificate       => $HasSSLCertificate,
            WebServerScriptNamePart => $WebServerScriptNamePart,
            IsTrialShop             => $IsTrialShop,
            IsInternalTestShop      => $IsInternalTestShop,
            IsClosed                => $IsClosed,
            ImportFiles             => \@ImportFiles,
            PublicFilesDir          => $PublicFiles,
            MerchantLogin           => $MerchantLogin,
            MerchantPassword        => $MerchantPassword,
            MerchantEMail           => $MerchantEMail
        });
    };
    if (ExistsError()) {
        my $Error = GetError();

        if ($Error->code eq 'WebServiceFault') {
            given ($Error->vars->{'FaultDetail'}->{'ErrorDetails'}->{'Code'}) {
                when (['UNITY_API_CALL_ERROR']) {
                    say 'Registering of the Unity shop failed.';
                }
            }
        }

        Error($Error);
    }

    printf "need %.1f seconds\n", time - $time1;
}

sub Usage {
    my ($Message) = @_;

    print STDERR "$Message\n\n" if defined $Message;

    print <<END_USAGE;
Description:
    Creates a new shop in the store database using a shop type of the
    provider specified by wsuser.
    Initial configuration data can be provided by XML import files.

Usage:
    perl $0 options [flags] [import files]

Options:
    -proxy             web service endpoint (default: local installation)
    -wsuser            web service user path
    -passwd            password of the web service user
    -alias             unique shop alias for the provider
    -shoptype          shop type alias
    -storename         store database alias
    -shopalias         shop alias within the store database (optional)
    -domainname        vanity domain name of the shop (optional)
    -scriptname        web server script name part (optional)
    -publicfiles       directory with public files, e.g. images (optional)
    -merchantlogin     merchant login
    -merchantpassword  merchant password
    -merchantemail     merchant e-mail addtress

Flags:
    -ssl        the domain specified by -domainname is SSL-enabled
    -trialshop  the shop is in trial period
    -testshop   the shop is used for test or support purposes
    -help       show the command line options
    -closed     created shop is closed

Examples:
    perl $0 -proxy http://myservername/epages/Site.soap -wsuser /Providers/Distributor/Users/admin -passwd admin -alias DemoShop -shoptype Demo -storename Store
    perl $0 -proxy http://myservername/epages/Site.soap -wsuser /Providers/Distributor/Users/admin -passwd admin -alias DemoShop -shoptype Demo -storename Store -shopalias DemoShop -scriptname demoshop -publicfiles $ENV{EPAGES_CARTRIDGES}/DE_EPAGES/DemoShop/Data/Public/Shops/DemoShop $ENV{EPAGES_CARTRIDGES}/DE_EPAGES/DemoShop/Database/XML/DemoShop.xml
END_USAGE
    exit 2;
}

RunScript( Sub => \&Main );
