#========================================================================================
# package      DE_EPAGES::Spreedly::UI::PaymentMethodSpreedlyStripe
# base         DE_EPAGES::Spreedly::UI::PaymentMethodSpreedly
# state        private
#----------------------------------------------------------------------------------------
# description  UI Controller for PaymentMethodSpreedlyStripe
#========================================================================================
package DE_EPAGES::Spreedly::UI::PaymentMethodSpreedlyStripe;
use base DE_EPAGES::Spreedly::UI::PaymentMethodSpreedly;

use DE_EPAGES::Core::API::IniConfig;
use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Shop::API::Url qw ( BuildShopUrl );
use DE_EPAGES::Spreedly::API::Constants qw ( OPERATIONAL_MODE_LIVE OPERATIONAL_MODE_TEST );

use strict;

use constant PAYMENT_TYPE_STRIPE => 'SpreedlyStripe';

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       new();
#----------------------------------------------------------------------------------------
# description  Class constructor.
#----------------------------------------------------------------------------------------
# input        $Servlet       | Settings page's Servlet              | object
# input        $PaymentMethod | Servlet's object                     | object
# return       $self          | PaymentMethodSpreedlyStripe instance | object
#========================================================================================
sub new {
    my $class = shift;
    my ($Servlet, $PaymentMethod) = @_;

    my $PaymentTypeAlias = $PaymentMethod->get('PaymentType')->alias();
    if ($PaymentTypeAlias ne PAYMENT_TYPE_STRIPE) {
        GetLog()->warn(sprintf('Expected PaymentType %s, found %s.', PAYMENT_TYPE_STRIPE, $PaymentTypeAlias));
    }

    my $self  = bless {
        'Servlet' => $Servlet,
        'PaymentMethod' => $PaymentMethod,
    }, $class;

    return $self;
}

#========================================================================================
# function     ViewSettings
# state        public
#----------------------------------------------------------------------------------------
# syntax       ViewSettings( $Servlet );
#----------------------------------------------------------------------------------------
# description  UI Controller to view settings on the MBO settings page.
#----------------------------------------------------------------------------------------
# input        $Servlet | Servlet | DE_EPAGES::Presentation::Servlet::BasePermission
#========================================================================================
sub ViewSettingsSpreedlyStripe {
    my $self = shift;

    my $URL = $self->_getCurrentURL();

    $self->{'Servlet'}->vars('StateURL', $URL);

    my $Shop = $self->{'PaymentMethod'}->getSite();
    my $IsInternalTestShop = $Shop->get('IsInternalTestShop');
    my $PaymentType = $self->{'PaymentMethod'}->get('PaymentType');
    my $ClientID = ($IsInternalTestShop) ? $PaymentType->get('TestAccountName') : $PaymentType->get('LiveAccountName');
    $self->{'Servlet'}->vars('ClientID', $ClientID);

    my $Form = $self->{'Servlet'}->form();
    my $AuthorizationCode = $Form->queryValue('code');

    if (isValid($AuthorizationCode)) {
        $self->_createGateway($AuthorizationCode, $IsInternalTestShop);
    }

    $self->{'Servlet'}->vars('SiteUrl', $self->_getSiteUrl());

    return $self->{'Servlet'};
}

#---------------------------------------------------------------------------
#  Private methods
#---------------------------------------------------------------------------
sub _createGateway {
    my $self = shift;
    my ($AuthorizationCode, $IsInternalTestShop) = @_;

    my $OperationalMode = $IsInternalTestShop ? OPERATIONAL_MODE_TEST : OPERATIONAL_MODE_LIVE;

    my $hAdditionalBodyParams = {
        'authorization_code' => $AuthorizationCode,
        'operational_mode'   => $OperationalMode,
    };

    my $PaymentMethodSpreedly = DE_EPAGES::Spreedly::UI::PaymentMethodSpreedly->new();
    $PaymentMethodSpreedly->SaveCredentials($self->{'Servlet'}, PAYMENT_TYPE_STRIPE, $hAdditionalBodyParams);

    return;
}

sub _getCurrentURL {
    my $self = shift;

    my $hFormValues = $self->{'Servlet'}->params();
    my $hParams = {
        'ViewAction' => $hFormValues->{'ViewAction'},
        'ObjectID'   => $self->{'PaymentMethod'}->id(),
    };
    my $hOptions = {
        'Type' => 'admin',
        'UseSSL' => 1,
    };

    return BuildShopUrl($self->{'PaymentMethod'}, $hParams, $hOptions, $self->{'Servlet'});
}

sub _getSiteUrl {
    my $self = shift;

    my $EpagesConf = DE_EPAGES::Core::API::IniConfig->new( FileName => $ENV{'EPAGES_CONFIG'}.'/epages.conf', CaseSensitive => 1 );

    return $EpagesConf->get( 'DE_EPAGES::Spreedly', 'SiteUrl');
}

1;
